:- use_module(counter).

% using assert/retract
count_to(1000000).
:- dynamic count/1.
bench_assert(X) :-
    assert(count(0)),
    bench_assert_aux(X).
bench_assert_aux(X) :-
    retract(count(C)),
    C < X, NC is C + 1,
    assert(count(NC)),
    bench_assert_aux(X).
bench_assert_aux(_).

% using C counter extension
bench_c(X) :-
    new_counter('my_counter'),
    bench_c_aux('my_counter',X).
bench_c_aux('my_counter',X) :-
    inc_counter('my_counter',C),
    C < X,
    bench_c_aux('my_counter',X).
bench_c_aux(_,_).

% using black-board operations
:- dynamic count/1.
bench_bb(X) :-
    bb_put(counter,0),
    bench_bb_aux(X).
bench_bb_aux(X) :-
    bb_get(counter,C),
    C < X, NC is C + 1,
    bb_put(counter,NC),
    bench_bb_aux(X).
bench_bb_aux(_).

bench :-
    counter_init,
    count_to(X),
    statistics(walltime,_),
    bench_assert(X),
    statistics(walltime,[_,TimeA]),
    bench_c(X),
    statistics(walltime,[_,TimeB]),
    bench_bb(X),
    statistics(walltime,[_,TimeC]),
    format('Counting to: ~w.~n Assert/Retract: ~w ms~n C-Based: ~w ms~n bb_put: ~w ms ~n', [X,TimeA,TimeB,TimeC]).

/*
Counting to: 1000000.
 Assert/Retract: 1091 ms
 C-Based: 285 ms
 bb_put: 727 ms
yes

*/
