% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(counter, [
                    counter_init/0,
                    new_counter/1,
                    get_counter/2,
                    set_counter/2,
                    inc_counter/1, inc_counter/2,
                    inc_counter_by/2,
                    reset_counter/1
                   ]).

foreign_resource('counter', [new_counter,get_counter,inc_counter,set_counter,reset_counter]).
foreign(new_counter, c, new_counter(+atom)).
foreign(get_counter, c, get_counter(+atom, [-integer])).
foreign(inc_counter, c, inc_counter(+atom, [-integer])).
foreign(set_counter, c, set_counter(+atom, +integer)).
foreign(reset_counter, c, reset_counter(+atom)).

% with this line below we need to adapt the Makefile to include the counter resource in the binary:
% (maybe this is a good idea)
% :- load_foreign_resource(counter).

:- dynamic loaded/0.

counter_init :-
  (loaded -> true
   ;  assert(loaded),
      %assert_dir,
      assert(user:library_directory('.')),
      safe_load_foreign_resource(counter)
   ).

inc_counter(Counter) :-
    inc_counter(Counter, _NewVal).

inc_counter_by(Counter,1) :- !,
    inc_counter(Counter, _NewVal).
inc_counter_by(Counter,Incr) :-
    get_counter(Counter,Old),
    New is Old+Incr,
    set_counter(Counter,New).

safe_load_foreign_resource(R) :-
   E=error(existence_error(_,_),_),
   % print(loading_foreign_resource(R)),nl,
   on_exception(E,
                load_foreign_resource(library(R)),
                (format(user_error,'~n! Could not load ~w library!~n! Check that it is available in the lib folder of ProB.~n~n',[R]),throw(E))).
