#include <sicstus/sicstus.h>
#include <stdlib.h> /* putenv() */

/*
 Example showing how to set environment variables from a userhook so
 they are seen by SICStus when it initializes itself.

 Note: The "right" way to pass options to SICStus, since SP 4.1, is
 to pass an option block to SP_initialize(). It has the advantage
 that the options does not pollute the process environment or the
 environment of sub-processes. However, using SU_initialize() is
 easier.
*/
#if _WIN32
 /* This would also work on MS Windows */
 #define PUTENV _putenv
#else
 /* This is the UNIX variant. It would also work on Windows but gives
    a warning suggesting _putenv. */
 #define PUTENV putenv
#endif

int SPCDECL SU_initialize(int argc, char *argv[])
{
 PUTENV("TRAILSTKSIZE=1M");
 PUTENV("PROLOGINCSIZE=50M");
 return 0;
}

#ifdef PROB_GLIBC_DOWNGRADE

/* CPP (preprocessor) madness -- just to put the value of PROB_GLIBC_DOWNGRADE into quotes (Daniel) */
#define STR_VALUE(arg)      #arg
#define AS_STRING(name) STR_VALUE(name)
#define DOWNGRADE_VERSION AS_STRING(PROB_GLIBC_DOWNGRADE)

/*
  This wrapper function is used to enforce that we have only a
  dependency to a pre-GLIBC 2.14 version.
 */
asm (".symver memcpy, memcpy@GLIBC_" DOWNGRADE_VERSION);
void *__wrap_memcpy(void *dest, const void *src, size_t n)
{
  return memcpy(dest, src, n);
}
#endif
