#ifndef helpers_included
#define helpers_included
#include "lib/include/zmq.h"
#include "lib/include/czmq.h" // FIXME update



enum SubMessageType {
    initial_state = 1,
    transition_group_list,
    variables_name_list,
    variables_type_list,
    state_label_matrix,
    may_write_matrix,
    must_write_matrix,
    reads_action_matrix,
    reads_guard_matrix,
    guard_info_matrix,
    guard_label_matrix,
    necessary_enabling_set,
    necessary_disabling_set,
    do_not_accord_matrix,
    may_be_coenabled_matrix,
    ltl_label_matrix,
    limit_message_types=128 // avoid undefined behaviour if we get new message types
};

typedef struct Chunk {
    char *data;
    int size;
} Chunk;

typedef struct ChunkArray {
    Chunk *chunks;
    size_t size;
} ChunkArray;

typedef ChunkArray State;

typedef struct MatrixRow {
    Chunk transition_group;
    ChunkArray variables;
} MatrixRow;

typedef struct Matrix {
    size_t nr_rows;
    MatrixRow *rows; // nr_rows many
} Matrix;

typedef struct InitialResponse {
    State initial_state;
    
    ChunkArray transition_groups;
    ChunkArray variables;
    ChunkArray variable_types;

    Matrix state_labels;

    Matrix may_write;
    Matrix must_write;
    Matrix reads_action;
    Matrix reads_guard;
} InitialResponse;


State get_state(zmsg_t *msg);
Matrix get_matrix(zmsg_t *msg);
InitialResponse get_init_response(zmsg_t *msg);

void put_state(zmsg_t *msg, State s);
void destroy_initial_response(InitialResponse *resp);
void destroy_state(State *s);
void destroy_chunk_array(ChunkArray *arr);
void destroy_matrix(Matrix *m);

void drop_frame(zmsg_t *msg);

#endif
