% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(graphiso,[graphiso_init/0,
          prob_start_graph/1,
          prob_add_edge/2,
          prob_get_number_of_colours/1,
          prob_set_colour_of_node/2,
          prob_exists_graph/1,
          prob_free_storage/0]).

:- use_module('../../src/module_information.pl').

:- module_info(group,symmetry).
:- module_info(description,'This is the interface to the C code for the symmetry reduction with nauty.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

% make -f MakePLlib
foreign(prob_init,prob_init).
foreign(prob_start_graph,prob_start_graph(+integer)).
foreign(prob_add_edge,prob_add_edge(+integer,+integer)).
foreign(prob_get_number_of_colours,prob_get_number_of_colours(+integer)).
foreign(prob_set_colour_of_node,prob_set_colour_of_node(+integer,+integer)).
foreign(prob_exists_graph,prob_exists_graph([-integer])).
foreign(prob_free_storage,prob_free_storage).
%foreign(prob_testrun,prob_testrun). %removed

foreign_resource(graphiso,
  [
   prob_init
  ,prob_start_graph
  ,prob_add_edge
  ,prob_get_number_of_colours
  ,prob_set_colour_of_node
  ,prob_exists_graph
  ,prob_free_storage
  ]).

:- dynamic loaded/0.

graphiso_init :- loadfr, prob_init.

loadfr :-
  (loaded -> true
   ;  assert(loaded),
      assert_dir,
      assert(user:library_directory(prob_home('.'))),
      safe_load_foreign_resource(graphiso)
   ).


assert_dir :- (user:library_directory('.') -> true ; assert(user:library_directory('.'))).

safe_load_foreign_resource(R) :-
   E=error(existence_error(_,_),_),
   % print(loading_foreign_resource(R)),nl,
   on_exception(E,
                load_foreign_resource(library(R)),
                (format(user_error,'~n! Could not load ~w library!~n! Check that it is available in the lib folder of ProB.~n~n',[R]),throw(E))).
