:- module(listsp, [
    % library(sets), should have been library(basic)
    select/3,       %  Elem <- Set -> Set
    selectchk/3,        %  Elem x Set -> Set
    % library(lists)
    append/2,           %   ListOfLists -> List
    append/5,           %   List x List x List x List x List
    correspond/4,           %   Elem <- List x List -> Elem
    delete/3,           %   List x Elem -> List
    delete/4,           %   List x Elem x Count -> List
    is_list/1,          %   List ->
    keys_and_values/3,      %   KeyValList -> KeyList x ValList
    last/2,             %   List -> Elem
    nextto/3,           %   Elem, Elem <- List
    nth0/3,             %   Integer x List -> Elem
    nth0/4,             %   Integer x List -> Elem x List
    nth1/3,             %   Integer x List -> Elem
    nth1/4,             %   Integer x List -> Elem x List
    one_longer/2,           %   List x List ->
    perm/2,             %   List -> List
    perm2/4,            %   Elem x Elem -> Elem x Elem
    permutation/2,          %   List <-> List
    proper_length/2,        %   List -> Length
    remove_dups/2,          %   List -> Set
    rev/2,              %   List -> List
    reverse/2,          %   List -> List
    same_length/2,          %   List x List ->
    same_length/3,          %   List x List x Integer ->
    select/4,           %   Elem x List x Elem -> List
    selectchk/4,            %   Elem x List x Elem -> List
    shorter_list/2,         %   List x List ->
    subseq/3,           %   List -> List x List
    subseq0/2,          %   List -> List
    subseq1/2,          %   List -> List
    sumlist/2,          %   List -> Integer
    transpose/2,            %   ListOfLists <-> ListOfLists
    % library(length)
    append_length/3,            % List x List x Length
    append_length/4,            % List x List x List x Length
    prefix_length/3,            % Whole x Part x Length
    proper_prefix_length/3,         % Whole x Part x Length
    proper_suffix_length/3,         % Whole x Part x Length
    rotate_list/2,              % List x List
    rotate_list/3,              % Integer -> (List x List)
    suffix_length/3,            % Whole x Part x Length
    sublist/3,              % Whole x Part x Length
    sublist/4,              % Whole x Part x Length^2
    sublist/5,              % Whole x Part x Length^3
    % library(list_parts)
    cons/3,         last/3,
    head/2,         tail/2,
    prefix/2,       proper_prefix/2,
    suffix/2,       proper_suffix/2,
    segment/2,      proper_segment/2,
    % library(maplist)
    cumlist/4,
    cumlist/5,
    cumlist/6,
    maplist/2,
    maplist/3,
    maplist/4,
    map_product/4,
    scanlist/4,
    scanlist/5,
    scanlist/6,
    some/2,
    some/3,
    some/4,
    somechk/2,
    somechk/3,
    somechk/4,
    % library(more_maps)
    convlist/3,
    exclude/3,
    exclude/4,
    exclude/5,
    include/3,
    include/4,
    include/5,
    partition/5,
    group/3,
    group/4,
    group/5,
    % library(ordered)
    ordered/1,
    ordered/2,
    max_member/2,
    min_member/2,
    max_member/3,
    min_member/3,
    select_max/3,
    select_max/4,
    select_min/3,
    select_min/4,
    % library(ordprefix)
    decreasing_prefix/3,
    decreasing_prefix/4,
    increasing_prefix/3,
    increasing_prefix/4,
    % library(clump)
    clumped/2,
    clumps/2,
    keyclumped/2,
    keyclumps/2
   ]).
/* use when overridden
:- meta_predicate
    cumlist(3, +, ?, ?),
    cumlist(4, ?, ?, ?, ?),
    cumlist(5, ?, ?, ?, ?, ?),
    maplist(1, ?),
    maplist(2, ?, ?),
    maplist(3, ?, ?, ?),
    map_product(3, +, +, ?),
    scanlist(3, +, ?, ?),
    scanlist(4, ?, ?, ?, ?),
    scanlist(5, ?, ?, ?, ?, ?),
    some(1, ?),
    some(2, ?, ?),
    some(3, ?, ?, ?),
    somechk(1, +),
    somechk(2, +, +),
    somechk(3, +, +, +),
    convlist(2, +, ?),
    exclude(1, +, ?),
    exclude(2, +, +, ?),
    exclude(3, +, +, +, ?),
    include(1, +, ?),
    include(2, +, +, ?),
    include(3, +, +, +, ?),
    partition(2, +, ?, ?, ?),
        partition_1(+, 2, ?, ?, ?),
            partition_1(+, +, ?, ?, ?, +, 2),
    group(2, +, ?),
        group1(+, ?, 2),
    group(1, +, ?, ?),
    group(2, +, +, ?, ?),
    ordered(2, +),
        ordered_0(+, 2),
        ordered_1(+, +, 2),
    max_member(2, +, ?),
    min_member(2, +, ?),
    select_min(2, ?, +, ?),
        sel_min_gen(+, ?, 2, ?),
        sel_min_gen(+, ?, +, 2, -, ?),
    select_max(2, ?, +, ?),
        sel_max_gen(+, ?, 2, ?),
        sel_max_gen(+, ?, +, 2, -, ?),
    decreasing_prefix(2, +, ?, ?),
    first_precedes(2, +, +),
    increasing_prefix(2, +, ?, ?),
    precedes_first(2, +, +).
*/

:- use_module(probsrc(module_information)).
:- module_info(group,tools).
:- module_info(description,'Patched lists library with runtime checks.').


/* plspec setup */
:- use_module(extension('plspec/plspec/plspec_core')).

 :- enable_all_spec_checks.
%:- plspec:enable_spec_check([avl_domain/2, avl_fetch/2]).


:- defspec(tl_list, one_of([compound([any|any]), atom([])])).
:- defspec(tl_list_var, one_of([compound([any|any]), atom([]), var])).
:- defspec_pred(is_list, is_list).
:- defspec(integer_or_var, one_of([integer, var])).

:- spec_pre(select/3,  [any,tl_list, tl_list_var]).
:- spec_pre(selectchk/3,  [any, tl_list, tl_list_var]).
:- spec_pre(append/2,  [list(tl_list_var), tl_list_var]).
:- spec_pre(delete/3,  [tl_list, any, tl_list_var]).
:- spec_pre(nth0/3,  [integer_or_var, tl_list_var, any]).
:- spec_pre(nth0/4,  [integer_or_var, tl_list_var, any, tl_list_var]).
:- spec_pre(nth1/3,  [integer_or_var, tl_list_var, any]).
:- spec_pre(nth1/4,  [integer_or_var, tl_list_var, any, tl_list_var]).
:- spec_pre(same_length/2,  [tl_list_var, tl_list_var]).
:- spec_pre(same_length/3,  [tl_list_var, tl_list_var,integer_or_var]).

select(A,B,C) :- lists:select(A,B,C).
selectchk(A,B,C) :- lists:selectchk(A,B,C).
append(A,C) :- lists:append(A,C).
delete(A,B,C) :- lists:delete(A,B,C).
nth0(A,B,C) :- lists:nth0(A,B,C).
nth0(A,B,C,D) :- lists:nth0(A,B,C,D).
nth1(A,B,C) :- lists:nth1(A,B,C).
nth1(A,B,C,D) :- lists:nth1(A,B,C,D).
same_length(A,B) :- lists:same_length(A,B).
same_length(A,B,C) :- lists:same_length(A,B,C).

:- use_module(library(~~(lists)),
   [
    %%select/3,     %  Elem <- Set -> Set
    %%selectchk/3,      %  Elem x Set -> Set
    % library(lists)
    %%append/2,         %   ListOfLists -> List
    append/5,           %   List x List x List x List x List
    correspond/4,           %   Elem <- List x List -> Elem
    %%delete/3,         %   List x Elem -> List
    delete/4,           %   List x Elem x Count -> List
    is_list/1,          %   List ->
    keys_and_values/3,      %   KeyValList -> KeyList x ValList
    last/2,             %   List -> Elem
    nextto/3,           %   Elem, Elem <- List
    nth0/3,             %   Integer x List -> Elem
    nth0/4,             %   Integer x List -> Elem x List
    nth1/3,             %   Integer x List -> Elem
    nth1/4,             %   Integer x List -> Elem x List
    one_longer/2,           %   List x List ->
    perm/2,             %   List -> List
    perm2/4,            %   Elem x Elem -> Elem x Elem
    permutation/2,          %   List <-> List
    proper_length/2,        %   List -> Length
    remove_dups/2,          %   List -> Set
    rev/2,              %   List -> List
    reverse/2,          %   List -> List
    same_length/2,          %   List x List ->
    same_length/3,          %   List x List x Integer ->
    select/4,           %   Elem x List x Elem -> List
    selectchk/4,            %   Elem x List x Elem -> List
    shorter_list/2,         %   List x List ->
    subseq/3,           %   List -> List x List
    subseq0/2,          %   List -> List
    subseq1/2,          %   List -> List
    sumlist/2,          %   List -> Integer
    transpose/2,            %   ListOfLists <-> ListOfLists
    % library(length)
    append_length/3,            % List x List x Length
    append_length/4,            % List x List x List x Length
    prefix_length/3,            % Whole x Part x Length
    proper_prefix_length/3,         % Whole x Part x Length
    proper_suffix_length/3,         % Whole x Part x Length
    rotate_list/2,              % List x List
    rotate_list/3,              % Integer -> (List x List)
    suffix_length/3,            % Whole x Part x Length
    sublist/3,              % Whole x Part x Length
    sublist/4,              % Whole x Part x Length^2
    sublist/5,              % Whole x Part x Length^3
    % library(list_parts)
    cons/3,         last/3,
    head/2,         tail/2,
    prefix/2,       proper_prefix/2,
    suffix/2,       proper_suffix/2,
    segment/2,      proper_segment/2,
    % library(maplist)
    cumlist/4,
    cumlist/5,
    cumlist/6,
    maplist/2,
    maplist/3,
    maplist/4,
    map_product/4,
    scanlist/4,
    scanlist/5,
    scanlist/6,
    some/2,
    some/3,
    some/4,
    somechk/2,
    somechk/3,
    somechk/4,
    % library(more_maps)
    convlist/3,
    exclude/3,
    exclude/4,
    exclude/5,
    include/3,
    include/4,
    include/5,
    partition/5,
    group/3,
    group/4,
    group/5,
    % library(ordered)
    ordered/1,
    ordered/2,
    max_member/2,
    min_member/2,
    max_member/3,
    min_member/3,
    select_max/3,
    select_max/4,
    select_min/3,
    select_min/4,
    % library(ordprefix)
    decreasing_prefix/3,
    decreasing_prefix/4,
    increasing_prefix/3,
    increasing_prefix/4,
    % library(clump)
    clumped/2,
    clumps/2,
    keyclumped/2,
    keyclumps/2
   ]).
