:- module(z3interface, [init_z3interface/0,
                        reset_z3interface/0, % TODO: make more lightweight once cvc 1.5 is out
                        pretty_print_smt1/0, % pretty print Z3's internal status and constraints
                        pretty_print_smt2/0, % pretty print Z3's internal status and constraints
                        pretty_print_smt2_for_id/1, % pretty print store + constraint pointed to by ID (previously created by z3_interface_call(s))
                        z3_interface_call/1]).

:- use_module('../../src/module_information.pl').
:- module_info(group,experimental).
:- module_info(description,'This is the interface between ProB and the Z3 SMT solver').

/*
:- use_module(probsrc(bsyntaxtree), [map_over_typed_bexpr/2,
                            find_typed_identifier_uses/3,
                            get_texpr_type/2,
                            get_texpr_id/2]).
*/
foreign_resource('z3interface', [init,
                                 pretty_print_smt1, pretty_print_smt2, pretty_print_smt2_for_id,
                                   mk_var,
                                   mk_bounded_var,
                                   mk_int_const,
                                   mk_real_const,
                                   mk_bool_const,
                                   mk_empty_set,
                                   mk_full_set,
                                   mk_record_const,
                                   mk_record_field,
                                   mk_set,
                                   mk_sort,
                                   mk_sort_with_cardinality,
                                   pop_frame, push_frame,
                                   setup_enumerated_set,
                                   mk_op,
                                   mk_op_arglist,
                                   mk_op_reverse,
                                   mk_op_domain,
                                   mk_op_range,
                                   mk_op_image,
                                   mk_op_comprehension_set_multi,
                                   mk_op_comprehension_set_singleton,
                                   smt_solve_query,
                                   reset,
                                   get_model_string,
                                   get_full_model_string,
                                   mk_quantifier,
                                   conjoin_negated_state,
                                   add_interpreter_constraint,
                                   mk_couple,
                                   get_header_version,
                                   get_full_version]).
% will be called by solver_interface:
:- public mk_var/3, mk_bounded_var/3, mk_empty_set/2, mk_int_const/2, mk_real_const/2, mk_bool_const/2. %, mk_string_const/2.
:- public mk_set/2, mk_sort/1, mk_sort_with_cardinality/2, mk_record_const/3, mk_record_field/4.
:- public setup_enumerated_set/1, mk_op/4, mk_quantifier/4, mk_op_arglist/3, smt_solve_query/4, get_model_string/2, conjoin_negated_state/3.
:- public pop_frame/0, push_frame/0, add_interpreter_constraint/4, mk_couple/4, mk_op_image/6, mk_op_reverse/3, mk_op_domain/5, mk_op_range/5.
:- public mk_op_comprehension_set_multi/4, mk_op_comprehension_set_singleton/3.

:- public mk_full_set/2, get_full_model_string/1, reset/0. % Z3 only

% function declarations
foreign(init, c, init).
foreign(pretty_print_smt1, c, pretty_print_smt1).
foreign(pretty_print_smt2, c, pretty_print_smt2).
foreign(pretty_print_smt2_for_id, c, pretty_print_smt2_for_id(+integer)).
foreign(mk_var, c, mk_var(+term, +string, [-integer])).  % creates a variable; first argument is an atom of the type,
                                                         % second argument is the identifier
foreign(mk_bounded_var, c, mk_bounded_var(+term, +string, [-integer])).
foreign(mk_empty_set, c, mk_empty_set(+term, [-integer])).
foreign(mk_full_set, c, mk_full_set(+term, [-integer])).
foreign(mk_int_const, c, mk_int_const(+integer, [-integer])).  % creates an int constant
foreign(mk_real_const, c, mk_real_const(+string, [-integer])).    % creates a real constant
foreign(mk_bool_const, c, mk_bool_const(+string, [-integer])). % creates a bool constant
foreign(mk_set, c, mk_set(+term, [-integer])).
foreign(mk_sort, c, mk_sort(+atom)).
foreign(mk_sort_with_cardinality, c, mk_sort_with_cardinality(+atom,+integer)).
foreign(mk_record_const, c, mk_record_const(+term, +term, [-integer])).
foreign(mk_record_field, c, mk_record_field(+integer, +integer, +atom, [-integer])).
foreign(setup_enumerated_set, c, setup_enumerated_set(+term)).
foreign(mk_op, c, mk_op(+string, +integer, +integer, [-integer])).
foreign(mk_quantifier, c, mk_quantifier(+string, +term, +integer, [-integer])).
foreign(mk_op_arglist, c, mk_op_arglist(+string, +term, [-integer])).
foreign(mk_op_reverse, c, mk_op_reverse(+integer, +term, [-integer])).
foreign(mk_op_domain, c, mk_op_domain(+integer, +term, +term, +term, [-integer])).
foreign(mk_op_range, c, mk_op_range(+integer, +term, +term, +term, [-integer])).
foreign(mk_op_image, c, mk_op_image(+integer, +integer, +term, +term, +term, [-integer])).
foreign(mk_op_comprehension_set_singleton, c, mk_op_comprehension_set_singleton(+integer, +integer, [-integer])).
foreign(mk_op_comprehension_set_multi, c, mk_op_comprehension_set_multi(+term, +term, +integer, [-integer])).
foreign(smt_solve_query, c, smt_solve_query(+atom,+integer,+integer,[-atom])).
foreign(get_model_string, c, get_model_string(+integer, [-atom])).
foreign(get_full_model_string, c, get_full_model_string([-atom])).
foreign(conjoin_negated_state, c, conjoin_negated_state(+term,+integer,[-integer])).
foreign(pop_frame, c, pop_frame).
foreign(push_frame, c, push_frame).
foreign(reset, c, reset).
foreign(add_interpreter_constraint, c, add_interpreter_constraint(+integer,+atom,+integer,[-term])).
foreign(mk_couple, c, mk_couple(+term,+integer,+integer,[-integer])).
foreign(get_header_version, c, get_header_version([-atom])).
foreign(get_full_version, c, get_full_version([-atom])).


:- dynamic is_initialised/0.

init_z3interface :- is_initialised,!.
init_z3interface :-
    catch(load_foreign_resource(library(z3interface)),E,
          (format(user_error,'*** LOADING Z3 library failed~n*** Be sure to set Z3_HOME and have libz3.dylib/so/dll on your dynamic library path.~n*** ~w~n',[E]),fail)),
    init,
    assert(is_initialised).

reset_z3interface :- is_initialised, !,
    reset.
reset_z3interface :- init_z3interface.

:- meta_predicate z3_interface_call(+).
z3_interface_call(Call) :-
    call(z3interface:Call).
