:- module(whash_send, []).

:- use_module(library(fastrw),[ fast_buf_read/2,fast_buf_write/3 ]).

foreign_resource('whash_send', [new_message,
                                new_results_message,
                                start_timer,
                                stop_timer,
                                msg_put_str,
                                put_succ,
                                add_stats,
                                send_and_destroy_msg]).
foreign(new_message, c, new_message([-address(zmsg_t)])).
foreign(new_results_message, c, new_results_message([-address(zmsg_t)])).
foreign(start_timer, c, start_timer).
foreign(stop_timer, c, stop_timer).
foreign(msg_put_str, c, msg_put_str(+address(zmsg_t), +string)).
foreign(put_succ, c, put_succ(+address(zmsg_t), +address(char), +integer)).
foreign(add_stats, c, add_stats(+address(zmsg_t))).
foreign(send_and_destroy_msg, c, send_and_destroy_msg(+address(zmsg_t))).

:- load_foreign_resource(whash_send).

put_results(WorkPackage, []). % no results -> don't create a message
put_results(WorkPackage, [H|T]) :-
    new_results_message(Msg),
    put_results2([H|T], Msg),
    pretty_print(WorkPackage, Pretty),
    msg_put_str(Msg, WorkPackage),
    send_and_destroy_msg(Msg).
put_results2([], _).
put_results2([H|T], Msg) :-
    % TODO: string representations of Prolog terms
    msg_put_str(Msg, H),
    put_results(T, Msg).

% TODO: state that has been checked should be included
put_successors([], _Msg).
put_successors([H|T], Msg) :-
    fast_buf_write(H, Len, Addr),
    put_succ(Msg, Addr, Len), % TODO: additional information?
    put_successors(T, Msg).

c_callback(ResultList, SuccessorList, HashMsg) :-
    put_results(ResultList, HashMsg),
    put_successors(SuccessorList, HashMsg),
    length(ResultList, S),
    send_hashes(HashMsg).


process(Addr, HashMsg) :-
    start_timer,
    fast_buf_read(WorkPackage, Addr),
    process(WorkPackage, Results, Successors),
    stop_timer,

    put_successors(Successors, HashMsg),
    put_results(Results),

    add_stats(HashMsg),
    send_and_destroy_msg(HashMsg).


