% (c) 2011-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(zmq, [init_zmq/0,
                master_main/10, alloc_term/3,
                setup_worker/5, work_reactor/1, teardown_worker/0,
                                new_results_message/1, start_timer/0, stop_timer/0,
                                msg_put_str/2, put_succ/4, add_stats/1, send_and_destroy_msg/1]).

:- use_module('../../src/module_information.pl').
:- module_info(group,experimental).
:- module_info(description,'This is the interface to C code for distributed model checking.').
:- module_info(revision,'$Rev: 9609 $').
:- module_info(lastchanged,'$LastChangedDate: 2011-11-18 10:43:02 +0100 (Fr, 18 Nov 2011) $').

foreign_resource('zmq', [master_main, alloc_term,
                         setup_worker, work_reactor, teardown_worker,
                         new_results_message,
                         start_timer, stop_timer,
                         msg_put_str, put_succ,
                         add_stats, send_and_destroy_msg]).

% master functions
foreign(master_main, c, master_main(+integer,+integer,+integer,+address(zchunk_t),+address(zchunk_t),+string,+string,+string,+integer,[-integer])).
foreign(alloc_term, c, alloc_term(+address(char), +integer, [-address(zchunk_t)])).

% worker functions
foreign(setup_worker, c, setup_worker(+integer, +integer, +string, +string, +string)).
foreign(work_reactor, c, work_reactor([-integer])).
foreign(teardown_worker, c, teardown_worker).
foreign(new_results_message, c, new_results_message([-address(zmsg_t)])).
foreign(start_timer, c, start_timer).
foreign(stop_timer, c, stop_timer).
foreign(msg_put_str, c, msg_put_str(+address(zmsg_t), +string)).
foreign(put_succ, c, put_succ(+address(zmsg_t), +address(char), +integer, +integer)).
foreign(add_stats, c, add_stats(+address(zmsg_t))).
foreign(send_and_destroy_msg, c, send_and_destroy_msg(+address(zmsg_t))).

:- dynamic is_initialised/0.

init_zmq :- is_initialised,!.
init_zmq :-
    safe_load_foreign_resource(zmq),
    assert(is_initialised).


safe_load_foreign_resource(R) :-
   E=error(existence_error(_,_),_),
   % print(loading_foreign_resource(R)),nl,
   on_exception(E,
                load_foreign_resource(library(R)),
                (format(user_error,'~n! Could not load ~w library!~n! Check that it is available in the lib folder of ProB.~n~n',[R]),throw(E))).
