% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

% bmachine provides access to the currently loaded B-machine and all its subsidiary machines

:- module(bmachine,[bmachine_is_precompiled/0,
                    b_set_initial_machine/0,
                    b_set_empty_machine/0,
                    b_set_machine/3,
                    b_set_typed_machine/2,
                    b_load_eventb_project/1, % used for loading .eventb files
                    b_set_eventb_project_flat/3, % set an Event-B project from terms

                    get_full_b_machine/2,
                    full_b_machine/1,
                    get_full_b_machine_sha_hash/1, % get SHA hash of loaded and type-checked B machine
                    b_machine_is_loaded/0,

                    get_proven_invariant/2,
                    get_unproven_invariants/1,
                    load_additional_information/1,
                    %% generate_specialized_invariants/0, %% no longer exists
                    get_operation_info/2,

                    b_load_machine_from_file/1, b_load_machine_from_file/2,
                    b_load_machine_probfile/1,
                    b_load_machine_from_list_of_facts/2,
                    b_load_machine_from_term/2, % used for Alloy
                    b_machine_precompile/0,  other_spec_precompile/0,
                    b_machine_reset/0,

                    b_machine_name/1,

                    b_get_all_used_filenames/1, b_get_main_filename/1,
                    b_get_main_filenumber/1, b_filenumber/4,
                    b_absolute_file_name_relative_to_main_machine/2,
                    b_get_animated_sections/1,

                    b_get_machine_set/1, b_get_machine_set/2,
                    b_get_named_machine_set/2,
                    b_get_disjoint_constants_of_type/3, b_get_constant_represented_inside_global_set/2,
                    b_get_machine_constants/1, b_get_machine_all_constants/1,
                    b_machine_has_constants/0, b_machine_has_constants_or_properties/0,
                    b_is_constant/2, b_is_constant/1, b_is_unused_constant/1,
                    b_machine_has_variables/0,
                    b_is_variable/1, b_is_variable/2,
                    b_get_constant_variable_description/2,
                    b_get_machine_variables/1,
                    b_get_all_used_identifiers/1, % Note: may contain var$0 instead of var for becomes_such
                    b_get_all_used_identifiers_in_section/2, % ditto
                    get_machine_identifiers/2,
                    source_code_for_identifier/6,

                    b_get_properties_from_machine/1,
                    b_get_invariant_from_machine/1,
                    b_get_linking_invariant_from_machine/1,
                    b_machine_has_assertions/0,
                    b_get_static_assertions_from_machine/1, b_machine_has_static_assertions/0,
                    b_get_unproven_static_assertions_from_machine/1,
                    b_get_dynamic_assertions_from_machine/1, b_machine_has_dynamic_assertions/0,
                    b_get_unproven_dynamic_assertions_from_machine/1,
                    b_get_assertions_from_main_machine/2, b_main_machine_has_no_assertions/0,
                    b_machine_has_unproven_assertions/0, get_assertions_from_machine/2,
                    get_all_assertions_from_machine/1,
                    b_machine_has_proven_invariants/0,
                    b_get_assertions/3,
                    b_get_assertion_count/3,
                    is_discharged_assertion/1,

                    b_get_initialisation_from_machine/2,
                    b_machine_has_operations/0,
                    b_get_machine_operation/4, b_get_machine_operation/6,
                    b_get_machine_operation_for_animation/6, % does some rewriting for ANY parameters
                    b_get_promoted_machine_operations/1,
                    b_top_level_operation/1, b_top_level_feasible_operation/1,
                    b_is_operation_name/1, b_get_operation_pos/2,
                    b_get_machine_operation_parameter_types/2,
                    b_get_machine_operation_parameter_names/2,
                    b_get_machine_operation_typed_parameters/2,
                    b_machine_operation_names_in_reverse_order/1,
                    b_get_operation_variant/3,

                    b_get_machine_setscope/2,
                    b_get_machine_operation_max/2,
                    b_get_machine_goal/1,
                    b_set_machine_goal/1, b_set_machine_goal/2,
                    b_set_machine_searchscope/1, b_set_machine_searchscope/2,
                    b_set_parsed_typed_machine_goal/1,
                    b_reset_machine_goal_from_DEFINITIONS/0,
                    b_parse_machine_expression_from_codes/2,
                    b_parse_machine_expression_from_codes_with_prob_ids/2,
                    b_parse_machine_expression_from_codes/5,
                    b_parse_machine_expression_from_codes/6,
                    b_parse_machine_formula/3, b_parse_machine_formula_from_codes/7,
                    b_parse_machine_predicate/2, b_parse_machine_predicate/3,
                    b_parse_optional_machine_predicate/2,
                    b_parse_machine_predicate_from_codes/3,
                    b_parse_machine_predicate_from_codes_open/5,
                    b_parse_machine_operation_pre_post_predicate/3,
                    get_machine_operation_additional_identifiers/2,

                    determine_type_of_formula/2,

                    parse_expression_raw_or_atom/2, parse_expression_raw_or_atom_with_prob_ids/2,

                    % a version where parsing and type checking are separate:
                    b_parse_wo_type_machine_predicate_from_codes_to_raw_expr/2,
                    b_type_check_raw_expr/4,

                    b_get_machine_searchscope/1,
                    b_get_machine_animation_function/2,
                    b_get_machine_heuristic_function/1,
                    b_get_machine_animation_expression/1,
                    b_get_machine_custom_edges_function/2,
                    b_get_machine_custom_nodes_function/2,
                    b_machine_temp_predicate/1, set_temp_predicate/2, assert_temp_typed_predicate/1,
                    reset_temp_predicate/0,
                    add_additional_property/1, b_machine_additional_property/1,

                    get_animation_image/2, get_animation_image_source_file/2,
                    b_get_definition/5,
                    b_get_definition_string_from_machine/2,
                    b_definition_prefixed/5,
                    % get typed definitions without parameters:
                    b_get_typed_definition/3, b_get_typed_predicate_definition/3,
                    b_get_typed_expression_definition/3, b_get_true_expression_definition/1,

                    b_specialized_invariant_for_op/2, b_operation_preserves_full_invariant/1,
                    tcltk_get_specialized_invariant_for_op/2, tcltk_get_specialized_invariants_for_ops/1,
                    b_get_operation_normalized_read_write_info/3, % used to be b_normalized_rwsets_for_op/3,
                    reuse_operation_effect/2, candidate_operation_for_reuse/1,
                    b_operation_cannot_modify_state/1,
                    b_operation_reads_output_variables/3,
                    b_get_operation_non_det_modifies/2,

                    b_extract_values_clause_assignment/3,

                    b_type_expression/5,b_type_expression_for_full_b_machine/6,
                    b_type_open_predicate/5,b_type_open_predicate_with_errors/4,
                    b_type_open_exists_predicate/3,
                    type_with_errors/4,

                    b_show_machine_representation/3, b_show_machine_representation/4,
                    b_write_machine_representation_to_file/2,
                    b_write_eventb_machine_to_classicalb_to_file/1,
                    b_show_eventb_as_classicalb/2,

                    b_get_machine_refinement_hierarchy/1,
                    b_get_refined_machine/1, b_get_refined_machine_name/1, b_get_refined_ancestors_names/1,
                    b_get_machine_header_position/2,
                    b_get_model_type/1,
                    b_machine_statistics/2,

                    % More Proof Information:
                    discharged_guard_strengthening/4,

                    % Flow information
                    wp_untyped/3, clear_wp/0,
                    nonchanging_guard/2,
                    typecheck_predicates/4
                   ]).

:- use_module(library(lists)).
:- use_module(library(ordsets)).

:- use_module(self_check).
:- use_module(error_manager).
:- use_module(debug).
:- use_module(bmachine_construction,[check_machine/4,type_in_machine_l/6,type_open_predicate_with_quantifier/6]).
:- use_module(bsyntaxtree).
:- use_module(bmachine_structure,[get_section/3,write_section/4,get_section_texprs/3]).
:- use_module(parsercall).
:- use_module(tools).
:- use_module(translate,[translate_machine/3,translate_eventb_to_classicalb/3]).
:- use_module(bmachine_eventb,[check_event_b_project/4,is_eventb_additional_info/1]).
:- use_module(kernel_freetypes,[reset_freetypes/0,add_freetype/2]).
:- use_module(b_machine_hierarchy).
:- use_module('kodkod/kodkod', [replace_by_kodkod/3]).

:- use_module(b_global_sets,[find_inequal_global_set_identifiers/4,
                             b_get_prob_deferred_set_elements/2,
                             b_check_and_precompile_global_set_type_definitions/0]).

% for applying transformations on the machines
:- use_module(record_detection,[replace_sets_by_records/2]).
:- use_module(partition_detection,[detect_partitions/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'This module provides access to the various parts of the loaded B machine.').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% bmachine: module for storing the currently loaded machine
%   Contains predicates for getting (already typed) information about
%   different parts of the machine
%  many predicates are precompiled by b_precompile_machine/0, see below for
%  details.

:- volatile bmachine_is_precompiled/0.
:- dynamic bmachine_is_precompiled/0.  /* TRUE if the precompiled predicates have been set */

assert_bmachine_is_precompiled :-
   (bmachine_is_precompiled -> true ; assert(bmachine_is_precompiled), debug_println(4,bmachine_is_precompiled)).

:- volatile machine/2.
:- dynamic machine/2.
% there is only one main machine fact

% true if a machine is loaded
b_machine_is_loaded :- (machine(_,_) -> true).

get_full_b_machine(Name,Res) :-
   (machine(Name,M) -> Res=M
     ;  (ground(Name) -> add_error(bmachine,'No B machine with this name loaded:',Name)
        ; add_error(bmachine,'No B machine loaded')),
        fail
   ).

full_b_machine(machine(Name,M)) :- get_full_b_machine(Name,M).

:- use_module(extension('probhash/probhash'),[raw_sha_hash/2]).
get_full_b_machine_sha_hash(Hash) :- full_b_machine(FM),  raw_sha_hash(FM,Hash).

b_set_machine(Main,Machines,Errors) :-
    b_machine_reset,
    check_machine(Main,Machines,M1,Errors1),!,
    apply_machine_transformations(M1,ResultMachine),
    ( no_real_perror_occurred(Errors1) ->
        %% get_memory_used(M1), %%
        assert_main_machine(ResultMachine)
       %%, get_memory_used(M2),print_memory_used(M2),  print_memory_used_difference(M1,M2),nl %%
    ; otherwise ->
        true),
    Errors=Errors1.

assert_main_machine(X) :- %tools_printing:trace_print(X),nl,
    nonvar(X), X=machine(A,B),
    !,
    assert(machine(A,B)). %, print(machine_asserted),nl.
assert_main_machine(X) :- add_internal_error('Illegal format: ',assert_main_machine(X)),fail.

:- meta_predicate apply_transformation_step(*,2,*,*).
% apply some transformations on the machine that require global information
% (other local transformations can be found in ast_cleanup)
apply_machine_transformations -->
    apply_transformation_step('record definition detection', replace_sets_by_records),
    apply_transformation_step('partition detection', detect_partitions). % is this necessary?? done in ast_cleanup

apply_transformation_step(Name,Pred,In,Out) :-
    ( call(Pred, In, Out) -> true
    ; otherwise ->
        ajoin(['applying ', Name, ' failed, skipping.'], Msg),
        add_error(bmachine,Msg),
        In=Out).

% try to translate (parts of) the properties to Kodkod
% if it succeeds, replace the translated predicates by
% calls to the Kodkod process
try_kodkod :-
    get_preference(try_kodkod_on_load,true),
    b_get_machine_constants(Constants),
    Constants = [_|_],!,
    b_get_properties_from_machine(OldPredicate),
    ( replace_by_kodkod(Constants,OldPredicate,NewPredicate) ->
        retract(machine(N,OldMachine)),
        retractall( b_get_properties_from_machine(_) ),
        write_section(properties,NewPredicate,machine(N,OldMachine),NewMachine),
        assert( NewMachine ),
        assert( b_get_properties_from_machine(NewPredicate) )
    ; otherwise ->
        true).
try_kodkod.

% set a machine term as current machine, without calling type checking
% (used for Z translations)
b_set_typed_machine(M1,FromFile) :-
    b_machine_clear,
    apply_machine_transformations(M1,ResultMachine),
    assert_main_machine(ResultMachine),
    set_all_used_filenames([FromFile]).

b_set_eventb_project_flat(Models,Contextes,Proofs) :-
    b_machine_clear,
    analyse_eventb_hierarchy(Models,Contextes),
    ( check_event_b_project(Models,Contextes,Proofs,M1) ->
        apply_machine_transformations(M1,ResultMachine),
        assert_main_machine(ResultMachine) % has the form machine/2
    ; otherwise ->
        % Note: if there are typecheck errors due to missing .ptm files for axiomatic operators it is not an internal error!
        fail).

:- volatile package/1, stored_preference/2.
:- dynamic package/1, stored_preference/2.
clear_package :- retractall(package(_)),retractall(stored_preference(_,_)).

b_load_eventb_project(Filename) :-
  clear_package,
  consult_without_redefine_warning(Filename),
  (package(load_event_b_project(Machines,Contexts,Proofs,Errors))
    -> true
    ; package(load_event_b_project(Machines,Contexts,Errors))
      -> Proofs=[], print('% Deprecated EventB Project without Proof Info.'),nl
      ; add_error_and_fail(b_load_eventb_project,'This is not an EventB Project file:',Filename)
  ),
  (Errors=[] -> true ; add_error(b_load_eventb_project,'Errors in Event-B Package: ',Errors)),
   read_stored_prefs,
   set_all_used_filenames([Filename]),
   b_set_eventb_project_flat(Machines,Contexts,Proofs),
   clear_wp,
   (load_additional_information(Proofs) -> true ; add_error(b_load_eventb_project,'Failed to load proof information',Proofs)).
%   generate_specialized_invariants.

read_stored_prefs :- stored_preference(Pref,Val), printsilent(stored_preference(Pref,Val)),nls,
      preferences:set_preference(Pref,Val),
      fail.
read_stored_prefs.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Collect Proof Information from Event B

:- volatile discharged/3.
:- volatile discharged_theorem/3, wd_po/3, wp_untyped/3.
:- volatile nonchanging_guard/2, discharged_guard_strengthening/4.
:- dynamic discharged/3.
% discharged(Model,Event,Name) : invariant Name in model Model is preserved by Event (no matter which Event at which refinement level)
:- dynamic discharged_theorem/3.
:- dynamic wd_po/3.
:- dynamic wp_untyped/3.
:- dynamic nonchanging_guard/2.
:- dynamic discharged_guard_strengthening/4.

:- volatile b_specialized_invariant_for_op/2.
:- dynamic b_specialized_invariant_for_op/2.
:- volatile get_proven_invariant/2.
:- dynamic get_proven_invariant/2.



clear_wp :-
    retractall( wp_untyped(_,_,_) ).
    %retractall( wp_typed(_,_,_) ). only in flow.pl

load_additional_information([]).
load_additional_information([H|T]) :-
      load_additional_information_fact(H),!,
      load_additional_information(T).
load_additional_information([H|T]) :-
      add_internal_error('Unknown additional info: ',load_additional_information_fact(H)),
      load_additional_information(T).

% old style proof info
load_additional_information_fact(discharged(Machine,Theorem)) :- % a discharged theorem
      debug_println(9,discharged_theorem(Machine,old_style_proof_info,Theorem)),
      assert(discharged_theorem(Machine,invariant,Theorem)).
load_additional_information_fact(discharged(Machine,Event,Invariant)) :-
      debug_println(9,discharged(Machine,Event,Invariant)),
      assert(discharged(Machine,Event,Invariant)).
load_additional_information_fact(Term) :- % theories are handled in bmachine_eventb.pl already
    is_eventb_additional_info(Term).

% new style proof info po(Machine,Info,SourceList,Discharged)

load_additional_information_fact(po(Machine,'Invariant  preservation',Source,true)) :- % invariant preservation
      add_discharged_po_infos(Source,Machine).

load_additional_information_fact(po(Machine,'Invariant  establishment',Source,true)) :- % invariant establishment
      add_discharged_po_infos(Source,Machine).

load_additional_information_fact(po(Machine,Text,Source,Proven)) :- % invariant or theorem wd
      ( Text = 'Well-definedness of Invariant'
       ; Text = 'Well-definedness of Theorem'),
      member(invariant(Invariant),Source),
      debug_println(9,wd_po(Machine,Invariant,Proven)),
      assert(wd_po(Machine,Invariant,Proven)).
load_additional_information_fact(po(Machine,Text,Source,Proven)) :- % axiom or axiom theorem wd
      ( Text = 'Well-definedness of Axiom'
       ; Text = 'Well-definedness of Theorem'),
      member(axiom(Axiom),Source),
      debug_println(9,wd_axiom(Machine,Axiom,Proven)).   % TO DO: assert and use
load_additional_information_fact(po(Machine,Text,Source,Proven)) :- % proven theorem as guard
      Text = 'Well-definedness of Theorem',
      member(guard(Grd),Source),
      member(event(Evt),Source),
      debug_println(9,wd_axiom(Machine,Evt/Grd,Proven)).   % TO DO: assert and use
load_additional_information_fact(po(Machine,Text,Source,Proven)) :- % wd of action
      Text = 'Well-definedness of action',
      (member(action(ACT),Source) % probably the guard in the last event with discharged info ??
       -> debug_println(9,wd_action(Machine,ACT,Proven))   % TO DO: assert and use
       ;  format('Illegal Rodin PO source in ~w for ~w: ~w~n',[Machine,Text,Source])
       ).
load_additional_information_fact(po(Machine,Text,Source,Proven)) :- % wd of guard
      Text = 'Well-definedness of Guard',
      (member(guard(GRD),Source) % probably the guard in the last event with discharged info ??
       -> debug_println(9,wd_guard(Machine,GRD,Proven))  % TO DO: assert and use
       ;  format('Illegal Rodin PO source in ~w for ~w: ~w~n',[Machine,Text,Source])
       ).
load_additional_information_fact(po(Machine,'Theorem',[invariant(Theorem)],true)) :- % a theorem
      debug_println(9,discharged_theorem(Machine,invariant,Theorem)),
      assert(discharged_theorem(Machine,invariant,Theorem)).
load_additional_information_fact(po(Machine,'Theorem',[axiom(Theorem)],true)) :- % a theorem
      debug_println(9,discharged_theorem(Machine,axiom,Theorem)),
      assert(discharged_theorem(Machine,axiom,Theorem)).
load_additional_information_fact(po(Machine,'Theorem',[guard(GuardTheorem),event(Event)],true)) :- % a theorem in a guard
      debug_println(9,discharged_theorem(Machine,guard(Event),GuardTheorem)),
      assert(discharged_theorem(Machine,guard(Event),GuardTheorem)).

% stuff we don't yet care about
load_additional_information_fact(po(Machine,Text,Infos,Proven)) :-
      ( Text = 'Guard strengthening (split)'
      ; Text = 'Guard strengthening (merge)'
      ; Text = 'Guard strengthening' ),
      !,
      (Proven=true, Infos = [event(AbsEvent),guard(Guard),event(ConcreteEvent)]
        -> assert(discharged_guard_strengthening(Machine,AbsEvent,Guard,ConcreteEvent)),
           debug_println(19,discharged_guard_strengthening(Machine,AbsEvent,Guard,ConcreteEvent))
        ;  true
      ).
load_additional_information_fact(po(_,'Variant of event',_,_)).
load_additional_information_fact(po(_,'Natural number variant of event',_,_)).
load_additional_information_fact(po(_,'Action simulation',_,_)).
load_additional_information_fact(po(_,'Feasibility of witness',_,_)).
load_additional_information_fact(po(_,'Feasibility of action',_,_)).
load_additional_information_fact(po(_,'Well-definedness of witness',_,_)).
load_additional_information_fact(po(_,'Well-definedness of variant',_,_)).
load_additional_information_fact(po(_,'Finiteness of variant',_,_)).
load_additional_information_fact(po(_,'Equality of common variables',_,_)).

load_additional_information_fact(po(_Machine,_Kind,_Source,false)). % ignore everything that is not proven

load_additional_information_fact(po(_Machine,_Kind,_Source,reviewed)). % ignore everything that is only reviewed

% ignore exporter version for now
% we are compatible to both versions 2 and 3 (added abstract constants)
load_additional_information_fact(exporter_version(V)) :- V=2 ; V=3.

% ignored because the pragmas are attached to the machine in load_event_b_project/4
load_additional_information_fact(pragma(_Type,_Source,_AttachedTo,_Content)).

load_additional_information_fact(wp(S,D,P)) :-
      assert( wp_untyped(S,D,P) ).

load_additional_information_fact(nonchanging_guard(E,P)) :-
      typecheck_predicates(P,TypedPred,E,_),
      assert(nonchanging_guard(E,TypedPred)).

load_additional_information_fact(invariant(_I)). %ignore

typecheck_predicates([],[],_,_).
typecheck_predicates([H|T],[TH|TT],S,D) :-
   (S='INITIALISATION' -> Scope1 = []; Scope1=[operation(S)]),
   ((D='INITIALISATION';var(D)) -> Scope2 = Scope1; Scope2=[operation(D)|Scope1]),
   Scope = [variables|Scope2],
   b_type_expression(H,Scope, _, TH, _),
   typecheck_predicates(T,TT,S,D).


add_discharged_po_infos(Source,Machine) :-
  % Source = [event(A1),...,event(CurMachineEvent),invariant(Inv)]
  % note that CurMachineEvent could be equal to A1; A1 can be split up into multiple events (one of which could have the same name as A1; see Ticket PROB-292)
      get_bottom_event(Source,Event),
      member(invariant(Invariant),Source),
      debug_println(9,discharged(Machine,Event,Invariant)),
      assert(discharged(Machine,Event,Invariant)),
      % TO DO: should we propagate info up to abstract events which are not split up or where all refinements preserve a certain invariant ??
      fail.
add_discharged_po_infos(_Source,_Machine).

% get the last event entry in proof info source list
get_bottom_event(Source,Event) :-
 (bottom_event(Source,'$NONE'(none),Event) -> true
   ; add_error(bmachine,'Could not determine event for proof info: ',Source),fail).
bottom_event([],Ev,Ev) :- Ev \= '$NONE'(none).
bottom_event([event(E)|T],_,Res) :- !,bottom_event(T,E,Res).
bottom_event([_|T],E,Res) :- bottom_event(T,E,Res).

:- use_module(debug).

%b_specialized_invariant_for_op_calc(Event,Invariant) :-
% b_get_invariant_from_machine(Invariant).

tcltk_get_specialized_invariants_for_ops(list(Res)) :-
   findall([Name,InvHeader,TI | Tail],
            ((b_top_level_operation(Name) ; Name='$initialise_machine'),
              tcltk_get_specialized_invariant_for_op_single(Name,TI,NrUnProven),
              tcltk_get_proven_invariant_for_op(Name,PI),
              ajoin([' INVARIANT (',NrUnProven,')'],InvHeader),
              (PI='none' -> Tail = ['END;']
               ; Tail = ['  NON-TRIVIAL PROVEN INVARIANT:', PI, 'END;']
              )
             ), Res).
tcltk_get_specialized_invariant_for_op_single(Event,TI,NrUnProven) :-
   (b_specialized_invariant_for_op(Event,Invariant) ->
     conjunction_to_list(Invariant,IL), length(IL,NrUnProven),
     translate:translate_bexpression(Invariant,TI)
     ; TI = 'not_simplified', NrUnProven='-').
tcltk_get_specialized_invariant_for_op(Event,list(TI)) :-
   (b_specialized_invariant_for_op(Event,Invariant) ->
     conjunction_to_list(Invariant,IL),
     (IL=[] -> TI = ['fully_proven']
      ; maplist(translate:translate_bexpression,IL,TI))
    ; TI = ['not_simplified']).
tcltk_get_proven_invariant_for_op(Event,TI) :- % get non-trivially proven invariants
   (get_proven_invariant(Event,Invariant), \+ is_truth(Invariant) ->
     translate:translate_bexpression(Invariant,TI)
    ; TI = 'none').

% use proof information to compute subset of invariant that needs to be checked for operations
:- public b_specialized_invariant_for_op_calc/2.
b_specialized_invariant_for_op_calc(Event,Invariant) :-
    preferences:get_preference(use_po,true),
    get_invariant_list_with_used_ids(IL,IL_Ids),
    % Note: we can have interference with CSE here, as lazy_let_pred at top-level can merge several conjuncts (but CSE is now disabled for the invariant if PROOF_INFO = TRUE)
    IL \= [], % otherwise: Invariant already trivial; no sense in keeping track of information
    get_op_or_init_and_syntax_filter(Event,ProofInfoEventName,IL,IL_Ids,RIL),
    exclude(is_proven_invariant(ProofInfoEventName),RIL,RL), % Warning !! The invariants are not all from Model; the discharged information could be for a proven invariant with the same label from another model, see examples/EventBPrologPackages/ProofDirected/TestR_label_reuse.eventb; this should now properly be dealt with by filter_proven_invariants
    \+ (IL=RL),  /* something filtered out */
    debug_println(9,simplified(Event)),
    conjunct_predicates(RL,Invariant). %, print(simplified),nl,translate:print_bexpr(Invariant),nl.


get_op_or_init_and_syntax_filter('$initialise_machine','INITIALISATION',IL,_,IL). % INIT always sets all variables
get_op_or_init_and_syntax_filter(Event,Event,IL,IL_Ids,RIL) :-
    b_is_operation_name(Event),
    filter_syntactically_relevant_invariants(IL,IL_Ids,Event,RIL). % filter out all invariants not modified by Event

% check if an operation preserves the full invariant; according to proof information
b_operation_preserves_full_invariant(Event) :-
    b_specialized_invariant_for_op(Event,Invariant),
    is_truth(Invariant).

:- public get_proven_invariant_calc/2.
% computes proven, relevant invariants (not trivial ones which are unaffected by Event) for main model
get_proven_invariant_calc(Event, Invariant) :-
    preferences:get_preference(use_po,true),
    b_machine_name(Model),
    get_invariant_list_with_used_ids(IL,IL_Ids),
    b_is_operation_name(Event),
    %print(filter_syntactically_relevant_invariants(Event,Model)),debug:nl_time,
    filter_syntactically_relevant_invariants(IL,IL_Ids,Event,RIL),
    %print(select_proven_invariants(Event,Model)),debug:nl_time,
    select_proven_invariants(RIL,Model,Event,RL),
    %print(conjunct_predicates(Event,Model)),debug:nl_time,
    conjunct_predicates(RL,Invariant).

get_invariant_list_with_used_ids(IL,IL_Ids) :-
    b_get_invariant_from_machine(I),
    conjunction_to_list(I,IL),
    maplist(find_id_uses_for_maplist,IL,IL_Ids).
find_id_uses_for_maplist(Inv,IDs) :- find_identifier_uses(Inv,[],IDs).

% compute all invariants which are not fully proven:
get_unproven_invariants(UnProvenList) :-
    findall(I,get_unproven_invariant_calc(I),UnProvenList).
get_unproven_invariant_calc(Invariant) :- % not precompiled at the moment
    b_machine_name(Model),
    b_get_invariant_from_machine(I),
    conjunction_to_list(I,IL),
    member(Invariant,IL),
    (is_unproven_inv_for_model(Invariant,Model,_) -> true
      ; debug_mode(on) -> print('PROVEN INVARIANT: '), translate:print_bexpr(Invariant),nl,fail
    ).
is_unproven_inv_for_model(_Invariant,_Model,_Event) :-
    preferences:get_preference(use_po,false),!.
is_unproven_inv_for_model(Invariant,Model,Event) :-
    b_is_operation_name(Event),
    is_syntactically_relevant_invariant_for_event(Event,Invariant),
    \+ is_proven_invariant_for_model(Model,Event,Invariant).


% filter_syntactically_relevant_invariants(+InvariantsList, +UsedIDs, +EventName, -RelevantItemsFromInvariantsList)
filter_syntactically_relevant_invariants(Invariants,IL_Ids,Event,Result) :-
    get_operation_info(Event,SubstitutionInfo),
    get_modifies_from_info(SubstitutionInfo,SortedModIds),
    filter_syntactically_relevant_invariants2(Invariants,IL_Ids,SortedModIds,Result).
    % noticeably faster than include, see e.g. public_examples/B/PerformanceTests/Generated/Generated4000.mch
    %include(is_syntactically_relevant_inv(SortedModIds),Invariants,IL_Ids,Result).

is_syntactically_relevant_inv(SortedModIds,_Inv,IDs) :-
    ord_intersect(SortedModIds,IDs).

filter_syntactically_relevant_invariants2([],_,_,[]).
filter_syntactically_relevant_invariants2([Inv|T],[IDs|TI],SortedModIds,R) :-
    (is_syntactically_relevant_inv(SortedModIds,Inv,IDs)
     -> R=[Inv|TR] ; R=TR),
    filter_syntactically_relevant_invariants2(T,TI,SortedModIds,TR).

is_syntactically_relevant_invariant_for_event(Event,Inv) :-
    get_operation_info(Event,SubstitutionInfo),
    get_modifies_from_info(SubstitutionInfo,SortedModIds),
    find_id_uses_for_maplist(Inv,IDs),
    is_syntactically_relevant_inv(SortedModIds,_Inv,IDs).

get_modifies_from_info(SubstitutionInfo,SortedModIds) :-
  ((memberchk(modifies(ModID),SubstitutionInfo) -> ground(ModID))
    ->  list_to_ord_set(ModID,SortedModIds) % is this necessary ?
   ;  SortedModIds=[]).

% filter_proven_invariants(+InvariantsList, +ModelName, +EventName, -RelevantItemsFromInvariantsList)
% this gets called during model checking time
is_proven_invariant(Event,E) :-
    % print('Filter Check: '), print(Event), print(' : '),translate:print_bexpr(E),nl,
   get_texpr_pos(E,POS),
   POS = rodinpos(Model,Name,_), % only accept new rodinpos information with three parameters !
   !,    % the old one with two parameters could lead to confusion as labels can be reused
   % TO DO: check that if E is not always well-defined that there are no unproven WD proof obligations !
  discharged_invariant(Model,Name,Event),
  check_wd_discharged_inv_or_thm(E,Model,Name,Event). %% print('DISCHARGED'(Model,Event,Name)),nl.

check_wd_discharged_inv_or_thm(E,_,_,_) :- always_well_defined(E),!.
check_wd_discharged_inv_or_thm(_,Model,Name,_Event) :- wd_po(Model,Name,ProofStatus),!,
    ProofStatus=true.
check_wd_discharged_inv_or_thm(_,Model,Name,Event) :-
    % no WD_PO associated or stored
    % TO DO: we could try ProB-WD prover; but this situation only happens with old-style exports; we could  use add_message
    format('% Assuming invariant/theorem ~w is well-defined for event ~w:~w (no WD PO found).~n',[Name,Model,Event]).

% select_proven_invariants(+InvariantsList, +ModelName, +EventName, -RelevantItemsFromInvariantsList)
select_proven_invariants(Invariants,Model,Event,Result) :-
   include(is_proven_invariant_for_model(Model,Event),Invariants,Result).

% Note: different from is_proven_invariant; TO DO: merge (here we assume main model)
is_proven_invariant_for_model(MainModel,Event,E) :- %% print(chck(MainModel,Event)),nl,
   get_rodin_name(E,Name), % if we do not have a rodin position we consider this as not proved
   (get_rodin_model_name(E,Model) -> true
     ; formatsilent('Could not get Rodin model name for invariant (deprecated .eventb format): ~w~n',[Name]),
       Model=MainModel,
       fail  % we have an old style .eventb file; safer to assume not proven as there can be clashes of labels
    ),
   discharged_invariant(Model,Name,Event).

is_discharged_assertion(Theorem) :-
   if( get_texpr_pos(Theorem,rodinpos(Model,Name,_)),
        (discharged_theorem(Model,_,Name), % does not matter if invariant or axiom
         check_wd_discharged_inv_or_thm(Theorem,Model,Name,assertion) ) % check that well-defined
      , (specfile:animation_minor_mode(eventb),
         (silent_mode(on) -> true
          ; print('Could not get position information for theorem: '),
            translate:print_bexpr(Theorem),nl,
            get_texpr_pos(Theorem,POS),debug_println(19,get_texpr_pos(POS))
         ),
         fail)
      ).
% derived predicate
discharged_invariant(Model,InvariantName,Event) :-
  % InvariantName from model Model is preserved by Event and any of its refinements
  discharged(Model,Event,InvariantName).
discharged_invariant(Model,InvariantName,Event) :-
  % InvariantName from model Model is preserved by Event and any of its refinements
  also_discharged(Model,Event,InvariantName).

b_machine_has_proven_invariants :-
  (discharged(_,_,_) -> true).

% complete the discharged/3 information: propagating information from abstract events
% down to refined top-level events
:- public complete_discharged_info_calc/0.
complete_discharged_info_calc :- debug_println(4,'completing discharged info: '),
    statistics(runtime,[Start,_]),
    (complete_discharged_info2 -> true ; true),
    statistics(runtime,[Stop,_]), T is Stop-Start, debug_print(4,T), debug_println(4,' ms').
:- volatile discharged2/2, also_discharged/3.
:- dynamic discharged2/2.
:- dynamic also_discharged/3.
discharged_info_exists :- discharged2(_,_),!.
complete_discharged_info2 :- retractall(also_discharged(_,_,_)),
    retractall(discharged2(_,_)),
    discharged(Model,Event,_), % find out which Model and Events are used in discharged info
   \+ discharged2(Model,Event), %print(discharged(Model,Event)),nl,
   assert(discharged2(Model,Event)),fail.
complete_discharged_info2 :- discharged_info_exists,
   b_get_animated_sections(Sections),
   discharged2(Model,Event),   % Event in Model was proven to preserve an invariant in Model
   b_get_abstract_events(Event,RModel,AbstractEvents), % search for a refinement of Model where the same event exists
   RModel \= Model,
   % i.e., invariant Name was proven in an abstract Model for Event
   \+ is_abstract_event_in_list(Event,Model,AbstractEvents), % this event is *NOT* a refinement of the proven one
   debug_println(4,removing_discharged(Model,Event)),
   retractall(discharged(Model,Event,_Name)),
   member(model(Model),Sections),  % only generate error message if the Model is also animated using b_get_animated_sections, see example  EventBPrologPackages/ProofDirected/TestInvPreserv_M2_ko_mch.eventb
   %  (otherwise the refinement may well exist but has been filtered out by restricting the animation levels)
   add_error(event_refinement,'An event does not refine an abstract event of the same name: ',Event:Model:RModel),
   % This will could cause errors in the proof information usage !
   %print(AbstractEvents),nl,
   fail.
complete_discharged_info2 :- discharged_info_exists,
   complete_discharged_info3(_Model,_Event,_Name).
complete_discharged_info3(InvNameModel,AbsEvent,Name) :-
   b_get_abstract_events(EventInRefinement,_RefModel,AbstrEvents),
   is_abstract_event_in_list(AbsEvent,InvNameModel,AbstrEvents), % Event in InvNameModel is an abstract event of EventInRefinement
   %print(check(InvNameModel,Event, EventInRefinement,RefModel)),nl,
   discharged(InvNameModel,AbsEvent,Name),  % print(chck2(Model,Event,Name)),nl,
   %InvNameModel:Name has been proven for AbsEvent *at all relevant* levels
   \+ discharged(InvNameModel,EventInRefinement,Name),
   \+ also_discharged(InvNameModel,EventInRefinement,Name),
   % this invariant has not yet been marked as proven for EventInRefinement
   debug_println(4,also_discharged_in_refinement(InvNameModel:Name,EventInRefinement,refined_from(AbsEvent))),
   assert(also_discharged(InvNameModel,EventInRefinement,Name)),
   fail.

%abstract_event_of_calc(TopLevelEvent,RModel,Event,Model) :-
%      b_get_abstract_events(TopLevelEvent,RModel,AbstrEvents),
%      is_abstract_event_in_list(Event,Model,AbstrEvents) , print(abs(TopLevelEvent,RModel,Event,Model)),nl.

is_abstract_event_in_list(Event,Model,AbstractEvents) :-
   member(b(rlevent(AEvent,AModel,_S,_P,_G,_Thm,_Act,_VW,_PW,_Unm,AbstractEvents2),_,_),AbstractEvents),
   (Event=AEvent, Model=AModel;
    is_abstract_event_in_list(Event,Model,AbstractEvents2)).

% get events that are refined by an event
b_get_abstract_events(Name,Section,AbsEvents) :-
    b_get_machine_operation(Name,_R,_P,Body,_OType,_OpPos),
    get_texpr_expr(Body,RLEVENT),
    RLEVENT = rlevent(_Name,Section,_Stat,_Parameters,_Grd,_Thm,_Act,_VWit,_PWit,_Unmod,AbsEvents).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- volatile b_operation_reads_output_variables/3, b_operation_cannot_modify_state/1.
:- dynamic b_operation_reads_output_variables/3, b_operation_cannot_modify_state/1.

% test if we have a query or skip operation which cannot modify the state
% (preconditions, assertions could still fail though !)
:- public b_operation_cannot_modify_state_calc/1.
b_operation_cannot_modify_state_calc(Event) :-
   b_get_operation_normalized_read_write_info(Event,_ReadVariables,[]).

:- use_module(b_read_write_info, [get_accessed_vars/4]).
% check if classical B operation reads its output variables
:- public b_operation_reads_output_variables_calc/3.
b_operation_reads_output_variables_calc(Event,ReadOutputVariablesIds,ReadOutputTypedInfo) :-
   b_get_machine_operation(Event,Res,_TParas,TBody,_OType,_Pos),
   def_get_texpr_ids(Res,ResIds),
   sort(ResIds,SResIds),
   get_operation_info(Event,SubstitutionInfo),
   member(reads_locals(Reads),SubstitutionInfo), % locals are either arguments or return output variables
   ord_intersection(Reads,SResIds,ReadOutputVariablesIds),
   ReadOutputVariablesIds \= [],
   % now recompute the information based on TBody after ast_cleanup: typing predicates have been removed
   get_accessed_vars(TBody,[],_,AllReadsAfterAstCleanup), % important to pass [] as local variables
   ord_intersection(AllReadsAfterAstCleanup,ReadOutputVariablesIds,ReadOutputVariablesIds2),
   ReadOutputVariablesIds2 \= [],
   (get_preference(allow_operations_to_read_outputs,true), debug_mode(off) -> true
     ; ajoin(['Operation ', Event, ' is possibly reading output parameters: '],Msg),
       add_message(bmachine,Msg,ReadOutputVariablesIds2,Res)
       % Note: analysis is not always precise: /public_examples/B/NewSyntax/CallOperationInExprSimple2.mch
   ),
   findall(reads_result(Index,ID,Type),
             (nth1(Index,Res,b(identifier(ID),Type,_)), ord_member(ID,ReadOutputVariablesIds2)),
             ReadOutputTypedInfo).


% Reuse information

:- volatile b_get_operation_normalized_read_write_info/3.
:- dynamic b_get_operation_normalized_read_write_info/3.
:- volatile reuse_operation_effect/2, candidate_operation_for_reuse/1.
:- dynamic reuse_operation_effect/2, candidate_operation_for_reuse/1.

:- use_module(bsyntaxtree,[get_global_identifiers/1]).

:- public b_get_operation_normalized_read_write_info_calc/3.
/* compute Read and Written Variables by an operation */
b_get_operation_normalized_read_write_info_calc(Event,SortedReadVariables,SortedWrittenVariables) :-
   get_global_identifiers(IS), % Note: find identifier uses was called before global sets precompiled
   get_operation_info(Event,SubstitutionInfo),
   (memberchk(modifies(WrittenVariables),SubstitutionInfo)
    -> sort(WrittenVariables,SWV), % normally not necessary; just to be sure
       ord_subtract(SWV,IS,SortedWrittenVariables)
     ; print(no_modifies_info(Event)),nl,fail),
   (memberchk(reads(ReadVariables),SubstitutionInfo)
    -> sort(ReadVariables,SWR),
       ord_subtract(SWR,IS,SortedReadVariables)
     ; debug_println(9,no_reads_info(Event)),fail). % happens currently in Event-B !
 %% ,  print(read_write_set(Event,ReadVariables,WrittenVariables)),nl.

:- use_module(specfile, [animation_minor_mode/1]).
:- public reuse_operation_effect_calc/2.
% True if after Executing Operation G:
%       the effect of the operation H (which writes variables in WriteH) is unaffected
reuse_operation_effect_calc(G,H) :- get_preference(try_operation_reuse,true),
   (specfile:animation_minor_mode(eventb) -> print('% OPERATION_REUSE not available for Event-B models'),nl,fail
    % this doesn't work ; see eg. prob_examples/examples/EventBPrologPackages/Advance_WP2/v6_Sep2014/ex_mch.eventb -animate 300 which generates deadlock
    ; true),
   b_is_operation_name(G),
   b_get_operation_normalized_read_write_info(G,_ReadG,WriteG), %print(analyse_reuse(G,_ReadG,WriteG)),nl,
   WriteG \= [], /* otherwise we have a skip operation; we will not recompute state anyway */
   b_get_machine_operation(H,_Results,_Parameters,_Body,OpType,_), % TO DO: check if _Body complicated enough
   b_get_operation_normalized_read_write_info(H,ReadH,WriteH),
   \+ ord_intersect(WriteG,ReadH),
   (OpType = classic
    -> \+ ord_intersect(WriteG,WriteH)  % because for the moment we do not yet detect whether a potentially written variable by H is really written !
   % TO DO: either store in reuse_operation the actual variables written, then we can remove the above line
   % or compute the variables that are potentially written PotWriteH and intersect with those
    ; true
    ),
   (debug:debug_mode(on) -> format('After: ~w reuse: ~w~n',[G,H]) ; true)
   %,format('~nAfter:  ~w (writes ~w)~n reuse: ~w (reads ~w, writes ~w)~n',[G,WriteG,H,ReadH,WriteH]) %%
   .
reuse_operation_effect_calc(_,_) :- get_preference(try_operation_reuse,true),
   findall(G,b_is_operation_name(G),LG), length(LG,NrOps), TotPairs is NrOps*NrOps,
   findall((G,H),reuse_operation_effect(G,H),Pairs), length(Pairs,NrReuse),
   formatsilent('OPERATION_REUSE Statistics: ~w reuse pairs (for ~w operations, i.e. ~w pairs)~n',[NrReuse,NrOps,TotPairs]),
   fail.

:- public candidate_operation_for_reuse_calc/1.
% True if for that operation we may be able sometimes to reuse previously calculated effects
candidate_operation_for_reuse_calc(Op) :-  reuse_operation_effect(_,Op). % no need to avoid multiple solutions; auto_precompile will remove them

:- public print_reuse/0.
print_reuse :-
  findall((G,H),reuse_operation_effect(G,H),L),
  print_reuse_info(L).

print_reuse_info([]).
print_reuse_info([(G,H)|T]) :- print('When applying '),print(H),print(' after '),
   print(G),print(' recalculate only '),
   b_get_operation_normalized_read_write_info(H,_ReadH,WriteH),
   print(WriteH),nl, print_reuse_info(T).

:- volatile b_get_operation_non_det_modifies/2.
:- dynamic b_get_operation_non_det_modifies/2.
:- public b_get_operation_non_det_modifies_calc/2.
:- use_module(b_read_write_info,[get_nondet_modified_vars/3]).
% determine which variables are non-deterministically written by an operation
b_get_operation_non_det_modifies_calc(OpName,NonDetModifies) :-
   b_get_machine_operation(OpName,Results,Parameters,Body,_OpType,_),
   append(Results,Parameters,TLocalVars),
   get_texpr_ids(TLocalVars,LocalVars),
   get_nondet_modified_vars(Body,LocalVars,NonDetModifies),
   debug_println(9,non_det_modifies(OpName,NonDetModifies)).
b_get_operation_non_det_modifies_calc('$initialise_machine',NonDetModifies) :-
   b_get_initialisation_from_machine(Body,_),
   get_nondet_modified_vars(Body,[],NonDetModifies),debug_println(19,init_not_det_modifies(NonDetModifies)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


consult_without_redefine_warning(File) :-
    prolog_flag(redefine_warnings, Old, off),
    (% load in dynamic mode: so that we can retract later
     my_consult(File)
      -> OK=true ; OK=false),
    prolog_flag(redefine_warnings, _, Old),
    (OK=true -> true ;
      add_error_and_fail(bmachine,'Consulting of File failed: ',File)).

my_consult(File) :- on_exception(error(existence_error(_,_),_),
                      load_files(File,[compilation_mode(assert_all)]),
                    add_error_fail(my_consult,'File does not exist: ',File)).

:- volatile  b_get_all_used_filenames/1, b_get_main_filename/1, b_get_main_filenumber/1.
:- dynamic b_get_all_used_filenames/1. % can be used, e.g., for second argument of pos info field
:- dynamic b_get_main_filename/1. % can be used in above list; there is also specfile:currently_opened_file
:- dynamic b_get_main_filenumber/1. % the number of the main_file in the b_get_all_used_filenames list (useful for pos information)
b_get_all_used_filenames([]).

b_load_machine_from_file(Filename) :- b_load_machine_from_file(Filename,[]).
b_load_machine_from_file(Filename,Options) :-
    (debug_mode(on)
     -> get_parser_version(Version),
        format('Java Parser Version: ~w~n',[Version]) % used to be called unconditionally to be sure to catch error if java or BParser.jar not available; but now we do not want the Java startup time unless really necessary
     ; true),
    b_reset_filenames_and_numbers,
    debug_stats(loading(Filename)),
    catch(load_b_machine_as_term(Filename,Machine,Options),
          parse_errors(Errors),
          (add_all_perrors_in_context_of_used_filenames(Errors,parse_error),fail)),
    b_load_machine_from_term(Filename,Machine).


% add parse errors in context of current machine and its filenames
add_all_perrors_in_context_of_used_filenames(Errors,TypeOfError) :- Errors\=[],
    b_get_all_used_filenames(Filenames), Filenames \= [],
    !,
    add_all_perrors(Errors,Filenames,TypeOfError).
add_all_perrors_in_context_of_used_filenames(Errors,TypeOfError) :-
    (Errors=[] -> true ; debug_println(19,'Unable to obtain filenames for parse errors, probably preparse_error')),
    add_all_perrors(Errors,[],TypeOfError).


% a way to load .prob files directly
b_load_machine_probfile(Filename) :- b_reset_filenames_and_numbers,
    load_b_machine_probfile_as_term(Filename,Machine),
    b_load_machine_from_term(Filename,Machine).

% a way to load form a list of Facts which would be found in the .prob file
b_load_machine_from_list_of_facts(Filename,ListOfFacts) :- % Filename only used below for find_main_file_name and later to find out about position infos if they are in the main file or not
    load_b_machine_list_of_facts_as_term(ListOfFacts,Machine),
    b_load_machine_from_term(Filename,Machine).


b_reset_filenames_and_numbers :-
    set_all_used_filenames([]),
    retractall(b_get_main_filename(_)),
    retractall(b_get_main_filenumber(_)),
    retractall(b_filenumber(_,_,_,_)).

set_all_used_filenames(AllFilenames) :-
    retractall(b_get_all_used_filenames(_)),
    assert(b_get_all_used_filenames(AllFilenames)).


b_load_machine_from_term(Filename,complete_machine(MainName,Machines,Filenames)) :-
    b_reset_filenames_and_numbers, % call in case we call this predicate directly from outside
    !,
    assert(b_get_main_filename(Filename)),
    set_all_used_filenames(Filenames),
    store_filenumbers(Filenames),
    (find_main_file_name(Nr,Filenames,Filename) -> assert(b_get_main_filenumber(Nr))
     ; ajoin(['Main filename ', Filename, ' does not appear in filenames list: '],Msg),
       add_warning(bmachine,Msg,Filenames)
    ),
    %print(find_main_file_name(Nr,Filenames,Filename)),nl,
    check_valid_machine_name(MainName),
    debug_stats(analyse_hierarchy(MainName,Filename)),
    (rewrite_complete_machine(MainName,Machines,Machines1) -> true ; Machines1 = Machines),
    (analyse_hierarchy(MainName,Machines1)
      -> true ; add_error(b_load_machine,'Analyse Hierarchy failed'),fail),
    debug_stats(type_check(MainName,Filename)),
    b_set_machine(MainName,Machines1,Errors), % Typechecking
    debug_stats(checked_machine(MainName,Filename)),
    add_all_perrors(Errors,Filenames,type_error), % TO DO: use type_check as source as this may also generate warnings
    no_real_perror_occurred(Errors).
b_load_machine_from_term(F,X) :- add_internal_error('Illegal machine: ',b_load_machine_from_term(F,X)),
    fail.

rewrite_complete_machine(MainName,[DefMachine],[NewMachine]) :-
    DefMachine=definition_file(Pos,Defs),
    % Generate virtual abstract machine when loading a .def DEFINITION file
    debug_println(19,creating_virtual_abstract_machine_for_definition_file),
    Header = machine_header(Pos,MainName,[]),
    NewMachine = abstract_machine(Pos,machine(Pos),Header,[Defs]).

check_valid_machine_name(Name) :- atom(Name),atom_codes(Name,Codes),
    invalid_chars(Invalid),
    (member(C,Codes),ord_member(C,Invalid)
      -> ajoin(['Machine name ', Name, ' contains illegal character: '],Msg),
         atom_codes(IChar,[C]),
         add_warning(bmachine,Msg,IChar)
      ;  true),
    !.
check_valid_machine_name(Name) :- add_internal_error('Illegal machine name:',check_valid_machine_name(Name)).

invalid_chars(".").

find_main_file_name(Nr,_Filenames,File) :- safe_absolute_file_name(File,AF),
   tools:get_modulename_filename(AF,TAF), tools:get_filename_extension(AF,ExtF),
   (ExtF = prob -> member(EXT,[mch,ref,imp,tla,rmch]) % any machine extension is ok for .prob file
     ; EXT =ExtF),
   b_filenumber(TAF,EXT,Nr,_).
   %find_main_file_name_aux(Nr,Filenames,TAF,EXT).
%find_main_file_name_aux(1,[H|_],TAF,ExtF) :- absolute_file_name(H,AF),tools:get_modulename_filename(AF,TAF),
%   tools:get_filename_extension(AF,ExtF).
%find_main_file_name_aux(Nr,[_|T],TAF,ExtF) :- find_main_file_name_aux(N1,T,TAF,ExtF), Nr is N1+1.


% remember as facts all B filenames, their extension and their number
:- volatile b_filenumber/4.
:- dynamic b_filenumber/4.
store_filenumbers(Filenames) :- retractall(b_filenumber(_,_,_,_)),
  store_filenumbers(Filenames,1).
store_filenumbers([],_).
store_filenumbers([H|T],Nr) :-
   safe_absolute_file_name(H,AF),tools:get_modulename_filename(AF,ModuleName),
   tools:get_filename_extension(AF,ExtF),
   assert(b_filenumber(ModuleName,ExtF,Nr,AF)),
   debug_println(9,b_filenumber(ModuleName,ExtF,Nr,AF)),
   N1 is Nr+1,
   store_filenumbers(T,N1).

:- use_module(tools,[safe_absolute_file_name/3, get_parent_directory/2]).
% open a file relateive to the main B file
b_absolute_file_name_relative_to_main_machine(File,AbsFileName) :-
   b_get_main_filename(MainFileName),
   get_parent_directory(MainFileName,Directory),
   % in Jupyter: MainFileName does not exist; SICStus then treats it as a directory
   safe_absolute_file_name(File,AbsFileName,[relative_to(Directory)]).

% call to clear all data stored for machine
b_machine_clear :-
   b_machine_reset,
   b_reset_filenames_and_numbers.

:- use_module(btypechecker,[reset_typechecker/0]).
:- use_module(b_global_sets,[b_clear_global_set_type_definitions/0]).
:- use_module(bmachine_construction,[reset_bmachine_construction/0]).
b_machine_reset :- % print(b_machine_reset),nl,
    retractall(machine(_,_)),
    retractall(discharged_theorem(_,_,_)),
    retractall(discharged_guard_strengthening(_,_,_,_)),
    retractall(discharged(_,_,_)),
    retractall(wd_po(_,_,_)),
    retractall(also_discharged(_,_,_)),
    reset_temp_predicate,
    retractall(b_machine_additional_property(_)),
    reset_typechecker,
    b_clear_global_set_type_definitions,
    reset_bmachine_construction,
    startup_precompiled. % change so that error messages generated if we call them

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(clear_specification,b_machine_clear,
                           'Reset bmachine facts.').

% :- use_module(library(random)).
:- use_module(preferences,[get_preference/2]).




% precompiled
:- volatile b_machine_name/1.
:- dynamic b_machine_name/1.
:- public b_machine_name_calc/1.
b_machine_name_calc(Name) :-
    get_full_b_machine(Name,_).

b_get_animated_sections(Sections) :-
    get_section_from_current_machine(meta,MetaInfos),
    ( member(animated_sections(Sections), MetaInfos) -> true
    ; otherwise -> add_error(bmachine,'No animated sections available'), fail).

:- volatile b_get_named_machine_set/3.
:- dynamic b_get_named_machine_set/3.
:- public b_get_named_machine_set_calc/3.
b_get_named_machine_set(Set,Members) :- b_get_named_machine_set(Set,_,Members).
% precompiled: all enumerated sets with their elements (all in atomic form)
b_get_named_machine_set_calc(Set,TExpr,Members) :-
    get_section_from_current_machine(enumerated_sets,Sets),
    get_section_from_current_machine(enumerated_elements,Elems),
    get_texpr_id(TExpr,Set),
    get_texpr_type(TExpr,set(Type)),
    get_texpr_type(TElem,Type),
    member(TExpr,Sets),
    findall(E, (member(TElem,Elems), get_texpr_id(TElem,E)), Members).

:- volatile b_get_machine_set/2.
:- dynamic b_get_machine_set/2.
:- public b_get_machine_set_calc/2.
b_get_machine_set(Set) :- b_get_machine_set(Set,_).
% precompiled: all deferred and enumerated sets (in atomic form)
b_get_machine_set_calc(Set,TExpr) :-
    get_section_from_current_machine(deferred_sets,Sets),
    get_texpr_id(TExpr,Set),
    member(TExpr,Sets).
b_get_machine_set_calc(Set,TExpr) :-
    b_get_named_machine_set(Set,TExpr,_).

:- volatile b_get_machine_constants/1, b_get_machine_all_constants/1, b_get_machine_variables/1, b_is_constant/2, b_is_variable/2, b_get_constant_variable_description/2.
:- dynamic b_get_machine_constants/1, b_get_machine_all_constants/1, b_get_machine_variables/1, b_is_constant/2, b_is_variable/2, b_get_constant_variable_description/2.
b_machine_has_constants :- b_get_machine_constants([_|_]).
:- public b_get_machine_all_constants_calc/1.
% precompiled: all constants (abstract and concrete) in typed form
b_get_machine_all_constants_calc(Csts) :-
    findall(C,get_multi_sections([abstract_constants,concrete_constants],C),Csts).
:- public b_get_machine_constants_calc/1.
% precompiled: all constants (abstract and concrete) in typed form; except for constant_represented inside global set
b_get_machine_constants_calc(Csts) :-
    findall(C,
      (get_multi_sections([abstract_constants,concrete_constants],C),
       get_texpr_id(C,CID),
       \+ b_get_constant_represented_inside_global_set(CID,_)), Csts).

b_machine_has_variables :- b_get_machine_variables([_|_]).
:- public b_get_machine_variables_calc/1.
% precompiled: all variables (abstract and concrete) in typed form
b_get_machine_variables_calc(Vars) :-
    findall(V,get_multi_sections([abstract_variables,concrete_variables],V),Vars).


b_is_constant(C) :- b_is_constant(C,_).
:- public b_is_constant_calc/2.
% precompiled: all constants (abstract and concrete) in atomic form
b_is_constant_calc(Cst,Type) :-
    b_get_machine_constants(Constants),
    get_texpr_id(Term,Cst),
    get_texpr_type(Term,Type),
    member(Term,Constants).
b_is_variable(V) :- b_is_variable(V,_).
:- public b_is_variable_calc/2.
% precompiled: all variables (abstract and concrete) in atomic form
b_is_variable_calc(Var,Type) :-
    b_get_machine_variables(Vars),
    get_texpr_id(Term,Var),
    get_texpr_type(Term,Type),
    member(Term,Vars).
% look up elements in several sections of the machine
get_multi_sections(Sections,Elem) :-
    member(Section,Sections),
    get_section_from_current_machine(Section,List),
    member(Elem,List).

:- public b_get_constant_variable_description_calc/2.
% get value of description pragmas @desc for variables and constants:
b_get_constant_variable_description_calc(ID,Desc) :-
    (b_get_machine_constants(L) ; b_get_machine_variables(L)),
    get_texpr_id(Term,ID),
    member(Term,L),
    get_texpr_description(Term,Desc).


:- volatile b_get_invariant_from_machine/1,
    b_get_linking_invariant_from_machine/1,
    b_get_properties_from_machine/1,
    b_get_static_assertions_from_machine/1,
    b_extract_values_clause_assignment/3,
    b_get_unproven_static_assertions_from_machine/1,
    b_get_dynamic_assertions_from_machine/1,
    b_get_unproven_dynamic_assertions_from_machine/1,
    b_get_assertions_from_main_machine/2,
    b_get_initialisation_from_machine/2,
    b_get_machine_operation/6,b_top_level_operation/1,
    b_get_machine_operation_for_animation/6,
    b_is_operation_name/1,
    b_get_promoted_machine_operations/1.
:- dynamic b_get_invariant_from_machine/1,
    b_get_linking_invariant_from_machine/1,
    b_get_properties_from_machine/1,
    b_get_static_assertions_from_machine/1,
    b_extract_values_clause_assignment/3,
    b_get_unproven_static_assertions_from_machine/1,
    b_get_dynamic_assertions_from_machine/1,
    b_get_unproven_dynamic_assertions_from_machine/1,
    b_get_assertions_from_main_machine/2,
    b_get_initialisation_from_machine/2,
    b_get_machine_operation/6,b_top_level_operation/1,
    b_get_machine_operation_for_animation/6,
    b_is_operation_name/1,
    b_get_promoted_machine_operations/1.

:- public b_get_invariant_from_machine_calc/1.
% precompiled: typed invariant (a predicate)
b_get_invariant_from_machine_calc(Inv) :-
    get_section_from_current_machine(invariant,Inv).

:- public b_get_linking_invariant_from_machine_calc/1.
% precompiled: typed linking invariant (a predicate)
b_get_linking_invariant_from_machine_calc(Inv) :-
    get_section_from_current_machine(linking_invariant,Inv).

:- public b_get_properties_from_machine_calc/1.
% precompiled: typed properties (a predicate)
b_get_properties_from_machine_calc(SortedProp) :-
    get_section_from_current_machine(properties,Prop),
    (sort_properties(Prop,SortedProp) -> true %(Prop=SortedProp->true ; print(sorted(SortedProp)))
     ;  add_error(sort_properties,'sort_properties failed: ',Prop),
        SortedProp=Prop).

sort_properties(Prop,SortedProp) :-  % print(sort_properties),nl,statistics(runtime, [R1,_]),
   weigh_properties(Prop,1,_,[],SProp),
   % statistics(runtime, [R2,_]), RDiff is R2-R1,print(done_weigh(RDiff,SProp)),nl,
    sort(SProp,SSProp),
   % statistics(runtime, [R3,_]), RDiff2 is R3-R2,print(done_sort(RDiff2)),nl,
   remove_weights(SSProp,PropList),
   conjunct_predicates(PropList,SortedProp).
  % translate:print_bexpr(SortedProp).

remove_weights([],[]).
remove_weights([prop(_W,_Nr,E)|T],[E|RT]) :- remove_weights(T,RT).

weigh_properties(Pred,Nr,ONr,InL,OutL) :-
   decompose_conjunct(Pred,LHS,RHS), % will propagate labels down to conjuncts LHS,RHS
   !,
   weigh_properties(LHS,Nr,N1,InL,L2),
   weigh_properties(RHS,N1,ONr,L2,OutL).
weigh_properties(b(Pred,S,I),Nr,N1,T,[prop(Weight,Nr,b(Pred,S,I))|T]) :-
   N1 is Nr+1,
   (member(contains_wd_condition,I) -> Weight=32000 % do not move properties with WD info
    ; weigh_properties2(Pred,Weight)).
   % translate:print_bexpr(b(Pred,S,I)), print(' WEIGHT: '), print(Weight),nl

weigh_properties2(truth,W) :- !, W=1.
weigh_properties2(falsity,W) :- !, W=1.
weigh_properties2(equal(L,R),W) :- weigh_term(L,W1), !,
  ( weigh_term(R,W2) -> W is W1*W2
  ; W is 3200).
%  ; W is W1*20). %causes issue with ParityFunction test
%weigh_properties2(equal(_,R),W) :- weigh_term(R,W2),  !,  W is W2*20.
weigh_properties2(SOp,W) :- comparison(SOp,L,R),!,
  ( (weigh_term(L,W1), weigh_term(R,W2))
    -> W is 7+W1+W2 % try and give things like x>10 priority over x>y (to overcome CLP(FD) limitations)
    ; W=10).
weigh_properties2(partition(LHS,RHS),W) :- weigh_term(LHS,W1), l_weigh_term(RHS,W2), !, W is W1*W2.
weigh_properties2(not_equal(L,R),W) :- ((id(L),empty(R)) ; (id(R),empty(L))),!, W=1.
weigh_properties2(not_equal(L,R),W) :- weigh_term(L,W1), weigh_term(R,W2),!,
   W is 20+W1*W2.
weigh_properties2(_Rest,W) :- W=32000.

comparison(less(L,R),L,R).
comparison(less_equal(L,R),L,R).
comparison(greater(L,R),L,R).
comparison(greater_equal(L,R),L,R).


id(b(identifier(_),_,_)).
empty(b(empty_set,_,_)).
empty(b(empty_sequence,_,_)).

l_weigh_term([],0).
l_weigh_term([H|T],R) :- l_weigh_term(T,TR), weigh_term(H,TH), R is TH+TR.

weigh_term(b(T,_,_),R) :- (weigh_term2(T,R) -> true ; fail). %R=20).

weigh_term2(value(_),1).
weigh_term2(boolean_true,1).
weigh_term2(boolean_false,1).
weigh_term2(empty_set,1).
weigh_term2(empty_sequence,1).
weigh_term2(integer(_),1).
weigh_term2(min_int,1).
weigh_term2(max_int,1).
weigh_term2(string(_),1).
weigh_term2(real(_),1).
weigh_term2(bool_set,2).
weigh_term2(identifier(_),2).
weigh_term2(set_extension(_),3).
weigh_term2(integer_set(_),3).
weigh_term2(interval(L,R),W) :- weigh_term(L,WL), weigh_term(R,WR), W is WL*WR.



b_machine_has_constants_or_properties :-
  (b_machine_has_constants -> true
   ; (b_get_properties_from_machine(Prop)), \+ is_truth(Prop)).

b_get_machine_operation_parameter_types(OpName,ParTypes) :-
   b_get_machine_operation_typed_parameters(OpName,Parameters),
   maplist(get_texpr_type,Parameters,ParTypes).

b_get_machine_operation_typed_parameters(Operation,Res) :-
   if(b_get_machine_operation_typed_parameters_aux(Operation,Paras),Res=Paras,
      (add_internal_error('Cannot get parameters: ',b_get_machine_operation_typed_parameters(Operation,Res)),fail)).
% also recognises SETUP_CONSTANTS & INITIALISE_MACHINE
b_get_machine_operation_typed_parameters_aux('$setup_constants',Constants) :-
   b_machine_has_constants_or_properties,
   b_get_machine_constants(Constants).
b_get_machine_operation_typed_parameters_aux('$initialise_machine',Variables) :-
   b_get_machine_variables(Variables).
b_get_machine_operation_typed_parameters_aux(OpName,Parameters) :-
   b_get_machine_operation(OpName,_R,Parameters,_Body,_OType,_OpPos).

b_get_machine_operation_parameter_names('@INITIALISATION',ParaNames) :- % for Tcl/Tk
   b_get_machine_operation_parameter_names('$initialise_machine',ParaNames).
b_get_machine_operation_parameter_names('@PROPERTIES',ParaNames) :- % for Tcl/Tk
   b_get_machine_operation_parameter_names('$setup_constants',ParaNames).
b_get_machine_operation_parameter_names(OpName,ParaNames) :-
   b_get_machine_operation_typed_parameters(OpName,Parameters),
   get_texpr_ids(Parameters,ParaNames).

b_get_machine_operation(Name,Results,Parameters,Body) :-
    b_get_machine_operation(Name,Results,Parameters,Body,_OType,_OpPos).

:- public b_get_assertions_from_main_machine_calc/2.
b_get_assertions_from_main_machine_calc(static,FProp) :-
    ((b_get_static_assertions_from_machine(Prop),
      exclude(is_not_main_assertion,Prop,FProp)) -> true
      ; add_error(bmachine,'b_get_assertions_from_main_machine_calc static fails'), FProp=[]).
b_get_assertions_from_main_machine_calc(dynamic,FProp) :-
    ((b_get_dynamic_assertions_from_machine(Prop),
      exclude(is_not_main_assertion,Prop,FProp)) -> true
      ; add_error(bmachine,'b_get_assertions_from_main_machine_calc dynamic fails'), FProp=[]).
is_not_main_assertion(Assertion) :-
    get_texpr_pos(Assertion,Pos),
    error_manager:position_is_not_in_main_file(Pos).

% get nr of assertions and nr of main assertions for static/dynamic
b_get_assertion_count(StaticOrDynamic,NrAll,NrMain) :-
   b_get_assertions(all,StaticOrDynamic,A), length(A,NrAll),
   b_get_assertions_from_main_machine(StaticOrDynamic,MA),length(MA,NrMain).

b_get_assertions(main,Mode,Ass) :- !,
   b_get_assertions_from_main_machine(Mode,Ass).
b_get_assertions(specific(Label),Mode,LabelAss) :- !,
   b_get_assertions(all,Mode,Ass),
   %maplist(get_texpr_label,Ass,Labels), print(labels(Labels)),nl,
   include(has_label(Label),Ass,LabelAss),
   (LabelAss=[] -> ajoin(['No ',Mode,' assertion has label: '],Msg),
                   add_warning(b_get_assertions,Msg,Label) ; true).
b_get_assertions(_,static,Ass) :- b_get_static_assertions_from_machine(Ass).
b_get_assertions(_,dynamic,Ass) :- b_get_dynamic_assertions_from_machine(Ass).
b_main_machine_has_no_assertions :-
  b_get_assertions_from_main_machine(static,[]),
  b_get_assertions_from_main_machine(dynamic,[]).


%has_label(Label,Expr) :- get_texpr_label(Expr,Label1), format('Labels ~w =? ~w~n',[Label1,Label]), fail.
has_label(Label,Expr) :- get_texpr_label(Expr,Label).

:- public b_get_static_assertions_from_machine_calc/1.
% precompiled: typed assertions (a list of predicates)
b_get_static_assertions_from_machine_calc(Prop) :-
    get_section_from_current_machine(assertions,Prop1),
    exclude(predicate_uses_variables,Prop1,Prop),!.
b_get_static_assertions_from_machine_calc(Prop) :-
    add_error(bmachine,'b_get_static_assertions_from_machine_calc fails'), Prop=[].

:- dynamic b_machine_has_static_assertions/0.
:- public b_machine_has_static_assertions_calc/0.
b_machine_has_static_assertions_calc :-
    b_get_static_assertions_from_machine([_|_]).

:- public b_get_unproven_static_assertions_from_machine_calc/1.
b_get_unproven_static_assertions_from_machine_calc(FAssertions) :-
    b_get_static_assertions_from_machine(Assertions),
    filter_out_proven_assertions(Assertions,FAssertions),!.
b_get_unproven_static_assertions_from_machine_calc(Prop) :-
    add_error(bmachine,'b_get_unproven_static_assertions_from_machine_calc fails'), Prop=[].

filter_out_proven_assertions([],[]).
filter_out_proven_assertions([A1|T],R) :-
     (is_discharged_assertion(A1)
      -> R = TR   %, print('Proven: '), translate:print_bexpr(A1),nl
      ; R = [A1|TR]),
      filter_out_proven_assertions(T,TR).

:- public b_get_dynamic_assertions_from_machine_calc/1.
b_get_dynamic_assertions_from_machine_calc(Prop) :-
    get_section_from_current_machine(assertions,Prop1),
    include(predicate_uses_variables,Prop1,Prop),!.
b_get_dynamic_assertions_from_machine_calc(Prop) :-
    add_error(bmachine,'b_get_dynamic_assertions_from_machine_calc fails'), Prop=[].


:- dynamic b_machine_has_dynamic_assertions/0.
:- public b_machine_has_dynamic_assertions_calc/0.
b_machine_has_dynamic_assertions_calc :-
    b_get_dynamic_assertions_from_machine([_|_]).

b_machine_has_assertions :- (b_machine_has_static_assertions -> true ; b_machine_has_dynamic_assertions).

b_machine_has_unproven_assertions :-
    b_get_unproven_dynamic_assertions_from_machine([_|_]) -> true
    ; b_get_unproven_static_assertions_from_machine([_|_]).


get_all_assertions_from_machine(Assertions) :-
    findall(A,((SD=static;SD=dynamic),get_assertions_from_machine(SD,SS),member(A,SS)),Assertions).

get_assertions_from_machine(Type,Assertions) :-
    ( Type=(dynamic) ->
        (get_preference(use_po,true)
         -> b_get_unproven_dynamic_assertions_from_machine(Assertions)
         ;  b_get_dynamic_assertions_from_machine(Assertions)
        )
    ; Type=(static)  ->
        (get_preference(use_po,true)
          -> b_get_unproven_static_assertions_from_machine(Assertions)
          ;  b_get_static_assertions_from_machine(Assertions)
        )
    ),
    Assertions = [_|_].

:- public b_get_unproven_dynamic_assertions_from_machine_calc/1.
b_get_unproven_dynamic_assertions_from_machine_calc(FAssertions) :-
    b_get_dynamic_assertions_from_machine(Assertions),
    filter_out_proven_assertions(Assertions,FAssertions),!.
b_get_unproven_dynamic_assertions_from_machine_calc(Prop) :-
    add_error(bmachine,'b_get_unproven_dynamic_assertions_from_machine_calc fails'), Prop=[].

predicate_uses_variables(Pred) :-
    b_get_machine_variables(Vars),
    get_texpr_ids(Vars,Ids),
    list_to_ord_set(Ids,OrdVars),
    find_identifier_uses(Pred,[],UsedVars),
    list_to_ord_set(UsedVars,OrdUsedVars),
    ord_intersect(OrdVars,OrdUsedVars).

:- public b_get_initialisation_from_machine_calc/2.
% precompiled: typed initialisation (a substitution)
b_get_initialisation_from_machine_calc(Init,OType) :-
    get_section_from_current_machine(initialisation,Init),
    ( get_texpr_expr(Init,rlevent(_Name,_Section,_Status,Params,_Guard,_Theorems,_Actions,_VWitnesses,_PWitnesses,_Unmod,_AbstractEvents)) ->
        get_operation_type(Init,'INITIALISATION',Params,OType)
    ; otherwise ->
        OType = classic).
% precompiled: operations, Name is in atomic form, Results and Parameters are lists
%   of typed identifiers, Body is a typed substitution

:- volatile b_machine_operation_names_in_reverse_order/1.
:- dynamic b_machine_operation_names_in_reverse_order/1.
% a list of operation names in reverse order; this way we can randomize the
% generation of the successor states
precompile_operation_names_in_reverse_order :-
   retractall(b_machine_operation_names_in_reverse_order(_)),
   b_get_machine_operation(Name,_Results,_Parameters,_Body,_OType,_OpPos),
   asserta(b_machine_operation_names_in_reverse_order(Name)),
   fail.
precompile_operation_names_in_reverse_order.

:- public b_get_machine_operation_calc/6.
b_get_machine_operation_calc(Name,Results,Parameters,OpBody,OType,OpPos) :-
    is_ground(Name,NameGround),
    get_section_from_current_machine(operation_bodies,Ops),
    get_texpr_expr(Op,operation(TId,Results,Parameters,Body)),
    get_texpr_id(TId,op(Name)),
    member(Op,Ops),
    % remove choicepoints when name was ground
    (NameGround==ground -> !; true),
    get_operation_type(Op,Name,Parameters,OType),
    (get_texpr_pos(TId,OpPos) -> debug_println(9,op_pos(Name,OpPos)) ; OpPos = unknown),
    ( get_preference(use_common_subexpression_also_for_substitutions,true),
      get_preference(use_common_subexpression_elimination,true),
      \+ specfile:animation_minor_mode(eventb) % CSE_SUBST not yet available for Event-B
       -> debug_println(19,applying_cse_to(Name)),
          b_expression_sharing:cse_optimize_substitution(Body,OpBody)
       ; OpBody = Body).
get_operation_type([],_Name,_Parameters,classic).
get_operation_type(Op,Name,Parameters,OType) :-
    get_texpr_info(Op,Info),
    ( member(eventb(_),Info) ->
        OType=eventb_operation(SortedChangeSet,Ps,OpPattern),
        member(modifies(ChangeSet),Info),!,
        sort(ChangeSet,SortedChangeSet), % maybe not necessary ?
        same_length(Parameters,Ps),
        OpPattern =.. [Name|Ps]
    ; otherwise ->
        OType=classic).

:- volatile get_operation_info/2.
:- dynamic get_operation_info/2.
:- public get_operation_info_calc/2.
% precompiled: operation infos
get_operation_info_calc(OpName,Info) :-
    get_section_from_current_machine(operation_bodies,Ops),
    create_texpr(operation(TId,_Results,_Parameters,_Body),_,Info,TOp),
    get_texpr_id(TId,op(OpName)),
    member(TOp,Ops).

:- public b_top_level_operation_calc/1.
% precompiled: operation names (in atomic form) of top-level operations (promoted operations)
b_top_level_operation_calc(Name) :-
    is_ground(Name,NameGround),
    get_section_from_current_machine(promoted,Promoted),
    get_texpr_id(TId,op(Name)),
    member(TId,Promoted),
    (NameGround==ground -> !; true).
:- public b_get_promoted_machine_operations_calc/1.
b_get_promoted_machine_operations_calc(Ops) :-
    get_section_from_current_machine(promoted,Ops).

:- volatile b_top_level_feasible_operation/1.
:- dynamic b_top_level_feasible_operation/1.
:- public b_top_level_feasible_operation_calc/1.
% filter out operations which are commented out statically by using 1=2 or similar in guard
b_top_level_feasible_operation_calc(Name) :- b_top_level_operation(Name),
   b_get_machine_operation(Name,_,_,Body,_,_Pos),
   \+ infeasible_tsubst(Body).
% TO DO: investigate how to link up with cbc feasibility analysis

infeasible_tsubst(b(S,_,_)) :- infeasible_subst(S).
infeasible_subst(precondition(G,_)) :- is_falsity(G).
infeasible_subst(select([b(select_when(G,_),_,_)])) :- is_falsity(G).
infeasible_subst(any(_,G,_)) :- is_falsity(G).

% this succeeds not just for top-level operations
%b_is_operation_name(Name) :- b_get_machine_operation(Name,_,_,_,_,_). is slower
:- public b_is_operation_name_calc/1.
b_is_operation_name_calc(Name) :-
   b_get_machine_operation(Name,_,_,_,_,_).

b_get_operation_pos(Name,Pos) :-
   b_get_machine_operation(Name,_,_,_,_,Pos).

b_get_operation_variant(Name,ConvOrAnt,Variant) :-
    b_get_machine_operation(Name,_,_,TBody),
    get_texpr_expr(TBody,Body),
    Body = rlevent(Name,_Section,Status,
                       _Params,_Guard,_Theorems,_Actions,_VWitnesses,_PWitnesses,_Unmod,_AbstractEvents),
    % TO DO: look maybe higher up in refinement hierarchy
    get_texpr_expr(Status,S),
    functor(S,ConvOrAnt,1),
    (ConvOrAnt = convergent ; ConvOrAnt = anticipated),
    arg(1,S,Variant).

b_machine_has_operations :-
    (b_is_operation_name(_) -> true).
% -----------------------------


% the names of definitions which have a given Prefix,
% Type is expression or predicate, Name is an atomic name
b_definition_prefixed(Type,Prefix,Suffix,Name,Pos) :-
    safe_atom_chars(Prefix,PrefixChars,b_definition_prefixed1),
    append(PrefixChars,SuffixChars,FullChars),
    (ground(Suffix) -> safe_atom_chars(Suffix,SuffixChars,b_definition_prefixed2),
                       safe_atom_chars(Name,FullChars,b_definition_prefixed3)
      ; true),
    b_get_definition_with_pos(Name,Type,Pos,_,_,_),
    safe_atom_chars(Name,FullChars,b_definition_prefixed4),
    safe_atom_chars(Suffix,SuffixChars,b_definition_prefixed5).

:- use_module(bmachine_construction,[type_open_formula/8]).
type_check_definitions :-
    temporary_set_preference(allow_untyped_identifiers,true_with_string_type,CHNG), % we may not have all identifiers visible
    get_error_context(Context),
    call_cleanup(type_check_definitions2,restore_error_context(Context)),
    reset_temporary_preference(allow_untyped_identifiers,CHNG).
type_check_definitions2 :-
    get_preference(type_check_definitions,true),
    full_b_machine(Machine),
    b_get_definition_with_pos(Name,Type,DefPos,Args,RawExpr,_Deps),
    % TO DO: do not re-analyse definitions already analysed for ANIMATION_FUNCTION,...
    (Type = predicate ; Type=expression), % can we also allow substitutions ?
    debug_format(9,'Type checking DEFINITION ~w~n',[Name]),
    set_error_context(checking_context('Type checking DEFINITION: ',Name)),
    Scope = [variables], % prob_ids(visible),
    %type_with_errors(RawExpr,Scope,_,_TExpr),
    AllowNewIds=true,
    % TO DO: improve solution below, e.g., delete Args from visible variables or rename Args to fresh ids!
    (Args=[] -> RawExpr2=RawExpr
     ; Type=predicate -> RawExpr2 = exists(none,Args,conjunct(none,truth(none),RawExpr)) % wrap Args into quantifier, avoid clash with variables,... ; conjunct used to avoid warnings about body of exists
     ; RawExpr2=exists(none,Args,equal(none,RawExpr,RawExpr)) % TODO: better solution to avoid duplicating RawExpr
    ),
    (type_open_formula(RawExpr2,Scope,AllowNewIds,Machine,FType,Identifiers,_TPred,Errors)
      ->  Errors = [_|_],
          format('Error for DEFINITION ~w of type ~w over identifiers ~w~n',[Name,FType,Identifiers]),
          add_all_perrors_in_context_of_used_filenames(Errors,definition_type_error),
          fail
          % TO DO: check for local clashes?
      ;   add_error(bmachine,'Type checking DEFINITION failed: ',Name,DefPos)
    ),
    fail.
type_check_definitions2 :- debug_format(9,'Finished type checking DEFINITIONs~n',[]).

b_get_definition(Name,DefType,Args,RawExpr,Deps) :-
   b_get_definition_with_pos(Name,DefType,_DefPos,Args,RawExpr,Deps).

b_get_definition_with_pos(Name,DefType,DefPos,Args,RawExpr,Deps) :-
    get_section_from_current_machine(definitions,Defs),
    member(definition_decl(Name,DefType,DefPos,Args,RawExpr,Deps),Defs).
% the next calls get typed definitions without parameters:
% (To treat definitions with parameters one would have to add the parameters to the type environment, see type_check_definitions2)
b_get_typed_predicate_definition(Name,Scope,ResTExpr) :-
    b_get_definition(Name,predicate,[],RawExpr,_Deps),
    type_with_errors_in_context(RawExpr,Scope,_,TExpr,machine_context),
    ResTExpr=TExpr. % Do unification after to not prevent generation of type errors.
b_get_typed_expression_definition(Name,Scope,ResTExpr) :-
    b_get_definition(Name,expression,[],RawExpr,_Deps),
    type_with_errors_in_context(RawExpr,Scope,_,TExpr,machine_context),
    ResTExpr=TExpr. % Do unification after to not prevent generation of type errors, e.g., test 1691
b_get_true_expression_definition(Name) :-
    b_get_typed_expression_definition(Name,[variables],PF),
    PF = b(boolean_true,boolean,_).
%b_get_typed_subst_definition(Name,Scope,Parameters,TExpr) :-
%    temporary_set_preference(allow_local_operation_calls,true,CHNG),
%    b_get_definition(Name,substitution,Parameters,RawExpr,_Deps),
%    type_with_errors(RawExpr,[operation_bodies|Scope],_,TExpr), % need to remove variables to avoid clash warnings
%    reset_temporary_preference(allow_local_operation_calls,CHNG).
b_get_typed_definition(Name,Scope,TExpr) :-
    b_get_definition(Name,_DefType,[],RawExpr,_Deps),
    type_with_errors_in_context(RawExpr,Scope,_,TExpr,machine_context).

type_with_errors(RawExpr,Scope,Type,TExpr) :-
    type_with_errors_in_context(RawExpr,Scope,Type,TExpr,not_in_machine_context).
type_with_errors_in_context(RawExpr,Scope,Type,TExpr,Context) :-  % print(scope(Scope)),nl,
    b_type_expression(RawExpr,Scope,Type,TExpr,Errors),!,
    (Context=not_in_machine_context
       -> add_all_perrors(Errors,[],type_expression_error) % no file name list available
       ;  add_all_perrors_in_context_of_used_filenames(Errors,type_expression_error)
    ),
    no_real_perror_occurred(Errors).

b_type_open_predicate_with_errors(OptionalQuantifier,RawPred,Scope,TPred) :-
    b_type_open_predicate(OptionalQuantifier,RawPred,Scope,TPred,Errors),
    add_all_perrors(Errors,[],type_expression_error),
    no_real_perror_occurred(Errors).
b_type_open_predicate(OptionalQuantifier,RawPred,Scope,TPred,Errors) :-
    b_type_open_predicate_for_full_b_machine(_Machine,OptionalQuantifier,RawPred,Scope,TPred,Errors).
b_type_open_predicate_for_full_b_machine(Machine,OptionalQuantifier,RawPred,Scope,TPred,Errors) :-
    (var(Machine) -> full_b_machine(Machine) ; true),
    replace_prob_set_elements_in_scope(Scope,Scope1),
    type_open_predicate_with_quantifier(OptionalQuantifier,RawPred,Scope1,Machine,TPred,Errors),!.
b_type_open_exists_predicate(Raw,Typed,Errors) :-
    b_type_open_predicate(open(exists),Raw,[prob_ids(visible),variables],Typed,Errors).

b_type_expression(RawExpr,Scope,Type,TExpr,Errors) :-
    b_type_expression_for_full_b_machine(_,RawExpr,Scope,Type,TExpr,Errors).

b_type_expression_for_full_b_machine(M,RawExpr,Scope,Type,TExpr,Errors) :-
    b_type_expressions_l(M,[RawExpr],Scope,[Type],[TExpr],Errors).

b_type_expressions_l(Machine,RawExprs,Scope,Types,TExprs,Errors) :-
   (var(Machine) -> full_b_machine(Machine) ; true),
   % in case you type check multiple expressions for a complex machine it is better to extract the machine once
    replace_prob_set_elements_in_scope(Scope,Scope1),
    % we need this, for example for tcltk_check_state_sequence_from_file, when deferred set elements written by TLC are in the file
    type_in_machine_l(RawExprs,Scope1,Machine,Types,TExprs,Errors),!.

replace_prob_set_elements_in_scope(Scope,Res) :- var(Scope),!,
    add_internal_error('Scope is a variable (use, e.g., [constants,variables,operation_bodies,...]):',replace_prob_set_elements_in_scope(Scope,Res)),
    Res=Scope.
replace_prob_set_elements_in_scope(Scope,NewScope) :-
    ( select(prob_ids(AllOrVisible),Scope,identifier(Ids),NewScope) ->   % replaces prob by deferred set element identifiers
     % for example if we have SETS PID then we would have PID1,... in Ids unless PID1 is known to be another identifier
        b_get_prob_deferred_set_elements(DefIDs,AllOrVisible),
        exclude(known_identifier,DefIDs,Ids) % exclude those ids that would clash
        % TO DO: add option to translate_identifiers using translate:keyword_to_id_cache
    ; otherwise ->
        Scope = NewScope).


known_identifier(X) :- get_texpr_id(X,ID),
     (bmachine:b_is_constant(ID,_) ;
      bmachine:b_is_variable(ID,_)).

% ------------------------

% a generic predicate to find all kinds of identifiers

:- use_module(b_global_sets,[b_global_set/1, is_b_global_constant/3]).
:- mode get_machine_identifiers(+Category, -ListOfIdentifiers).
get_machine_identifiers(machines,MN) :-
      findall(FID, (b_filenumber(FID,Type,_,_),Type \= def),MN).
get_machine_identifiers(definition_files,DFN) :-
      findall(FID, b_filenumber(FID,def,_,_),DFN).
get_machine_identifiers(definitions,DN) :-
      findall(DefID,b_get_definition(DefID,_,_,_,_),DN).
get_machine_identifiers(sets,Sets) :-
      findall(GS,b_global_set(GS),Sets).
get_machine_identifiers(set_constants,Csts) :-
      findall(Cst,is_b_global_constant(_GS,_,Cst),Csts).
get_machine_identifiers(constants,CN) :-
       b_get_machine_constants(Constants),
       get_texpr_ids(Constants,CN).
get_machine_identifiers(variables,VN) :-
       b_get_machine_variables(Variables),
       get_texpr_ids(Variables,VN).
get_machine_identifiers(operations,Ops) :-
       findall(Op,b_top_level_operation(Op),Ops).

:- use_module(b_global_sets,[b_global_deferred_set/1,all_elements_of_type/2]).
:- use_module(input_syntax_tree,[get_raw_position_info/2]).
% try and get source code for an identifier
source_code_for_identifier(ID,constant,Type,OriginStr,OriginTerm,Source) :- b_is_constant(ID),!,
      get_constant_type_and_origin(ID,Type,OriginStr,OriginTerm),
      findall(Def,get_constant_definition(ID,Def),AllDefs),
      (AllDefs=[_|_] -> conjunct_predicates(AllDefs,Source)
        ; % try and find other type of predicates defining ID
          findall(Pred,get_constant_predicate(ID,Pred),AllPreds),
          conjunct_predicates(AllPreds,Source)).
source_code_for_identifier(ID,variable,Type,OriginStr,OriginTerm,Source) :- b_is_variable(ID),!,
      get_variable_type_and_origin(ID,Type,OriginStr,OriginTerm),
      Source = b(truth,pred,[]). % TO DO: improve
source_code_for_identifier(ID,Kind,Type,OriginStr,OriginTerm,Source) :- b_get_machine_set(ID,TID),!,
      (b_global_deferred_set(ID) -> Kind = deferred_set ; Kind = enumerated_set),
      Type = set(global(ID)),
      all_elements_of_type(ID,All),
      translate:translate_span_with_filename(TID,OriginStr), OriginTerm=TID,
      Source = b(equal(TID,b(value(All),Type,[])),pred,[]).
source_code_for_identifier(ID,Kind,Type,OriginStr,OriginTerm,Source) :-
      global_set_element(ID,GID),!,
      (b_global_deferred_set(GID)
       -> Kind = deferred_set_element ; Kind = enumerated_set_element),
      b_get_machine_set(GID,TGID),
      Type = global(ID),
      all_elements_of_type(GID,All),
      translate:translate_span_with_filename(TGID,OriginStr), OriginTerm=TGID,
      Source = b(equal(TGID,b(value(All),set(Type),[])),pred,[]).
source_code_for_identifier(ID,definition,Type,OriginStr,OriginTerm,Source) :-
      b_get_definition(ID,DefType,Args,RawExpr,_Deps),!,
      Type = DefType,
      % Definitions are only present in RAW form
      get_raw_position_info(RawExpr,PosInfo),
      translate:translate_span_with_filename(PosInfo,OriginStr), % the position info does not always seem to be valid (when coming from DEFINITION FILE) !??! TO DO fix
      OriginTerm=PosInfo,
      source_code_for_definition(DefType,ID,Args,RawExpr,Source).
source_code_for_identifier(ID,file,Type,OriginStr,OriginTerm,b(string(Source),string,[])) :-
      b_filenumber(ID,Kind,_,Filename),
      extension_kind(Kind,Type),
      !,
      ajoin([Type,' ',ID],Source),
      OriginTerm =src_position_with_filename(1,0,0,Filename),
      OriginStr = Filename.
source_code_for_identifier(ID,Kind,subst,OriginStr,OriginTerm,Source) :-
      b_find_operation(ID,Res,TParas,Body,OType),
      (OType=classic -> Kind = operation ; Kind=event),
      translate:translate_span_with_filename(Body,OriginStr),Body=OriginTerm,
      get_texpr_info(Body,Info),
      Source=b(operation(b(identifier(ID),subst,[]),Res,TParas,Body),subst,Info).
source_code_for_identifier(ID,assertion,pred,OriginStr,OriginTerm,Assertion) :-
      b_get_assertions(all,_,Ass), member(Assertion,Ass),
      get_texpr_label(Assertion,ID), OriginTerm = Assertion,
      translate:translate_span_with_filename(Assertion,OriginStr).
source_code_for_identifier(ID,invariant,pred,OriginStr,OriginTerm,Invariant) :-
      b_get_invariant_from_machine(C), conjunction_to_list(C,L), member(Invariant,L),
      get_texpr_label(Invariant,ID), OriginTerm = Invariant,
      translate:translate_span_with_filename(Invariant,OriginStr).
% TO DO: look for invariant labels as ids ?

:- use_module(tools,[ split_last/4 ]).
b_find_operation(ID,Res,TParas,Body,OType) :-
   b_get_machine_operation(ID,Res,TParas,Body,OType,_Pos).
b_find_operation(ID,Res,TParas,Body,OType) :-
   b_get_machine_operation(ID2,Res,TParas,Body,OType,_Pos),
   split_last(ID2, '.',  _, ID).

extension_kind(mch,'MACHINE').
extension_kind(ref,'REFINEMENT').
extension_kind(def,'DEFINITIONS FILE').
extension_kind(imp,'IMPLEMENTATION').

source_code_for_definition(predicate,ID,_,RawExpr,Source) :-
     b_type_open_exists_predicate(RawExpr,Typed,Errors),
     (no_real_perror_occurred(Errors) -> true ; debug_println(9,errs(ID,Errors))),
     Source=Typed.
source_code_for_definition(expression,ID,Args,RawExpr,Source) :-
     % TO DO: add parameters to avoid type errors; but we get Typed anyway
     b_type_expression(RawExpr,[prob_ids(visible),variables],Type,Typed,Errors),
     (no_real_perror_occurred(Errors) -> true ; debug_println(9,errs(ID,Args,Type,Errors))),
     Source=Typed.
% TO DO: treate substitutions
source_code_for_definition(DefType,ID,_Args,_RawExpr,Source) :-
     Source=b(identifier(ID),DefType,[]).

:- use_module(b_global_sets,[lookup_global_constant/2,prob_deferred_set_element/4]).
global_set_element(ID,GlobalSetID) :- lookup_global_constant(ID,fd(_,GlobalSetID)).
global_set_element(ID,GlobalSetID) :-
       prob_deferred_set_element(GlobalSetID,_Elem,ID,all).


get_constant_type_and_origin(ID,Type,OriginStr,OriginTerm) :-
      b_get_machine_constants(Constants),
      TID = b(identifier(ID),Type,_Infos),
      member(TID,Constants),
      % we could also use member(origin(Origin),Infos)
      translate:translate_span_with_filename(TID,OriginStr), OriginTerm=TID.
get_variable_type_and_origin(ID,Type,OriginStr,OriginTerm) :-
      b_get_machine_variables(Vars),
      TID = b(identifier(ID),Type,_Infos),
      member(TID,Vars),
      % we could also use member(origin(Origin),Infos)
      translate:translate_span_with_filename(TID,OriginStr), OriginTerm=TID.

get_constant_definition(ID,FullDef) :-
      b_get_properties_from_machine(Prop),
      EqDef = b(equal(LHS,RHS),pred,_),
      member_in_conjunction_cse(FullDef,EqDef,Prop),
      (get_texpr_id(LHS,ID) ; get_texpr_id(RHS,ID)).

get_constant_predicate(ID,FullPred) :-
      b_get_properties_from_machine(Prop),
      member_in_conjunction_cse(FullPred,Pred,Prop),
      involves_id(Pred,ID).
involves_id(ID,b(E,_,_)) :- involves_id_aux(E,ID).
involves_id_aux(member(LHS,_),ID) :- get_texpr_id(LHS,ID).
involves_id_aux(subset(LHS,_),ID) :- get_texpr_id(LHS,ID).
involves_id_aux(strict_subset(LHS,_),ID) :- get_texpr_id(LHS,ID).

:- volatile b_get_machine_setscope/2.
:- dynamic b_get_machine_setscope/2.
:- public b_get_machine_setscope_calc/2.
% precompiled: definitions of the form "scope_SET == Expr",
%   Set is the name of the Set, TExpr the typed expression
b_get_machine_setscope_calc(Set,TExpr) :-
    b_definition_prefixed(expression, scope_, Set, Name,_),
    % TO DO: check if it is a SET
    b_get_typed_definition(Name,[],TExpr).

% get a specific MAX_OPERATIONS Value for an operation
:- volatile b_get_machine_operation_max/2.
:- dynamic b_get_machine_operation_max/2.
:- public b_get_machine_operation_max_calc/2.
% precompiled: definitions of the form "scope_SET == Expr",
%   Set is the name of the Set, TExpr the typed expression
b_get_machine_operation_max_calc(Set,MaxOp) :-
    b_definition_prefixed(expression, 'MAX_OPERATIONS_', Set, Name,DefPos),
    (b_get_typed_definition(Name,[constants],DEF), % scope=constants allows to use other definitions
     DEF=b(integer(MaxOp),integer,_)
    % TO DO: check if it is an operation
     %, MaxOp >= 0  we now use negative numbers for randomised restart
     -> true
     ; add_warning(b_get_machine_operation_max_calc,'Definition must specify an integer (negative numbers means using randomised restart) ',Name,DefPos),
       fail
     ).

:- volatile b_get_machine_goal/1.
:- dynamic b_get_machine_goal/1.
:- public b_get_machine_goal_calc/1.
% precompiled: definition of GOAL, a typed predicate
b_get_machine_goal_calc(TPred) :- get_goal(TPred).

b_reset_machine_goal_from_DEFINITIONS :-
    get_goal(TPred),
    b_set_parsed_typed_machine_goal(TPred).
get_goal(Goal) :-
    get_proz_settings(Settings),memberchk(goal(Goal),Settings),!.
get_goal(Goal) :- % print(getting_goal),nl,
    get_texpr_type(Goal,pred),
    b_get_typed_predicate_definition('GOAL',[variables],Goal).

:- volatile b_get_machine_searchscope/1.
:- dynamic b_get_machine_searchscope/1.
:- public b_get_machine_searchscope_calc/1.
% precompiled: definition of SCOPE, a typed predicate
b_get_machine_searchscope_calc(TPred) :-
    get_texpr_type(TPred,pred),
    b_get_typed_definition('SCOPE',[variables],TPred).

:- volatile b_get_definition_string_from_machine/3.
:- dynamic b_get_definition_string_from_machine/3.
:- public b_get_definition_string_from_machine_calc/3.
% precompiled: string definitions, Name == "String"
b_get_definition_string_from_machine(Name,String) :-
   b_get_definition_string_from_machine(Name,_,String).
b_get_definition_string_from_machine_calc(Name,Pos,String) :-
    b_get_definition(Name,expression,[],string(Pos,Str),_Deps),
    get_texpr_expr(TExpr,string(String)),
    type_with_errors_in_context(string(Pos,Str),[],_,TExpr,machine_context).

:- use_module(specfile,[animation_minor_mode/1]).
get_animation_image(Nr,S) :-
    animation_minor_mode(z),
    get_proz_animation_function(_,_,Images),!,
    nth1(Nr,Images,S).
get_animation_image(Nr,S) :- number(Nr),!,
    number_codes(Nr,TailCodes),
    /* ANIMATION_IMG */
    atom_codes(Def_Name,[65,78,73,77,65,84,73,79,78,95,73,77,71|TailCodes]),
    b_get_definition_string_from_machine(Def_Name,S).
get_animation_image(Nr,S) :-
    b_get_definition_string_from_machine(Def_Name,S),
    /* ANIMATION_IMG */
    atom_codes(Def_Name,[65,78,73,77,65,84,73,79,78,95,73,77,71|TailCodes]),
    number_codes(Nr,TailCodes).

:- use_module(error_manager,[extract_file_number_and_name/3]).
:- use_module(specfile,[b_or_z_mode/0]).
get_animation_image_source_file(Nr,File) :- b_or_z_mode,
    % get source file where Animation image is defined; relevant for finding images
    b_get_definition_string_from_machine(Def_Name,Pos,_),
    atom_codes(Def_Name,[65,78,73,77,65,84,73,79,78,95,73,77,71|TailCodes]),
    number_codes(Nr,TailCodes),
    extract_file_number_and_name(Pos,_,File).

get_proz_animation_function(Function,Default,Images) :-
    get_proz_settings(Settings),
    memberchk(animation_function(Function,Default,Images),Settings).
get_proz_settings(Settings) :-
    machine(_,Sections),memberchk(meta/Meta,Sections),
    memberchk(proz_settings(Settings),Meta).


:- volatile b_get_machine_animation_function/2.
:- dynamic b_get_machine_animation_function/2.
:- public b_get_machine_animation_function_calc/2.
:- use_module(tools,[safe_number_codes/2]).
:- use_module(bsyntaxtree,[get_texpr_set_type/2]).
% precompiled: definition of ANIMATION_FUNCTION, a typed expression
b_get_machine_animation_function_calc(AFun,Nr) :-
    get_proz_animation_function(TFun,Default,_),!,
    ( Nr = -1, AFun = Default
    ; Nr = 0, TFun \= none, AFun=TFun
    ).
b_get_machine_animation_function_calc(TFun,Nr) :-
    get_definition_with_nr_suffix("ANIMATION_FUNCTION",Nr,Def_Name),
    b_get_typed_expression_definition(Def_Name,[variables],TFun), % only type it afterwards; avoid throwing type errors for other definitions
    (get_texpr_set_type(TFun,couple(_,_)) -> true
     ; get_texpr_type(TFun,Type),
       add_warning(b_get_machine_animation_function_calc,'Illegal type for ANIMATION_FUNCTION: ',Def_Name:Type),fail).


:- volatile b_get_machine_heuristic_function/1.
:- dynamic b_get_machine_heuristic_function/1.
:- public b_get_machine_heuristic_function_calc/1.
b_get_machine_heuristic_function_calc(TFun) :- %print(extracting_heuristic_fun(TFun)),nl,
    get_texpr_type(TFun,integer),
    b_get_typed_expression_definition('HEURISTIC_FUNCTION',[variables],TFun).

:- volatile b_get_machine_animation_expression/1.
:- dynamic b_get_machine_animation_expression/1.
:- public b_get_machine_animation_expression_calc/1.
b_get_machine_animation_expression_calc(AExpr) :- %print(extracting_heuristic_fun(TFun)),nl,
    b_get_typed_expression_definition('ANIMATION_EXPRESSION',[variables],AExpr).

% Custom nodes and edges for a custom graph representation of a state
:- volatile b_get_machine_custom_nodes_function/2, b_get_machine_custom_edges_function/2.
:- dynamic b_get_machine_custom_nodes_function/2, b_get_machine_custom_edges_function/2.
:- public b_get_machine_custom_nodes_function_calc/2.
b_get_machine_custom_nodes_function_calc(TFun,Nr) :-
    get_definition_with_nr_suffix("CUSTOM_GRAPH_NODES",Nr,Def_Name),
    %get_texpr_set_type(TFun,_), % it is a set; we now accept more complex expressions
    b_get_typed_expression_definition(Def_Name,[variables],TFun).
:- public b_get_machine_custom_edges_function_calc/2.
b_get_machine_custom_edges_function_calc(TFun,Nr) :- %print(extracting_heuristic_fun(TFun)),nl,
    get_definition_with_nr_suffix("CUSTOM_GRAPH_EDGES",Nr,Def_Name),
    %get_texpr_set_type(TFun,couple(_,_)), % it is a function; we now accept more complex expressions
    b_get_typed_expression_definition(Def_Name,[variables],TFun).


get_definition_with_nr_suffix(Prefix,Nr,Def_Name) :-
    b_get_definition(Def_Name,expression,[],_RawExpr,_Deps),
    atom_codes(Def_Name,AC),
    append(Prefix,TailCodes,AC),
    (TailCodes=[] -> Nr=0
     ; TailCodes = "_DEFAULT" -> Nr = -1
     ; safe_number_codes(Nr,TailCodes)).


% ------------------------

:- volatile b_get_constant_represented_inside_global_set/2.
:- dynamic b_get_constant_represented_inside_global_set/2.

:- public b_get_constant_represented_inside_global_set_calc/2.
b_get_constant_represented_inside_global_set_calc(X,GlobalSetName) :-
  b_get_disjoint_constants_of_type(GlobalSetName, DisjointConstants,_),
  member(X,DisjointConstants). /* X will be integrated into the global_set */


:- volatile b_get_disjoint_constants_of_type/3.
:- dynamic b_get_disjoint_constants_of_type/3.
:- public b_get_disjoint_constants_of_type_calc/3.
b_get_disjoint_constants_of_type_calc(GlobalSetName, DisjointConstants, AllConstantsIDs) :-
    b_get_machine_all_constants(Constants), /* get all the constants */
    b_get_properties_from_machine(Properties),
    b_get_machine_set(GlobalSetName),
    \+ b_get_named_machine_set(GlobalSetName,_), % not an explicitly enumerated set
    find_constants_of_type(Constants,GlobalSetName,AllConstantsIDs,_GSConstants),
    find_inequal_global_set_identifiers(AllConstantsIDs,global(GlobalSetName),Properties,DisjointConstants),
    DisjointConstants \= [],
    debug_println(9,disjoint_enumerated_constants(GlobalSetName,DisjointConstants)).
   % TO DO: remove from Properties irrelevant conjuncts; sort AllConstantsIDs by having first those with max disequalities

find_constants_of_type([],_GS,[],[]) :- !.
find_constants_of_type([b(identifier(ID),global(GS),INFO)|T], GS, [ID|IT],
                       [b(identifier(ID),global(GS),INFO)|TT]) :- !,
          find_constants_of_type(T,GS,IT,TT).
find_constants_of_type([b(_,_,_)|T], GS, IT, TT) :- !,
          find_constants_of_type(T,GS,IT,TT).
find_constants_of_type(X,GS,_,_) :- print(find_constants_of_type_error(X,GS)),nl,fail.


% ------------------------
:- volatile b_get_all_used_identifiers_in_section/2.
:- dynamic b_get_all_used_identifiers_in_section/2.
:- public b_get_all_used_identifiers_in_section_calc/2.

% Note: may contain var$0 instead of var for becomes_such
b_get_all_used_identifiers_in_section_calc(SECTION,Identifiers) :-
    full_b_machine(BMachine),
    member(SECTION,[constraints,properties,invariant,assertions,
                initialisation,operation_bodies]),
    findall(I, find_used_identifier_in_machine_section(BMachine,SECTION,I), Ids),
    %print(section_ids(SECTION,Ids)),nl,
    sort(Ids,Identifiers).

:- volatile b_get_all_used_identifiers/1.
:- dynamic b_get_all_used_identifiers/1.
:- public b_get_all_used_identifiers_calc/1.
:- use_module(library(ordsets),[ord_union/2]).
% Precompiled: b_get_all_used_identifiers_calc/1 returns a list of all
% identifiers used in machine sections that contain expressions and predicates
b_get_all_used_identifiers_calc(Identifiers) :-
    findall(I, b_get_all_used_identifiers_in_section(_,I), ListOfList),
    ord_union(ListOfList,Identifiers). % ,print(all(Identifiers)),nl.

find_used_identifier_in_machine_section(Machine,Sec,Identifier) :-
    % treat it like a list of expressions
    get_section_texprs(Sec,Machine,TExprs),
    % and find all identifiers in each expression
    member(TExpr,TExprs),
    find_identifier_uses(TExpr,[],Ids),
    member(Identifier,Ids).

:- volatile b_is_unused_constant/1.
:- dynamic b_is_unused_constant/1.
:- public b_is_unused_constant_calc/1.
b_is_unused_constant_calc(UnusedConstant) :-
      findall(C,b_is_constant(C),CL), %b_get_machine_constants(Constants), gives typed ids
      CL\=[] ,sort(CL,Constants),
      b_get_all_used_identifiers_in_section(invariant,I1),
      b_get_all_used_identifiers_in_section(assertions,I2),
      b_get_all_used_identifiers_in_section(initialisation,I3),
      b_get_all_used_identifiers_in_section(operation_bodies,I4),
      ord_union([I1,I2,I3,I4],UsedIds),
      %print(used(UsedIds)),nl,
      ord_subtract(Constants,UsedIds,UnusedConstants),
      UnusedConstants\=[],
      (b_get_machine_variables([]) -> true
       ; length(UnusedConstants,NrUC),
         (silent_mode(on) -> true
         ; (NrUC>10,debug_mode(off),UnusedConstants=[C1,C2|_]) ->
             print_message_with_max_depth(unused_constants(NrUC,[C1,C2,'...']),500)
           ; print_message_with_max_depth(unused_constants(NrUC,UnusedConstants),500))
      ),
      member(UnusedConstant,UnusedConstants).


% precompiled: Skel is the operation skeleton, SpecializedInv a typed predicate
% (Skel,SpecializedInv) :-
%  b_get_machine_operation(Name,_Res,Params,_Body),
%  atom_concat('SIMPLIFIED_INV_OP_',Name,ASNC),
%  b_get_typed_definition(ASNC,[variables,identifiers(Params)],SpecializedInv),
%  length(Params,Len),
%  functor(Skel,Name,Len),
%  print(b_specialized_invariant_for_op(Skel)).


get_section_from_current_machine(Section,Content) :-
    full_b_machine(BMachine),
    get_section(Section,BMachine,Content).

is_ground(X,R) :- ground(X),!,R=ground.
is_ground(_,nonground).

b_set_empty_machine :- debug_println(9,setting_empty_machine),
    empty_machine(Main,Machines),
    b_set_machine(Main,Machines,[]).

b_get_machine_refinement_hierarchy(Hierarchy) :-
    \+ machine(_Name,_M),!,Hierarchy=[]. %No B Machine loaded
b_get_machine_refinement_hierarchy(Hierarchy) :-
    get_section_from_current_machine(meta,Infos),
    member(hierarchy/Hierarchy,Infos),!.

b_get_refined_machine_name(Machine) :-
   b_get_machine_refinement_hierarchy([_Main,Machine|_Rest]).
b_get_refined_ancestors_names(list(AncestorList)) :-
   b_get_machine_refinement_hierarchy([_Main|AncestorList]).
b_get_refined_machine(M) :-
    get_section_from_current_machine(meta,Infos),
    member(refined_machine/M,Infos), !.



b_get_machine_header_position(MachName,Position) :-
    get_section_from_current_machine(meta,Infos),
    member(header_pos/HeaderPosList,Infos),
    % List with elements MachineName/PositionOfMachineHeader
    member(MachName/Position,HeaderPosList).

% model type can be system for Atelier-B Event-B
b_get_model_type(M) :-
    get_section_from_current_machine(meta,Infos),
    member(model_type/M,Infos), !.

% TODO(DP,6.8.2008)
:- volatile b_machine_temp_predicate/1.
:- dynamic b_machine_temp_predicate/1.

set_temp_predicate(CustomPredicate) :-
  b_parse_machine_predicate(CustomPredicate,TypedPred),
  assert_temp_typed_predicate(TypedPred).

% set a temporary predicate as additional guard
set_temp_predicate(CustomPredicate,'$initialise_machine') :- !,set_temp_predicate(CustomPredicate).
set_temp_predicate(CustomPredicate,'$setup_constants') :- !,set_temp_predicate(CustomPredicate).
set_temp_predicate(CustomPredicate,OpName) :-
  b_parse_machine_operation_pre_post_predicate(CustomPredicate,TypedPred, OpName),
  assert_temp_typed_predicate(TypedPred).

assert_temp_typed_predicate(TypedPred) :-
  reset_temp_predicate,
  assert(b_machine_temp_predicate(TypedPred)).


reset_temp_predicate :- retractall(b_machine_temp_predicate(_)).

:- volatile b_machine_additional_property/1.
:- dynamic b_machine_additional_property/1.

% additional properties for set_up_constants
% TO DO: check if there are constants/properties; otherwise raise warning
add_additional_property(CustomPredicate) :-
  format('Adding Property: ~w~n',[CustomPredicate]),
  b_parse_machine_predicate(CustomPredicate,TypedPred),
  assert(b_machine_additional_property(TypedPred)).

:- use_module(bsyntaxtree, [safe_create_texpr/4]).

:- public b_extract_values_clause_assignment_calc/3.
b_extract_values_clause_assignment_calc(ID,Type,TVal) :-
    % extract equalities from VALUES clause
    get_section_from_current_machine(values,Values),
    member(b(values_entry(TID,TVal),_,Info),Values),
    get_texpr_id(TID,ID),
    get_texpr_type(TID,Type),
    debug_println(20,values_entry(ID,Type,TVal)),
    (b_is_constant(ID)
     -> safe_create_texpr(equal(TID,TVal),pred,Info,Equality),
        assert(b_machine_additional_property(Equality)),
        (b_extract_values_clause_assignment(ID,_,_)
          -> add_error(bmachine,'Multiple VALUES entries for constant: ',ID,TID)
          ; true)
      ; b_get_machine_set(ID) -> true % will be dealt with in b_global_sets
      ; add_error(bmachine,'VALUES clause assigns to unknown constant: ',ID,TID),
        fail
     ).

% -------------------

:- use_module(specfile,[unset_animation_minor_modes/1,reset_animation_minor_modes/1]).
:- use_module(translate,[set_unicode_mode/0, unset_unicode_mode/0]).
b_show_machine_representation(Rep,AdditionalInfo,UnsetMinorModes,Unicode) :-
   Unicode=true,!,
   set_unicode_mode,
   call_cleanup(b_show_machine_representation(Rep,AdditionalInfo,UnsetMinorModes),unset_unicode_mode).
b_show_machine_representation(Rep,AdditionalInfo,UnsetMinorModes,_) :-
   b_show_machine_representation(Rep,AdditionalInfo,UnsetMinorModes).

% AdditionalInfo=true means show additional information
% UnsetMinorModes will print in default classical B style
b_show_machine_representation(Rep,AdditionalInfo,UnsetMinorModes) :-
    full_b_machine(BMachine),
    (UnsetMinorModes==true -> unset_animation_minor_modes(L) ; L=[]),
    translate_machine(BMachine,Rep,AdditionalInfo),
    (UnsetMinorModes==true -> reset_animation_minor_modes(L) ; true).

:- use_module(tools,[get_tail_filename/2,split_filename/3,safe_atom_chars/3]).
b_write_machine_representation_to_file(Level,PPFILE) :-
    %get_full_b_machine(Name,M),
    get_tail_filename(PPFILE,Tail),
    split_filename(Tail,MachName,_Ext),
    b_write_machine_representation_to_file3(MachName,Level,PPFILE).

:- use_module(translate,[set_print_type_infos/1]).
:- use_module(tools_files, [write_to_utf8_file_or_user_output/2]).
b_write_machine_representation_to_file3(MachName,Level,PPFILE) :-
    debug_println(20,'% Pretty-Printing Internal Representation to File: '),
    debug_println(20,PPFILE),
    translate:set_print_type_infos(Level), % needed or all
    get_full_b_machine(_Name,M),
    (translate_machine_to_classicalb(MachName,M,Rep)
      -> translate:set_print_type_infos(none)
       ; add_internal_error('Translating machine failed: ',b_write_machine_representation_to_file3(MachName,Level,PPFILE)),
         set_print_type_infos(none),
        fail),
    write_to_utf8_file_or_user_output(PPFILE,Rep).

:- use_module(specfile).
translate_machine_to_classicalb(MachName,M,Rep) :-
   temporary_set_preference(expand_avl_upto,-1,CHNG), % avoid printing sets using condensed #
   (animation_minor_mode(X)
    -> remove_animation_minor_mode,
       call_cleanup(translate_machine(machine(MachName,M),Rep,false), set_animation_minor_mode(X))
     ; translate_machine(machine(MachName,M),Rep,false)),
    reset_temporary_preference(expand_avl_upto,CHNG).

b_write_eventb_machine_to_classicalb_to_file(PPFILE) :-
  get_tail_filename(PPFILE,Tail),
  split_filename(Tail,MachName,_Ext),
  b_show_eventb_as_classicalb1(MachName,Rep,_AdditionalInfo),
  write_to_utf8_file_or_user_output(PPFILE,Rep).

:- use_module(preferences,[temporary_set_preference/3,reset_temporary_preference/2]).
b_show_eventb_as_classicalb1(MachName,Rep,AdditionalInfo) :-
  full_b_machine(machine(_Name,MachineBody)),
  % limit abstract level to 0
  b_limit_abstract_level_to_zero(MachineBody,MachineBodyLevel0),
  temporary_set_preference(translate_print_typing_infos,true,CHNG),
  temporary_set_preference(translate_ids_to_parseable_format,true,CHNG2),
   temporary_set_preference(expand_avl_upto,-1,CHNG3), % avoid printing sets using condensed #
  (translate_eventb_to_classicalb(machine(MachName,MachineBodyLevel0),AdditionalInfo,Rep)
    -> reset_temporary_preference(translate_print_typing_infos,CHNG),
       reset_temporary_preference(translate_ids_to_parseable_format,CHNG2),
       reset_temporary_preference(expand_avl_upto,CHNG3)
    ;  add_internal_error('Translating Event-B machine to Classical B failed: ',b_show_eventb_as_classicalb1(MachName,Rep,AdditionalInfo)),
       reset_temporary_preference(translate_print_typing_infos,CHNG),
       reset_temporary_preference(translate_ids_to_parseable_format,CHNG2),
       reset_temporary_preference(expand_avl_upto,CHNG3),
       fail).

b_limit_abstract_level_to_zero(MachineBody,MachineBodyLevel0) :-
  maplist(b_set_to_level_zero,MachineBody,MachineBodyLevel0).

% TODO: set to level zero also the other events of the Event-B machine
b_set_to_level_zero(initialisation/InitEvent,initialisation/InitEventLevelZero) :-
  get_texpr_expr(InitEvent,rlevent(Id,Sec,St,Par,Grd,Thms,Act,VW,PW,Ums,_Abs)),
  get_texpr_info(InitEvent,Info),
  create_texpr(rlevent(Id,Sec,St,Par,Grd,Thms,Act,VW,PW,Ums,[]),subst,Info,InitEventLevelZero).
b_set_to_level_zero(X,X).

b_show_eventb_as_classicalb(Rep,AdditionalInfo) :-
    full_b_machine(BMachine),
    translate_eventb_to_classicalb(BMachine,AdditionalInfo,Rep).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% precompilation: predicates are precompiled by calling the ..._calc predicate

:- use_module(state_packing).
:- use_module(tools,[start_ms_timer/1,stop_ms_walltimer_with_msg/2]).
% maybe this predicate should be directly called by b_load_machine
b_machine_precompile :- debug_println(9,'Precompiling B Machine'),
    start_ms_timer(T1),
    get_full_b_machine(_,_),
    auto_precompile,
    (debug_mode(on) -> stop_ms_walltimer_with_msg(T1,'auto_precompile: ') ; true),
    precompile_operation_names_in_reverse_order,
    % TODO(DP, 14.8.2008): remove reference to b_global_sets
    start_ms_timer(T2),
    b_check_and_precompile_global_set_type_definitions,
    (debug_mode(on) -> stop_ms_walltimer_with_msg(T2,'precompiling global sets: ') ; true),
    % TODO(DP, 12.9.2008): remove reference to kernel_freetypes
    reset_freetypes,
    get_section_from_current_machine(freetypes,Freetypes),
    add_all_freetypes(Freetypes),
    auto_precompile_phase2,
    state_packing:precompile_state_packing,
    do_machine_consistency_check,
    try_kodkod,
    (debug_mode(on) -> stop_ms_walltimer_with_msg(T1,'complete precompiling B machine: ') ; true).
add_all_freetypes([]).
add_all_freetypes([freetype(Name,Cases)|Rest]) :-
    add_freetype(Name,Cases),
    add_all_freetypes(Rest).


other_spec_precompile :- retractall(bmachine_is_precompiled),
  /* call if you do not animate a B specification */
    is_precompiled_predicate(Pred/Arity),
    functor(Pattern,Pred,Arity),
    retractall(Pattern),
    (precompiled_predicate_no_error_when_not_precompiled(Pred/Arity)
      -> assert( (Pattern :- fail ))
      ;  assert( (Pattern :- print_message('No B machine available for '),print_message(Pattern),fail) )
    ),
    fail.
other_spec_precompile :-
    retractall(b_get_machine_constants(_)),
    assert( b_get_machine_constants([]) ),
    retractall(b_get_machine_variables(_)),
    assert( b_get_machine_variables([]) ),
    retractall(b_get_static_assertions_from_machine(_)),
    assert( b_get_static_assertions_from_machine([]) ),
    retractall(b_machine_has_static_assertions),
    retractall(b_get_unproven_static_assertions_from_machine(_)),
    assert( b_get_unproven_static_assertions_from_machine([]) ),
    retractall(b_get_dynamic_assertions_from_machine(_)),
    assert( b_get_dynamic_assertions_from_machine([]) ),
    retractall(b_machine_has_dynamic_assertions),
    retractall(b_get_unproven_dynamic_assertions_from_machine(_)),
    assert( b_get_unproven_dynamic_assertions_from_machine([]) ),
    retractall( b_get_machine_searchscope(_) ),
    retractall( b_get_machine_goal(_) ),
    assert_bmachine_is_precompiled, debug_println(4,other_spec_precompile).

is_precompiled_predicate(P) :- precompiled_predicate(P).
is_precompiled_predicate(P) :- precompiled_predicate_phase2(P).

% phase 1 precompilation
precompiled_predicate(b_machine_name/1).
precompiled_predicate(b_get_named_machine_set/3).
precompiled_predicate(b_get_properties_from_machine/1).
precompiled_predicate(b_get_machine_all_constants/1).
precompiled_predicate(b_get_machine_set/2).
precompiled_predicate(b_get_disjoint_constants_of_type/3).
precompiled_predicate(b_get_constant_represented_inside_global_set/2).
precompiled_predicate(b_get_machine_constants/1).
precompiled_predicate(b_get_machine_variables/1).
precompiled_predicate(b_get_invariant_from_machine/1).
precompiled_predicate(b_get_linking_invariant_from_machine/1).
precompiled_predicate(b_is_constant/2).
precompiled_predicate(b_is_variable/2).
precompiled_predicate(b_get_constant_variable_description/2).
precompiled_predicate(b_get_static_assertions_from_machine/1).
precompiled_predicate(b_machine_has_static_assertions/0).
precompiled_predicate(b_get_unproven_static_assertions_from_machine/1).
precompiled_predicate(b_get_dynamic_assertions_from_machine/1).
precompiled_predicate(b_machine_has_dynamic_assertions/0).
precompiled_predicate(b_get_unproven_dynamic_assertions_from_machine/1).
precompiled_predicate(b_get_assertions_from_main_machine/2).
precompiled_predicate(b_get_initialisation_from_machine/2).
precompiled_predicate(b_get_machine_operation/6).
precompiled_predicate(b_top_level_operation/1).
precompiled_predicate(b_is_operation_name/1).
precompiled_predicate(b_top_level_feasible_operation/1).
precompiled_predicate(b_get_promoted_machine_operations/1).
precompiled_predicate(get_operation_info/2).
precompiled_predicate(b_get_machine_operation_for_animation/6).
precompiled_predicate(b_get_machine_goal/1).
precompiled_predicate(b_get_machine_setscope/2).
precompiled_predicate(b_get_machine_operation_max/2).
precompiled_predicate(b_get_machine_searchscope/1).
precompiled_predicate(b_get_definition_string_from_machine/3).
precompiled_predicate(b_get_machine_animation_function/2).
precompiled_predicate(b_get_machine_heuristic_function/1).
precompiled_predicate(b_get_machine_animation_expression/1).
precompiled_predicate(b_get_machine_custom_nodes_function/2).
precompiled_predicate(b_get_machine_custom_edges_function/2).
precompiled_predicate(complete_discharged_info/0). % just call it before calling b_specialized_invariant_for_op_calc !
precompiled_predicate(get_proven_invariant/2).
precompiled_predicate(b_specialized_invariant_for_op/2).
precompiled_predicate(b_get_all_used_identifiers_in_section/2).
precompiled_predicate(b_get_all_used_identifiers/1).
precompiled_predicate(b_is_unused_constant/1).
precompiled_predicate(b_extract_values_clause_assignment/3).
precompiled_predicate(b_machine_statistics/2).

% require b_global_sets to be initialised
precompiled_predicate_phase2(b_get_operation_normalized_read_write_info/3).
precompiled_predicate_phase2(b_operation_cannot_modify_state/1).
precompiled_predicate_phase2(b_operation_reads_output_variables/3).
precompiled_predicate_phase2(reuse_operation_effect/2).
precompiled_predicate_phase2(b_get_operation_non_det_modifies/2).
precompiled_predicate_phase2(candidate_operation_for_reuse/1).

% these functions sometimes get called also in csp mode; simply fail silently for those cases in other_spec_precompile
precompiled_predicate_no_error_when_not_precompiled(b_get_machine_animation_function/2).
precompiled_predicate_no_error_when_not_precompiled(b_get_machine_heuristic_function/1).
precompiled_predicate_no_error_when_not_precompiled(b_get_machine_animation_expression/1).
precompiled_predicate_no_error_when_not_precompiled(b_get_machine_custom_nodes_function/2).
precompiled_predicate_no_error_when_not_precompiled(b_get_machine_custom_edges_function/2).

% on startup, all precompiled predicates yield error message
startup_precompiled :- retractall(bmachine_is_precompiled),
    precompiled_predicate(Pred/Arity),
    functor(Pattern,Pred,Arity),
    retractall(Pattern),
    assert( (Pattern :- add_error(bmachine,'B machine not precompiled for ',Pattern),fail) ),
    fail.
startup_precompiled :- debug_println(4,startup_precompiled).

:- startup_precompiled.

% for each predicate in "precompiled_predicate", auto_precompile computes
% all solutions by calling the predicate with an appended _calc
auto_precompile :-
    % failure-driven loop
    precompiled_predicate(Pred/Arity), %print(precompile(Pred/Arity)),nl,
    %statistics(walltime,[Tot,Delta]), format('Precompiling ~w/~w  [~w ms (Delta ~w) ms]~n',[Pred,Arity,Tot,Delta]),
    auto_precompile2(Pred,Arity,Pattern),
    (Pattern -> true /* this pattern already asserted */
             ; assert(Pattern)), fail.
auto_precompile :- assert_bmachine_is_precompiled, debug_println(4,auto_precompile).
auto_precompile_phase2 :-
    % failure-driven loop
    precompiled_predicate_phase2(Pred/Arity), %print(precompile(Pred/Arity)),nl,
    auto_precompile2(Pred,Arity,Pattern),
    (Pattern -> true /* this pattern already asserted */
             ; assert(Pattern)), fail.
auto_precompile_phase2.
% if the precompiled predicates is e.g. foo/2 then
% Pattern=foo(A,B) and Call=foo_calc(A,B)
auto_precompile2(Pred,Arity,Pattern) :-
    atom_concat(Pred,'_calc',CallFunctor),
    functor(Pattern,Pred,Arity),
    functor(Call,CallFunctor,Arity),
    unify_args(Arity,Call,Pattern),
    % delete all existing solutions
    retractall(Pattern), %print(retractall(Pattern)),nl,
    % the failure-driven loop for one predicate
    call(Call).
unify_args(0,_,_) :- !.
unify_args(N,A,B) :-
    arg(N,A,Arg), arg(N,B,Arg),
    N1 is N-1, unify_args(N1,A,B).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% set machine goal

b_set_machine_goal(Goal) :- b_set_machine_goal(Goal,false).
b_set_machine_goal(Goal,WithDeferredSets) :-
    (WithDeferredSets=with_deferred -> Scope = [prob_ids(visible),variables] ; Scope = [variables]),
    b_parse_machine_predicate(Goal,Scope,TypedPred),
    b_set_parsed_typed_machine_goal(TypedPred).
b_set_parsed_typed_machine_goal(TypedPred) :-
    retractall(b_get_machine_goal(_)),
    assert(b_get_machine_goal(TypedPred)).


b_set_machine_searchscope(Goal) :- b_set_machine_searchscope(Goal,with_deferred).
b_set_machine_searchscope(Goal,WithDeferredSets) :-
    (WithDeferredSets=with_deferred -> Scope = [prob_ids(visible),variables] ; Scope = [variables]),
    b_parse_machine_predicate(Goal,Scope,TypedPred),
    b_set_parsed_typed_machine_searchscope(TypedPred).
b_set_parsed_typed_machine_searchscope(TypedPred) :-
    retractall(b_get_machine_searchscope(_)),
    assert(b_get_machine_searchscope(TypedPred)).

% also allow empty string, and #invariant and #not_invariant
b_parse_optional_machine_predicate(TargetString,Target) :-
    is_empty_string(TargetString),!,create_texpr(truth,pred,[],Target).
b_parse_optional_machine_predicate('#invariant',Invariant) :- !,
   b_get_invariant_from_machine(Invariant).
b_parse_optional_machine_predicate('#not_invariant',TargetPredicate) :- !,
   b_get_invariant_from_machine(Invariant),
   bsyntaxtree:create_negation(Invariant,TargetPredicate).
b_parse_optional_machine_predicate(TargetString,Target) :-
    b_parse_machine_predicate(TargetString,Target).

is_empty_string(Atom) :-
    atom_codes(Atom,Codes),
    is_empty_string2(Codes).
is_empty_string2([]).
is_empty_string2([32|Codes]) :- is_empty_string2(Codes).


% parsing a string as predicate with deferred set elements
b_parse_machine_predicate(Pred,TypedPred) :- b_parse_machine_predicate(Pred,[variables],TypedPred).
b_parse_machine_predicate(Pred,Scope,TypedPred) :-
    atom_codes(Pred,Codes),
    b_parse_machine_predicate_from_codes(Codes,Scope,TypedPred).
b_parse_machine_predicate_from_codes(Codes,Scope,TypedPred) :-
    b_parse_machine_predicate_from_codes2(Codes,[],Scope,TypedPred,closed).
b_parse_machine_predicate_from_codes_open(OptionalQuantifier,Codes,Defs,Scope,TypedPred) :-
    b_parse_machine_predicate_from_codes2(Codes,Defs,Scope,TypedPred,open(OptionalQuantifier)).

b_parse_machine_predicate_from_codes2(Codes,Defs,Scope,TypedPred,Mode) :-
    debug_println(9,parse_predicate(Mode)),
    (Defs=[] -> true ; debug_println(9,'% Ignoring DEFINITIONS')),
    b_parse_wo_type_machine_predicate_from_codes_to_raw_expr(Codes,Parsed),
    %format('Raw Parsed AST = ~w~n',[Parsed]),
    b_type_check_raw_expr(Parsed,Scope,TypedPred,Mode).

check_codes(X) :- (X=[] ; X=[H|_], number(H)),!.
check_codes(X) :- add_error(bmachine,'Error while parsing: not a code list: ',X),fail.

b_parse_wo_type_machine_predicate_from_codes_to_raw_expr(Codes,ParsedRaw) :-
    check_codes(Codes),
    catch(
               parse_predicate(Codes,ParsedRaw), % ,print(ParsedRaw),nl),
           Exception,
           ( Exception = parse_errors(Errors) -> %print(errors(Errors)),nl,
               add_all_perrors(Errors,[],parse_machine_predicate_error),fail
           ; otherwise ->
               add_error(bmachine,'Exception while parsing predicate: ',Exception),
               fail)).

b_type_check_raw_expr(ParsedRaw,Scope,TypedPred,Mode) :-
    debug_println(9,type_with_errors(ParsedRaw)),
    ( Mode == closed ->
        type_with_errors(ParsedRaw,Scope,pred,TypedPred)
    ; Mode = open(OptionalQuantifier) ->
        b_type_open_predicate_with_errors(OptionalQuantifier,ParsedRaw,Scope,TypedPred)).


b_parse_machine_operation_pre_post_predicate(Pred,TypedPred,OpName) :-
    atom_codes(Pred,Codes),
    catch( parse_predicate(Codes,Parsed),
           Exception,
           ( Exception = parse_errors(Errors) ->
               add_all_perrors(Errors,[],parse_pre_predicate_error),fail
           ; otherwise ->
               add_error(bmachine,'Exception while parsing predicate: ',Exception),
               fail)),
    ( is_initialisation_op(OpName) -> V=[]
     ; get_machine_operation_additional_identifiers(OpName,V) -> true
     ; add_error_fail(b_parse_machine_operation_pre_post_predicate,'Unknown operation:', OpName)),
    Scope = [prob_ids(visible),identifier(V),variables,external_library(all_available_libraries)],
    %print(typing(Parsed,V)),nl,
    type_with_errors(Parsed,Scope,pred,TypedPred).

% Note: B and Event-B have different way of priming; this is aimed at classical B
% and supposes predicate used together with execute_operation_by_predicate_in_state
% in classical B x$0 refers to the value before a substitution (e.g., becomes_such)
get_primed_machine_variables(Ids) :-
    b_get_machine_variables(Vars),
    maplist(prime_variable,Vars,Ids).
:- use_module(btypechecker,[prime_atom0/2]). % add $0 at end of variable
prime_variable(b(identifier(ID),T,I),b(identifier(PID),T,I)) :- prime_atom0(ID,PID).

is_initialisation_op('$setup_constants').
is_initialisation_op('$initialise_machine').

% get additional ids to add to Scope for typing with identifier(V)
get_machine_operation_additional_identifiers(OpName,V) :-
     b_get_machine_operation_for_animation(OpName,Results,Paras,_,_,true),
     append(Paras,Results,PR), % b_get_machine_operation_typed_parameters
     get_primed_machine_variables(PV),
     append(PV,PR,V).

:- public b_get_machine_operation_for_animation_calc/6.
% adds some additional ANY parameters if preference is set
b_get_machine_operation_for_animation_calc('$initialise_machine',[],Parameters,Body,OType,_TopLevel) :-
     b_get_machine_variables(DeclaredVars), Parameters=DeclaredVars,
     b_get_initialisation_from_machine(Body,OType).
b_get_machine_operation_for_animation_calc(Name,Results,Parameters,Body,OType,TopLevel) :-
    get_preference(show_eventb_any_arguments,EVENTB),
    b_get_machine_operation(Name,Results,RealParameters,RealBody,OType,OpPos),
    % TO DO: check whether we should only keep b_top_level_operation(Name) in case TopLevel==true !
    %length(RealParameters,P), format('Get Mach Op ~w ; paras ~w ; EB ~w, TopLevel=~w~n',[Name,P,EVENTB,TopLevel]),
    ((Results=[_|_] ; RealParameters=[_|_] ; EVENTB=false)
     -> Parameters=RealParameters, Body=RealBody
     ; translate_any_into_parameters(RealBody,FakeParameters,FakeBody),
       length(FakeParameters,Len),
       (Len > 255
        -> add_message(bmachine,'Not adding any parameters to operation, max_arity (255) exceeded: ',Name:Len,OpPos),
           Parameters=RealParameters, Body=RealBody
        ; (    TopLevel=true,  Parameters=FakeParameters, Body=FakeBody
            ;  TopLevel=false, Parameters=RealParameters, Body=RealBody
          )
      )
     ; Parameters=RealParameters, Body=RealBody
    ).

% TO DO: limit to max_arity
translate_any_into_parameters(b(E,Type,Info),Parameters,NewSubst) :-
   translate_any_into_parameters_aux(E,Type,Info,Parameters,NewSubst).
translate_any_into_parameters_aux(any(Parameters,PreCond,ABody),Type,Info,Parameters,NewSubst) :-
    NewSubst=b(select([b(select_when(PreCond,ABody),Type,Info)]),Type,Info). % TO DO: go deeper ? (nested ANY ?)
translate_any_into_parameters_aux(let(Parameters,Defs,LBody),Type,Info,Parameters,NewSubst) :-
    NewSubst=b(select([b(select_when(Defs,LBody),Type,Info)]),Type,Info). % TO DO: go deeper ? ANY could have been translated into several LETs ?
translate_any_into_parameters_aux(lazy_let_subst(Id,IdExpr,Body),Type,Info,Parameters,NewSubst) :-
    NewSubst=b(lazy_let_subst(Id,IdExpr,NewInnerSubst),Type,Info),
    translate_any_into_parameters(Body,Parameters,NewInnerSubst).


% add deferred set prob_ids and external libraries as well
b_parse_machine_expression_from_codes_with_prob_ids(Codes,TypedExpr) :-
    b_parse_machine_expression_from_codes3(Codes,
                  [prob_ids(visible),variables,external_library(all_available_libraries)],TypedExpr).
b_parse_machine_expression_from_codes(Codes,TypedExpr) :-
    b_parse_machine_expression_from_codes3(Codes,[variables],TypedExpr).

b_parse_machine_expression_from_codes3(Codes,Scope,TypedExpr) :-
    b_parse_machine_expression_from_codes(Codes,Scope,TypedExpr,_Type,true,Error),!,
    (Error=none -> true
     ; add_error(b_parse_machine_expression_from_codes,'Errors occured during parsing: ', Error),
       % format(user_error,'Parsing: ~s~n',[Codes]),
       fail).


%b_parse_machine_expression_from_codes(Codes,TypedExpr) :-
%    b_parse_machine_expression_from_codes(Codes,TypedExpr,_Type,false,Error), Error=none.

b_parse_machine_expression_from_codes(Codes,Typed,Type,GenerateParseErrors,Error) :-
    b_parse_machine_expression_from_codes(Codes,[variables],
                                          Typed,Type,GenerateParseErrors,Error).
b_parse_machine_expression_from_codes(Codes,Scope,
                                      Typed,Type,GenerateParseErrors,Error) :-
    b_parse_machine_formula_from_codes(expression,Codes,Scope,
                                      Typed,Type,GenerateParseErrors,Error).

% a variant of b_parse_machine_predicate that allows also expressions
b_parse_machine_formula(PredOrExpr,Scope,TypedPredOrExpr) :-
   atom_codes(PredOrExpr,Codes),
   b_parse_machine_formula_from_codes(formula,Codes,Scope,
       TypedPredOrExpr,_Type,true,Error), Error=none.

% Kind = expression, predicate, formula, substitution
b_parse_machine_formula_from_codes(Kind,Codes,Scope,
                                      Typed,Type,GenerateParseErrors,Error) :-
    b_parse_only(Kind,Codes, ParsedAST, GenerateParseErrors, Error),
    %print(parsed(Parsed,Error)),nl,
    ( nonvar(Error) -> true /* either exception or parse error occured */
      ; type_with_errors(ParsedAST,Scope,Type,Typed) -> Error=none
      ; otherwise -> Error=type_error).

b_parse_only(Kind,Codes, ParsedAST, _,Error) :- (Codes \= [] , Codes \= [_|_]),!,
   add_internal_error('Not a codes list: ',b_parse_only(Kind,Codes, ParsedAST, _, Error)),
   Error = internal_error.
b_parse_only(Kind,Codes, ParsedAST, GenerateParseErrors,Error) :-
    catch( parse(Kind,Codes,ParsedAST),
           Exception,
           ( GenerateParseErrors=true,
               ( Exception = parse_errors(Errors) ->
                   add_all_perrors(Errors,[],parse_machine_formula_error),
                   Error=parse_error
               ; otherwise ->
                   ajoin(['Exception while parsing ',Kind,': '],Msg),
                   add_error(bmachine,Msg,Exception),
                   Error=exception))).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% a flexible parsing predicate, useful for dot and table tools to be called from
% either probcli or prob2 with either atoms or with raw ASTs:

parse_expression_raw_or_atom(Expr,TypedExpr) :-
    parse_expression_raw_or_atom3(Expr,[variables],TypedExpr).
parse_expression_raw_or_atom_with_prob_ids(Expr,TypedExpr) :-
    parse_expression_raw_or_atom3(Expr,[prob_ids(visible),variables],TypedExpr).

parse_expression_raw_or_atom3('$VARIABLES',_,TypedExpr) :- !,
    bmachine:b_get_machine_variables(DeclaredVars),
    gen_couple_from_list(DeclaredVars,TypedExpr).
parse_expression_raw_or_atom3(RawAST,Scope,TypedExpr) :- compound(RawAST),!, % allow to pass raw AST as well
    type_with_errors(RawAST,Scope,Type,TypedExpr),
    ((Type=pred ; Type=subst)
      -> add_error(parse_expression,'Expected expression formula but obtained: ',Type),fail
      ; true ).
parse_expression_raw_or_atom3(VorE,Scope,TypedExpr) :-
    atom_codes(VorE,VorECodes),
    % to do: maybe support b_parse_machine_expression_from_codes_with_prob_ids
    b_parse_machine_expression_from_codes3(VorECodes,Scope,TypedExpr).

:- use_module(bsyntaxtree, [create_couple/3]).
gen_couple_from_list([],string('NO VARIABLES')).
gen_couple_from_list([TID],Res) :- !, Res=TID.
gen_couple_from_list([TID|T],Couple) :-
   gen_couple_from_list(T,Rest),
   create_couple(TID,Rest,Couple).


:- use_module(bsyntaxtree, [find_typed_identifier_uses/2,get_texpr_id/2]).
% compute requirements for evaluating formula
determine_type_of_formula(TypedExpr,Class) :-
    find_typed_identifier_uses(TypedExpr,Identifiers),
    determine_type_of_formula2(Identifiers,Class).
determine_type_of_formula2([],requires_nothing).
determine_type_of_formula2(Ids,Res) :-
   member(TID,Ids), get_texpr_id(TID,ID),
   b_is_variable(ID,_),
   !, Res = requires_variables.
determine_type_of_formula2(Ids,Res) :-
   member(TID,Ids), get_texpr_id(TID,ID),
   b_is_constant(ID,_),
   !, Res = requires_constants.
determine_type_of_formula2(_,requires_nothing).
% only open, implicitly existentially quantified variables
% also, TO DO: check that insert_before_substitution_variables / $0 variables are properly treated

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% consistency check to spot some errors
do_machine_consistency_check :-
    specfile:animation_mode(b),            % TODO (dp,27.09.2012): this is just an ugly hack
    type_check_definitions,
    minor_mode_to_check,
    %debug:time(bmachine:all_operations_have_valid_read_modifies_infos).
    !,
    all_operations_have_valid_read_modifies_infos.
do_machine_consistency_check.

% In future, all formalisms should provide the relevent data
minor_mode_to_check :- specfile:animation_minor_mode(eventb),!.
minor_mode_to_check :- specfile:animation_minor_mode(_),!,fail.
minor_mode_to_check. % classical B

all_operations_have_valid_read_modifies_infos :-
    %findall(O,b_get_machine_operation(O,_,_,_,_),OpNames),
    %maplist(operation_has_valid_read_modifies_infos,OpNames).
    b_is_operation_name(OpName),
    operation_has_valid_read_modifies_infos(OpName),fail.
all_operations_have_valid_read_modifies_infos.

operation_has_valid_read_modifies_infos(OpName) :- % print(check(OpName)),nl,
    ( operation_has_valid_read_modifies_infos2(OpName) -> true
    ; otherwise ->
        add_failed_call_error(operation_has_valid_read_modifies_infos2(OpName))).
operation_has_valid_read_modifies_infos2(OpName) :-
    b_get_operation_normalized_read_write_info(OpName,Reads,Writes),
    maplist(check_read(OpName),Reads),
    %print(operation(OpName,reads(Reads),writes(Writes))),nl,
    maplist(check_write(OpName),Writes).


check_read(OpName,Id) :-
    ( (check_is_var(Id);check_is_constant(Id) ; check_is_operation(Id)) -> true
    ; otherwise ->
        ajoin(['Unknown identifier ',Id,' in "reads" information of operation:'],Msg),
        add_internal_error(bmachine,Msg,OpName,unknown)).
check_write(OpName,Id) :-
    ( check_is_var(Id) -> true
    ; otherwise ->
        ajoin(['Unknown variable ',Id,' in "modifies" information of operation:'],Msg),
        add_internal_error(bmachine,Msg,OpName,unknown)).

check_is_var(Id) :-
    b_is_variable(Id,_).
    %get_texpr_id(E,Id),
    %b_get_machine_variables(Vars),
    %memberchk(E,Vars).
check_is_constant(Id) :- b_is_constant(Id,_).
check_is_constant(Id) :- b_get_constant_represented_inside_global_set(Id,_).
   % b_get_machine_all_constants(Constants),
   % get_texpr_id(TId,Id),
   % memberchk(TId,Constants).
check_is_constant(Id) :- b_get_machine_set(Id).
check_is_constant(Id) :- b_get_named_machine_set(Id,_).
check_is_constant(Id) :-
    b_get_named_machine_set(_,Elems),member(Id,Elems).

check_is_operation(op(ID)) :-
    b_is_operation_name(ID),  % TO DO: check that other predicates can deal with op(Id) infos
    get_preference(allow_operation_calls_in_expr,true).

% provide access to number of constants, ....:

:- volatile b_machine_statistics/2.
:- dynamic b_machine_statistics/2.
:- public b_machine_statistics_calc/2.
b_machine_statistics_calc(files,Nr) :-
      b_get_all_used_filenames(L),length(L,Nr).
b_machine_statistics_calc(deferred_sets,Nr) :-
      get_section_from_current_machine(deferred_sets,L),length(L,Nr).
b_machine_statistics_calc(enumerated_sets,Nr) :-
      get_section_from_current_machine(enumerated_sets,L),length(L,Nr).
b_machine_statistics_calc(definitions,Nr) :-
      get_section_from_current_machine(definitions,L),length(L,Nr).
b_machine_statistics_calc(constants,Nr) :- b_get_machine_constants(L), length(L,Nr).
b_machine_statistics_calc(variables,Nr) :- b_get_machine_variables(L), length(L,Nr).
b_machine_statistics_calc(properties,Nr) :-
   b_get_properties_from_machine(C), conjunction_to_list(C,L),length(L,Nr).
b_machine_statistics_calc(invariants,Nr) :-
   b_get_invariant_from_machine(C), conjunction_to_list(C,L),length(L,Nr).
b_machine_statistics_calc(static_assertions,Nr) :-
   b_get_static_assertions_from_machine(L), length(L,Nr).
b_machine_statistics_calc(dynamic_assertions,Nr) :-
   b_get_dynamic_assertions_from_machine(L), length(L,Nr).
b_machine_statistics_calc(operations,Nr) :- findall(N,b_is_operation_name(N),L),length(L,Nr).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% default and initial machines
b_set_initial_machine :-
    initial_machine(Main,Machines),
    b_set_machine(Main,Machines,[]).

initial_machine(phonebook,[M]) :-
    M = abstract_machine(none,machine(none),machine_header(none,phonebook,[]),Body),
    Body = [sets(none, [deferred_set(none,'Name'),
                        deferred_set(none,'Code')]),
            definitions(none,[ScopeName,ScopeCode,TraceTest0]),
            variables(none, [identifier(none, db)]),
            invariant(none, member(none,
                                   identifier(none, db),
                                   partial_function(none,
                                                    identifier(none, 'Name'),
                                                    identifier(none, 'Code')))),
            initialisation(none,
                           assign(none, [identifier(none, db)], [empty_set(none)])),
            operations(none, [Add,Lookup,Update,Reset])],

    % definitions
    ScopeName = expression_definition(none,
                                      scope_Name,
                                      [],
                                      interval(none, integer(none, 1), integer(none, 3))),
    ScopeCode = expression_definition(none,
                                      scope_Code,
                                      [],
                                      interval(none, integer(none, 1), integer(none, 2))),
    TraceTest0 = expression_definition(none, trace_Test0, [], empty_sequence(none)),

    % add operation
    Add = operation(none,
                    identifier(none, add),
                    [],
                    [identifier(none,cc),identifier(none,nn)],
                    precondition(none, AddPre, AddAssign)),
    AddPre = conjunct(none,
                      conjunct(none,
                               member(none, identifier(none,nn), identifier(none,'Name')),
                               member(none, identifier(none,cc), identifier(none,'Code'))),
                      not_member(none,
                                 identifier(none,nn),
                                 domain(none, identifier(none,db)))),
    AddAssign = assign(none,
                       [identifier(none,db)],
                       [union(none,
                              identifier(none,db),
                              set_extension(none,
                                            [couple(none,[identifier(none,nn),
                                                          identifier(none,cc)])]))]),

    % lookup operation
    Lookup = operation(none,
                       identifier(none, lookup),
                       [identifier(none, cc)],
                       [identifier(none, nn)],
                       precondition(none, LookupPre, LookupAssign)),
    LookupPre = member(none, identifier(none,nn), domain(none,identifier(none,db))),
    LookupAssign = assign(none,
                          [identifier(none, cc)],
                          [function(none, identifier(none, db), identifier(none, nn))]),

    % update operation
    Update = operation(none,
                       identifier(none, update),
                       [],
                       [identifier(none,nn),identifier(none,cc)],
                       precondition(none, UpdatePre, UpdateAssign)),
    UpdatePre = conjunct(none,
                         conjunct(none,
                                  member(none, identifier(none,nn), identifier(none,'Name')),
                                  member(none, identifier(none,cc), identifier(none,'Code'))),
                         member(none,
                                identifier(none,nn),
                                domain(none, identifier(none,db)))),
    UpdateAssign = assign(none,
                          [function(none, identifier(none,db), identifier(none,nn))],
                          [identifier(none, cc)]),

    % reset operation
    Reset = operation(none,
                      identifier(none,reset),
                      [],
                      [],
                      precondition(none, ResetPre, ResetAssign)),
    ResetPre = member(none,
                      identifier(none, db),
                      total_function(none, identifier(none,'Name'), identifier(none,'Code'))),
    ResetAssign = assign(none,[identifier(none,db)],[empty_set(none)]).

empty_machine(empty_machine,[M]) :-
    M = abstract_machine(none,machine(none),machine_header(none,empty_machine,[]),Body),
    Body = [].
