:- module(dpllt_settings, [restart_policy/1,
                           static_smt_symmetry_breaking/1,
                           luby_restart_unit_length/1,
                           glucose_restart_recent_lbds_threshold/1,
                           glucose_restart_margin_ratio/1,
                           glucose_restart_trail_threshold/1,
                           glucose_restart_stack_avg_factor/1,
                           discard_clause_greater_lbd/1,
                           discard_learned_clauses_frequency_constant/1,
                           discard_learned_clauses_constant_factor/1,
                           decision_heuristic/1,
                           vsids_decay_value/1,
                           vsids_decay_frequency/1,
                           evsids_f_value/1]).

:- use_module(probsrc(module_information), [module_info/2]).

:- module_info(group, dpllt).
:- module_info(description,'This module provides settings for the DPLL(T) solver.').

static_smt_symmetry_breaking(true).
%% Random Restarts
% Available restart policies: luby, glucose.
restart_policy(glucose).
% The length of a unit in the Luby sequence.
luby_restart_unit_length(32).
% The amount of conflict clauses to be considered "recent" in Glucose restarts. Small value leads to more restarts.
glucose_restart_recent_lbds_threshold(50).
% Do restart if defined proportion of recent LBDs average is greater than total average. Large value leads to less restarts.
glucose_restart_margin_ratio(0.8).
% The amount of conflicts to enable possible emptying of the recent LBD queue to favor SAT instances too.
glucose_restart_trail_threshold(5000).
% For factor f, the last stack size needs to be larger than f * recentSizes.avg() to empty the recent LBD queue.
glucose_restart_stack_avg_factor(1.4).
%%

%% Reducing Learned Clauses
% Maximum LBD score for learned clauses.
% Each learned clause is propagated at least once independent of its actual LBD score.
discard_clause_greater_lbd(5).
% Remove half of the learned clauses every "frequency_constant + constant_factor * x" conflicts.
discard_learned_clauses_frequency_constant(20000).
discard_learned_clauses_constant_factor(500).
%%

%% Decision Heuristic
% Available decision heuristics:  vsids, evsids.
decision_heuristic(evsids).
% Decision score s is updated by s/v.
vsids_decay_value(2).
% Amount of found conflicts when to decay scores.
vsids_decay_frequency(128).
% A value between 0 and 1. Decision score is updated by s + f^(-k).
evsids_f_value(0.9).
%%
