
%:- module(external_functions_reals,['STRING_TO_REAL'/3, 'RADD'/3,'RSUB'/3,'RMUL'/3,'RDIV'/3,'RSIN'/3]).


% -------------------------------

expects_waitflag('STRING_TO_REAL').
:- public 'STRING_TO_REAL'/3.
:- block 'STRING_TO_REAL'(-,?,?).

:- use_module(kernel_reals,[construct_real/2]).
% allows to call construct_real/2; also works for numbers without decimal point
'STRING_TO_REAL'(string(A),Result,_) :-
   construct_real(A,Result).

% -------------------------------

:- use_module(kernel_reals,[real_addition_wf/4, real_subtraction_wf/4,
                            real_multiplication_wf/4, real_division_wf/4, real_power_of_wf/5,
                            real_round_wf/3,
                            real_unop_wf/4, real_binop_wf/5,
                            real_comp_wf/5]).

expects_waitflag('RADD').
external_fun_type('RADD',[],[real,real,real]).
:- public 'RADD'/4.
'RADD'(RX,RY,RR,WF) :-
  real_addition_wf(RX,RY,RR,WF).

expects_waitflag('RSUB').
external_fun_type('RSUB',[],[real,real,real]).
:- public 'RSUB'/4.
'RSUB'(RX,RY,RR,WF) :-
  real_subtraction_wf(RX,RY,RR,WF).

expects_waitflag('RMUL').
external_fun_type('RMUL',[],[real,real,real]).
:- public 'RMUL'/4.
'RMUL'(RX,RY,RR,WF) :-
  real_multiplication_wf(RX,RY,RR,WF).

expects_waitflag('RDIV').
external_fun_type('RDIV',[],[real,real,real]).
:- public 'RDIV'/4.
'RDIV'(RX,RY,RR,WF) :-
  real_division_wf(RX,RY,RR,WF).

external_fun_type('RPI',[],[real]).
:- public 'RPI'/1.
'RPI'(term(floating(R))) :- R is pi.

external_fun_type('RZERO',[],[real]).
:- public 'RZERO'/1.
'RZERO'(term(floating(R))) :- R = 0.0.

external_fun_type('RONE',[],[real]).
:- public 'RONE'/1.
'RONE'(term(floating(R))) :- R = 1.0.

external_fun_type('REULER',[],[real]).
:- public 'REULER'/1.
'REULER'(term(floating(R))) :- R is exp(1.0).

% ---- unary operators

% --- Trigonometric

external_fun_type('RSIN',[],[real,real]).
expects_waitflag('RSIN').
:- public 'RSIN'/3.
:- block 'RSIN'(-,?,?).
'RSIN'(X,R,WF) :-
  real_unop_wf('sin',X,R,WF).

external_fun_type('RCOS',[],[real,real]).
expects_waitflag('RCOS').
:- public 'RCOS'/3.
:- block 'RCOS'(-,?,?).
'RCOS'(X,R,WF) :-
  real_unop_wf('cos',X,R,WF).

external_fun_type('RTAN',[],[real,real]).
expects_waitflag('RTAN').
:- public 'RTAN'/3.
:- block 'RTAN'(-,?,?).
'RTAN'(X,R,WF) :-
  real_unop_wf('tan',X,R,WF).

external_fun_type('RCOT',[],[real,real]).
expects_waitflag('RCOT').
:- public 'RCOT'/3.
:- block 'RCOT'(-,?,?).
'RCOT'(X,R,WF) :-
  real_unop_wf('cot',X,R,WF).

external_fun_type('RSINH',[],[real,real]).
expects_waitflag('RSINH').
:- public 'RSINH'/3.
:- block 'RSINH'(-,?,?).
'RSINH'(X,R,WF) :-
  real_unop_wf('sinh',X,R,WF).

external_fun_type('RCOSH',[],[real,real]).
expects_waitflag('RCOSH').
:- public 'RCOSH'/3.
:- block 'RCOSH'(-,?,?).
'RCOSH'(X,R,WF) :-
  real_unop_wf('cosh',X,R,WF).

external_fun_type('RTANH',[],[real,real]).
expects_waitflag('RTANH').
:- public 'RTANH'/3.
:- block 'RTANH'(-,?,?).
'RTANH'(X,R,WF) :-
  real_unop_wf('tanh',X,R,WF).

external_fun_type('RCOTH',[],[real,real]).
expects_waitflag('RCOTH').
:- public 'RCOTH'/3.
:- block 'RCOTH'(-,?,?).
'RCOTH'(X,R,WF) :-
  real_unop_wf('coth',X,R,WF).

external_fun_type('RASIN',[],[real,real]).
expects_waitflag('RASIN').
:- public 'RASIN'/3.
:- block 'RASIN'(-,?,?).
'RASIN'(X,R,WF) :-
  real_unop_wf('asin',X,R,WF).

external_fun_type('RACOS',[],[real,real]).
expects_waitflag('RACOS').
:- public 'RACOS'/3.
:- block 'RACOS'(-,?,?).
'RACOS'(X,R,WF) :-
  real_unop_wf('acos',X,R,WF).

external_fun_type('RATAN',[],[real,real]).
expects_waitflag('RATAN').
:- public 'RATAN'/3.
:- block 'RATAN'(-,?,?).
'RATAN'(X,R,WF) :-
  real_unop_wf('atan',X,R,WF).

external_fun_type('RACOT',[],[real,real]).
expects_waitflag('RACOT').
:- public 'RACOT'/3.
:- block 'RACOT'(-,?,?).
'RACOT'(X,R,WF) :-
  real_unop_wf('acot',X,R,WF).

external_fun_type('RASINH',[],[real,real]).
expects_waitflag('RASINH').
:- public 'RASINH'/3.
:- block 'RASINH'(-,?,?).
'RASINH'(X,R,WF) :-
  real_unop_wf('asinh',X,R,WF).

external_fun_type('RACOSH',[],[real,real]).
expects_waitflag('RACOSH').
:- public 'RACOSH'/3.
:- block 'RACOSH'(-,?,?).
'RACOSH'(X,R,WF) :-
  real_unop_wf('acosh',X,R,WF).

external_fun_type('RATANH',[],[real,real]).
expects_waitflag('RATANH').
:- public 'RATANH'/3.
:- block 'RATANH'(-,?,?).
'RATANH'(X,R,WF) :-
  real_unop_wf('atanh',X,R,WF).

external_fun_type('RACOTH',[],[real,real]).
expects_waitflag('RACOTH').
:- public 'RACOTH'/3.
:- block 'RACOTH'(-,?,?).
'RACOTH'(X,R,WF) :-
  real_unop_wf('acoth',X,R,WF).

% -----------------------

external_fun_type('REXP',[],[real,real]).
expects_waitflag('REXP').
:- public 'REXP'/3.
:- block 'REXP'(-,?,?).
'REXP'(X,R,WF) :-
  real_unop_wf('exp',X,R,WF).

external_fun_type('RLOGe',[],[real,real]).
expects_waitflag('RLOGe').
:- public 'RLOGe'/3.
:- block 'RLOGe'(-,?,?).
'RLOGe'(X,R,WF) :-
  real_unop_wf('log',X,R,WF).

external_fun_type('RSQRT',[],[real,real]).
expects_waitflag('RSQRT').
:- public 'RSQRT'/3.
:- block 'RSQRT'(-,?,?).
'RSQRT'(X,R,WF) :-
  real_unop_wf('sqrt',X,R,WF).

external_fun_type('RABS',[],[real,real]).
expects_waitflag('RABS').
:- public 'RABS'/3.
:- block 'RABS'(-,?,?).
'RABS'(X,R,WF) :-
  real_unop_wf('abs',X,R,WF).

external_fun_type('ROUND',[],[real,real]).
expects_waitflag('ROUND').
:- public 'ROUND'/3.
:- block 'ROUND'(-,?,?).
'ROUND'(X,R,WF) :-
  real_round_wf(X,R,WF).

external_fun_type('RSIGN',[],[real,real]).
expects_waitflag('RSIGN').
:- public 'RSIGN'/3.
:- block 'RSIGN'(-,?,?).
'RSIGN'(X,R,WF) :-
  real_unop_wf('sign',X,R,WF).

external_fun_type('RINTEGER',[],[real,real]).
expects_waitflag('RINTEGER').
:- public 'RINTEGER'/3.
:- block 'RINTEGER'(-,?,?).
'RINTEGER'(X,R,WF) :-
  real_unop_wf('float_integer_part',X,R,WF).

external_fun_type('RFRACTION',[],[real,real]).
expects_waitflag('RFRACTION').
:- public 'RFRACTION'/3.
:- block 'RFRACTION'(-,?,?).
'RFRACTION'(X,R,WF) :-
  real_unop_wf('float_fractional_part',X,R,WF).

% ---- other binary operators
expects_waitflag('RMAX').
external_fun_type('RMAX',[],[real,real,real]).
:- public 'RMAX'/4.
'RMAX'(RX,RY,RR,WF) :-
  real_binop_wf(max,RX,RY,RR,WF).

expects_waitflag('RMIN').
external_fun_type('RMIN',[],[real,real,real]).
:- public 'RMIN'/4.
'RMIN'(RX,RY,RR,WF) :-
  real_binop_wf(min,RX,RY,RR,WF).

expects_waitflag('RPOW').
external_fun_type('RPOW',[],[real,real,real]).
expects_spaninfo('RPOW').
:- public 'RPOW'/5.
'RPOW'(RX,RY,RR,Span,WF) :-
  real_power_of_wf(RX,RY,RR,Span,WF).

expects_waitflag('RLOG').
external_fun_type('RLOG',[],[real,real,real]).
:- public 'RLOG'/4.
'RLOG'(RX,RY,RR,WF) :-
  real_binop_wf(log,RX,RY,RR,WF).

% ---- other binary predicates

expects_waitflag('RLT').
external_fun_type('RLT',[],[real,real,boolean]).
:- public 'RLT'/4.
'RLT'(RX,RY,RR,WF) :-
  real_comp_wf('<',RX,RY,RR,WF).

expects_waitflag('REQ').
external_fun_type('REQ',[],[real,real,boolean]).
:- public 'REQ'/4.
'REQ'(RX,RY,RR,WF) :-
  real_comp_wf('=:=',RX,RY,RR,WF).

expects_waitflag('RNEQ').
external_fun_type('RNEQ',[],[real,real,boolean]).
:- public 'RNEQ'/4.
'RNEQ'(RX,RY,RR,WF) :-
  real_comp_wf('=/=',RX,RY,RR,WF).

expects_waitflag('RLEQ').
external_fun_type('RLEQ',[],[real,real,boolean]).
:- public 'RLEQ'/4.
'RLEQ'(RX,RY,RR,WF) :-
  real_comp_wf('=<',RX,RY,RR,WF).

expects_waitflag('RGT').
external_fun_type('RGT',[],[real,real,boolean]).
:- public 'RGT'/4.
'RGT'(RY,RX,RR,WF) :-'RLT'(RX,RY,RR,WF).
expects_waitflag('RGEQ').
external_fun_type('RGEQ',[],[real,real,boolean]).
:- public 'RGEQ'/4.
'RGEQ'(RY,RX,RR,WF) :-'RLEQ'(RX,RY,RR,WF).


