% (c) 2016-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(performance_messages,
  [
   performance_monitoring_on/0,
   perfmessage/1, perfmessage/2, perfmessage/3,
   cond_perfmessage/2,
   perfmessagecall/3, perfmessagecall/4,
   perf_format/2,
   toggle_perfmessages/0
   ]).

:- meta_predicate perfmessagecall(*,0,*).
:- meta_predicate perfmessagecall(*,*,0,*).

:- use_module(module_information,[module_info/2]).
:- module_info(group,profiling).
:- module_info(description,'This module allows to generate performance messages.').

:- use_module(preferences).
:- use_module(error_manager).
:- use_module(tools_printing,[format_with_colour/4]).

% -------------------------------------------

:- dynamic performance_monitoring_on/0. % should we use a preference
%performance_monitoring_on.

performance_monitoring_on :- get_preference(performance_monitoring_on,true).

toggle_perfmessages :- print('TURNING PERF-MESSAGES '),
  (performance_monitoring_on
   -> set_preference(performance_monitoring_on,false), print('OFF')
    ; set_preference(performance_monitoring_on,true), print('ON')),nl.

perf_format(M,Args) :- preference(performance_monitoring_on,true),!,
   format_with_colour(user_output,[blue],'==PERF-MESSAGE==> ',[]),
   format_with_colour(user_output,[blue],M,Args).
perf_format(_,_).

perfmessage(M) :- preference(performance_monitoring_on,true),!,
   format_with_colour(user_output,[blue],'==PERF-MESSAGE==> ~w~n',[M]).
perfmessage(_).

:- use_module(library(lists),[maplist/2]).
% conditional perfmessage
cond_perfmessage(Prefs,M) :- preference(performance_monitoring_on,true),
    maplist(prefset,Prefs),
    !,
   format_with_colour(user_output,[blue],'==PERF-MESSAGE==> ~w~n',[M]).
cond_perfmessage(_,_).

prefset(P/V) :- get_preference(P,V).

perfmessage(M,Location) :- preference(performance_monitoring_on,true),!,
   add_message(performance_message,'==PERF-MESSAGE==> ',M,Location).
perfmessage(_,_).

perfmessage(Category,M,Location) :- preference(performance_monitoring_on,true),
   show_message(Category),
   !,
   add_message(performance_message,'==PERF-MESSAGE==> ',M,Location).
perfmessage(_,_,_).


perfmessagecall(M,Call,Location) :- preference(performance_monitoring_on,true),!,
   add_message(performance_message,'==PERF-MESSAGE==> ',M,Location),
   call(Call),nl(user_output).
perfmessagecall(_,_,_).


perfmessagecall(Category,M,Call,Location) :- preference(performance_monitoring_on,true),
   show_message(Category),
   !,
   add_message(performance_message,'==PERF-MESSAGE==> ',M,Location),
   call(Call),nl(user_output).
perfmessagecall(_,_,_,_).

% decide based upon category and current preferences whether to show the message or not
show_message(reify) :- !,
   get_preference(data_validation_mode,false).
show_message(_Category).
