% (c) 2009 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(schemavars,[schema_vars/3,schema_vars_decl/3]).

:- use_module(library(lists)).

:- use_module(probsrc(tools)).

:- use_module(z_tools).
:- use_module(zenvironment).

:- use_module(probsrc(module_information)).
:- module_info(group,proz).
:- module_info(description,'This module contains code to extract the variables declared and visible in a schema declaration').

%********************************************************************************
% extract vars from declaration

schema_vars(sref(Name,Deco,_,Renamings),Env,Vars) :-
    zlookup_schema(Env,Name,_,Expr),
    schema_vars(Expr,Env,Vars1),
    decorate_vars(Vars1,Deco,DVars),
    rename(DVars,Renamings,Vars).

schema_vars(sand(A,B),Env,Vars) :- schema_vars_logbin(A,B,Env,Vars).
schema_vars(sor(A,B),Env,Vars) :- schema_vars_logbin(A,B,Env,Vars).
schema_vars(simplies(A,B),Env,Vars) :- schema_vars_logbin(A,B,Env,Vars).
schema_vars(sequiv(A,B),Env,Vars) :- schema_vars_logbin(A,B,Env,Vars).
schema_vars(snot(A),Env,Vars) :- schema_vars(A,Env,Vars).
schema_vars(sforall(B,S),Env,Vars) :- schema_vars_quantifier(B,S,Env,Vars).
schema_vars(sexists(B,S),Env,Vars) :- schema_vars_quantifier(B,S,Env,Vars).
schema_vars(sexists1(B,S),Env,Vars) :- schema_vars_quantifier(B,S,Env,Vars).
schema_vars(hide(S,Hide),Env,Vars) :-
    schema_vars(S,Env,All),remove_all(All,Hide,Vars).

schema_vars(fatsemi(A,B),Env,Vars) :-
    schema_vars(A,Env,AVars),
    findall(name(Name,'\''),member(name(Name,'\''),AVars),InterA),
    remove_all(AVars,InterA,Aall),

    schema_vars(B,Env,BVars),
    findall(name(Name,''),member(name(Name,''),BVars),InterB),
    remove_all(BVars,InterB,Ball),

    append_not_duplicate(Aall,Ball,Vars).
schema_vars(project(_,B),Env,Vars) :- schema_vars(B,Env,Vars).
schema_vars(pre(Schema),Env,Vars) :-
    schema_vars(Schema,Env,AllVars),
    findall(name(Name,Deco),
            (member(name(Name,Deco),AllVars), \+ member(Deco,['\'','!'])),
            Vars).
schema_vars(pipe(A,B),Env,Vars) :-
    schema_vars(A,Env,AVars),
    schema_vars(B,Env,BVars),
    findall(name(Name,'!'),
            (member(name(Name,'!'),AVars),member(name(Name,'?'),BVars)),
            Out),
    findall(name(Name,'?'),member(name(Name,'!'),Out),In),
    remove_all(AVars,Out,AwoO),
    remove_all(BVars,In,BwoI),
    append_not_duplicate(AwoO,BwoI,Vars).

schema_vars(text(Body),Env,Vars) :- schema_vars(Body,Env,Vars).
schema_vars(body(Decl,_),Env,Vars) :- schema_vars_decl(Decl,Env,Vars).

% Helper predicates

% vars of a schema binary logical expression
schema_vars_logbin(A,B,Env,Vars) :-
    schema_vars(A,Env,AVars),
    schema_vars(B,Env,BVars),
    append_not_duplicate(AVars,BVars,Vars).

% vars of a schema quantifier expression
schema_vars_quantifier(Schema,Body,Env,Vars) :-
    schema_vars(Body,Env,Hide),
    schema_vars(hide(Schema,Hide),Env,Vars).

% vars of the declaration part of a body
schema_vars_decl(Decls,Env,Vars) :-
    maplist(schema_vars_decl2(Env),Decls,LVars),
    append(LVars,Vars).
schema_vars_decl2(_,decl(Vars,_),Vars).
schema_vars_decl2(Env,sdecl(Sexpr),Vars) :- schema_vars(Sexpr,Env,Vars).

% renames variables
rename([],_,[]).
rename([NameA|Rest],Renamings,[NameB|RenamedRest]) :-
    (member(rename(NameB,NameA),Renamings) ->
        true;
        NameA = NameB),
    rename(Rest,Renamings,RenamedRest).
