% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

% :- prolog_flag(source_info,_,on).
% comment in if you want source location, e.g., for exceptions,...
% you can also comment in a line in remove_debugging_calls in debugging_calls_te.pl to see location of nl/0
% with ?- trace, leash(off). one can creep without entering return in the debugger; with leash([redo]) the debugger only stops at the Redo-Port. Use @<RETURN> to enter Prolog commands in the debugger.

:- module(prob_cli, [go_cli/0, go_cli/2, reset_cli/0, recognised_cli_option/4, recognised_option/2, go_proxy/0]).

user:portray_message(informational, _).

% patch for SICStus 4.3.3 performance issue
% sprm_14972_patch.pl BEGIN
:- if((current_prolog_flag(dialect, sicstus),
      current_prolog_flag(version_data, sicstus(4,3,3,_,_)))).
prolog:wf_call_like_arg(A, B, C, D, E, F, G, H, I, _) :-
  prolog:wellformed_body_iso(A, B, C, D, E, F, G, H, I, quiet),
  !.
prolog:wf_call_like_arg(A, B, C, A, D, _, E, _, _, _) :-
  F=E:A,
  prolog:condense_layout(B, G),
  prolog:comp_layout2(B, G, B, H),
  C=call(F),
  prolog:comp_layout1(B, H, D).
:- endif.
% sprm_14972_patch.pl END

%% sp_4_1_3_nested_exception_patch.pl BEGIN
%% Compile this to patch a bug in 4.1.3 affecting nested exception handlers.
%% This source code is a no-op on other versions of sicstus than 4.1.3.
:- if(current_prolog_flag(dialect,sicstus)).
:- if(current_prolog_flag(version_data,sicstus(4,1,3,0,_))).

prolog:(matching_exception(internal, Ball, Templ) :- !,
        Templ = Ball).
prolog:(matching_exception(prolog, '$reserved'(_), _) :- !,
        fail).
prolog:(matching_exception(prolog, time_out, _) :- !,
        fail).
prolog:(matching_exception(_, Ball, Template) :-
        compound(Ball),
        sysball_iso_equiv(Ball,Iso1Exc), !,
        Template = error(Iso1Exc, Ball)).
prolog:(matching_exception(_, Ball, Ball)).

:- endif.                       % sicstus 4.1.3
:- endif.                       % sicstus
%% sp_4_1_3_nested_exception_patch.pl END

%:- include('self_check_off.pl').

% use module_info/3 because we are not in a module here
:- use_module(module_information).
:- module_info(prob_cli,group,infrastructure).
:- module_info(prob_cli,description,'ProB start file in cli mode.').

%:- use_module(debugging_calls).
%:- register_debugging_call pp_mnf(*), pp_cll(*), mnf(*), mnf(-,*), det_call(*).
%:- disable_debugging_calls.

:- use_module(prob_startup, [startup_prob/0]).
%:- use_module(pathes,[set_search_pathes/0]). % called first to set_compile_time_search_pathes
:- use_module(tools,[string_concatenate/3,arg_is_number/2]).
:- use_module(tools_printing,[print_error/1]).


% :- use_module(extension('timer/assert_profiler.pl')).


:- use_module(tcltk_interface).
%:- compile(gui_tcltk).

:- use_module(eclipse_interface).
:- use_module(prob2_interface,[start_animation/0, is_initialised_state/1, reset_animator/0,
                               set_eclipse_preference/2, update_preferences_from_spec/1,
                               load_cspm_spec_from_cspm_file/1]).


:- use_module(prob_socketserver,[start_prob_socketserver/2]).

:- use_module(coverage_statistics,[compute_the_coverage/5]).
:- use_module(value_persistance, [set_storage_directory/1,unset_storage_directory/0]).

do_one_gui_event.
tk_do_all_events.

:- dynamic junit_mode/1.

:- use_module(b_trace_checking,[check_default_trace_for_specfile/1, tcltk_check_state_sequence_from_file/1,
                                tcltk_save_history_as_trace_file/2]).
%:- use_module(cspParserHs,'cia/haskell/java_api/cspParserHs.pl',
%               [initCspParser/0]).

:- use_module(library(file_systems),[file_exists/1]).
:- use_module(debug).
:- use_module(preferences).
:- use_module(probltlsrc(ltl),[ltl_check_assertions/2,ltl_model_check/4]).
:- use_module(probltlsrc(ctl),[ctl_model_check/4]).
:- use_module(test_typechecker,[run_typecheck_testcase/2]).

:- use_module(basic_unit_tests). % basic unit tests
:- use_module(self_check,[disable_interaction_on_errors/0]).
:- use_module(error_manager).

:- use_module(extension('zmq/master/master')).
:- use_module(extension('zmq/worker/worker')).
:- use_module(extension('ltsmin/ltsmin')).

:- use_module(self_check,[perform_self_check/2,turn_off_run_time_type_checks/0,turn_on_run_time_type_checks/0]).

:- use_module(library(system)).
:- use_module(library(codesio)).

:- use_module(logger).
:- use_module(specfile).
:- use_module(bmachine,[b_write_machine_representation_to_file/2,full_b_machine/1, b_write_eventb_machine_to_classicalb_to_file/1]).

:- use_module(xtl_interface,[set_cspm_main_process/1]).
:- use_module(meta_interface,[is_dot_command/1, call_dot_command/3,
                              is_dot_command_for_expr/1,call_dot_command_for_expr/4,
                              command_description/3]).
:- use_module('kodkod/kodkod_test').
:- use_module(predicate_analysis,[test_predicate_analysis/0]).
:- use_module(b_show_history,[write_history_to_file/2,write_values_to_file/1,write_all_values_to_dir/1,write_history_to_user_output/1]).
:- use_module(sap, [write_all_deadlocking_paths_to_xml/1, test_generation_by_xml_description/1]).

:- use_module(smtlib_solver(smtlib2_cli)).
:- use_module(disprover_test_runner, [run_disprover_on_all_pos/1, load_po_file/1,print_disprover_stats/0, set_disprover_timeout/1, set_disprover_options/1, reset_disprover_timeout/0]).
:- use_module(latex_processor, [process_latex_file/2]).


:- meta_predicate if_option_set(-,0).
:- meta_predicate if_option_set(-,0,0).
:- meta_predicate if_options_set(-,0).
:- meta_predicate if_option_set_loaded(-,-,0).
:- meta_predicate ifm_option_set(-,0).
:- meta_predicate ifm_option_set(-,-,0).
:- meta_predicate ifm_option_set_loaded(-,-,0).

start_probcli_timer(timer(T1,WT1)) :-
                   statistics(runtime,[T1,_]),
                   statistics(walltime,[WT1,_]).
stop_probcli_debug_timer(Timer,Msg) :- (debug_mode(on) -> stop_probcli_timer(Timer,Msg) ; true).
stop_probcli_timer(timer(T1,WT1),Msg) :- stop_probcli_timer(timer(T1,WT1),Msg,_).
stop_probcli_timer(timer(T1,WT1),Msg,WTotTime) :-
        statistics(runtime,[T2,_]), TotTime is T2-T1,
        statistics(walltime,[WT2,_]), WTotTime is WT2-WT1,
        format('~w ~w ms walltime (~w ms runtime), since start: ~w ms~n',[Msg,WTotTime,TotTime,WT2]),
        !.
stop_probcli_timer(Timer,Msg,_) :- add_internal_error('Illegal timer call: ',stop_probcli_timer(Timer,Msg)).
print_total_probcli_timer :-
        statistics(runtime,[T2,_]),
        statistics(walltime,[WT2,_]),
        format('Since start of probcli: ~w ms walltime (~w ms runtime)~n',[WT2,T2]).
get_probcli_elapsed_walltime(timer(_,WT1),WTotTime) :-
        statistics(walltime,[WT2,_]), WTotTime is WT2-WT1.

:- use_module(tools_meta,[safe_time_out/3]).
:- meta_predicate timeout_call(0,-).
timeout_call(Call,NOW) :- option(timeout(TO)),!,
   statistics(runtime,[T1,_]),
   safe_time_out(Call,TO,Res),
   statistics(runtime,[T2,_]), Runtime is T2-T1,
   printsilent('Runtime: '), printsilent(Runtime), printsilent(' ms'),nls,
   (Res=time_out -> print('*** Timeout occurred: '), print(TO),nl,
                    print('*** Call: '), print(Call),nl,
                    nl,
                    writeln_log(timeout_occurred(NOW,Call))
                 ;  true).
timeout_call(Call,_NOW) :-
   statistics(runtime,[T1,_]),
   call(Call),
   statistics(runtime,[T2,_]), Runtime is T2-T1,
   printsilent('Runtime: '), printsilent(Runtime), printsilent(' ms'),nls.

:- use_module(junit_tests,[set_junit_dir/1, print_junit/2, create_junit_result/6]).
set_junit_mode(X) :- set_junit_dir(X), statistics(runtime,[Start,_]), assert(junit_mode(Start)).

go_proxy :-
    catch( go_cli(['-s','8888'],[proxy]), halt(ExitCode),
           ( nl,write('CLI halt prevented, exit code '),write(ExitCode),nl) ).

go_cli :-
    prolog_flag(argv,ArgV),
    go_cli_with_junit_check(ArgV).

initialise_cli :- counter_init,
  new_counter(cli_execute_inits),
  new_counter(cli_errors), new_counter(cli_warnings),
  new_counter(cli_expected_errors).

% called from test_runner.pl:
reset_cli :-
  tcltk_turn_debugging_off,
  reset_errors,
  reset_expected_error_occurred,
  reset_optional_errors_or_warnings,
  reset_counter(cli_errors), reset_counter(cli_warnings),
  reset_counter(cli_expected_errors),
  preferences:reset_to_defaults,
  reset_logger,
  unset_storage_directory,
  % now done via event handling: clear_dynamic_predicates_for_POR,
  %     retractall(refinement_checker: generated_predeterministic_refinement(_,_,_)),
  %     retractall(refinement_checker: determinism_check_div_found(_)),
  %     reset_cspm_main_process,
  retractall(accumulated_infos(_,_,_)),
  retractall(individual_file_infos(_,_,_)),
  (file_loaded(_)
   -> clear_loaded_machines,
      retractall(file_loaded(_,_)),
      retractall(loaded_main_file(_,_))
   ;  true).

go_cli_with_junit_check(ArgV) :-
    catch( go_cli(ArgV,[junit]),
           halt(ExitCode),
           ( ExitCode = 0 ->
               true
           ; junit_mode(S) ->
               statistics(runtime,[E,_]), T is E - S,
	       create_junit_result(ArgV,'Integration Tests','Integration_Test',T,error([ExitCode]),Result),
               print_junit([Result],'Integration_Test')
           ; throw(halt(ExitCode)))).

go_cli(ArgV,Context) :-
     (on_exception(Exc,
         go_cli2(ArgV),
         process_exception(Exc,Context)
         )
     -> true
     ; print_error('INTERNAL ERROR OCCURRED !'),nl,
       print_error(' (go_cli failed)'),nl, halt_exception(1)).

process_exception(halt(A),_) :- !, throw(halt(A)).
process_exception(user_interrupt_signal,Context) :- !,
   %add_error(probcli,'probcli interrupted by user (CTRL-C)').
   format_with_colour(user_error,[red],'~nprobcli interrupted by user (CTRL-C)~n',[]),
   (member(test_runner,Context)
     -> throw(user_interrupt_signal) % will be caught by test_runner
     ;  error_occurred_with_msg(user_interrupt_signal,'probcli interrupted by user (CTRL-C)')
   ).
process_exception(Exc,_) :- error_occurred(internal_error(exception(Exc))),fail.

no_command_issued :- \+ command_option(_).
command_option(X) :- option(X), \+ not_command_option(X).
not_command_option(verbose).
not_command_option(very_verbose).
not_command_option(profiling_on).
not_command_option(set_pref(_,_)).
not_command_option(set_card(_,_)).
not_command_option(set_argv(_)).
not_command_option(silent).

probcli_startup :-
    %print_total_probcli_timer,
    startup_prob,
    %myheap_init,
    initialise_cli,
    init_eclipse_preferences. % startup_prob will already call init_preferences; TO DO: simplify

 :- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(prob_logging_mode,true)).
cli_set_options(ArgV,RemArgV) :-
    cli_init_options(['-ll'|ArgV],RemArgV). %% adds -ll to have an automatically logging probcli to /tmp/prob_cli_debug.log
:- else.
cli_set_options(ArgV,RemArgV) :- cli_init_options(ArgV,RemArgV).
:- endif.

%:- use_module(extension('myheap/myheap')).
go_cli2(ArgV) :- %print(probcli_startup),nl_time,
    probcli_startup,
    external_functions:reset_argv,
    cli_set_options(ArgV,RemArgV),
    maplist(prob_cli:check_remaining_arg_looks_like_file,RemArgV),
    %% cli_set_options(['-vv','-version'|ArgV],RemArgV),  %% comment in to have an automatically verbose probcli
    !,
    ( RemArgV=[File],no_command_issued,
      get_filename_extension(File,Ext),
      \+ do_not_execute_automatically(Ext)
       % then assume we want to do -execute_all:
      -> assert_option(execute(2147483647,false,current_state))
       ; true),
    if_option_set(test_mode,set_random_seed_to_deterministic_start_seed, set_new_random_seed),
    if_option_set(verbose,
                  verbose, %% set_verbose_mode -> does tcltk_turn_debugging_on(19),
                  tcltk_turn_debugging_off),
    if_option_set(very_verbose,
                  very_verbose), %tcltk_turn_debugging_on(5)),
    if_option_set(profiling_on,profiling_on),
    debug_print(9,'Command Line Arguments: '),debug_println(9,ArgV),
    if_option_set(very_verbose,
                  print_options),
    debug_print(6,'Command Line File Args: '),debug_println(6,RemArgV),
    if_option_set(cache_storage(StorageDir), set_storage_directory(StorageDir)),
    if_option_set(parsercp(ParserLoc),
                  assert(parsercall:parser_location(ParserLoc))),
    generate_time_stamp(Datime,NOW),
    if_option_set(log(LogF,Mode),
                  cli_start_logging(LogF,Mode,NOW,Datime,RemArgV)),
    if_option_set(runtimechecking,
                  turn_on_run_time_type_checks,
                  turn_off_run_time_type_checks),
    if_option_set(junit(JUX),
                  set_junit_mode(JUX)),
    set_prefs,
    set_optional_errors,
    cli_show_help(ArgV,RemArgV),
    if_option_set(set_argv(ArgVStr),
                  set_argv(ArgVStr)),
    if_option_set(selfcheck(_),
                  cli_start_selfcheck),
    if_option_set(typechecker_test(Filename),
                  (run_typecheck_testcase(Filename,typesok) -> halt_prob(NOW,0); halt_prob(NOW,1))),
    if_option_set(print_version, print_version),
    if_option_set(print_short_version, print_short_version),
    if_option_set(check_java_version, check_java_version),
    if_option_set(very_verbose,
                  preferences:print_preferences),
    if_option_set(zmq_worker(Identifier), zmq_start_worker(Identifier, NOW)),
    %if_option_set(zmq_worker2(MasterIP, Port, ProxyID, Logfile), zmq_start_worker(MasterIP,Port,ProxyID,Logfile,NOW)),
    % process remaining arguments as files to load
    debug_println(6,processing(RemArgV)),
    cli_load_files(RemArgV,NOW), % all CLI arguments which are not understood are supposed to be files to be treated
    debug_println(6,finished_loading(RemArgV)),
    if_option_set(socket(Port,Loopback),
                  cli_start_socketserver(Port,Loopback)),
    % check_all_expected_errors_occurred(NOW),  % is now checked for each file; socket_server should not generate errors ?
    debug_println(20,'% ProB Finished'),
    cli_print_junit_results(ArgV),
    stop_xml_probcli_run(NOW).

% finish logxml file by writing total number of errors and warnings
stop_xml_probcli_run(NOW) :-
    get_counter(cli_errors,CErrs), get_counter(cli_warnings,CWarns), get_counter(cli_expected_errors,EErrs),
    writeln_log_time(prob_finished(NOW,CErrs,CWarns)),
    (EErrs>0
     -> write_xml_element_to_log('probcli-errors',[errors/CErrs,warnings/CWarns,expected_errors/EErrs])
      ; write_xml_element_to_log('probcli-errors',[errors/CErrs,warnings/CWarns])
     ),
    ((CErrs>0 ; CWarns>0) -> format(user_error,'! Total Errors: ~w, Warnings:~w~n',[CErrs,CWarns]) ; true),
    stop_xml_group_in_log('probcli-run'). % Generating this tag means probcli ran to completion without segfault,...

% check if a remaining argument looks suspicious (e.g., like an unknown command)
check_remaining_arg_looks_like_file(File) :- atom_codes(File,Codes), check_remaining_arg_codes(Codes,File).
check_remaining_arg_codes([45|_],Arg) :- !,
   add_message(probcli,'Command-line argument looks like an unknown probcli command: ',Arg).
check_remaining_arg_codes([D|T],Arg) :- D >= 0'0, D =< 0'9,
   nonmember(0'/,T), % detect things like 01_Jan/a.mch
   !,
   add_message(probcli,'Command-line argument looks like a number: ',Arg).
check_remaining_arg_codes(_,_).

:- use_module(tools,[get_options/5]).
cli_init_options(ArgV,RemArgV) :- %print(argv(ArgV)),nl,
    append(ProBArgV,['--'|BArgV],ArgV),!, % pass arguments after -- to B via external_functions
    cli_init_options2(ProBArgV,RemArgV),
    debug_println(20,set_argv_from_list(BArgV)),
    external_functions:set_argv_from_list(BArgV).
cli_init_options(ArgV,RemArgV) :- cli_init_options2(ArgV,RemArgV).
cli_init_options2(ArgV,RemArgV) :-
    reset_options,
    %%assert(option(log('/tmp/ProBLog.log'))), print('LOGGING'),nl, %% coment in to build a version of probcli that automatically logs
    ( get_options(ArgV,recognised_cli_option,Options,RemArgV,throw(halt(1))) ->
        assert_all_options(Options)
    ; otherwise ->
        print_error(get_options_failed(ArgV)),definite_error_occurred).
cli_show_help(ArgV,RemArgV) :-
    ( (option(help) ; ArgV=[]) ->
        print_help, (RemArgV=[] -> halt_exception ; true)
    ; true).
cli_start_logging(F,Mode,NOW,Datime,RemArgV) :-
    debug_print(20,'%logging to: '), debug_println(20,F),
    set_log_file(F), set_logging_mode(Mode),
    start_xml_group_in_log('probcli-run'),
    writeln_log(start_logging(NOW,F)),
    version(V1,V2,V3,Suffix), revision(Rev),  lastchangeddate(LCD),
    writeln_log(version(NOW,V1,V2,V3,Suffix,Rev,LCD)), % still used by log_analyser
    prolog_flag(version,PV),
    write_xml_element_to_log(version,[major/V1,minor/V2,patch/V3,suffix/Suffix,revision/Rev,lastchanged/LCD,prolog/PV]),
    findall(Opt, option(Opt), Options),
    write_prolog_term_as_xml_to_log(options(NOW,Options)),
    write_prolog_term_as_xml_to_log(files(NOW,RemArgV)), %
    datime(Datime,DateRecord),
    writeln_log(date(NOW,DateRecord)),
    (DateRecord=datime(Yr,Mon,Day,Hr,Min,Sec)
     -> write_xml_element_to_log(date,[year/Yr,month/Mon,day/Day,hour/Hr,minutes/Min,seconds/Sec]) ; true).

:- use_module(eventhandling,[store_virtual_event/1]).
:- use_module(bmachine,[b_set_initial_machine/0]).
cli_start_selfcheck :-
      %clear_loaded_machines_wo_errors,
      b_set_initial_machine,
      set_animation_mode(b),
      store_virtual_event(clear_specification), % TO DO: try and get rid of the need for this
      start_animation,
      option(selfcheck(Module)),
      (option(silent) -> Opt=silent ; option(verbose) -> Opt=verbose ; Opt=normal),
      (perform_self_check(Module,Opt) -> true ; true),
      (test_error_occurred(_) -> error_occurred(selfcheck) ; true),
      fail.
cli_start_selfcheck.


:- dynamic file_loaded/2.
file_loaded(Status) :- file_loaded(Status,_File).


cli_load_files([],NOW) :- % no files are provided
   !,
   ( options_allow_start_without_file -> cli_set_empty_machine
     ; we_did_something -> true
     ; otherwise -> print('No file to process'),nl),
   writeln_log_time(start_processing_empty_machine(NOW)),
   start_xml_feature(process_file,filename,'$EMPTY_MACHINE',FINFO),
   cli_process_loaded_file(NOW,'$EMPTY_MACHINE'),
   check_all_expected_errors_occurred(NOW), % check that all expected errors occurred; below they will be checked for each file
	stop_xml_feature(process_file,FINFO).
cli_load_files(RemArgV,NOW) :- cli_load_files2(RemArgV,NOW,0).


cli_set_empty_machine :- % TO DO: do this more properly here and for initialise_required
%       announce_event(clear_specification),
       set_animation_mode(b),
%       announce_event(start_initialising_specification),
       bmachine:b_set_empty_machine,
       assert(file_loaded(true,'$$empty_machine')).
       %announce_event(specification_initialised).


options_allow_start_without_file :- option(eval_repl(_)).
options_allow_start_without_file :- option(eval_string_or_file(_,_,_,_,_)).
options_allow_start_without_file :- option(check_log(_)).
options_allow_start_without_file :- option(process_latex_file(_,_)).

we_did_something :- option(print_version).
we_did_something :- option(print_short_version).
we_did_something :- option(check_java_version).

option_only_works_for_single_file(zmq_assertion(_Identifier)).
option_only_works_for_single_file(zmq_master(_Identifier)).

:- use_module(tools,[safe_absolute_file_name/2, safe_absolute_file_name/3]).
clear_loaded_files :-
   (file_loaded(_) -> clear_loaded_machines_wo_errors ; true).

% called if we have at least one file
cli_load_files2([],_,NrFilesProcessed) :- !,
    print_accumulated_infos(NrFilesProcessed). % print summary of all runs for different files
cli_load_files2([F1,F2|T],NOW,_NrFilesProcessed) :-
    option(Option),
    option_only_works_for_single_file(Option),!,
    add_error(probcli,'The following option can only be used for a single file: ',Option),
    add_error(probcli,'Multiple files provided: ',[F1,F2|T]),
    halt_prob(NOW,0).
cli_load_files2(RemArgV,NOW,NrFilesProcessed) :-
    %print_total_probcli_timer,
    clear_loaded_files,
    retractall(file_loaded(_,_)),
    RemArgV = [MainFile0|Rest],!,
    N1 is NrFilesProcessed+1,
    cli_load_files3(MainFile0,Rest,NOW,N1).
cli_load_files3(MainFile0,Rest,NOW,NrOfFile) :-
    safe_absolute_file_name(MainFile0,MainFile,[access(none)]), % converts Windows slash into Unix slash,...
    if_option_set(file_info,print_file_info(MainFile)),
    ((Rest=[_|_] ; NrOfFile>1)
     -> length(Rest,RLen), Tot is NrOfFile+RLen,
        format('~n~n% Processing file ~w/~w: ~w~n',[NrOfFile,Tot,MainFile]) % was formatsilent
      ; true),
    start_xml_feature(process_file,filename,MainFile,FINFO),
    ( file_exists(MainFile) ->
        debug_println(6,file_exists(MainFile)),
        ( load_main_file(MainFile,NOW,Already_FullyProcessed) ->
          (Already_FullyProcessed==true -> true
           ; assert(file_loaded(true,MainFile)),
             trimcore_if_useful(Rest),
             writeln_log_time(start_processing(NOW)),
             start_probcli_timer(Timer),
             (cli_process_loaded_file(NOW,MainFile)
               -> stop_probcli_debug_timer(Timer,'% Finished processing file after')
               ;  print_error('### Processing or loading file failed: '), print_error(MainFile)
             ),
             writeln_log_time(finished_processing(NOW))
           )
        ; otherwise ->
            assert(file_loaded(error,MainFile)),
            print_error('### Loading Specificaton Failed'),
            writeln_log_time(loading_failed(NOW,MainFile)),
            error_occurred(load_main_file)
        ),
        nls
    ; otherwise ->
        nl, assert(file_loaded(error,MainFile)),
        (number(MainFile0)
            -> add_error(load_main_file,'Command-line argument is a number which is not associated with a command and does not exist as file: ',MainFile0)
         ; atom_codes(MainFile0,[45|_]) % starts with a dash - : probably an illegal command-line option
             -> add_error(load_main_file,'Specified Option or File does not exist: ',MainFile0)
          ; add_error(load_main_file,'Specified File does not exist: ',MainFile)
        )
    ),
    check_all_expected_errors_occurred(NOW),
	stop_xml_feature(process_file,FINFO),

    reset_expected_error_occurred, % reset for next file
    reset_errors,
    NOW1 is NOW+1, update_time_stamp(NOW1),
    cli_load_files2(Rest,NOW1,NrOfFile).

print_file_info(F) :-
   print('Specification_File('), print(F), print(')'),nl.

trimcore_if_useful(_) :- option(release_java_parser),!, prob_trimcore.
  % if user wants to release java parser, this is an indication that the .prob files are big and it can be good to free memory
trimcore_if_useful([]) :- % try and reduce Prologs memory consumption, see also PROLOGKEEPSIZE parameter
    % a lot of memory can be consumed loading .prob file and doing machine construction
    (option(X), memory_intensive_option(X)
       -> debug_format(9,'Not trimming memory usage because of memory intensive option: ~w~n',[X])
    ; statistics(memory_used,M), M< 300000000 % less than 300 MB used
       -> debug_format(9,'Not trimming memory usage because of memory used is already low: ~w~n',[M])
       ;  prob_trimcore
    ).
trimcore_if_useful(_) :- debug_println(9,'Not trimming memory usage as there are still files to process').

prob_trimcore :- (option(verbose) ; option(release_java_parser)),!,prob_trimcore_verbose.
prob_trimcore :- prob_trimcore_silent.

prob_trimcore_verbose :-
   print('Memory used before trimming: '),tools:print_memory_used_wo_gc,flush_output, nl_time,
   prob_trimcore_silent,
   print('Memory used after trimming : '),tools:print_memory_used_wo_gc,flush_output, nl_time.
prob_trimcore_silent :-
   garbage_collect, % is important, otherwise trimming may achieve very little
   trimcore.

memory_intensive_option(cli_mc(_)).
memory_intensive_option(ltl_formula_model_check(_,_)).
memory_intensive_option(ctl_formula_model_check(_,_)).
memory_intensive_option(refinement_check(_,_,_)).

% ---------------------

% process all the commands for a loaded file:
cli_process_loaded_file(NOW,MainFile) :-
    (real_error_occurred -> print_error('% *** Errors occurred while loading ! ***'),nl,nl ; true),
    get_errors, reset_errors,
    if_option_set(kodkod_performance(KPFile,Iterations),
                  compare_kodkod_performance1(KPFile,Iterations,NOW)),
    if_option_set(kodkod_comparision(MaxResiduePreds),
                  test_kodkod_and_exit(MaxResiduePreds,NOW)),
    % if_option_set(add_csp_guide(CspGuide), tcltk_add_csp_file(CspGuide)), %% moved to later to ensure B machine is precompiled; allows e.g. type_check_csp_and_b to run

    if_option_set(csp_main(MAINPROC),
                  set_cspm_main_process(MAINPROC)),

    if_option_set(zmq_master(Identifier), zmq_start_master(invariant,Identifier)),
    %if_option_set(zmq_master(IP, Logfile), zmq_start_master(invariant,200,-1,5000,0,IP,Logfile)),

    % STARTING ANIMATION/MODEL CHECKING
    cli_start_animation(NOW),


    if_option_set_loaded(trace_check,trace_check,
                         cli_start_trace_check(MainFile)),

    cli_process_loaded_file_afer_start_animation(NOW).

cli_process_loaded_file_afer_start_animation(NOW) :-
    ifm_option_set(cli_print_machine_info(IKind),
                  cli_print_machine_info(IKind)),
    ifm_option_set(pretty_print_internal_rep(PPFILE1,TYPES,ASCII),
                  pretty_print_internal_rep(PPFILE1,TYPES,ASCII)),
    ifm_option_set(pretty_print_internal_rep_to_B(PPFILE3),
                  b_write_eventb_machine_to_classicalb_to_file(PPFILE3)),

    if_option_set_loaded(state_trace(TraceFile),state_trace,
                         cli_start_trace_state_check(TraceFile)),

    if_option_set(evaldot(EvalDotF),
                  set_eval_dot_file(EvalDotF)),

    (initialise_required
      -> check_loaded(initialise),
         cli_start_initialisation(NOW),
         writeln_log_time(initialised(NOW))
      ;  true),

    if_option_set(zmq_assertion(Identifier),
                  zmq_start_master(assertion,Identifier)),

    if_option_set(cli_lint,cli_lint),
    if_option_set(cli_wd_check(Disch,TotPos),cli_wd_check(Disch,TotPos)),
    if_option_set(cli_start_mc_with_tlc,cli_start_mc_with_tlc),
    if_option_set(cli_start_sym_mc_with_lts(LType),cli_start_sym_mc_with_lts(LType)),

    if_option_set(cli_symbolic_model_check(Algorithm),cli_symbolic_model_check(Algorithm)),

    if_option_set_loaded(cli_check_properties,check_properties,
                         cli_check_properties(NOW)),
    ifm_option_set_loaded(cli_check_assertions(ALL,ReqInfos),check_assertions,
                         cli_check_assertions(ALL,ReqInfos,NOW)),
    if_option_set(set_goal(GOAL),
                  cli_set_goal(GOAL)),
    if_option_set(set_searchscope(SCOPE),
                  cli_set_searchscope(SCOPE)),
    ifm_option_set_loaded(cli_mc(Nr),model_check,
                         cli_start_model_check(Nr,NOW)),
    ifm_option_set_loaded(cli_random_animate(Steps,ErrOnDeadlock),animate,
                         cli_random_animate(NOW,Steps,ErrOnDeadlock)),
    ifm_option_set_loaded(execute(ESteps,ErrOnDeadlock,From),execute,
                         cli_execute(ESteps,ErrOnDeadlock,From)),
    if_option_set_loaded(pa_check,predicate_analysis,
                         test_predicate_analysis),

    cbc_check(NOW),

    ifm_option_set_loaded(logxml_write_ids(Prefix,IDScope),logxml_write_ids,
                          logxml_write_ids(Prefix,IDScope)),

    if_options_set(generate_read_write_matrix_csv(RWCsvFile),
                   generate_read_write_matrix(RWCsvFile)),
    if_options_set(feasibility_analysis_csv(TimeOut,EnablingCsvFile),
                   do_feasibility_analysis(TimeOut,EnablingCsvFile)),
    ifm_option_set_loaded(mcm_tests(ADepth1,AMaxS,ATarget1,Output1),mcm_test_cases,
                         mcm_test_case_generation(ADepth1,AMaxS,ATarget1,Output1)),
    ifm_option_set_loaded(all_deadlocking_paths(File),all_deadlocking_paths,
                         write_all_deadlocking_paths_to_xml(File)),
    ifm_option_set_loaded(cbc_tests(ADepth2,ATarget2,Output2),cb_test_cases,
                         cbc_test_case_generation(ADepth2,ATarget2,Output2)),
    ifm_option_set_loaded(test_description(TestDescFile),cb_test_cases,
                         test_generation_by_xml_description(TestDescFile)),
    if_options_set(csp_in_situ_refinement_check(RP,RType,RQ),
                  cli_csp_in_situ_refinement_check(RP,RType,RQ,NOW)),
    if_options_set(csp_checkAssertion(Proc,Model,AssertionType),
                   cli_checkAssertion(Proc,Model,AssertionType,NOW)),
    if_options_set(check_csp_assertion(Assertion),
                   cli_check_csp_assertion(Assertion,NOW)),
    if_options_set(refinement_check(RefFile,PerformSingleFailures,RefNrNodes),
                  cli_start_refinement_check(RefFile,PerformSingleFailures,RefNrNodes,NOW)),
    if_options_set(ctl_formula_model_check(Formula,Expected),cli_ctl_model_check(Formula,Expected,_)),
    % TO DO print ctl/ltl statistics
    if_options_set(csp_get_assertions,cli_csp_get_assertions),
    if_options_set(eval_csp_expression(CspExpr),cli_eval_csp_expression(CspExpr)),
    if_options_set(csp_translate_to_file(PlFile),cli_csp_translate_to_file(PlFile)),
    if_options_set(get_min_max_coverage(MinMxFileName),cli_get_min_max_coverage(MinMxFileName)),
    if_options_set(get_coverage_information(CovFileName),cli_get_coverage_information(CovFileName)),
    if_options_set(vacuity_check,cli_vacuity_check),
    if_option_set_loaded(check_goal,check_goal,cli_check_goal),
    if_option_set_loaded(animate,animate,
                         (interactive_animate_machine -> true ; true)),
    if_option_set(ltsmin, start_ltsmin_srv('/tmp/ltsmin.probz', NOW)),
    if_option_set(ltsmin2(EndpointPath), start_ltsmin_srv(EndpointPath, NOW)),
    if_option_set(ltsmin_ltl_output(Path), ltsmin_ltl_output(Path, NOW)),
    evaluate_from_commandline,
    if_option_set_loaded(ltl_assertions,check_ltl_assertions,
                         (timeout_call(ltl_check_assertions,NOW) -> true; true)),
    ifm_option_set_loaded(ltl_formula_model_check(Formula,Expected),check_ltl_assertions,
                         (option(cli_start_sym_mc_with_lts(_))-> true % we request LTSMin, do not start prob model check
                          ; timeout_call(cli_ltl_model_check(Formula,Expected,_),NOW) -> true; true)),
    ifm_option_set_loaded(ltl_file(LtlFilename),check_ltl_file,
                         (ltl_check_file(LtlFilename) -> true; true)),
    ifm_option_set_loaded(history(HistoryFilename),history,
                         cli_print_history(HistoryFilename)),
    ifm_option_set_loaded(print_values(ValuesFilename),values,
                         cli_print_values(ValuesFilename)),
    ifm_option_set_loaded(print_all_values(ValuesDirname),values,
                         cli_print_all_values(ValuesDirname)),
    if_options_set(save_state_for_refinement(SaveRefF),
                  refinement_checker:tcltk_save_specification_state_for_refinement(SaveRefF)),
    if_options_set(dot_command(DCommand,DotFile),
                  dot_command(DCommand,DotFile)),
    if_options_set(dot_command_for_expr(DECommand,Expr,DotFile,Opts),
                  dot_command_for_expr(DECommand,Expr,DotFile,Opts)),
    if_options_set(evaluate_expression_over_history_to_csv_file(HistExpr,HistDotFile),
                  tcltk_interface:evaluate_expression_over_history_to_csv_file(HistExpr,HistDotFile)),
    if_options_set(enabling_analysis_csv(EnablingCsvFile),
                   do_enabling_analysis_csv(EnablingCsvFile,NOW)),
    if_options_set(process_latex_file(LatexF1,LatexF2),
                   process_latex_file(LatexF1,LatexF2)),
    ifm_option_set(coverage(Nodes,Operations,ShowEnabledInfo),
                  cli_show_coverage(Nodes,Operations,ShowEnabledInfo,NOW)),
    ifm_option_set(check_statespace_hash(ExpectedHash,Kind),
                  cli_check_statespace_hash(ExpectedHash,Kind)),
    ifm_option_set(variable_coverage,
                  cli_show_variable_coverage(NOW)),
    ifm_option_set(coverage(ShowEnabledInfo),
                  cli_show_coverage(ShowEnabledInfo,NOW)),
    if_option_set(save_state(StateFile),
                  save_state(StateFile)),
    if_option_set(prob_profile,
                  prob_profile),
    if_option_set(statistics,
                  cli_print_statistics),
    if_option_set(profile_statistics,
                  profile_statistics),
    if_option_set(show_cache,
                  show_cache),
    if_option_set(check_complete, check_complete),
    if_option_set(check_complete_operation_coverage, check_complete_operation_coverage),
    if_option_set(check_scc_for_ltl_formula(LtlFormula,SCC),cli_check_scc_for_ltl_formula(LtlFormula,SCC)).

:- use_module(value_persistance,[show_cache_file_contents/0,show_cache_file_constants/0]).
show_cache :- option(verbose),!,show_cache_file_contents,nl.
show_cache :- show_cache_file_constants,nl.

% new profiler
%:- use_module('../extensions/profiler/profiler.pl').
%cli_print_statistics :- pen,nl,garbage_collect,statistics,nl,state_space:state_space_initialise_with_stats.

% old profiler
:- use_module(hit_profiler).
:- use_module(b_operation_cache,[print_op_cache/0]).
:- use_module(memoization,[reset_memo_with_statistics/0]).
cli_print_statistics :- format('PROB STATISTICS~n',[]),
    profile_statistics,
    garbage_collect,
    statistics,
    nl,
    print_op_cache,
    state_space:state_space_initialise_with_stats,
    reset_memo_with_statistics.

profile_statistics :-
    (print_hit_profile_statistics -> true ; true),
    (option(profiling_on)
      -> on_exception(error(existence_error(_,_),_),print_profile,
                       print_red('SICStus Prolog Profiler can only be used when running from source'))
      ; true).


:- use_module(state_space,[not_all_transitions_added/1, not_invariant_checked/1, get_operation_name_coverage_infos/4]).
check_complete :-
  (tcltk_find_max_reached_node(Node1) ->
     add_error(check_complete,'Maximum number of transitions reached for at least one state: ',Node1) ; true),
  (not_all_transitions_added(Node2) ->
     add_error(check_complete,'At least one state was not examined: ',Node2) ; true),
  (not_invariant_checked(Node3) ->
     add_error(check_complete,'The invariant was not checked for at least one state: ',Node3) ; true).

check_complete_operation_coverage :-
  (state_space: operation_name_not_yet_covered(OpName) ->
     add_error(check_complete_operation_coverage,'At least one operation is not covered: ', OpName)
  ; true).

show_operation_coverage_summary(NOW) :-
   get_operation_name_coverage_infos(PossibleNr,FeasibleNr,UncovNr,UncoveredList),
   writeln_log(uncovered_info(NOW,PossibleNr,UncoveredList)),
   (UncovNr=0 -> format(' All ~w possible operations have been covered',[PossibleNr]),nl
     ; (FeasibleNr=PossibleNr
        -> format(' The following ~w operations (out of ~w) were not covered:~n   ~w~n',
                  [UncovNr, PossibleNr,UncoveredList])
        ;  INr is PossibleNr-FeasibleNr,
           format(' The following ~w operations (out of ~w with ~w infeasible) were not covered:~n   ~w~n',
                  [UncovNr, PossibleNr, INr,UncoveredList])
    )).
show_initialisation_summary(NOW) :-
   findall(ID,state_space:is_concrete_constants_state_id(ID),L),
   length(L,Nr), N1 is Nr+1, % for root
   writeln_log(uninitialised_states(NOW,N1)),
   format(' Uninitialised states: ~w (root and constants only)~n',[N1]).

% ---------------------

animation_mode_does_not_support_animation(File) :-
    loaded_main_file(smt2,File).
cli_start_animation(NOW) :-
    file_loaded(true,LoadedFile),
    \+ animation_mode_does_not_support_animation(LoadedFile),
    !,
    debug_println(20,'% Starting Animation'),
    writeln_log_time(start_animation(NOW)),
    start_probcli_timer(Timer1),
    start_animation_without_computing,
    stop_probcli_debug_timer(Timer1,'% Finished Starting Animation'),
    if_option_set(add_csp_guide(CspGuide), tcltk_add_csp_file(CspGuide)),

    xml_log_machine_statistics,
    getAllOperations(Ops),
    debug_print(20,'Operations: '), debug_println(20,Ops),

    (we_need_only_static_assertions(ALL)
      -> debug_println(20,'% Projecting on static ASSERTIONS'),
         b_interpreter:set_projection_on_static_assertions(ALL) ; true),

    (option(load_state(File))
      -> debug_println(20,'% Loading stored state from file'),
         state_space:tcltk_load_state(File)
       ; computeOperations_for_root_required ->
         debug_println(20,'% Searching for valid initial states'),
         start_probcli_timer(Timer2),
         cli_computeOperations(EO),
         stop_probcli_debug_timer(Timer2,'% Finished searching for valid initial states'),
         debug_println(10,EO)
       ; debug_println(20,'% No initialisation required')
    ).
cli_start_animation(_NOW).

start_animation_without_computing :-
    update_preferences_from_spec(ListOfPrefs),
    (ListOfPrefs=[] -> true ; write_prolog_term_as_xml_to_log(b_machine_preferences(ListOfPrefs))),
    set_prefs, % override SET_PREF in DEFINITIONS with values from command-line;
    start_animation,
    ifm_option_set(add_additional_property(PROP),
                   cli_add_additional_property(PROP)),
    get_errors.

:- use_module(state_space,[current_expression/2]).
:- use_module(extension('counter/counter')).
:- use_module(tools_lists,[count_occurences/2]).
:- use_module(tools,[start_ms_timer/1, stop_ms_timer/1, stop_ms_timer/2, stop_ms_timer_with_msg/2]).
% an execution engine with minimal overhead: states are not stored in visited_expression database, only first enabled operation is taken
cli_execute(Steps,ErrorOnDeadlock,FromWhere) :-
   FromWhere=from_all_initial_states,!,
  % try out all initial states and from each of those perform deterministic execution
    start_ms_timer(Start),
    format('Running execute from all initial states~n',[]),
    reset_counter(cli_execute_inits),
	findall(Result,
	 (cli_trans(root,Action,CurState,0,'$NO_OPERATION'), %print(Action),nl,
	     (\+ functor(Action,'$setup_constants',_)
		   -> inc_counter(cli_execute_inits,Nr),
		      format('~nExecuting model from initial state ~w~n',[Nr]),
			  (option(animate_stats) -> print_state_silent(CurState) ; true),
			  (cli_execute_from(CurState,Steps,ErrorOnDeadlock,1,Result) -> true)
		   ;  format('~nInitialising state~n',[]), % we need to execute initialise_machine
			  cli_trans(CurState,_ActionName,NewState,0,'$NO_OPERATION'),
			  inc_counter(cli_execute_inits,Nr),
			  format('~nExecuting model from initial state ~w with constants~n',[Nr]),
			  (option(animate_stats) -> print_state_silent(NewState) ; true),
			  (cli_execute_from(NewState,Steps,ErrorOnDeadlock,2,Result) -> true)
	  )), Results),
    get_counter(cli_execute_inits,Nr),
    format('---------~nTotal runtime for all ~w executions:',[Nr]),nl,
    stop_ms_timer(Start),
    count_occurences(Results,Occs), format('Results: ~w~n',[Occs]).
cli_execute(Steps,ErrorOnDeadlock,_) :-
    current_expression(ID,CurState),
    start_xml_feature(execute,max_steps,Steps,FINFO),
    debug_format(19,'Starting execute from state ~w with maximum number of steps ~w~n',[ID,Steps]),
    cli_execute_from(CurState,Steps,ErrorOnDeadlock,1,_Result),
    stop_xml_feature(execute,FINFO).
% TO DO: add transition from root to executed state ?

:- use_module(bmachine,[b_top_level_operation/1]).
allow_filter_unused_constants :-
    b_or_z_mode,
    b_top_level_operation(_), % we can filter out unused constants, unless there are no operations in which case the user probably wants to see the constant values
    \+ options_can_eval_any_cst.

options_can_eval_any_cst :- option(eval_repl(_)).
options_can_eval_any_cst :- option(eval_string_or_file(_,_,_,_,_)).
options_can_eval_any_cst :- option(dot_command(_,_)).
options_can_eval_any_cst :- option(dot_command_for_expr(_,_,_,_)).
options_can_eval_any_cst :- option(process_latex_file(_,_)).
options_can_eval_any_cst :- option(logxml_write_ids(all,_)).  % the user writes out constants (not just variables) to file: also do not filter

:- dynamic execute_timeout_occurred/0.

cli_execute_from(CurState,Steps,ErrorOnDeadlock,FirstStepNr,Result) :-
    retractall(max_walltime(_,_,_)),
    retractall(execute_timeout_occurred),
    start_ms_timer(Start),
    (allow_filter_unused_constants -> temporary_set_preference(filter_unused_constants,true,CHNG) ; true),
    cli_execute_aux(FirstStepNr,Steps,CurState,_MEMO,ErrorOnDeadlock,'$NO_OPERATION',Result),
    (allow_filter_unused_constants -> reset_temporary_preference(filter_unused_constants,CHNG) ; true),
    (option(silent) -> true ; stop_ms_timer_with_msg(Start,'-execute')),
    print_max_walltime.

:- use_module(tools_strings,[ajoin/2, ajoin_with_sep/3]).
:- use_module(external_functions,[reset_side_effect_occurred/0, side_effect_occurred/1]).
cli_execute_aux(Nr,Steps,CurState,_,_ErrorOnDeadlock,_LastActionName,Result) :- Nr>Steps,!,
    formatsilent('Stopping execution after ~w steps~n',[Steps]), Result = stopped,
    print_state_silent(CurState),
    cli_execute_add_virtual_transition(Steps,CurState,Result).
cli_execute_aux(Nr,Steps,CurState0,MEMO,ErrorOnDeadlock,LastActionName,Result) :-
    (Nr mod 5000 =:= 0, \+option(animate_stats), \+option(silent)
     -> (var(LastActionName) -> format('Step ~w~n',[Nr])
         ; format('Step ~w (after ~w)~n',[Nr,LastActionName])),
        (option(verbose) -> print_state_silent(CurState0) ; true),
        %copy_term(CurState0,CurState), tools_printing:print_term_summary((CurState0)),nl,
        !,
        garbage_collect,
        !,
        print('Memory used: '),tools:print_memory_used_wo_gc,flush_output,nl %nl_time
     ; true),
    prepare_state_for_specfile_trans(CurState0,MEMO,CurState), % ensure we memoize expanded constants in MEMO
       % avoid re-expanding constants in every state !
       % relevant e.g. for probcli -execute 20001 DataValidationTestSmallStep.mch -init
    (cli_invariant_ko(CurState,LastActionName,InvStatus,CliErr) -> % also recognises no_inv command
           N1 is Nr-1,
           ajoin(['INVARIANT ',InvStatus,' after ',N1,' steps (after ',LastActionName,').'],ErrMsg),
           format('~w~n',[ErrMsg]),!,
           print_state_silent(CurState),
           error_occurred_with_msg(CliErr,ErrMsg),
           Result=CliErr,
           cli_execute_add_virtual_transition(N1,CurState,Result)
    ; \+ cli_assertions_ok(CurState,LastActionName) -> % also recognises no_inv command
           N1 is Nr-1,
           format('ASSERTION VIOLATED after ~w steps (after ~w).~n',[N1,LastActionName]),!,
           print_state_silent(CurState),
           error_occurred(assertion_violation), Result=assertion_violation,
           cli_execute_add_virtual_transition(N1,CurState,Result)
     ; cli_goal_found(CurState) -> % also recognizes no_goal command
           N1 is Nr-1,
           format('GOAL FOUND after ~w steps (after ~w).~n',[N1,LastActionName]),!,
           print_state_silent(CurState), Result=goal_found,
           cli_execute_add_virtual_transition(N1,CurState,Result)
    ;      reset_side_effect_occurred,
           cli_trans(CurState,ActionName,NewState,Nr,LastActionName),
           !,
           N1 is Nr+1,
           (NewState=CurState0, % could be expensive for large states; states are expanded ! % TO DO: look only at written variables ?!
            \+ side_effect_occurred(file)
             -> formatsilent('Infinite loop reached after ~w steps (looping on ~w).~n',[N1,ActionName]),
                Result=loop,
                cli_execute_add_virtual_transition(N1,CurState,Result)
              ; cli_execute_aux(N1,Steps,NewState,MEMO,ErrorOnDeadlock,ActionName,Result))
    ).
cli_execute_aux(Nr,_Steps,CurState,_,_ErrorOnDeadlock,LastActionName,Result) :- execute_timeout_occurred,!,
    N1 is Nr-1,
    formatsilent('Timeout occurred after ~w steps (after ~w).~n',[N1,LastActionName]),
    Result=time_out,
    print_state_silent(CurState),
    cli_execute_add_virtual_transition(N1,CurState,Result).
cli_execute_aux(Nr,_Steps,CurState,_,ErrorOnDeadlock,LastActionName,Result) :- N1 is Nr-1,
    formatsilent('Deadlock reached after ~w steps (after ~w).~n',[N1,LastActionName]),
    Result=deadlock,
    (ErrorOnDeadlock=true,\+ option(no_deadlocks) -> error_occurred(deadlock) ; true),
    print_state_silent(CurState),
    cli_execute_add_virtual_transition(N1,CurState,Result).

cli_execute_add_virtual_transition(Steps,CurState,Result) :-
    getCurrentStateID(CurID),
    write_xml_element_to_log(executed,[steps/Steps,result/Result]),
    (Steps>0 -> tcltk_interface:tcltk_add_new_transition(CurID,'$execute'(Steps),ToID,CurState,[]),
                debug_format(19,'Added virtual transtion ~w -> ~w for ~w steps~n',[CurID,ToID,Steps]),
                tcltk_goto_state('$execute'(Steps),ToID)
             ; true).

:- dynamic max_walltime/3.
print_max_walltime :- (max_walltime(Action,Nr,WT),option_verbose
      -> format('% Maximum walltime ~w ms at step ~w for ~w.~n',[WT,Nr,Action]) ; true).

% will be used to compute a single successor
% b_operation_cannot_modify_state
cli_trans(CurState,ActionName,NewState,Nr,LastActionName) :-
    option(animate_stats),!, % provide statistics about the animation
    start_probcli_timer(Timer),
    cli_trans_aux(CurState,ActionName,Act,NewState,Nr,LastActionName),
    (option_verbose -> translate:translate_event(Act,TStr)
      ; translate:translate_event_with_limit(Act,100,TStr)),
    format('~w~5|: ~w~n',[Nr,TStr]),
    format(' ~5|',[]),stop_probcli_timer(Timer,' '),
    (option_verbose -> format(' ~5|',[]),tools:print_memory_used_wo_gc,nl ; true),
    flush_output,
    get_probcli_elapsed_walltime(Timer,WallTime),
    (max_walltime(_,_,MWT), MWT >= WallTime -> true
      ; retractall(max_walltime(_,_,_)),
        assert(max_walltime(ActionName,Nr,WallTime))).
cli_trans(CurState,ActionName,NewState,Nr,LastActionName) :-
    cli_trans_aux(CurState,ActionName,_,NewState,Nr,LastActionName).

cli_trans_aux(CurState,ActionName,Act,NewState,Nr,LastActionName) :-
   option(timeout(TO)),!,
   safe_time_out(cli_trans_aux2(CurState,ActionName,Act,NewState,LastActionName),TO,Res),
   (Res=time_out -> format_error('*** Timeout occurred while performing step ~w of execute: ~w ms~n',[Nr,TO]),
                    % TO DO: try and obtain operation name in which time-out occured
                    error_occurred(time_out,execute),
                    assert(execute_timeout_occurred),
                    fail
                 ; true).
cli_trans_aux(CurState,ActionName,Act,NewState,_Nr,LastActionName) :-
    cli_trans_aux2(CurState,ActionName,Act,NewState,LastActionName).

cli_trans_aux2(CurState,ActionName,Act,NewState,LastActionName) :-
     catch_enumeration_warning_exceptions(
        (throw_enumeration_warnings_in_current_scope,
         cli_execute_trans(CurState,ActionName,Act,NewState,LastActionName), % no time-out !
         (error_occurred_in_error_scope -> ErrorEvent=true ; ErrorEvent=false)
         ),
             (error_occurred(virtual_time_out_execute),
              ActionName = '*** VIRTUAL_TIME_OUT ***', Act=ActionName,
              CurState=NewState) % this forces loop detection above; not very elegant way of signalling
              ),
     (ErrorEvent==true,option(strict_raise_error)
       -> print('*** ERROR OCCURED DURING EXECUTE ***'),nl, error_occurred(execute),fail
        ; true).

:- use_module(runtime_profiler,[profile_single_call/3]).
:- use_module(specfile,[get_specification_description/2]).
cli_execute_trans(CurState,ActionName,Act,NewState,LastActionName) :-
    statistics(runtime,[StartExecuteForState,_]),
    get_possible_next_operation_for_execute(CurState,LastActionName,ActionName),
    start_check_disabled(ActionName,StartActionTime),
    on_exception(EXC,
        profile_single_call(ActionName,
          unknown, % state Unknown
          specfile_trans_with_check(CurState,ActionName,Act,NewState,Residue) % no time-out !
          ),
        (translate_exception(EXC,EMSG),
         (nonvar(ActionName)
          -> get_specification_description(operation,OP),
             format_with_colour(user_error,[red,bold],'~n*** ~w while executing ~w "~w"~n~n',[EMSG,OP,ActionName])
          ;  get_specification_description(operations,OP),
             format_with_colour(user_error,[red,bold],'~n*** ~w while computing ~w~n~n',[EMSG,OP])
         ),
         throw(EXC))),
    (Residue=[]
     -> check_trans_time(StartExecuteForState,StartActionTime,ActionName)
     ; error_occurred(cli_execute_residue(ActionName,Residue,Act))).

:- use_module(enabling_analysis,[static_cannot_enable/2]).
% compute all possible next operations to try out in -execute, based on LastActionName
% the predicate makes use of the fact that operations are tried in order
get_possible_next_operation_for_execute(CurState,LastActionName,ActionName) :-
    get_preference(randomise_operation_order,false),
    get_preference(use_po,true), % PROOF_INFO: should we use a proper preference for use_po_for_execute ?
    b_or_z_mode,
    b_top_level_operation(LastActionName),
    % prevent trying ActionName which we know is infeasible according to previous LastActionName
    % we rely on the fact that execute tries the operations in order
    findall(AN,specfile_possible_trans_name_for_successors(CurState,AN),ANS),
    !,
    member_with_last(ActionName,ANS,LastActionName,FoundLast),
    (var(FoundLast)
      -> % ActionName occurs before LastActionName, this means
         % we did try this ActionName at the last execution step and it was not possible
         % format('Checking operation ~w against ~w~n',[ActionName,LastActionName]),debug:nl_time,
         %Note: we could use result of cbc enabling analysis if it was performed
         (static_cannot_enable(LastActionName,ActionName)
           -> %format('Operation ~w : ~w cannot be enabled by ~w : ~w~n',[Nr2,ActionName,Nr1,LastActionName]),
              %print('- '),
              fail
            ; true
         )
       ; true % we did not try this ActionName at the last execution step
    ).
get_possible_next_operation_for_execute(CurState,_LastActionName,ActionName) :-
    specfile_possible_trans_name_for_successors(CurState,ActionName).

% like member but instantiate FoundLast when we pass the element Last
member_with_last(X,[H|T],Last,FoundLast) :-
  (H=Last -> !,FoundLast=true, member(X,[H|T]) ; X=H).
member_with_last(X,[_|T],Last,FoundLast) :- member_with_last(X,T,Last,FoundLast).

% also checks whether a setup_constants_inconsistent error should be raised:
specfile_trans_with_check(CurState,ActionName,Act,NewState,Residue) :-
    if(specfile_trans_or_partial_trans(CurState,ActionName,Act,NewState,_TransInfo,Residue,Partial), % no time-out !
       check_partial_trans(Partial,ActionName),
       check_deadlock_fail(ActionName)
    ).

% check whether we should raise errors on deadlock in -execute:
check_deadlock_fail('$setup_constants') :- !, error_occurred(setup_constants_inconsistent),fail.

% check whether we should raise errors due to partial transitions in -execute:
check_partial_trans(true,'$setup_constants') :- !, error_occurred(setup_constants_inconsistent).
check_partial_trans(true,ActionName) :- !,format('Unknown partial transition: ~w~n',[ActionName]),
   error_occurred(setup_constants_inconsistent).
check_partial_trans(_,_).

% check if we did not spend too much time on disabled operations and print warning if we do
check_trans_time(StartExecuteForState,StartActionTime,ActionName) :-
    option(execute_monitoring),
    statistics(runtime,[CurrentTime,_]),
    Delta1 is StartActionTime-StartExecuteForState,
    Delta2 is CurrentTime-StartActionTime,
    Delta1 > 100, Delta1 > Delta2,
    !,
    format_with_colour(user_output,[blue],'~n ~5|: WARNING from -execute_monitor: ~w ms for disabled operations and ~w ms for operation ~w itself~n',[Delta1,Delta2,ActionName]).
check_trans_time(_,_,_).

% check if we did not spend too much time on a single disabled operations and print warning if we do
start_check_disabled(ActionName,StartActionTime) :- option(execute_monitoring),!,
    statistics(runtime,[StartActionTime,_]),
    (true
     ; statistics(runtime,[EndActionTime,_]),
       Delta is EndActionTime - StartActionTime,
       Delta > 50,
       format_with_colour(user_output,[blue],'~n ~5|: WARNING from -execute_monitor: ~w ms for disabled operation ~w~n',[Delta,ActionName]),
       fail
    ).
start_check_disabled(_,0).

translate_exception(user_interrupt_signal,'User-Interrupt (CTRL-C)').
translate_exception(enumeration_warning(_,_,_,_,_),'Enumeration Warning').
translate_exception(E,E).

:- use_module(bmachine,[b_machine_has_constants/0]).
print_state_silent(_) :- option(silent),!.
print_state_silent(CurState) :- (option(verbose);\+ b_machine_has_constants),!,
   translate:print_state(CurState),nl.
print_state_silent(CurState) :- remove_constants(CurState,VarState),
   % only print variables
   format('VARIABLES (use -v to see constants or -silent to suppress output):~n',[]),
   translate:print_state(VarState),nl.

:- use_module(bmachine,[b_is_constant/1]).
is_constant_binding(bind(C,_)) :- b_is_constant(C).
remove_constants(const_and_vars(_,Vars),Res) :- !,Res=Vars.
remove_constants([H|T],Res) :- !,exclude(prob_cli:is_constant_binding,[H|T],Res).
remove_constants(root,Res) :- !,Res=[].
remove_constants(concrete_constants(_),Res) :- !, Res=[].
remove_constants(X,X).

% write vars to xml log file if they start with a given prefix
logxml_write_ids(variables,Prefix) :- !,
    current_expression(_,CurState),
    remove_constants(CurState,VarState),
    % TO DO: store also final state in xml_log
    write_bstate_to_log(VarState,Prefix).
logxml_write_ids(_,Prefix) :- !,
    current_expression(_,CurState),
    expand_const_and_vars_to_full_store(CurState,EState),
    write_bstate_to_log(EState,Prefix).

:- use_module(bmachine,[b_get_invariant_from_machine/1, b_specialized_invariant_for_op/2, b_machine_has_constants/0]).
cli_invariant_ko(_,_,_,_) :- option(no_invariant_violations),!,fail. % user asks not to check it
cli_invariant_ko(_,_,_,_) :- get_preference(do_invariant_checking,false),!,fail. % user asks not to check it via preference
cli_invariant_ko(CurState,LastActionName,ResInvStatus,CliError) :-
    state_corresponds_to_initialised_b_machine(CurState,BState),!,
    start_probcli_timer(InvTimer),
    (b_specialized_invariant_for_op(LastActionName,Invariant) -> true
       %, print('Specialized invariant: '),translate:print_bexpr(Invariant),nl
      ; b_get_invariant_from_machine(Invariant)),
    cli_test_pred(BState,'INVARIANT',Invariant,ResInvStatus),
    stop_probcli_debug_timer(InvTimer,'Finished Invariant Checking'),
    ResInvStatus \= 'TRUE',
    (ResInvStatus == 'FALSE' -> CliError = invariant_violation
    ; ResInvStatus == 'UNKNOWN' -> CliError = invariant_unknown
    ;  format_error('Unexpected invariant status: ~w~n',[ResInvStatus]),
       CliError= invariant_unknown).
%cli_invariant_ko(_,_,_,_) :- fail. % not yet initialised

cli_test_pred(BState,PredKind,Pred) :- cli_test_pred(BState,PredKind,Pred,'TRUE').
cli_test_pred(BState,PredKind,Pred,Res) :-
    % currently does not do a time-out: % b_interpreter calls: time_out_with_enum_warning_one_solution_no_new_error_scope
     on_exception(E,
          on_enumeration_warning(b_interpreter:b_test_boolean_expression_for_ground_state(Pred,[],BState ),
            (add_error(cli_test_pred,'Enumeration warning while testing',PredKind,Pred),
            cli_print_pred_info(Pred),
            Res='UNKNOWN')),
         (string_concatenate('VIRTUAL TIME-OUT while testing ',PredKind,Msg),
          add_error(cli_test_pred,Msg,E,Pred),
          cli_print_pred_info(Pred),
          Res='UNKNOWN')),
      !,
      (var(Res) -> Res = 'TRUE' ; true).
cli_test_pred(_BState,_PredKind,_Pred,'FALSE').

cli_print_pred_info(Pred) :- bsyntaxtree:get_texpr_label(Pred,Label),
    format('Label = ~w~n',[Label]),fail.
cli_print_pred_info(Pred) :- bsyntaxtree:get_texpr_description(Pred,Desc),
    format('Description = ~w~n',[Desc]),fail.
%cli_print_pred_info(Pred) :- bsyntaxtree:get_texpr_pos(Pred,Pos), Pos \= none, translate_span(Pos,Str),
%    format('Location = ~w~n',[Str]),fail.
cli_print_pred_info(_).

:- use_module(bmachine,[get_assertions_from_machine/2]).
% TO DO: also check static assertions
cli_assertions_ok(_,_) :- option(no_assertion_violations),!. % user asks not to check it
cli_assertions_ok(CurState,_LastActionName) :-
    state_corresponds_to_initialised_b_machine(CurState,BState),
    get_assertions_from_machine(dynamic,Assertions), % TO DO: do something similar to b_specialized_invariant_for_op
    !,
    start_probcli_timer(AssTimer),
    %nl,nl,print(check),nl,maplist(translate:print_bexpr,Assertions),nl,
    maplist(prob_cli:cli_test_pred(BState,'ASSERTION'),Assertions),
    stop_probcli_debug_timer(AssTimer,'Finished Checking Assertions').
cli_assertions_ok(_,_). % not yet initialised or no assertions


:- use_module(bmachine,[b_get_machine_goal/1]).
cli_goal_found(_):- option(no_goal),!,fail.
cli_goal_found(CurState) :-
   b_get_machine_goal(Goal),
   state_corresponds_to_initialised_b_machine(CurState,BState),
   b_interpreter:b_test_boolean_expression_for_ground_state(Goal,[],BState ).


% random animation
:- public cli_random_animate/2. % for repl to use -animate command
cli_random_animate(Steps,Err) :- probcli_time_stamp(NOW),
    cli_random_animate(NOW,Steps,Err).
cli_random_animate(_NOW,Steps,ErrorOnDeadlock) :-
    start_xml_feature(random_animate,max_steps,Steps,FINFO),
    start_ms_timer(S),
    perform_random_steps(Steps,ErrorOnDeadlock),
    stop_ms_timer(S,Res),
    %tcltk_save_history_as_trace_file(prolog,user),
    printsilent(finished_random_animate(Steps,Res)),nls,
    stop_xml_feature(random_animate,FINFO).

:- use_module(bmachine,[b_get_assertions/3]).
we_need_only_static_assertions(ALL) :- specfile:b_or_z_mode,
  (option(cli_check_assertions(main,_)) -> ALL=main
     ; option(cli_check_assertions(ALL,_))),
   % we do an assertion check
   \+ ((option(A), option_requires_all_properties(A))),
   b_get_assertions(ALL,dynamic,[]). % the assertions do not reference variables

% do we need all properties/constants of the machine, or only certain ones (e.g., static)
option_requires_all_properties(cli_mc(_)).
option_requires_all_properties(cli_check_properties(_)). % we may want to check all properties
option_requires_all_properties(cli_random_animate(_)).
option_requires_all_properties(trace_check).
option_requires_all_properties(state_trace(_)).
option_requires_all_properties(mcm_tests(_,_,_,_)).
option_requires_all_properties(cbc_tests(_,_,_)).
option_requires_all_properties(animate).
option_requires_all_properties(initialise). % INITIALISATION may access constants
option_requires_all_properties(eval_repl(_)).
option_requires_all_properties(eval_string_or_file(_,_,_,_,_)).
option_requires_all_properties(ltl_assertions).
option_requires_all_properties(ltl_file(_)).
option_requires_all_properties(refinement_check(_,_,_)).
option_requires_all_properties(cli_start_mc_with_tlc).
option_requires_all_properties(cli_symbolic_model_check(_)).
option_requires_all_properties(process_latex_file(_,_)).
option_requires_all_properties(cli_wd_check(_,_)).
option_requires_all_properties(cli_lint).

:- public trace_check/0.
trace_check :- loaded_main_file(_,MainFile),
     cli_start_trace_check(MainFile).
cli_start_trace_check(MainFile) :-
     debug_println(20,'% Starting Trace Check: '),
     ( check_default_trace_for_specfile(MainFile) -> true
     ; otherwise -> error_occurred(trace_check)).
cli_start_trace_state_check(File) :-
     debug_println(20,'% Starting Trace Check: '),
     ( tcltk_check_state_sequence_from_file(File) -> true
     ; otherwise -> error_occurred(state_trace)).

% is it necessary to compute enabled operations for root state
computeOperations_for_root_required :- initialise_required.
computeOperations_for_root_required :- option(trace_check).
computeOperations_for_root_required :- option(state_trace(_)).
computeOperations_for_root_required :- option(ltl_assertions).
computeOperations_for_root_required :- option(cli_random_animate(_,_)).
computeOperations_for_root_required :- option(socket(_,_)).
computeOperations_for_root_required :- option(cli_mc(_)).
computeOperations_for_root_required :- option(ltl_file(_)).
computeOperations_for_root_required :- option(ltl_formula_model_check(_,_)).
computeOperations_for_root_required :- option(ctl_formula_model_check(_,_)).
computeOperations_for_root_required :- option(refinement_check(_,_,_)).
computeOperations_for_root_required :- option(csp_in_situ_refinement_check(_,_)).
computeOperations_for_root_required :- option(csp_checkAssertion(_,_)).
computeOperations_for_root_required :- option(mcm_tests(_,_,_,_)).
computeOperations_for_root_required :- option(mcm_cover(_)).

% is an initialisation mandatory:
initialise_required :- option(initialise), \+ file_loaded(true,'$$empty_machine').
initialise_required :- \+ option(trace_check), \+ option(state_trace(_)),
                       \+ option(load_state(_)),
                       \+ file_loaded(true,'$$empty_machine'),
                       \+ (option(execute(Nr,_,_)), Nr>=2), % execute will also initialise machine
                       init_req2.
init_req2 :- option(cli_check_properties).
init_req2 :- option(zmq_assertion(_,_,_)).
init_req2 :- option(cli_check_assertions(_,_)).
init_req2 :- option(process_latex_file(_,_)).
init_req2 :- option(eval_string_or_file(_,_,_,_,_)). % ensure that we initialise/precompile empty machine in case no main file specified; currently no longer required

:- public initialise/0. % for REPL
initialise :- probcli_time_stamp(NOW),cli_start_initialisation(NOW).

cli_start_initialisation(NOW) :-
    debug_println(20,'% Performing INITIALISATION: '),
    (perform_random_initialisation -> true ;
      writeln_log_time(cli_start_initialisation_failed(NOW)),
      fail).

:- use_module(wdsrc(well_def_analyser),[analyse_wd_for_machine/4]).
:- public cli_wd_check/2. % for REPL
cli_wd_check(ExpectedDis,ExpectedTot) :-
   (option(timeout(TO)) -> true ; TO=5000),
   (option(silent) -> Opts=[discharge_po,ignore_wd_infos,reorder_conjuncts]
     ; Opts=[create_not_discharged_msg(warning),discharge_po,ignore_wd_infos,reorder_conjuncts]),
   statistics(walltime,[W1,_]),
   safe_time_out(analyse_wd_for_machine(NrDischarged,NrTot,_Res,Opts),TO,Res),
   statistics(walltime,[W2,_]), WT is W2-W1,
   (Res=time_out -> accumulate_infos(wd_check,[timeout-1,walltime-WT]), % discharged and total are unknown
       add_error(cli_wd_check,'TIME-OUT in WD Analysis (use -timeout X to increase it)')
    ;  format(user_output,'WD Analysis Result: discharged ~w / ~w',[NrDischarged,NrTot]),
       (NrTot >0
        -> Perc is 100*NrDischarged/NrTot,
           (NrDischarged=NrTot -> Col=[green,bold] ; Col=[red])
        ; Perc = 100, Col=[]),
       format_with_colour(user_output,Col,' (~2F %)~n',[Perc]),
       WDInfos = [discharged-NrDischarged,timeout-0,total-NrTot,walltime-WT],
       accumulate_infos(wd_check,WDInfos),
       (ExpectedDis==ExpectedTot, ExpectedTot = NrTot -> true ; true), % for -wd-check-all: bind ExpectedDis
       check_required_infos([discharged-ExpectedDis,total-ExpectedTot],WDInfos,cli_wd_check)
  ).

:- use_module(bmachine_static_checks,[extended_static_check_machine/0]).
% perform some additional static checks
:- public cli_lint/0. % for REPL
cli_lint :-
   extended_static_check_machine.



:- use_module(predicate_debugger,[tcltk_debug_properties/3]).
:- use_module(state_space,[current_state_corresponds_to_setup_constants_b_machine/0]).
:- public cli_check_properties/0. % for REPL
cli_check_properties :- probcli_time_stamp(NOW),
    cli_check_properties(NOW).
cli_check_properties(NOW) :-
    printsilent('% Checking PROPERTIES: '),nls,
    writeln_log_time(starting_check_properties(NOW)),
    ( current_state_corresponds_to_setup_constants_b_machine ->
        set_analyse_hook('_P'),
        predicate_evaluator:tcltk_analyse_properties(_PROPRES,PROPInfos),
        unset_analyse_hook,
        printsilent(PROPInfos),nls,   % ex: [total/33,true/29,false/0,unknown/4,timeout/4,runtime/49950]
        accumulate_infos(properties,PROPInfos),
        write_important_xml_element_to_log(check_properties,PROPInfos),
        (predicate_evaluator:check_summary_all_true(PROPInfos) -> true
          ; print_error('### Not all PROPERTIES true'), error_occurred(check_properties))
    ; otherwise ->
        (tcltk_debug_properties(list(PROPRES),false,Satisfiable) ->
         printsilent(PROPRES),nls,
         printsilent(Satisfiable),nls
         ; error_occurred(debug_properties_failed))
    ),
    writeln_log_time(finished_check_properties(NOW,PROPInfos)),
    loaded_root_filename(RootName),
    formatsilent('% Finished checking PROPERTIES of ~w~n',[RootName]).
:- public cli_check_assertions/2. % for REPL
cli_check_assertions(ALL,RI) :-
    probcli_time_stamp(NOW),
    cli_check_assertions(ALL,RI,NOW).
cli_check_assertions(ALL,ReqInfos,NOW) :-
    printsilent('% Checking ASSERTIONS: '),nls,
    writeln_log_time(starting_check_assertions(NOW)),
    set_analyse_hook('_A'), % for dot output, in case users wants to generate dot files for assertions
    predicate_evaluator:tcltk_analyse_assertions(ALL,_ASSRES,Infos),
    unset_analyse_hook,
    printsilent(Infos),nls,
    accumulate_infos(assertions,Infos),
    write_important_xml_element_to_log(check_assertions,Infos),
    check_required_infos(ReqInfos,Infos,check_assertions),
    writeln_log_time(finished_check_assertions(NOW,Infos)),
    loaded_root_filename(RootName),
    formatsilent('% Finished checking ASSERTIONS of ~w~n',[RootName]),!.
cli_check_assertions(ALL,ReqInfos,NOW) :-
    add_internal_error('Analyse ASSERTIONS unexpectedly failed',cli_check_assertions(ALL,ReqInfos,NOW)),
    error_occurred(internal_error).
cli_set_goal(GOAL) :-
    debug_println(20,set_goal(GOAL)), %print(set_goal(GOAL)), nl,
    bmachine:b_set_machine_goal(GOAL).
cli_add_additional_property(PROP) :-
    debug_println(20,add_additional_property(PROP)),
    bmachine:add_additional_property(PROP).
cli_set_searchscope(GOAL) :-
    debug_println(20,set_searchscope(GOAL)),
    format('Setting SCOPE for verification: ~w~n (Only states satsifying this predicate will be examined)~n',[GOAL]),
    bmachine:b_set_machine_searchscope(GOAL).
cli_check_goal :- \+ b_get_machine_goal(_),!,
    add_error(cli_check_goal,'No GOAL DEFINITION found'),
    error_occurred(cli_check_goal).
cli_check_goal :-
    printsilent('% Checking GOAL predicate: '),nls,
    tcltk_analyse_goal(_List,Summary),
    debug_println(20,Summary),
    accumulate_infos(check_goal,Summary),
    write_important_xml_element_to_log(check_goal,Summary),
    check_required_infos([false/0,unknown/0],Summary,check_goal).
:- public cli_mc/1. % for REPL
cli_mc(Nr) :- probcli_time_stamp(NOW), cli_start_model_check(Nr,NOW).
cli_start_model_check(Nr,NOW) :-
    start_xml_feature(model_check,max_states,Nr,FINFO),
    model_check(Nr,Time,MCRes,NOW),
    %nl,
    stop_xml_feature(model_check,FINFO),
    accumulate_infos(model_check,[runtime-Time]), % for bench_csv output
    writeln_log_time(model_check(NOW,Nr,Time,MCRes)).
cli_start_mc_with_tlc :-
    (animation_mode(b), \+ animation_minor_mode(eventb) -> true; error_manager: add_error_and_fail(mc_with_tlc,'TLC4B tool can be used only for classical B models.')),
    % TO DO: use b_write_eventb_machine_to_classicalb_to_file to do conversion
    on_exception(error(E,_),
                 safe_absolute_file_name(prob_lib('TLC4B.jar'),TLC4BTool),
                 error_manager: add_error_fail(get_tlc_command,'Could not find TLC4B.jar file.',E)),
    start_xml_feature(model_check_with_tlc,tlc4bjar,TLC4BTool,FINFO),
    construct_and_execute_tlc_command(TLC4BTool),
    stop_xml_feature(model_check_with_tlc,FINFO).

:- use_module(system_call,[system_call/4]).
construct_and_execute_tlc_command(TLC4BTool) :-
    parsercall: get_java_command_path(JavaCmd),
    loaded_main_file(File),
    (get_preference(tlc_number_of_workers,TLCWorkers), TLCWorkers>1
     -> number_codes(TLCWorkers,CC), atom_codes(TLAWA,CC), WW = ['-workers',TLAWA]
      ; WW=[]),
    (option(no_assertion_violations) -> WA = ['-noass'] ; WA=[]),
    (option(no_deadlocks) -> WD = ['-nodead'] ; WD=[]),
    (option(no_invariant_violations) -> WI = ['-noinv'] ; WI=[]),
    (option(no_goal) -> WG = ['-nogoal'] ; WG=[]),
    (option(no_ltl) -> WL = ['-noltl'] ; WL=[]),
    append([WW,WA,WD,WI,WG,WL],XtraTLCArgs),
    append(XtraTLCArgs,[File],TLCArgs),
    debug_println(19,tlc_args(TLCArgs)),
    statistics(walltime,[W1,_]),
    system_call(JavaCmd, ['-jar', TLC4BTool | TLCArgs], Text,JExit),
    statistics(walltime,[W2,_]),
    WTime is W2-W1,
    formatsilent('exit : ~w  walltime: ~w ms~n',[JExit,WTime]),
    (JExit=exit(0)
     ->  true
     ;   error_manager: add_error(construct_and_execute_tlc_command,'Error while model checking with TLC: ',TLC4BTool/File),
         atom_codes(T,Text),
         error_manager: add_error_fail(construct_and_execute_tlc_command,'Std error: ',T)
    ).

% SymbolicOrSequential = symbolic or sequential
cli_start_sym_mc_with_lts(SymbolicOrSequential) :-
    (option(no_deadlocks) -> NoDead = true ; NoDead = false),
    (option(no_invariant_violations) -> NoInv = true ; NoInv = false), % does LTSMin support goal checking
    findall(Option,option(ltsmin_option(Option)),MoreFlags1),
    findall(ltl_formula(LTLF),option(ltl_formula_model_check(LTLF,_)),MoreFlags2),
    append(MoreFlags1,MoreFlags2,MoreFlags),
    (NoDead = false, NoInv = false ->
        print_error('ERROR: cannot start LTSmin with both deadlock and invariant checking'),
        print_error('       use either the -noinv or -nodead flag'),
        flush_output(user_error)
      ; true),
    formatsilent('starting prob2lts-sym/seq (flags nodead=~w, noinv=~w, moreflags=~w)~n',[NoDead,NoInv,MoreFlags]),
    start_ltsmin(SymbolicOrSequential, [NoDead, NoInv], MoreFlags,Result),
    process_ltsmin_result(Result).
    % TO DO: start lts-sym + start start_ltsmin_srv('/tmp/ltsmin.probz', NOW) + print output

:- use_module(extension('ltsmin/ltsmin_trace'),[csv_to_trace/3]).
process_ltsmin_result(ltsmin_model_checking_ok) :-
    print_green('LTSMin found no counter example\n').
process_ltsmin_result(ltsmin_model_checking_aborted) :-
    add_warning(ltsmin_model_checking_aborted,'LTSMin was aborted (e.g., by CTRL-C)').
process_ltsmin_result(ltsmin_counter_example_found(CsvFile)) :-
    add_error(ltsmin_counter_example_found,'LTSMin found a counter example, written to:',CsvFile),
    (option(silent) -> true
      ;  csv_to_trace(CsvFile,_States,Transitions),
         print('*** TRACE: '),nl,print_list(Transitions) % ,print(_States),nl
    ).

:- use_module(symbolic_model_checker(ic3), [ic3_symbolic_model_check/1]).
:- use_module(symbolic_model_checker(ctigar), [ctigar_symbolic_model_check/1]).
:- use_module(symbolic_model_checker(kinduction), [kinduction_symbolic_model_check/1,
                                                   tinduction_symbolic_model_check/1]).
:- use_module(symbolic_model_checker(bmc), [bmc_symbolic_model_check/1]).
cli_symbolic_model_check(Algorithm) :-
    debug_format(20,'% Starting Symbolic Model Check. Using ~w Algorithm', [Algorithm]),
    start_xml_feature(model_check,algorithm,Algorithm,FINFO),
    (animation_mode(b)
     -> true
     ;  error_manager:add_error_and_fail(cli_symbolic_model_check,'Symbolic Model Checking is currently only available for B and Event-B.')),
    perform_symbolic_model_checking(Algorithm,Result),
    handle_symbolic_model_check_result(Result),
    stop_xml_feature(model_check,FINFO).

perform_symbolic_model_checking(ic3,Result) :- !, ic3_symbolic_model_check(Result).
perform_symbolic_model_checking(ctigar,Result) :- !, ctigar_symbolic_model_check(Result).
perform_symbolic_model_checking(kinduction,Result) :- !, kinduction_symbolic_model_check(Result).
perform_symbolic_model_checking(tinduction,Result) :- !, tinduction_symbolic_model_check(Result).
perform_symbolic_model_checking(bmc,Result) :- !, bmc_symbolic_model_check(Result).
perform_symbolic_model_checking(Alg,_) :- add_error_fail(cli_symbolic_model_check,'Invalid symbolic model checking algorithm: ',Alg).

handle_symbolic_model_check_result(counterexample_found) :- !, error_occurred(invariant_violation).
handle_symbolic_model_check_result(property_holds) :- !,
    format('Model checking complete, invariant holds~n',[]).
handle_symbolic_model_check_result(solver_and_provers_too_weak) :- !,
    format('Model checking incomplete because a constraint could not be solved in time~n',[]),
    error_occurred(model_check_incomplete).
handle_symbolic_model_check_result(limit_reached) :- !,
    format('Model checking incomplete because an iteration limit was reached~n',[]),
    error_occurred(model_check_incomplete).

zmq_start_master(invariant,Identifier) :-
    start_animation_without_computing,
    zmq_get_initialisation_term(InitTerm),
    (option(strict_raise_error) -> Strict = 1 ; Strict = 0),
    get_preference(port, PortStart),
    get_preference(max_states, Max),
    get_preference(ip, IP),
    get_preference(logdir, LogDir),
    get_preference(tmpdir, TmpDir),
    get_preference(hash_cycle, HashCycle),
    atom_concat(LogDir, '/distb-', ATmp),
    atom_concat(ATmp, Identifier, Logfile),
    atom_concat(TmpDir, '/db-distb-', TTmp),
    atom_concat(TTmp, Identifier, TmpDir2),
    start_master(InitTerm,Max,PortStart,Strict,IP,Logfile,TmpDir2,HashCycle),
    halt.
zmq_start_master(assertion,Identifier) :-
    get_preference(port, PortStart),
    get_preference(logdir, LogDir),
    get_preference(ip, IP),
    get_preference(tmpdir, TmpDir),
    get_preference(hash_cycle, HashCycle),
    atom_concat(LogDir, '/distb-', ATmp),
    atom_concat(ATmp, Identifier, Logfile),
    atom_concat(TmpDir, '/db-distb-', TTmp),
    atom_concat(TTmp, Identifier, TmpDir2),
    current_state_corresponds_to_setup_constants_b_machine,
    animation_mode(b),
    full_b_machine(Machine),
    b_get_assertions(_,static,SAss),
    b_get_assertions(_,dynamic,DAss),
    append(SAss,DAss,Ass),
    count_assertions(Ass,0,N),
    assert(master:assertion_count(N)),
    current_expression(_,State1),
    specfile:state_corresponds_to_set_up_constants(State1,State),
    zmq_get_important_options(Options),
    (option(strict_raise_error) -> Strict = 1 ; Strict = 0),
    %start_master(assertions(classical_b(Machine,Options),State,Ass),2,-1,PortStart,0,Strict,IP,Logfile,TmpDir2),
    start_master(assertions(classical_b(Machine,Options),State,Ass),2,PortStart,Strict,IP,Logfile,TmpDir2,HashCycle),
    halt.

zmq_get_initialisation_term(Term) :-
    (animation_mode(b) ; animation_mode(csp_and_b)), % CSP file not yet added when ZMQ master starts working
    \+ animation_minor_mode(eventb),
    option(add_csp_guide(CspGuide)),
    !, % Classical B + CSP
    debug_println(20,'ZMQ: Transferring CSP || B model'),
    full_b_machine(Machine),
    % TO DO: extract CSP Term rather than file name: will not work for distribution on other file-systems
    zmq_get_important_options(Options),
    Term = classical_b_with_csp(Machine,CspGuide,Options).
zmq_get_initialisation_term(Term) :-
    debug_println(20,'Generating ZMQ Worker Initialisation'),
    animation_mode(b), \+ animation_minor_mode(eventb), !, % Classical B
    debug_println(20,'ZMQ: Transferring Classical-B model'),
    full_b_machine(Machine),
    zmq_get_important_options(Options),
    Term = classical_b(Machine,Options).
zmq_get_initialisation_term(Term) :-
    animation_mode(b), animation_minor_mode(eventb), !, % Event-B
    debug_println(20,'ZMQ: Transferring Event-B model'),
    full_b_machine(Machine),
    zmq_get_important_options(Options),
    Term = eventb(Machine,Options).
zmq_get_initialisation_term(Term) :-
    animation_mode(cspm),
    loaded_main_file(MainCSPFile),!, % TO DO: pass CSP Prolog term rather than file name (for distribution)
    zmq_get_important_options(Options),
    debug_println(20,'ZMQ: Transferring CSP specification'),
    Term = csp_specification(MainCSPFile,Options).
zmq_get_initialisation_term(_Term) :-
    \+ real_error_occurred, % otherwise error occured while loading
    animation_mode(Mode),
    add_internal_error('Unsupported formalism for ZMQ', Mode),
    fail.

zmq_get_initialisation_term(filename(FN)) :-
    loaded_main_file(FN).

% get important command-line options to be transmitted to probcli worker
zmq_get_important_options(Options) :- findall(O, (option(O), zmq_important_option(O)), Options),
   debug_println(20,transferring_zmq_options_to_workers(Options)).
zmq_important_option(coverage(_)).
zmq_important_option(expect_error(_)).
zmq_important_option(optional_error(_)).
zmq_important_option(file_info).
zmq_important_option(log(_)).
zmq_important_option(print_version).
zmq_important_option(print_short_version).
zmq_important_option(profiling_on).
zmq_important_option(set_card(_,_)).
zmq_important_option(set_pref(_,_)).
zmq_important_option(set_preference_group(_,_)).
zmq_important_option(verbose).
zmq_important_option(statistics).
zmq_important_option(variable_coverage).
zmq_important_option(very_verbose).
% we could consider also supporting: -argv, -cache, -prefs FILE  csp_main(ProcessName) profiling_on prob_profile runtimechecking

% set options received by a zmq worker
zmq_set_important_options(Options) :- debug_println(20,setting_zmq_options(Options)),
     maplist(prob_cli:zmq_set_option,Options).
zmq_set_option(file_info) :- !, file_info.
zmq_set_option(log(F)) :- !,
    generate_time_stamp(Datime,NOW),
    cli_start_logging(F,ascii,NOW,Datime,[zmq_worker]).
zmq_set_option(print_version) :- !, print_version.
zmq_set_option(print_short_version) :- !, print_short_version.
zmq_set_option(profiling_on) :- !, profiling_on.
zmq_set_option(set_card(P,V)) :- !, set_eclipse_card(P,V).
zmq_set_option(set_pref(P,V)) :- !, set_pref(P,V).
zmq_set_option(set_preference_group(P,V)) :- !, set_preference_group(P,V).
zmq_set_option(verbose) :- !, verbose.
zmq_set_option(very_verbose) :- !, very_verbose.
zmq_set_option(O) :- zmq_delayed_option(O),!, assert_option(O). % DO IT LATER
zmq_set_option(O) :- add_internal_error('Unsupported option for ZMQ worker: ',zmq_set_option(O)).

zmq_delayed_option(coverage(_)).
zmq_delayed_option(expect_error(_)).
zmq_delayed_option(expect_error_pos(_,_,_)).
zmq_delayed_option(optional_error(_)).
zmq_delayed_option(statistics).
zmq_delayed_option(variable_coverage).

ltsmin_ltl_output(Filename, NOW) :-
    if_option_set(ltl_formula_model_check(Formula, _),true),
    ltsmin:ltsmin_generate_ltlfile(Formula, Filename),
    halt_prob(NOW,0). % if we additionally specify -ltsformula, we do not want to model check it


start_ltsmin_srv(X, NOW) :-
    nls,println_silent('Starting LTSMin Server...'),
    if_option_set(ltl_formula_model_check(Formula, _),true),
    ltsmin:ltsmin_init(X, Zocket, Formula),
    ltsmin:ltsmin_loop(Zocket),
    ltsmin:ltsmin_teardown(Zocket, X),
    nls,println_silent('Stopped LTSMin Server.'),
    halt_prob(NOW,0). % if we additionally specify -ltsformula, we do not want to model check it

zmq_start_worker(Identifier, NOW) :-
    get_preference(port, Port),
    get_preference(logdir, LogDir),
    get_preference(tmpdir, TmpDir),
    get_preference(proxynumber, ProxyNumber),
    /* TODO: ensure directory exists (pk, 09.01.2018) */
    atom_concat(LogDir, '/worker-', ATmp),
    atom_concat(ATmp, Identifier, Logfile),
    % TODO: tmp dir currently not used
    atom_concat(TmpDir, '/db-worker-', TTmp),
    atom_concat(TTmp, Identifier, TmpDir2),
    start_worker(Port,ProxyNumber,Logfile,TmpDir2,zmq_worker_load_model),
    formatsilent('ZMQ worker finished (Port:~w)~n',[Port]),
    cli_process_loaded_file_afer_start_animation(NOW),
    println_silent('Exiting probcli worker'),
    halt_prob(NOW,0).

zmq_worker_load_model(classical_b(Machine,Options)) :- !,
    debug_println(20,'ZMQ WORKER: Loading classical B model'),
    zmq_set_important_options(Options),
    bmachine:b_machine_reset, bmachine:assert_main_machine(Machine),
    set_animation_mode(b),
    prob2_interface:start_animation, cli_computeOperations(_)
    .
zmq_worker_load_model(classical_b_with_csp(Machine,CspGuide,Options)) :- !,
    debug_println(20,'ZMQ WORKER: Loading CSP || B model'),
    zmq_set_important_options(Options),
    bmachine:b_machine_reset, bmachine:assert_main_machine(Machine),
    set_animation_mode(b),
    prob2_interface:start_animation,
    tcltk_add_csp_file(CspGuide), % TO DO: use CSP Prolog term rather than filename <----------------
    prob2_interface:start_animation,
    cli_computeOperations(_)
    .
zmq_worker_load_model(eventb(Machine,Options)) :- !,
    zmq_set_important_options(Options),
    bmachine:b_machine_reset, bmachine:assert_main_machine(Machine),
    set_animation_mode(b), set_animation_minor_mode(eventb),
    prob2_interface:start_animation, cli_computeOperations(_).
zmq_worker_load_model(csp_specification(CSPFile,Options)) :-
    zmq_set_important_options(Options),
    load_cspm_spec_from_cspm_file(CSPFile), % TO DO: pass CSP Prolog term rather than filename
    prob2_interface:start_animation, cli_computeOperations(_).
zmq_worker_load_model(filename(FN)) :- !,
    printsilent('loading file by filename\n'),flush_output,
    ( is_eventb_b(FN) ->
        eclipse_interface:load_eventb_file(FN)
    ; otherwise ->
        bmachine:b_load_machine_probfile(FN)),
    prob2_interface:start_animation.
zmq_worker_load_model(assertions(Machine,State,Assertions)) :- !,
   assert(assertion_counter(-1)),
   println_silent(loaded_model_for_assertion_checking),
   zmq_worker_load_model(Machine),
   assert(worker:assertion_state(State)),
   make_assertionwps(Assertions).
   % assert current state
zmq_worker_load_model(Other) :-
    add_internal_error('ZMQ worker: Unexpected machine description', zmq_worker_load_model(Other)),
    fail.


:-dynamic assertion_counter/1.

count_assertions([],A,A).
count_assertions([H|T],A,R) :- bsyntaxtree:conjunction_to_list(H,HL),
                               length(HL,N1),
                               NN is A + N1,
                               count_assertions(T,NN,R).

make_assertionwps([]).
make_assertionwps([H|T]) :- bsyntaxtree:conjunction_to_list(H,HL),
                            sort_assertions(HL,SL),
                            append_assertion(SL),
                            make_assertionwps(T).

append_assertion([]).
append_assertion([H|T]) :- assertion_counter(N),
                           retractall(assertion_counter(_)),
                           N1 is N + 1,
                           assert(assertion_counter(N1)),
                           assert(worker:assertion_task(N1,H)),
                           append_assertion(T).

%assertions_order(A,B) :- term_size(A,NA),term_size(B,NB), NA > NB.
sort_assertions(X,X).
% :- samsort(assertions_order,X,Y).



is_eventb_b(FN) :- append(_,FN,".eventb").
 % load_model(Initialisation)


:- use_module(predicate_evaluator).
:- use_module(bmachine,[b_machine_name/1]).
set_analyse_hook(AddPrefix) :- % set a hook to write false/unknown expressions into a dot file
   reset_dot_file_number,
   if_options_set(dot_analyse_output_prefix(_Path),
                  (set_dot_file_prefix_if_option_set(AddPrefix),
                   register_conjunct_error_hook(prob_cli:pred_eval_hook))).
unset_analyse_hook :- predicate_evaluator:reset_conjunct_error_hook.

:- use_module(tools,[get_modulename_filename/2]).
loaded_root_filename(RootName) :- loaded_main_file(MainFile),
                   get_modulename_filename(MainFile,RootName).

set_dot_file_prefix_if_option_set(AddPrefix) :-
   if_options_set(dot_analyse_output_prefix(Path),
                  (loaded_root_filename(RootName),
                   % we could also use b_machine_hierarchy:main_machine_name(RootName)
                   string_concatenate(Path,RootName,P1),
                   string_concatenate(P1,AddPrefix,FullPath),
                   set_dot_file_prefix(FullPath),
                   debug_println(9,dot_file_prefix(FullPath)))).

% Status: true, false, unknown
:- public pred_eval_hook/5.
pred_eval_hook(_Conjunct,true,_EnumWarning,_IsExpanded, _CS) :-
  \+ option(dot_generate_for_all_formulas),!. % don't generate .dot for true formulas, unless explicitly requested
pred_eval_hook(Conjunct,Status,_EnumWarning,_IsExpanded, CS) :-
   printsilent('Generating dotfile for: '),printsilent(CS),nls,
   (write_dot_graph_to_new_file(Status,Conjunct) -> true
     ; add_error(dot_output,'Writing dot to file failed: ',CS)).


:- dynamic dot_file_prefix/1.
:- dynamic dot_file_number/1.

dot_file_prefix('~/Desktop/dot').
set_dot_file_prefix(F) :- retractall(dot_file_prefix(_)), assert(dot_file_prefix(F)).
dot_file_number(0).
reset_dot_file_number :- retractall(dot_file_number(_)), assert(dot_file_number(0)).
get_next_nr(GNr) :- retract(dot_file_number(Nr)), N1 is Nr+1,
  assert(dot_file_number(N1)), GNr = Nr.
write_dot_graph_to_new_file(Status,BExpr) :-
    dot_file_prefix(Dir),get_next_nr(Nr),
    string_concatenate('_',Status,Str1),
    string_concatenate(Nr,Str1,NS),
    string_concatenate(Dir,NS,F1),
    atom_concat(F1,'.dot',FileName),
    tcltk_interface:write_dot_file_for_pred_expr(BExpr,FileName).

% get dot file name if dot_output has been set
get_dot_file(Type,FileName) :- option(dot_analyse_output_prefix(_)),
    set_dot_file_prefix_if_option_set(Type),
    dot_file_prefix(Dir),
    string_concatenate('_',Type,Str1),
    string_concatenate(Dir,Str1,F1),
    atom_concat(F1,'.dot',FileName).

:- use_module(refinement_checker).
cli_csp_in_situ_refinement_check(P,Type,Q,NOW) :-
    debug_println(20,'% Starting CSP Refinement Check'),
    loaded_main_file(CSPFile),
    ajoin_with_sep(['assert',P,Type,Q], ' ',Assertion),
    start_xml_feature(csp_refinement_check,assertion,Assertion,FINFO),
    ( timeout_call(tcltk_interface:tcltk_check_csp_assertion(Assertion,CSPFile,'False',_PlTerm,RefTrace),NOW)
      -> check_ref_result(RefTrace)
      ; otherwise -> true),
    stop_xml_feature(csp_refinement_check,FINFO).
cli_start_refinement_check(RefFile,PerformSingleFailures,RefNrNodes,NOW) :-
    start_xml_feature(refinement_check,file,RefFile,FINFO),
    refinement_checker:tcltk_load_refine_spec_file(RefFile),
    ( timeout_call(tcltk_refinement_search(RefTrace,PerformSingleFailures,RefNrNodes),NOW)
      -> check_ref_result(RefTrace)
      ; otherwise -> true),
    stop_xml_feature(refinement_check,FINFO).
check_ref_result(RefTrace) :-
   ( RefTrace==no_counter_example ->
            print('==> Refinement Check Successful'),nl
        ; otherwise ->
            print('*** Refinement Check Counter-Example: ***'),nl, print(RefTrace),nl,
            print('*** Refinement Check Failed ***'),nl,
            error_occurred(refinement_check_fails)).
cli_checkAssertion(Proc,Model,AssertionType,_NOW) :-
    loaded_main_file(CSPFile),
    ajoin(['assert ',Proc,' :[ ',AssertionType,'[',Model,']',' ]'],Assertion),
    start_xml_feature(csp_deadlock_check,assertion,Assertion,FINFO),
    ( /*timeout_call(*/tcltk_interface:tcltk_check_csp_assertion(Assertion,CSPFile,'False',_PlTerm,ResTrace)/*,NOW)*/
      -> check_model_result(Assertion,ResTrace)
      ; otherwise -> true),
    stop_xml_feature(csp_deadlock_check,FINFO).
cli_check_csp_assertion(Assertion,NOW) :-
    start_xml_feature(csp_assertion_check,assertion,Assertion,FINFO),
    loaded_main_file(CSPFile),
    ajoin(['assert ',Assertion],AssertionFull),
    ( timeout_call(tcltk_interface:tcltk_check_csp_assertion(AssertionFull,CSPFile,_Negated,PlTerm,ResTrace),NOW)
      -> check_model_result(PlTerm,ResTrace)
      ; otherwise -> true),
    stop_xml_feature(csp_assertion_check,FINFO).



check_model_result(AssertionPlTerm,ResTrace) :-
  ( ResTrace==no_counter_example ->
    printsilent('==> Model Check Successful'),nls
  ;otherwise ->
    (functor(AssertionPlTerm,assertRef,_Arity) ->
      print('*** Refinement Check Counter-Example: ***'),nl, print(ResTrace),nl,
      print('*** Refinement Check Failed ***'),nl,
      error_occurred(refinement_check_fails)
    ; otherwise ->
      print('*** Model Check Counter-Example: ***'),nl,print(ResTrace),nl,
      print('*** Model Check Failed ***'),nl,
      error_occurred(model_check_fails))
  ).
:- use_module(probcspsrc(haskell_csp),[get_csp_assertions_as_string/2,
             parse_and_load_cspm_file_into_specific_pl_file/2,
             evaluate_csp_expression/2, evaluate_csp_expression/3]).
cli_csp_get_assertions :-
    loaded_main_file(CSPFile),
    get_csp_assertions_as_string(CSPFile,String),
    print('*** Assertions in File (separated by $) ***'),nl,print(String),nl.
cli_eval_csp_expression(E) :-
    (loaded_main_file(CSPFile) ->
	       evaluate_csp_expression(E, CSPFile, Res)
	    ;  evaluate_csp_expression(E,Res)
    ), print('Evaluated Expression: '),nl,print(Res),nl.
cli_csp_translate_to_file(PlFile) :-
    loaded_main_file(CSPFile),
    parse_and_load_cspm_file_into_specific_pl_file(CSPFile,PlFile).
cli_check_scc_for_ltl_formula(LtlFormula,SCC) :-
    ltl_fairness: check_scc_ce(LtlFormula,SCC).
:- use_module(tools_files,[write_lines_to_file/2]).
cli_get_min_max_coverage(Filename) :-
    coverage_statistics:tcltk_compute_min_max(list(R)),
    tools:filter(tools: is_a_comment('/*','*/'),R,_Comments,Res),
    (Filename=user_output -> maplist(prob_cli:print_line,['Min-Max-Coverage'|Res])
     ; write_lines_to_file(Filename,Res)).
%    pretty_print_min_max_coverage_to_file(FileName).
print_line(L) :- format(user_output,'~w~n',[L]).
:- use_module(coverage_statistics,[pretty_print_coverage_information_to_file/1]).
cli_get_coverage_information(FileName) :-
    pretty_print_coverage_information_to_file(FileName).
cli_vacuity_check :-
    eclipse_interface:get_vacuous_invariants(L),
    (L=[] -> print('No vacuous invariants'),nl
     ; maplist(prob_cli:add_vacuous_invariant,L)).
add_vacuous_invariant(Inv) :-
    translate:translate_bexpression(Inv,TI),
    add_error(vacuity_check,'Vacuous invariant: ',TI).
cli_start_socketserver(Port,Loopback) :-
    printsilent('Starting Socket Server'),nls,
    safe_absolute_file_name(prob_home('.'),AppDir),
    printsilent('Application Path: '),printsilent(AppDir),nls,
    disable_interaction_on_errors,
    ( start_prob_socketserver(Port,Loopback) -> true
    ; otherwise ->
        print('Starting socket server failed, Port: '), print(Port),nl),
    printsilent('Finished Socket Server'),nls.
:- use_module(tools,[platform_is_64_bit/0]).
cli_check_statespace_hash(Expected,Kind) :-
    printsilent('Computing hash of entire statespace: '),
    state_space:compute_full_state_space_hash(Hash),
    printsilent(Hash),nls, % TO DO: maybe also compute hash for transitions and check that
    (Hash=Expected -> true
      ; Kind=='64bit', \+ platform_is_64_bit -> format('Hash does not match ~w (but was computed on 64-bit system)~n',[Expected])
      ; Kind=='32bit', platform_is_64_bit -> format('Hash does not match ~w (but was computed on 32-bit system)~n',[Expected])
      ; add_error(hash,'Expected Statespace Hash to be: ',Expected)).
cli_show_coverage(ShowEnabledInfo,NOW) :-
    cli_show_coverage(_Nodes,_Operations,ShowEnabledInfo,NOW).
cli_show_coverage(Nodes,Operations,ShowEnabledInfo,NOW) :-
    ShowEnabledInfo==just_check_stats,!, % no printing of individual transition coverage
    state_space:get_state_space_stats(TotalNodeSum,TotalTransSum,_ProcessedTotal), % no computation overhead
    writeln_log(computed_coverage(NOW,TotalNodeSum,TotalTransSum)),
    check_totals(Nodes,Operations,TotalNodeSum,TotalTransSum).
cli_show_coverage(Nodes,Operations,ShowEnabledInfo,NOW) :-
    ShowEnabledInfo== just_summary,
    !, % no printing of detailed transition coverage (avoid traversing state space)
    state_space:get_state_space_stats(TotalNodeSum,TotalTransSum,_ProcessedTotal), % no computation overhead
    writeln_log(computed_coverage(NOW,TotalNodeSum,TotalTransSum)),
    format('Coverage:~n States: ~w~n Transitions: ~w~n',[TotalNodeSum,TotalTransSum]),
    show_initialisation_summary(NOW),
    show_operation_coverage_summary(NOW),
    (invariant_violated(ID) -> format('At least one state violates the invariant (~w) ~n',[ID]) ; true),
    check_totals(Nodes,Operations,TotalNodeSum,TotalTransSum).
cli_show_coverage(Nodes,Operations,ShowEnabledInfo,NOW) :-
    print('Coverage:'),nl,
    compute_the_coverage(Res,TotalNodeSum,TotalTransSum,ShowEnabledInfo,false),
    writeln_log(computed_coverage(NOW,TotalNodeSum,TotalTransSum)),
    print(Res),nl,
    check_totals(Nodes,Operations,TotalNodeSum,TotalTransSum).
check_totals(Nodes,Operations,TotalNodeSum,TotalTransSum) :-
    ( Nodes=TotalNodeSum -> true
    ; otherwise ->
        add_error(probcli,'Unexpected number of nodes: ',TotalNodeSum),
        add_error(probcli,'Expected: ',Nodes),error_occurred(coverage)),
    ( Operations=TotalTransSum -> true
    ; otherwise ->
        add_error(probcli,'Unexpected number of transitions: ',TotalTransSum),
        add_error(probcli,'Expected: ',Operations),error_occurred(coverage)).
cli_show_variable_coverage(NOW) :-
    print('Variable (min/max) coverage: '),nl,
    coverage_statistics:tcltk_compute_min_max(list(R)),
    print_list(R),nl,
    writeln_log(computed_variable_coverage(NOW,R)).

:- use_module(bmachine,[b_machine_statistics/2, b_get_main_filename/1, b_get_all_used_filenames/1,get_full_b_machine_sha_hash/1]).
:- use_module(tools_strings,[get_hex_bytes/2]).
cli_print_machine_info(statistics) :-
    b_machine_name(Name),
    %(b_get_main_filename(File) -> true ; File=unknown),
    format('Machine statistics for ~w:~n',[Name]),
    findall(Key/Nr,b_machine_statistics(Key,Nr),L),
    maplist(prob_cli:print_keynr,L),!.
cli_print_machine_info(files) :-
    b_machine_name(Name),
    format('Files used for machine ~w:~n',[Name]),
    b_get_all_used_filenames(Files),
    maplist(prob_cli:print_file,Files),!.
cli_print_machine_info(hash(Expected)) :-
    b_machine_name(MainName), % to do: findall machines and hashes
    get_full_b_machine_sha_hash(HashBytes),
    get_hex_bytes(HashBytes,Hash),
    format('SHA hash for machine ~w = ~s~n',[MainName,Hash]),!,
    (var(Expected) -> true
      ; atom_codes(Expected,Hash)
        -> format_with_colour(user_output,[green],'Machine hash for ~w matches provided hash.~n',[MainName])
      ; add_error(machine_hash_check,'Unexpected machine hash, expected: ',Expected)).
cli_print_machine_info(Kind) :- add_error(machine_stats,'Could not obtain machine information:',Kind).
print_keynr(Key/Nr) :- format(' ~w : ~w~n',[Key,Nr]).
print_file(File) :- format(' ~w~n',[File]).

:- use_module(tools,[get_tail_filename/2]).
xml_log_machine_statistics :-
    animation_mode(Major),
    (animation_minor_mode(Minor) -> true ; Minor=none),
    write_xml_element_to_log(animation_mode,[major/Major,minor/Minor]),
    (b_or_z_mode, b_machine_name(Main)
     ->  findall(Key/Nr,b_machine_statistics(Key,Nr),BMachStats),
         (b_get_main_filename(MainFile) -> get_tail_filename(MainFile,TailFile) ; TailFile = unknown),
         write_xml_element_to_log(b_machine_statistics,[machine_name/Main, tail_filename/TailFile|BMachStats])
       ; true).

cli_print_junit_results(ArgV) :-
    junit_mode(S),!,
    statistics(runtime,[E,_]),
    T is E - S,
    create_junit_result(ArgV,'Integration Tests','Integration_Test',T,pass,Result),
    print_junit([Result],'Integration_Test').
cli_print_junit_results(_).

cli_print_history(HistFile) :-
    findall( O, option(history_option(O)), Options),
    debug_println(9,writing_history_to_file(HistFile)),
    (Options=[trace_file] -> tcltk_save_history_as_trace_file(prolog,HistFile)  % save as Prolog trace file for replay with -t
     ; Options=[json] -> tcltk_save_history_as_trace_file(json,HistFile)  % save for replay with ProB2 UI
     ; write_history_to_file(HistFile,Options) -> true
     ; add_error(history,'Writing history to file failed: ',HistFile)).

cli_print_values(ValuesFilename) :-
    (write_values_to_file(ValuesFilename) -> true ; add_error(sptxt,'Writing values to file failed: ',ValuesFilename)).
cli_print_all_values(ValuesDirname) :-
    (write_all_values_to_dir(ValuesDirname) -> true ; add_error(sstxt,'Writing all values to directory failed: ',ValuesDirname)).

:- use_module(library(terms),[term_hash/2]).
:- dynamic probcli_time_stamp/1.
generate_time_stamp(NOW,TS) :- retractall(probcli_time_stamp(_)),
  now(NOW),
  prolog_flag(argv,ArgV),term_hash(ArgV,Hash),
  Rnd is Hash mod 1000,
  % random(0,1000,Rnd),  always returns 216 % TO DO: try to get milliseconds from some library function
  TS is (NOW*1000)+Rnd,
  assert(probcli_time_stamp(TS)).
update_time_stamp(NOW1) :- retractall(probcli_time_stamp(_)),
  assert(probcli_time_stamp(NOW1)).

%get_errors :- \+ real_error_occurred,!, (get_error(_Source,_Msg) -> print('*** Warnings occurred'),nl ; true), reset_errors.
get_errors :-
   (get_preference(view_probcli_errors_using_bbresults,true)
     -> tools_commands:show_errors_with_bb_results([current]) ; true),
   get_error_sources.

get_error_sources :- get_error_with_span(ErrSource,Msg,Span), !,
      error_occurred_with_msg(ErrSource,Msg,Span),
      findall(1,get_error(ErrSource,_),L), length(L,Nr),
      (Nr>0 -> N1 is Nr+1, get_error_category_and_type(ErrSource,Cat,Type),
               (Type=error -> print_error('*** Occurences of this error: ')
                 ; print_error('*** Occurences of this warning: ')),
               print_error(N1),
               write_xml_element_to_log(multiple_errors_occurred,[category/Cat,type/Type,number/N1])
       ; true),
      get_error_sources.
get_error_sources.

:- use_module(state_space,[state_error/3, invariant_violated/1, time_out_for_invariant/1, time_out_for_assertions/1, time_out_for_node/3]).
get_state_space_errors :- option(strict_raise_error),
      !,
	  (\+ option(no_invariant_violations),invariant_violated(ID)
		  -> (option_verbose ->
		        format('Invariant violation in state with id = ~w~n',[ID]),
		        b_interpreter:analyse_invariant_for_state(ID) % caused issue for test 1076
		      ; format('Invariant violation in state with id = ~w (use -v to print more details)~n',[ID])
		     ),
			   error_occurred(invariant_violation)
		 ; true),
	  (state_error(_,_,abort_error(TYPE,Msg,_,Span)) -> error_occurred(TYPE,error,Span,Msg) ; true),
	  get_state_errors(_).
get_state_space_errors.

get_state_errors(ID) :- state_error(ID,_,X), X\=invariant_violated, X\=abort_error(_,_,_,_),
  create_state_error_description(X,Msg),error_occurred(Msg),fail.
get_state_errors(ID) :- time_out_for_invariant(ID),error_occurred(time_out_for_invariant),fail.
get_state_errors(ID) :- time_out_for_assertions(ID),error_occurred(time_out_for_assertions),fail.
get_state_errors(ID) :- time_out_for_node(ID,_,time_out),error_occurred(time_out),fail.
get_state_errors(ID) :-
   time_out_for_node(ID,_,virtual_time_out(_)), %print(virtual_time_out_for_node(ID)),nl,
   error_occurred(virtual_time_out),fail.
get_state_errors(_).


create_state_error_description(eventerror(Event,Error,_),Description) :- !,
    functor(Error,Functor,_),
    ajoin(['event_error:',Event,':',Functor],Description).
create_state_error_description(StateError,Description) :-
    functor(StateError,Functor,_),
    atom_concat('state_error:',Functor,Description).

% require a real machine to be loaded
check_loaded_not_empty(Action) :-
    file_loaded(true,'$$empty_machine'),!,
    add_error(probcli,'No file specified; cannot perform action: ',Action),
    error_occurred(loading),fail.
check_loaded_not_empty(Action) :- check_loaded(Action).

check_loaded(Action) :-
    ( file_loaded(true)  -> true
    ; file_loaded(error) -> fail /* we have already generated error message */
    ; otherwise ->
        add_error(probcli,'No file specified; cannot perform action: ',Action),
        error_occurred(loading),fail).

:- dynamic loaded_main_file/2.
loaded_main_file(File) :- loaded_main_file(_Ext,File).

:- use_module(tools,[get_filename_extension/2]).
load_main_file(MainFile,NOW,Already_FullyProcessed) :- retractall(loaded_main_file(_,_)),
               debug_print(20,'% Loading: '), debug_println(20,MainFile),
               writeln_log_time(loading(NOW,MainFile)),
               get_filename_extension(MainFile,Ext),
               debug_println(6,file_extension(Ext)),
               start_probcli_timer(Timer),
               load_spec_file(Ext,MainFile,Already_FullyProcessed),
               stop_probcli_debug_timer(Timer,'% Finished loading'),
               (Already_FullyProcessed==true -> true
                ; assert(loaded_main_file(Ext,MainFile))).
%load_spec_file('pl',MainFile) :- !, load_cspm_spec_from_pl_file(MainFile). % no longer needed ?
load_spec_file('csp',MainFile) :- !, load_cspm_spec_from_cspm_file(MainFile).
load_spec_file('cspm',MainFile) :- !, load_cspm_spec_from_cspm_file(MainFile).
load_spec_file('P',MainFile) :- !, load_xtl_file(MainFile).
load_spec_file('p',MainFile) :- !, load_xtl_file(MainFile). % sometimes windows is confused about the upper case letter....
load_spec_file('eventb',MainFile) :- !, load_eventb_file(MainFile).
load_spec_file('v',MainFile) :- !,
    print('---------------------------------'),nl,
    print('Loading: '),print(MainFile),nl,
    load_b_file_with_options(MainFile). % Siemens Rule File; maybe in future use -eval_rule_file
load_spec_file('prob',MainFile) :- !,load_prob_file(MainFile). % .prob files
load_spec_file('mch',MainFile) :- !,load_b_file_with_options(MainFile).
load_spec_file('sys',MainFile) :- !,load_b_file_with_options(MainFile).
load_spec_file('ref',MainFile) :- !,load_b_file_with_options(MainFile).
load_spec_file('imp',MainFile) :- !,load_b_file_with_options(MainFile).
load_spec_file('rmch',MainFile) :- !,load_b_file_with_options(MainFile).
load_spec_file('def',MainFile) :- !,load_b_file_with_options(MainFile). % .def DEFINITIONS file
load_spec_file('fuzz',MainFile) :- !,tcltk_open_z_file(MainFile).
load_spec_file('tex',MainFile) :- !,tcltk_open_z_tex_file(MainFile).
load_spec_file('zed',MainFile) :- !,tcltk_open_z_tex_file(MainFile).
load_spec_file('als',MainFile) :- !,tcltk_open_alloy_file(MainFile).
load_spec_file('tla',MainFile) :- !, load_tla_file(MainFile).
load_spec_file('eval',File) :- !, cli_set_empty_machine, cli_start_animation(eval_file), eval_string_or_file(file,File,exists,_,norecheck).
load_spec_file('pb',File) :- !, cli_set_empty_machine, cli_start_animation(pb_file), eval_string_or_file(file,File,exists,_,norecheck).
%load_spec_file('pml',MainFile) :- !,parsercall:call_promela_parser(MainFile),
%   parsercall:promela_prolog_filename(MainFile,PrologFile),
%   println_silent(consulting(PrologFile)),
%   tcltk_open_promela_file(PrologFile).
load_spec_file(EXT,MainFile) :- print_error('### Unknown file extension, assuming B machine:'),
                                 print_error(EXT),
                                 load_b_file_with_options(MainFile).

load_spec_file('pl',MainFile, Already_FullyProcessed) :- !, Already_FullyProcessed=true,
   printsilent('Processing PO file: '),printsilent(MainFile),nls,
   load_po_file(MainFile),
   (option(timeout(TO)) -> set_disprover_timeout(TO) ; reset_disprover_timeout),
   (option(disprover_options(L)) -> set_disprover_options(L) ; set_disprover_options([])),
   println_silent('Running ProB Disprover'),
   run_disprover_on_all_pos(Summary),
   print_disprover_stats,
   accumulate_infos(disprover,[po_files-1|Summary]),
   get_errors,
   (option(cli_check_disprover_result(Infos)) -> check_required_infos(Infos,Summary,load_po_file)
    ; option(strict_raise_error) -> check_required_infos([false-0,unknown-0,failure-0],Summary,load_po_file)
    % TO DO: provide way for user to specify expected info
     ; true),
     clear_loaded_machines.
load_spec_file('smt2',MainFile,Already_FullyProcessed) :- !, Already_FullyProcessed=true,
   printsilent('Processing SMT file: '),printsilent(MainFile),nls,
   (option(eval_repl([])) -> Opts = [repl] ; Opts=[]),
   smtlib2_file(MainFile,Opts).
load_spec_file(EXT,F,false) :- load_spec_file(EXT,F).

load_b_file_with_options(File) :-
   (option(release_java_parser) -> Options = [release_java_parser,use_fastread] ; Options = []),
   % TO DO: automatically release if no option requires parsing and no more file uses it; or print warning if release will affect other options like -repl (DEFINITIONS not available,...)
   load_b_file(File,Options).

% do not perform -execute_all if no parameters provided
do_not_execute_automatically('pl').
do_not_execute_automatically('smt2').

test_kodkod_and_exit(MaxResiduePreds,NOW) :-
    start_animation_without_computing,
    test_kodkod(MaxResiduePreds),
    halt_prob(NOW,0).

compare_kodkod_performance1(KPFile,Iterations,NOW) :-
    start_animation_without_computing,
    compare_kodkod_performance(KPFile,Iterations),
    halt_prob(NOW,0).

:- use_module(parsercall,[check_java_version/2,get_parser_version/1, ensure_console_parser_launched/0]).
check_java_version :- check_java_version(V,Result),
    format('Result of checking Java version:~n  ~w~n',[V]),
    (Result=compatible -> check_parser_version
      ; add_error(check_java_version,V)).

check_parser_version :- get_parser_version(PV),!,
    format('  ProB B Java Parser available in version: ~w.~n ',[PV]). % will also launch parser
check_parser_version :- add_error(check_parser_version,'Cannot start Java B Parser to obtain version number').

:- use_module(version).
print_short_version :-
    version(V1,V2,V3,Suffix),revision(Rev),
    format('VERSION ~p.~p.~p-~p (~p)~N',[V1,V2,V3,Suffix,Rev]).

:- use_module(parsercall,[get_parser_version/1, get_java_command_path/1, get_java_fullversion/1]).
print_version :-
    print('ProB Command Line Interface'),nl,
    print_probcli_version(user_output),
    ( option_verbose ->
        prolog_flag(system_type,SysType),
        format('  Prolog System Type: ~p~N', [SysType]), % development or runtime
        safe_absolute_file_name(prob_home('.'),AppDir),
        format('  Application Path: ~p~N', [AppDir]),
        (get_java_command_path(JavaPath)
          -> format('  Java Runtime: ~p~N', [JavaPath]),
             (get_java_fullversion(JavaVersion)
               ->  format('  Java Version: ~s~N', [JavaVersion])
               ;   format('  Java Version: *** not available ***~N',[])
             ),
             (get_parser_version(ParserVersion)
               -> format('  Java Parser: ~p~N', [ParserVersion])
               ;  format('  Java Parser: *** not available ***~N',[])
             )
          ;  format('  Java Runtime: *** not available ***~N',[])
        ),
        (option(very_verbose) -> print_prolog_flags ; true)
     ;  true
    ), print_compile_time_flags.

print_probcli_version(Stream) :-
    version(V1,V2,V3,Suffix),
    revision(Rev),  lastchangeddate(LCD), prolog_flag(version,PV),
    format(Stream,'  VERSION ~p.~p.~p-~p (~p)~N  ~p~N  Prolog: ~p~N',
           [V1,V2,V3,Suffix,Rev,LCD,PV]).

:- use_module(compile_time_flags,[compile_time_flags/1, relevant_prolog_flags/1]).
print_compile_time_flags :-
   compile_time_flags(list(Flags)),
   (Flags=[], \+ option_verbose -> true ; format('  COMPILE TIME FLAGS: ~w~N',[Flags])).
print_prolog_flags :-
   relevant_prolog_flags(Flags),
   format('  PROLOG FLAGS: ~w~N',[Flags]).

print_help :-
          print_version,
          print('Usage: probcli FILE [OPTIONS]'),nl,
          print('  OPTIONS are: '),nl,
          print('     -mc Nr       model check; checking at most Nr states'),nl,
          print('     -model_check model check without limit on states explored'),nl,
          ( \+ option_verbose ->
          print('     -noXXX       XXX=dead,inv,goal,ass (for model check)'),nl % -nodead, -noinv, -nogoal, -noass
          ;
          print('     -nodead      do not look for deadlocks (for model check, animate, execute)'),nl,
          print('     -noinv       do not look for invariant violations (for model check, animate, execute)'),nl,
          print('     -nogoal      do not look for GOAL predicate (for model check, execute)'),nl,
          print('     -noass       do not look for ASSERTION violations (for model check, execute)'),nl
          ),
          print('     -bf          proceed breadth-first (default is mixed bf/df)'),nl,
          print('     -df          proceed depth-first'),nl,
          print('     -mc_mode M   M=hash,heuristic,random,dlk,breadth-first,depth-first,mixed'),nl, % dlk stands for out_degree_hash
          print('     -timeout N       total timeout in ms for model/refinement checking and'),nl,
          print('                      and execute steps and disprover checks'),nl,
          print('     -disable_timeout disable timeouts for operations, invariants,....'),nl, % speeds up mc
          print('     -t               trace check (associated .trace file must exist)'),nl,
          print('     -init            initialise specification'),nl,
          print('     -cbc OPNAME      constraint-based invariant checking for an operation'),nl,
          print('                      (you can also use OPNAME=all)'),nl,
          print('     -cbc_deadlock    constraint-based deadlock checking'),nl,
          ( \+ option_verbose -> true ;
          print('     -cbc_deadlock_pred PRED  as above but with additional predicate'),nl
          ),
          print('     -cbc_assertions  constraint-based static assertion checking'),nl,
          print('     -cbc_refinement  constraint-based static refinement checking'),nl,
          print('     -cbc_sequence S  constraint-based search for sequence of operations'),nl,
          print('     -strict          raise error if model-checking finds counter example'),nl,
          print('                      or trace checking fails or any error state found'),nl,
          print('     -expcterr ERR    expect error to occur (ERR=cbc,mc,ltl,...)'),nl,
          print('     -animate Nr      random animation (max. Nr steps)'),nl,
          print('     -animate_all     random animation until a deadlock is reached'),nl,
          print('     -animate_stats   provide feedback which operations are animated or executed'),nl,
          print('     -execute Nr      execute specification (maximally Nr steps)'),nl,
          print('                      in contrast to -animate: stops at first operation found, is deterministic,'),nl,
          print('                      does not store intermediate states and does not use TIME_OUT preference'),nl,
          print('     -execute_all     execute until a deadlock, direct loop, goal or error is reached'),nl,
          print('     -execute_monitor monitor performance of execute'),nl,
          print('     -his File        write history to File'),nl,
          print('     -his_option O    additional option when writing a history (show_init,show_states,json,trace_file)'),nl,
          print('     -sptxt File      save constants and variable values of last discovered state to File'),nl,
          print('     -sstxt Dir       save constants and variable values of all discovered states to files in Dir'),nl,
          print('     -cache Directory automatically save constants to files and avoid recomputation'),nl,
          print('     -det_check       check if animation steps are deterministic'),nl,
          print('     -det_constants   only check if SETUP_CONSTANTS step is deterministic'),nl,
          ( \+ option_verbose -> true ;
          print('     -i               interactive animation. Only for interactive sessions,'),nl,
          print('                      the output can arbitrarily change in future versions. '),nl,
          print('                      Do not build automatic tools using the interactive mode'),nl
          ),
          print('     -repl            start interactive read-eval-loop'),nl,
          print('     -eval "E"        evaluate expression or predicate'),nl,
          print('     -eval_file FILE  evaluate expression or predicate from file'),nl,
          print('     -c               print coverage statistics'),nl,
          print('     -cc Nr Nr        print and check coverage statistics'),nl,
          print('     -vacuity_check   look for vacuous implications in invariant'),nl,
          print('     -cbc_redundant_invariants Nr  find redundant invariants, expecting Nr'),nl, % Nr exepcted
          print('     -statistics      print memory and other statistics at the end'),nl,
          print('     -p PREF Val      set preference to value'),nl,
          print('     -prefs FILE      set preferences from Prolog file'),nl,
          print('     -pref_group G S  set group G of preferences to predefined value set S'),nl,
          print('     -card GS Val     set cardinality (aka scope) of B deferred set'),nl,
          print('     -goal "PRED"     set GOAL predicate for model checker'),nl,
          print('     -check_goal      check GOAL (after -mc, -t, or -animate)'),nl,
          print('     -scope "PRED"    set scope predicate for model checker'),nl,
          print('                      (only states satsifying this predicate will be examined)'),nl,
          print('     -s Port          start socket server on given port'),nl,
          print('     -ss              start socket server on port 9000'),nl,
          print('     -sf              start socket server on some free port'),nl,
          print('     -l LogFile       log activities in LogFile'),nl,
          print('     -ll              log activities in /tmp/prob_cli_debug.log'),nl,
          ( \+ option_verbose -> true ;
          print('     -lg LogFile      analyse logfile using gnuplot'),nl
          ),
          print('     -logxml LogFile      log activities in XML LogFile'),nl,
          print('     -logxml_write_ids P  write variables/constants starting with P to XML LogFile'),nl,
          print('     -pp FILE         pretty-print internal representation to file (or user_output)'), nl,
          print('     -ppf FILE        like -pp, but force printing of all type infos'),nl,
          print('     -ppB FILE        pretty-print Event-B model to file in valid B syntax'),nl,
          print('     -v               verbose'),nl,
          ( \+ option_verbose -> true ;
          print('     -vv          very verbose'),nl
          ),
          print('     -mc_with_tlc        model check using the model checker TLC'),nl,
          print('     -mc_with_lts_sym    model check using the model checker LTSmin (symbolic)'),nl,
          print('     -mc_with_lts_seq    model check using the model checker LTSmin (sequential)'),nl,

          ( \+ option_verbose -> true ;
          print('     -ltsmin_option OPT  set option for LTSmin (e.g, por)'),nl,
          print('     -ltsmin_ltl_output FILE      set output file for LTSMin'),nl,
          print('     -symbolic_model_check ALGO   ALGO is bmc, kinduction, ctigar, ic3'),nl,
          print('     -enabling_analysis_csv FILE   perform operation enabling analysis'),nl,
          print('     -feasibility_analysis   perform operation feasibility analysis'),nl,
          print('     -read_write_matrix   show read/write matrix for operations'),nl
          ),
          print('     -version            print version information (-svers for short info)'),nl,
          print('     -check_java_version check that Java version compatible with ProB parser'),nl,
          print('     -assertions         check ASSERTIONS'),nl,
          print('     -main_assertions    check ASSERTIONS from main file only'),nl,
          print('     -properties         check PROPERTIES'),nl,
          print('     -cache Dir          use directory "Dir" to cache constants and variables'),nl,
          print('     -ltlfile F          check LTL formulas in file F'),nl,
          print('     -ltlassertions      check LTL assertions (in DEFINITIONS)'),nl,
          print('     -ltllimit L         explore at most L states when model-checking LTL or CTL'),nl,
          print('     -ltlformula \"F\"   check the LTL formula F'),nl,
          print('     -ctlformula \"F\"   check the CTL formula F'),nl,
          print('     -save File          save state space for later refinement check'),nl,
          print('     -refchk File        refinement check against previous saved state space'),nl,
          print('     -mcm_tests Depth MaxStates EndPredicate File'),nl,
          print('                  generate test cases with maximum length Depth, explore'),nl,
          print('                  maximally MaxStates, the last state satisfies EndPredicate'),nl,
          print('                  and the test cases are written to File'),nl,
          print('     -mcm_cover Operation'),nl,
          print('                  when generating MCM test cases, Operation should be covered'),nl,
          print('     -cbc_tests Depth EndPredicate File'),nl,
          print('                  generate test cases by constraint solving with maximum'),nl,
          print('                  length Depth, the last state satisfies EndPredicate'),nl,
          print('                  and the test cases are written to File'),nl,
          print('     -cbc_cover Operation'),nl,
          print('                  when generating CBC test cases, Operation should be covered'),nl,
%          print('     -cbc_cover_all   try and cover all operations'),nl, % is now default if no cbc_cover provided
          print('     -test_description File'),nl,
          print('                      read information for test generation from File'),nl,
          print('     -dot CMD File         write a graph to a dot file, with CMD being one of:'),nl,
          (is_dot_command(Cmd),command_description(Cmd,_,Desc),
           format('          ~w : ~w~n',[Cmd,Desc]),fail
            ; true),
          print('     -dotexpr CMD Expr File write a graph for Expr to a dot file, with CMD:'),nl,
          (is_dot_command_for_expr(Cmd),command_description(Cmd,_,Desc),
           format('          ~w : ~w~n',[Cmd,Desc]),fail
            ; true),
          print('     -dot_output      Path generate dot files for false assertions/properties'),nl,
          print('     -dot_all         also generate dot files for true assertions/properties'),nl,
          print('     -csvhist E File  evaluate expression over history and generate CSV file.'),nl,
          print('     -load_state File load state from a saved state (generated by ProB Tcl/Tk or -save_state)'),nl,
       % For Eclipse Version only
       %%  print('     -parsercp  CP class path of the B Parser, this has to be a valid Java class path'),nl,
       %%   print('     -cspm         load CSP-M .csp file rather than B Machine .mch/.ref/.imp File'),nl,
       %%   print('     -csp         load CSP-M .pl file rather than B Machine File'),nl,

          /* Options -cspref, -cspdeadlock, -cspdeterministic, and -csplivelock are deprecated, should be excluded in favor of -csp_assertion */
          print('     -cspref Spec [m= Impl File'),nl,
          print('                  checks a refinement statement,'),nl,
          print('                  where Spec and Impl are processes from File, and \'m\' the type of the refinement:'),nl,
          print('                  \'T\' for traces, \'F\' for failures, or \'FD\' for failures-divergences.'),nl,
          print('     -cspdeadlock P m File'),nl,
          print('                  checks a process for deadlock,'),nl,
          print('                  where \'P\' is a process from File, and \'m\' the type of the model:'),nl,
          print('                  \'F\' for failures and \'FD\' for failures-divergences.'),nl,
          print('     -cspdeterministic P m File'),nl,
          print('                  checks a process for determinism,'),nl,
          print('                  where \'P\' is a process from File, and \'m\' the type of the model:'),nl,
          print('                  \'F\' for failures and \'FD\' for failures-divergences.'),nl,
          print('     -csplivelock P File'),nl,
          print('                  checks a process for divergence,'),nl,
          print('                  where \'P\' is a process from File.'),nl,
          /* Options -cspref, -cspdeadlock, -cspdeterministic, and -csplivelock are deprecated, should be excluded in favor of -csp_assertion */

          print('     -csp_assertion \"A\" File'),nl,
          print('                  checks the CSP assertion \'A\' on file \'File\''),nl,
	        print('     -csp_eval "E" evaluate CSP-M expression.'),nl,
          print('     -csp-guide File   CSP||B: Use the CSP File to control the B machine'),nl,
          print('     '),nl,
          ( \+ option_verbose -> true
           ;
	        print('     -test_mode       set random seed to the Prolog\'s current random state'),nl,
          print('     -rc          runtime checking of types/pre-/post-conditions'),nl,
          print('     -state_trace File      read a file of B predicates (one per line) and try find a matching trace.'),nl

          ),
          print(' FILE extensions are: '),nl,
          print('      .mch         for B abstract machines'),nl,
          print('      .ref         for B refinement machines'),nl,
          print('      .imp         for B implementation machines'),nl,
          print('      .sys         for Event-B abstract machines'),nl,
          print('      .rmch        for B Rule DSL machines'),nl,
          print('      .csp, .cspm  for CSP-M files, same format as FDR'),nl,
          print('      .eventb      for Event-B packages exported from Rodin ProB Plugin'),nl,
          print('      .tex, .zed   for Z models'),nl,
          print('      .tla         for TLA+ models'),nl,
          print('      .als         for Alloy models'),nl,
          print('      .P           for Prolog XTL models'),nl,
          ( option_verbose ->
              print(' Preferences PREF are: '),nl,
              print_eclipse_prefs
          ; otherwise ->
              print(' Use --help -v to print available preferences PREF'),nl
          ),
          print(' Set NO_COLOR environment variable to disable terminal colors'),nl,
          print(' More info at: http://www.stups.hhu.de/ProB/'),nl,
          print('       and at: http://www.prob2.de/probcli/'),nl,
          % http://www.stups.hhu.de/ProB/index.php5/Using_the_Command-Line_Version_of_ProB
          nl.


set_argv(V) :-
   debug_println(20,set_argv(V)),
   external_functions:set_argv_from_atom(V).

:- use_module(state_space_exploration_modes,[set_depth_breadth_first_mode/1]).
set_prefs :-
   if_option_set(socket(_,_), % then we may need the event(.) transition_info for the Java API
       preferences:set_preference(store_event_transinfo,true)),
   option(set_prefs_from_file(File)),
   debug_println(20,load_preferences(File)),
   preferences:load_preferences(File),
   fail.
set_prefs :-
   option(set_preference_group(P,V)),
   debug_println(20,set_preference_group(P,V)),
   set_preference_group(P,V),
   fail.
% eclipse preference or 'normal preference'
set_prefs :-
   option(set_pref(P,V)),
   set_pref(P,V),
   fail.
set_prefs :- option(set_card(P,V)),
   debug_println(20,set_card(P,V)),
   set_eclipse_card(P,V),
   fail.
set_prefs :-
    ( option(breadth_first) -> set_depth_breadth_first_mode(breadth_first)
    ; option(depth_first)   -> set_depth_breadth_first_mode(depth_first)
    ; option(depth_breadth_first_mode(M)) -> set_depth_breadth_first_mode(M)
    ; otherwise -> true
    ).
:- use_module(tools_strings, [convert_cli_arg/2]).
set_pref(P,V) :-
   debug_println(20,set_pref(P,V)),
   ( eclipse_preference(P,_)
     -> set_eclipse_preference(P,V)
     ; obsolete_eclipse_preference(P) -> probcli_add_light_warning('Obsolete preference: ',P)
     ; obsolete_preference(P) -> probcli_add_light_warning('Obsolete preference: ',P)
     ; % might be a term if its a plugin preference
     atom_codes(P,Codes),
     append(Codes,".",Codes2), % to make term readable by read_from_codes
     read_from_codes(Codes2,Preference),
     (nonvar(Preference),preference_val_type(Preference,_)
        -> convert_cli_arg(V,Value),
           set_preference(Preference,Value)
        ;  add_error(probcli,'Illegal preference:',P)
     )
   ).

% add non severe warning:
probcli_add_light_warning(Msg,Term) :- option(strict_raise_error),!,
    add_warning(probcli,Msg,Term). % does not write on user_error
probcli_add_light_warning(Msg,Term) :- add_message(probcli,Msg,Term).

set_optional_errors :- % register optional/expected errors in the error_manager; avoid printing on stderr
   reset_optional_errors_or_warnings,
   (option(optional_error(Type)) ; option(expect_error(Type)) ; option(expect_error_pos(Type,_Line,_Col))),
   register_optional_error_or_warning(Type),
   fail.
set_optional_errors.

model_check(Nr,Time,ErrRes,NOW) :-
   option(timeout(TO)),!,
   safe_time_out(model_check(Nr,Time,ErrRes),TO,Res),
   (Res=time_out -> print('*** Timeout occurred while model checking: '), print(TO),nl,nl,
                    writeln_log(timeout_occurred(NOW,model_check(Nr,Time,ErrRes))),
                    Time=time_out(TO), ErrRes=time_out
                 ; true).
model_check(Nr,Time,ErrRes,_NOW) :- model_check(Nr,Time,ErrRes).

:- use_module(model_checker,[model_checking_is_incomplete/6]).

add_model_checking_warnings(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations) :-
   %print(check(model_checking_is_incomplete(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations,Msg,Term))),nl,
   model_checking_is_incomplete(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations,Msg,Term),
   nl,
   add_warning(model_check_incomplete,Msg,Term),
   fail.
add_model_checking_warnings(_,_,_,_).

:- use_module(state_space,[current_state_id/1]).
model_check(Nr,Time,ErrRes) :-
   statistics(runtime,[T1,_]),
   statistics(walltime,[W1,_]),
   on_exception(user_interrupt_signal, model_check_aux(Nr,T1,W1,Time,ErrRes),
      (statistics(walltime,[W2,_]),  WTime is W2-W1,
       format_with_colour(user_error,[red],'~nmodel checking interrupted after ~w ms by user (CTRL-C)~n',[WTime]),
       coverage(just_summary),
       throw(user_interrupt_signal)
      )).

model_check_aux(Nr,T1,W1,Time,ErrRes) :-
   (option(no_deadlocks) -> FindDeadlocks=0 ; FindDeadlocks=1),
   (option(no_invariant_violations) -> FindInvViolations=0 ; FindInvViolations=1),
   (option(no_goal) -> FindGoal=0 ; FindGoal=1),
   (option(no_state_errors) -> FindStateErrors=0 ; FindStateErrors=1),
   (option(no_assertion_violations)
     -> FindAssViolations=0
      ; FindAssViolations=1
    ),
   get_preference(por,POR),
   get_preference(pge,PGE),
   StopAtFullCoverage=0,
   %STOPMCAFTER =  86400000, /* 86400000 = 1 day timeout */
   STOPMCAFTER = 1152921504606846975, /* equals 13,343,998,895 days */
   InspectExistingNodes = 1,
   write_xml_element_to_log(model_checking_options,[find_deadlocks/FindDeadlocks,
        find_invariant_violations/FindInvViolations, find_goal/FindGoal,
        find_assertion_violations/FindAssViolations,
        find_state_errors/FindStateErrors,
        partial_order_reduction/POR,
        partial_guard_evaluation/PGE,
        inspect_existing_nodes/InspectExistingNodes]),
   (tcltk_interface:do_model_check(Nr,NodesAnalysed,STOPMCAFTER,ErrRes,
             FindDeadlocks,FindInvViolations,FindGoal,
             FindAssViolations,FindStateErrors,StopAtFullCoverage,POR,PGE, InspectExistingNodes)
      -> (statistics(runtime,[T2,_]), statistics(walltime,[W2,_]),
           Time1 is T2-T1, WTime is W2-W1,
           (model_checker: expired_static_analysis_time(AnalysisTime) ->
               Time is Time1 - AnalysisTime
           ;   Time = Time1),
           formatsilent('Model checking time: ~w ms (~w ms walltime)~n',[Time,WTime]),
           formatsilent('States analysed: ~w~n',[NodesAnalysed]),
           state_space: get_state_space_stats(_,NrTransitions,_),
           printsilent('Transitions fired: '),printsilent(NrTransitions),nls,
           write_xml_element_to_log(model_checking_statistics,
                   [result/ErrRes,runtime/Time,walltime/WTime,states/NodesAnalysed,transitions/NrTransitions]),
          (ErrRes = no
           -> print('No counter example Found, not all nodes visited'),nl,
              add_warning(model_check_incomplete,'Not all nodes examined due to limit set by user: ',Nr)
           ;  (ErrRes=all
                -> (tcltk_find_max_reached_node
                     -> print('No counter example found. However, not all transitions were computed !')
                     ;  print_green('No counter example found. ALL nodes visited.')
                   ),
                   add_model_checking_warnings(FindInvViolations,FindDeadlocks,FindGoal,FindAssViolations)
                ;   print_red('*** COUNTER EXAMPLE FOUND ***'),nl,
                    debug_println(20,ErrRes),nl,
                    tcltk_interface:translate_error_for_tclk(ErrRes,TclTkRes),
                    print(TclTkRes),nl,
                    (option(silent) -> true
                      ; option(no_counter_examples) -> true % -nocounter
                      ; tcltk_interface:tcltk_get_history(list(Hist)),
                        length(Hist,Len),
                        format('*** TRACE (length=~w):~n',[Len]),
                        reverse(Hist,Trace),
                        print_nr_list(Trace),
                        (option_verbose,
                         current_state_id(ID),invariant_violated(ID)
                         -> print('*** '),
                            b_interpreter:analyse_invariant_for_state(ID)
                          ; true)
                    ),
                    error_occurred(TclTkRes)
              )
          ),nl
         )
      ;  ( statistics(runtime,[T2,_]),
           Time1 is T2-T1,
           (model_checker: expired_static_analysis_time(AnalysisTime) ->
               Time is Time1 - AnalysisTime
           ;   Time = Time1),
           printsilent('Model checking time: '), printsilent(Time), printsilent(' ms'),nls,
           print_error('*** Model checking FAILED '),nl,
           ErrRes=fail,
           definite_error_occurred)
   ).

% perform all cbc checks on current machine
cbc_check(_NOW) :-
     option(cbc_deadlock_check(DeadlockGoalPred)),
     cbc_deadlock_check(DeadlockGoalPred),
     fail.
cbc_check(_NOW) :-
     option(constraint_based_check(OpName)),
     constraint_based_check(OpName),
	   fail.
cbc_check(_NOW) :- option(cbc_assertions(AllowEnumWarning,Options)),
	    cbc_assertions(AllowEnumWarning,Options),
      fail.
%cbc_check(NOW) :-
%     option(cbc_pred(TargetPredString)),
%     check_loaded(cbc_pred),
%     print('% Starting Constraint-Based Check for Predicate: '), print(TargetPredString),nl,
%     b_set_up_valid_state_with_pred(NormalisedState,Pred)  TO DO: add this feature
cbc_check(_NOW) :- option(cbc_sequence(Sequence,TargetPredString,Findall)),
      cbc_sequence(Sequence,TargetPredString,Findall),
      fail.
cbc_check(_NOW) :- option(cbc_refinement),
     cbc_refinement,
      fail.
cbc_check(_NOW) :- option(cbc_redundant_invariants(NrExpected)),
      cbc_redundant_invariants(NrExpected),
      fail.
cbc_check(_).

:- use_module(tcltk_interface,[tcltk_constraint_based_check/2,
   tcltk_constraint_based_check_with_timeout/2,
   tcltk_constraint_find_deadlock_state_with_goal/3,
   tcltk_cbc_find_trace/4,
   tcltk_cbc_refinement_check/2]).
:- use_module(probsrc(bmachine),[b_is_operation_name/1]).

constraint_based_check(all) :-
     check_loaded_not_empty(constraint_based_check),
     print_repl_prompt_s('% Starting Constraint-Based Check for all Operations: '),nl,
     start_xml_feature(cbc_operation_check,all_operations,true,FINFO),
          (tcltk_constraint_based_check(list(Result),ErrorsWereFound)
            ->  print('% Constraint-Based Check Result: '),nl,
                print(Result),nl,
                write_result_to_file(Result),
                (ErrorsWereFound=true
                  ->  print_red('*** CONSTRAINT-BASED CHECK FOUND ERRORS ***'),nl, error_occurred(cbc)
                  ;  (ErrorsWereFound=false -> print_green('NO ERRORS FOUND'),nl)
                       ; print_red('*** TIMEOUT OCCURRED ***'),nl,error_occurred(cbc)
                )
            ;  write_result_to_file(cbc_check_failed), Result=internal_error, ErrorsWereFound=false,
               add_internal_error('ConstraintBasedCheck unexpectedly failed. ',cbc_check(all)),definite_error_occurred
          ),nl,
      write_xml_element_to_log(cbc_check_result,[result/Result,errors_were_found/ErrorsWereFound]),
      stop_xml_feature(cbc_operation_check,FINFO).
constraint_based_check(OpName) :- OpName\=all,  % -cbc OpName
     check_loaded_not_empty(constraint_based_check),
     print_repl_prompt_s('% Starting Constraint-Based Check for Operation: '), print(OpName),nl,
     start_xml_feature(cbc_operation_check,operation,OpName,FINFO),
	   (tcltk_constraint_based_check_with_timeout(OpName,Result)
       ->  print('% Constraint-Based Check Result: '),nl, print(Result),nl,
        write_result_to_file(Result),
        (Result=time_out
          ->  print_red('*** TIMEOUT OCCURRED ***'),nl, error_occurred(cbc)
          ;  (Result=ok -> print_green('NO ERRORS FOUND'),nl)
             ; print_red('*** CONSTRAINT-BASED CHECK FOUND ERRORS ***'),nl,error_occurred(cbc) )
       ;  write_result_to_file(constraint_based_check_failed), Result=internal_error,
         add_error(probcli,'ConstraintBasedCheck unexpectedly failed'),
         (b_is_operation_name(OpName) -> true
           ; add_error(probcli,'Unknown Operation Name: ',OpName)),
         definite_error_occurred
	    ),nl,
      write_xml_element_to_log(cbc_check_result,[result/Result]),
      stop_xml_feature(cbc_operation_check,FINFO).
cbc_deadlock_check(DeadlockGoalPred) :-
     print_repl_prompt_s('% Starting Constraint-Based DEADLOCK check '),nl,
     start_xml_feature(cbc_deadlock_check,FINFO),
     (tcltk_constraint_find_deadlock_state_with_goal(DeadlockGoalPred,false,Res)
       -> write_result_to_file(Res),
         (Res=time_out ->
            print_red('*** TIME_OUT occurred ***'),nl,
            error_occurred(cbc_deadlock_check_time_out)
          ; print_red('*** DEADLOCK state found ***'),nl,
            error_occurred(cbc_deadlock_check),
            (silent_mode(on) -> true
             ; print('*** STATE = '),nl,
               current_b_expression(DBState), translate:print_bstate(DBState),nl,
               print('*** END DEADLOCKING STATE '),nl
            )
         )
       ;  write_result_to_file(no_deadlock_found), Res=no_deadlock_found,
          print_green('No DEADLOCK state found'),nl
     ),
      write_xml_element_to_log(cbc_check_result,[result/Res]),
      stop_xml_feature(cbc_deadlock_check,FINFO).
cbc_assertions(AllowEnumWarning,Options) :-
     findall(OPT,option(cbc_option(OPT)),FullOptions,Options),
     check_loaded_not_empty(cbc_assertions),
     print_repl_prompt_s('% Starting Constraint-Based static ASSERTIONS check '),nl,
     start_xml_feature(cbc_assertion_check,allow_enumeration_warning,AllowEnumWarning,FINFO),
     write_prolog_term_as_xml_to_log(options(Options)),
     (cbc_constraint_find_static_assertion_violation(Res,FullOptions)
       -> process_cbc_assertion_result(Res,AllowEnumWarning)
       ;  write_result_to_file(cbc_assertions_failed), Res=internal_error,
          print_red('CBC Check failed'),nl,
          error_occurred(cbc_assertions_failure)
     ),
      write_xml_element_to_log(cbc_check_result,[result/Res]),
      stop_xml_feature(cbc_assertion_check,FINFO).
cbc_sequence(Sequence,TargetPredString,Findall) :-
     check_loaded_not_empty(cbc_sequence),
     print_repl_prompt_s('% Starting Constraint-Based Check for Sequence: '), print_repl_prompt_s(Sequence),
     start_xml_feature(cbc_sequence_check,sequence,Sequence,FINFO),
     (TargetPredString='' -> true ; print(' with target: '), print(TargetPredString)),
     nl,
     write_xml_element_to_log(options,[target_predicate/TargetPredString]),
     (tcltk_cbc_find_trace(Sequence,TargetPredString,Findall,Res)
       -> (Res=ok -> print_green('Sequence found and executed'),nl
           ; Res=time_out -> error_occurred(cbc_sequence_time_out)
           ; Res=no_solution_found -> print_red('*** NO SOLUTION FOUND '),error_occurred(cbc_sequence_no_solution_found)
           ; Res=nr_cbc_sols(NrSols) -> print('*** # SOLUTIONS FOUND: '),print(NrSols),nl
           ; print_red('*** Unknown result: '), print(Res),nl,
             error_occurred(cbc_sequence)
          )
       ;  print('*** Internal error: Check failed '), error_occurred(cbc_sequence), Res=internal_error
     ),
      write_xml_element_to_log(cbc_check_result,[result/Res]),
      stop_xml_feature(cbc_sequence_check,FINFO).
cbc_refinement :-
     check_loaded_not_empty(cbc_refinement),
     print_repl_prompt_s('% Starting Constraint-Based static refinement check '),nl,
     start_xml_feature(cbc_refinement_check,FINFO),
     tcltk_cbc_refinement_check(list(Result),ErrorsWereFound),
     print('% Constraint-Based Refinement Check Result: '),nl,print(Result),nl,
     (ErrorsWereFound = time_out -> print_red('*** TIME_OUT occurred ***'),nl,error_occurred(cbc_refinement_time_out) ;
      ErrorsWereFound = true -> print_red('*** Refinement Violation found ***'),nl,error_occurred(cbc_refinement) ;
      otherwise -> print_green('No static Refinement Violation found'),nl
     ),
      write_xml_element_to_log(cbc_check_result,[errors_were_found/ErrorsWereFound]),
      stop_xml_feature(cbc_refinement_check,FINFO).
:- use_module(b_state_model_check,[cbc_find_redundant_invariants/2]).
cbc_redundant_invariants(NrExpected) :-
     check_loaded_not_empty(cbc_redundant_invariants),
     print_repl_prompt_s('% Starting Constraint-Based invariant redundancy check'),nl,
     start_xml_feature(cbc_redundant_invariants,FINFO),
     cbc_find_redundant_invariants(Res,TimeoutOccured),
     length(Res,NrInvs),
     (Res = [] -> print_green('No redundant invariants found'),nl
      ; format('*** REDUNDANT INVARIANTS (~w) ***~n',[NrInvs]),
        prnt(1,Res), nl
      ),
      (NrExpected = NrInvs -> true
       ;  format_with_colour(user_error,[red],'*** Expected ~w redundant invariants (instead of ~w).~n',[NrExpected,NrInvs]),
          error_occurred(cbc_redundant_invariants)),
      write_xml_element_to_log(cbc_redundant_invariants,[redundant_invariants/NrInvs, timeout_occured/TimeoutOccured]),
      stop_xml_feature(cbc_redundant_invariants,FINFO).

prnt(_,[]).
prnt(N,[H|T]) :- format(' ~w : ~w~n',[N,H]), N1 is N+1, prnt(N1,T).

:- use_module(solver_interface,[predicate_uses_unfixed_deferred_set/2, unfixed_typed_id_in_list/3]).
:- use_module(bsyntaxtree,[get_texpr_id/2]).
:- use_module(translate,[pretty_type/2]).
process_cbc_assertion_result(time_out,_) :- !,
        write_result_to_file(no_counterexample_found('"TIME_OUT"')),
	    print_red('*** TIME_OUT occurred ***'),nl,
	    error_occurred(cbc_assertions_time_out).
process_cbc_assertion_result(no_counterexample_exists(Constants,TotPredicate,OtherInfo),AllowEnumWarning) :- !,
		print_green('No counter-example to ASSERTION exists '),(OtherInfo=[] -> true ; print(OtherInfo)),nl,
		(unfixed_typed_id_in_list(TID,CType,Constants) % TO DO: look only at component
		  -> write_deferred_set_used(AllowEnumWarning),
			 get_texpr_id(TID,CID),pretty_type(CType,CTypeS),
		     format('Warning: Some constants use deferred sets (e.g., ~w:~w) which have only been checked for a single cardinality!~n',[CID,CTypeS])
		  ; predicate_uses_unfixed_deferred_set(TotPredicate,CType)
		    -> write_deferred_set_used(AllowEnumWarning),pretty_type(CType,CTypeS),
		       format('Warning: Some quantified variables use deferred sets (e.g., ~w) which have only been checked for a single cardinality!~n',[CTypeS]) % happens for tests 1173, 1174
		  ; write_result_to_file(no_counterexample_exists)
		    %,print('Computing unsat core: '),nl,unsat_cores:unsat_core(TotPredicate,Core),print('CORE: '),translate:print_bexpr(Core),nl
		). % WE HAVE A PROOF
process_cbc_assertion_result(no_counterexample_found,AllowEnumWarning) :- !,
		write_result_to_file(no_counterexample_found('"Enumeration Warning"')),
		print('No counter-example for ASSERTION found (*enumeration warning occured*)'),nl,
		(AllowEnumWarning=true -> true ; error_occurred(cbc_assertions_enumeration_warning)).
process_cbc_assertion_result(counterexample_found,_) :- !,
		write_result_to_file(counterexample_found),
    print_red('*** Counter-example for ASSERTION found ***'),nl,
		error_occurred(cbc_assertions),
		(silent_mode(on) -> true
             ; print('*** STATE = '),nl,
		           current_b_expression(DBState), translate:print_bstate(DBState),nl,
		           print('*** END ASSERTION counter-example STATE '),nl
		),
		(get_dot_file('cbc_assertions',DFile) -> generate_dot_from_assertions(DFile) ; true).
process_cbc_assertion_result(Res,A) :-
		write_result_to_file(Res),
		add_internal_error('Unknown: ',process_cbc_assertion_result(Res,A)).


write_deferred_set_used(AllowEnumWarning) :-
             write_result_to_file(no_counterexample_found('"Deferred Sets Used"')),
			 (AllowEnumWarning=true -> true ; error_occurred(cbc_assertions_enumeration_warning)).

:- use_module(tools_io,[safe_open_file/4]).
write_result_to_file(Result) :- option(cbc_result_file(FILE)),
   safe_open_file(FILE,write,Stream,[]),
   !,
   write(Stream,Result),
   close(Stream).
write_result_to_file(_).



if_option_set(Option,Call) :-
    if_option_set(Option,Call,true).
if_option_set(Option,Then,Else) :-
    ( option(Option) -> call(Then)
    ; otherwise      -> call(Else)).
ifm_option_set(Option,Call) :-
    ifm_option_set(Option,Call,true).
ifm_option_set(Option,Then,Else) :- % can perform multiple options
   findall(Then,option(Option),As),
   (As=[] -> call(Else) ; perform(As)).
perform([]).
perform([A|T]) :- call(A), perform(T).
if_option_set_loaded(Option,Action,Call) :-
    ( (option(Option),check_loaded_not_empty(Action)) ->
        call(Call)
    ; otherwise -> true).
ifm_option_set_loaded(Option,Action,Call) :- % can perform multiple options
   findall(Call,(option(Option),check_loaded_not_empty(Action)),As), perform(As).



if_options_set(Option,Call) :- % allow multiple solutions for Option
   option(Option),call(Call),fail.
if_options_set(_,_).

print_options :- print('CLI OPTIONS: '),nl,
                 option(Option), print(Option), nl, fail.
print_options :- nl.

:- use_module(enabling_analysis,[infeasible_operation_cache/1]).
:- use_module(sap,[explore_and_generate_testcases/7,cbc_gen_test_cases_from_string/5, tcl_get_stored_test_cases/1]).
:- use_module(translate,[print_bexpr/1]).

mcm_test_case_generation(ADepth,AMaxStates,ATarget,Output) :-
    arg_is_number(ADepth,MaxDepth),
    arg_is_number(AMaxStates,MaxStates),
    bmachine:b_parse_machine_predicate(ATarget,Target),!,
    get_comma_or_space_separated_options(mcm_cover,Events),
    (option(silent) -> true
     ; print('mcm test case generation, maximum search depth: '),print(MaxDepth),nl,
       print('mcm test case generation, maximum number of states: '),print(MaxStates),nl,
       print('mcm test case generation, target state predicate: '),print_bexpr(Target),nl,
       print('mcm test case generation, output file: '),print(Output),nl,
       print('mcm test case generation, events to cover: '),print_list(Events),nl
    ),
    explore_and_generate_testcases(Events,Target,MaxDepth,MaxStates,Output,NumTests,Uncovered),
    printsilent('mcm test case generation, generated test cases: '),printsilent(NumTests),nls,
    print_uncovered('mcm test case generation, ',Uncovered).
mcm_test_case_generation(_ADepth,_AMaxStates,_ATarget,_Output) :-
    print_error('### MCM Test Case Generation failed'),
    error_occurred(mcm_tests).

cbc_test_case_generation(ADepth,TargetString,Output) :-
    arg_is_number(ADepth,MaxDepth),
    ( option(cbc_cover_all) -> Events=all
       ; (get_comma_or_space_separated_options(cbc_cover,Events), Events \= []) -> true
       ; Events=all ),
    (\+ option(cbc_cover_final) -> FEvents = Events
     ; Events=all -> FEvents=all,
         add_error(cbc_cover_final,'Option cbc_cover_final not compatible with trying to cover all events')
     ; FEvents = final(Events),
       println_silent('constraint based test case generation, target events considered final')),
    printsilent('constraint based test case generation, maximum search depth: '),printsilent(MaxDepth),nls,
    printsilent('constraint based test case generation, target state predicate: '),printsilent(TargetString),nls,
    printsilent('constraint based test case generation, output file: '),printsilent(Output),nls,
    (TargetString = '#not_invariant' -> BMC=invariant_violation
     ; TargetString = '#deadlock' -> BMC=deadlock
     ; BMC = 'none'),
    (BMC \= 'none' ->
       printsilent('constraint based test case generation, performing bounded model checking'),nls
    ;  option(silent) -> true
    ;  print('constraint based test case generation, events to cover: '),print_list(Events),nl),
    cbc_gen_test_cases_from_string(FEvents,TargetString,MaxDepth,Output,Uncovered),
    !,
    %print('mcm test case generation, generated test cases: '),print(NumTests),nl,
    (BMC \= 'none'
      -> tcl_get_stored_test_cases(list(Tests)), %print(tests(Tests)),nl,
        (Tests=[] -> print_green('No counterexample found'),nl
         ; Tests = [_|_], BMC=deadlock -> add_error(deadlock,'Deadlock found by bmc')
         ; Tests = [_|_] -> add_error(invariant_violation,'Invariant violation found by bmc')
         ; add_internal_error('Unexpected bmc result: ',Tests)
        )
      ; Uncovered=[_|_],option(strict_raise_error)
        -> add_error(cbc_tests,'Uncovered events: ',Uncovered)
      ; print_uncovered('constraint based test case generation, ',Uncovered)
    ).
cbc_test_case_generation(_ADepth,_ATarget,_Output) :-
    print_error('### CBC Test Case Generation failed'),
    error_occurred(cbc_tests).

print_uncovered(Msg,Uncovered) :-
   include(infeasible_operation_cache,Uncovered,Infeasible),
    (Infeasible=[]
      -> format('~wuncovered events: ',[Msg]),print_list(Uncovered),nl
      ;  format('~winfeasible uncovered events: ',[Msg]),print_list(Infeasible),nl,
         exclude(infeasible_operation_cache,Uncovered,Feasible),
         format('~wuncovered events: ',[Msg]),print_list(Feasible),nl
    ).

print_list(all) :- print('** all **').
print_list(list(L)) :- print_list(L). % possibly not used
print_list([]) :- print('** none **').
print_list([H|T]) :- length([H|T],Len), format('(~w) ',[Len]),
    print(H),print(' '),print_list2(T).
print_list2([]).
print_list2([H|T]) :- print(H),print(' '),print_list2(T).

get_comma_or_space_separated_options(Option,Selection) :-
    functor(O,Option,1),
    findall(E, (option(O),arg(1,O,CommaSep),
                   split_by_seperator(CommaSep,Es),
                   member(E,Es)),
            Selection).

split_by_seperator(NonAtomic,Res) :- \+ atomic(NonAtomic),!, Res=[NonAtomic].
split_by_seperator(String,Strings) :-
    atom_chars(String,Chars),
    split_by_seperator2(Chars,Strings).
split_by_seperator2(Chars,Result) :-
    append(AChars,[X|B],Chars),seperator(X),!,
    (AChars=[] -> Result=Rest ; atom_chars(A,AChars), Result=[A|Rest]),
    split_by_seperator2(B,Rest).
split_by_seperator2(Chars,[String]) :- atom_chars(String,Chars).

seperator(',').
seperator(' ').
seperator(';').

ltl_check_assertions :-
    (option(ltl_limit(Limit)) -> true; Limit= -1), % -1 means no limit
    print('LTL Model checking assertions'),nl,
    ltl_check_assertions(Limit,Outcome),
    ( Outcome = pass -> print_green('LTL check passed'),nl
    ; Outcome = fail -> print_red('*** LTL check failed'),nl,error_occurred(ltl)
    ; Outcome = no_tests -> print_red('*** No LTL assertions found, test failed'),nl,definite_error_occurred
    ; otherwise -> print_red('*** An error occurred in the LTL assertion test'),nl,
	definite_error_occurred).

:- use_module(probltlsrc(ltl),[parse_ltlfile/2]).
ltl_check_file(Filename) :-
    (option(ltl_limit(Limit)) -> true; Limit= -1), % -1 means no limit
    print_repl_prompt_s('LTL Model checking assertions'),nl,
    parse_ltlfile(Filename, Formulas),
    ( real_error_occurred ->
        print_red('An error occurred while parsing the LTL file.\n'), fail
    ; otherwise ->
        ltl_check_formulas(Formulas,Limit)).

:- use_module(probltlsrc(ltl),[ltl_model_check2/4]).
ltl_check_formulas([],_) :-
    print_green('All LTL formulas checked.\n').
ltl_check_formulas([formula(Name,F)|Rest],Limit) :-
    print('Checking formula '),print(Name),print(':\n'),
    ltl_model_check2(F,Limit,init,Status),
    ( Status == no ->
        print_red('Counter-example found for formula \"'),print_red(Name),
        print_red('\", saving trace file.\n'),
        ajoin(['ltlce_', Name, '.trace'], Tracefile),
        tcltk_save_history_as_trace_file(prolog,Tracefile),
        add_error(ltl_counterexample,'Counter-example was found')
    ; Status == ok ->
        ltl_check_formulas(Rest,Limit)
    ; Status == incomplete ->
        ajoin(['Model was not completly model-checked, aborted after ',Limit,' new states'],
              Msg),
        add_error(ltl,Msg)
    ; otherwise ->
        ajoin(['Model checker returns unexpected result (',Status,')'],Msg),
        add_error(ltl,Msg)).

cli_ltl_model_check(Formula,ExpectedStatus,Status) :-
    (option(ltl_limit(Max)) -> true; Max = -1), % -1 means no limit
    start_xml_feature(ltl_model_check,formula,Formula,FINFO),
    ltl_model_check(Formula,Max,init,Status),
    write_xml_element_to_log(model_check_result,[status/Status,expected_status/ExpectedStatus]),
    check_status(Status,ExpectedStatus,Formula,ltl),
    stop_xml_feature(ltl_model_check,FINFO).

cli_ctl_model_check(Formula,ExpectedStatus,Status) :-
    (option(ltl_limit(Max)) -> true; Max = -1), % -1 means no limit
    start_xml_feature(ctl_model_check,formula,Formula,FINFO),
    ctl_model_check(Formula,Max,init,Status),
    write_xml_element_to_log(model_check_result,[status/Status,expected_status/ExpectedStatus]),
    check_status(Status,ExpectedStatus,Formula,ctl),
    stop_xml_feature(ctl_model_check,FINFO).

check_expected(St,Exp,Mode) :-
   (St=Exp -> true ; add_error(Mode,'Unexpected model checking result: ',Mode:St)).

check_status(ok,Expected,Formula,ltl) :- !, % TO DO: make uniform ? CTL returns true; LTL returns ok
    format_with_colour(user_output,[green],'LTL Formula TRUE.~nNo counter example found for ~w.~n',[Formula]),
    flush_output(user_output),
    check_expected(true,Expected,ltl).
check_status(true,Expected,Formula,ctl) :- !,
    format_with_colour(user_output,[green],'CTL Formula TRUE.~nNo counter example found for ~w.~n',[Formula]),
    flush_output(user_output),
    check_expected(true,Expected,ctl).
check_status(incomplete,Expected,Formula,LTLorCTL) :- !,
    incomplete_warning(LTLorCTL,Warning),
    add_warning(Warning, 'Warning: Model Check incomplete for: ', Formula),nl,
    format('No counter example found so far for ~w.~n',[Formula]),
    check_expected(incomplete,Expected,LTLorCTL).
check_status(NO,Expected,Formula,LTLorCTL) :- (NO=no ; NO=false),!, % TO DO: make uniform
    (Expected==false -> format_with_colour(user_error,[red],'Model Check Counter-Example found for: ~w~n',[Formula])
      ; add_error(LTLorCTL, 'Model Check Counter-Example found for: ', Formula),nl
    ),
    print('Formula '), print('FALSE.'),nl,
        %print('Counter example found for '),print(Formula),print('.'),nl,
    nl,
    check_expected(false,Expected,LTLorCTL).
check_status(Status,Expected,Formula,LTLorCTL) :-
    add_internal_error('Unknown status: ', check_status(Status,Expected,Formula,LTLorCTL)).

incomplete_warning(ltl,ltl_incomplete) :- !.
incomplete_warning(ctl,ctl_incomplete) :- !.
incomplete_warning(X,X).

evaluate_from_commandline :-
    retractall(eval_result(_,_)),
    option(eval_string_or_file(A,B,Q,E,Rchk)),   %print(eval(A,B,Q,E)),nl,
    % treat eval_string and eval_file together to ensure proper order of evaluation
    %  (only possible side-effect at the moment: formula can add new machine_string facts)
    eval_string_or_file(A,B,Q,E,Rchk),
    fail.
evaluate_from_commandline :- print_eval_results,
    % treat -repl option or -replay File option
    (option(eval_repl([File1|TF]))
      -> (repl_evaluate_expressions([File1|TF]) -> true ; true)
      ; option(eval_repl([]))
      -> (repl_evaluate_expressions([]) -> true ; true)
      ; true).

:- dynamic eval_result/2.
add_eval_result(R) :- retract(eval_result(R,N)),!,
   N1 is N+1, assert(eval_result(R,N1)).
add_eval_result(R) :- assert(eval_result(R,1)).
print_eval_results :- findall(R/N, eval_result(R,N), L), sort(L,SL),
   (SL=[] -> true ; format('Evaluation results: ~w~n',[SL])).

:- use_module(tools_printing,[print_error/1, format_error/2]).
%eval_string_or_file(string,_String,_,'FALSE',_Recheck) :- !. % comment in to skip evalf
eval_string_or_file(string,String,_,Expected,Recheck) :-
    set_current_probcli_command(eval_string(String)),
    (option(silent),nonvar(Expected) -> true
      ; nonvar(Expected) -> format('eval(~w): ~w~n',[Expected,String])
      ; format('eval: ~w~n',[String])
    ),
    reset_error_spans, % avoid underlining previous errors in eval_string
    eval_string_with_time_out(String,StringResult,EnumWarning,_LS),
    add_eval_result(StringResult),
    eval_check_result(StringResult,Expected,EnumWarning,eval_string,String),
    (Recheck=recheck(Mode) -> recheck_pp_of_last_expression(Mode,_,_) ; true),
   unset_current_probcli_command.
eval_string_or_file(file,File,Quantifier,Expected,_) :-
    % evaluate a single formula stored in a file
    set_current_probcli_command(eval_file(File)),
    turn_show_error_source_off, % reduce clutter in user feedback; eval_file used in ProB Logic Calculator for example
    formatsilent('~nEvaluating file: ~w~n',[File]),
    error_manager:reset_error_scopes, % TO DO: avoid that exceptions mess up error scopes in eval_string/file
    statistics(runtime,[Start,_]),nl,
    (Expected=='TRUE' -> TypeInfo=predicate(_) % avoids parsing as expression
     ; true),
    (eval_file(File,Quantifier,Result,EnumWarning,TypeInfo)
       -> add_eval_result(Result),
          eval_check_result(Result,Expected,EnumWarning,eval_file,File)
       ;  add_eval_result(eval_file_failed),
          print_error('### Eval from file failed: '), print_error(File),
          error_occurred(eval_file)
    ),
   (debug_mode(on) -> statistics(runtime,[Stop,_]), Time is Stop - Start,
                      print(File),print(' time: '), print(Time), print(' ms'),nl
      ; true),
  turn_show_error_source_on,
  unset_current_probcli_command.

eval_check_result(StringResult,Expected,_,Origin,File) :- Expected\=StringResult,!,
   format_error('### Evaluation error, expected result to be: ~w (but was ~w) in ~w~n',[Expected,StringResult,File]),
   error_occurred(Origin).
eval_check_result('NOT-WELL-DEFINED',Expected,_,Origin,File) :- var(Expected),!,
   format_error('### Evaluation NOT-WELL-DEFINED in ~w~n',[File]),
   error_occurred(Origin).
eval_check_result(_,_,EnumWarning,_,_) :- eval_gen_enum_warning(EnumWarning).

eval_gen_enum_warning(false) :- !.
eval_gen_enum_warning(time_out) :- !,error_occurred(eval_string_time_out).
eval_gen_enum_warning(_) :- print_error('### Enumeration warning occurred'),
                       error_occurred(eval_string_enum_warning,warning).
%repl :- repl_evaluate_expressions([]).
:- use_module(parsercall,[ensure_console_parser_launched/0]).
repl_evaluate_expressions(StartFiles) :-
                        get_errors, % first clear any errors from earlier commands
                        nl,
                        print('ProB Interactive Expression and Predicate Evaluator '), nl,
                        print('Type ":help" for more information.'),nl,
                        turn_show_error_source_off, % reduce clutter in user feedback
                        (option(evaldot(File))
                          -> print('Solutions written to dot file: '), print(File),nl
                          ;  true
                        ),
                        print_repl_prompt, % >>>
                        (ensure_console_parser_launched
                           -> maplist(prob_cli:set_repl_input_file(verbose),StartFiles),
                              top_level_eval
                            ; print('>>> ABORTING REPL'),nl),
                        turn_show_error_source_on.

:- use_module(user_interrupts,[interruptable_call/1]).
top_level_eval :-
  catch(top_level_eval1, halt(0), (format('~s', ["Bye."]), nl)).

:- use_module(tools_printing,[reset_terminal_colour/1, print_red/1, print_green/1]).
print_repl_prompt :- reset_terminal_colour(user_output), write('>>> ').
print_repl_prompt_s(_) :- option(silent),!.
print_repl_prompt_s(P) :- print_repl_prompt(P).
print_repl_prompt(P) :- reset_terminal_colour(user_output), write(P).
%print_repl_prompt(P) :- tools_printing:start_terminal_colour(dark_gray,user_output), write(P), reset_terminal_colour(user_output).

top_level_eval1 :-
 (interruptable_call(eval1) -> true
  ; print_red('Evaluation failed or interrupted'),nl,
    print_repl_prompt('Use :q to quit REPL'),nl),
  reset_errors,
  print_repl_prompt, % >>>
  top_level_eval1.
eval0 :- store_last_error_location_for_repl,
        reset_errors, % get_errors prints errors again and quits in -strict mode
        % However, reset_errors means that in subsequent REPL runs errors are not printed again!!
        print_repl_prompt, % >>>
        garbage_collect, eval1.
eval1 :- repl_multi_read_line(Expr), eval_probcli_repl_line(Expr).

:- dynamic last_repl_error/2.
store_last_error_location_for_repl :-
        retractall(last_repl_error(_,_)),
        check_error_span_file_linecol(_,File,Line,_,_,_),!,
        assert(last_repl_error(File,Line)).
store_last_error_location_for_repl.

:- dynamic current_repl_input_stream/2.
close_repl_input_stream(file_closed) :- retract(current_repl_input_stream(X,File)),!,
   format(":replayed ~w~n",[File]),
   close(X).
close_repl_input_stream(no_file).
:- use_module(tools_io,[safe_open_file/4]).
set_repl_input_file(_,File) :- current_repl_input_stream(_,File),!,
   add_error(set_repl_input_file,'Cyclic file replay: ',File).
set_repl_input_file(Verbose,File) :-
   % close_repl_input_stream, % this way we allow one REPL file to call another
   safe_open_file(File,read,Stream,[]),!,
   (Verbose=verbose -> format('Replaying REPL commands in file: ~w~n',[File]) ; true),
   asserta(current_repl_input_stream(Stream,File)).
set_repl_input_file(_,_).

repl_multi_read_line(Line) :-
   (current_repl_input_stream(Stream,_)
    -> repl_multi_read_line_aux(Stream,[],Line),
       format('~s~n',[Line])
     ; repl_multi_read_line_aux(user_input,[],Line)
    ).
repl_multi_read_line_aux(Stream,SoFar,Line) :-
  read_line(Stream,L),
  (L=end_of_file -> close_repl_input_stream(FileC),
                    (SoFar=[], FileC = file_closed
                     -> print_repl_prompt, % >>>
                        repl_multi_read_line(Line) % last line of file empty; do not process
                     ; FileC = file_closed -> Line=SoFar
                     ; Line=end_of_file) % user pressed CTRL-D
    ; append(LFront,[92],L)  % line ends with slash
	   -> append(SoFar,LFront,NewSoFar), print_repl_prompt('... '),
		  repl_multi_read_line_aux(Stream,NewSoFar,Line)
	; append(SoFar,L,Line)).

:- use_module(eval_strings).
:- dynamic trace_eval/0.

generate_atom_list([],[],R) :- !, R=[].
generate_atom_list([],Last,[NewAtom]) :-
   reverse(Last,RL),
   atom_codes(NewAtom,RL).
generate_atom_list([39|X],[],[QuotedAtom|T]) :- !,
   get_quoted_atom(X,[],QuotedAtom,Rest),
   strip_leading_ws(Rest,X2),
   generate_atom_list(X2,[],T).
generate_atom_list([32|X],Last,[NewAtom|T]) :- !,
   reverse(Last,RL),
   atom_codes(NewAtom,RL),
   strip_leading_ws(X,X2),
   generate_atom_list(X2,[],T).
generate_atom_list([H|X],Last,Res) :- generate_atom_list(X,[H|Last],Res).

get_quoted_atom([],Acc,QuotedAtom,[]) :- reverse(Acc,R), atom_codes(QuotedAtom,R).
get_quoted_atom([39|T],Acc,QuotedAtom,T) :- !, reverse(Acc,R), atom_codes(QuotedAtom,R).
get_quoted_atom([H|T],Acc,QuotedAtom,Rest) :- get_quoted_atom(T,[H|Acc],QuotedAtom,Rest).


strip_leading_ws([32|X],R) :- !, strip_leading_ws(X,R).
strip_leading_ws(X,X).

:- meta_predicate call_probcli_option(0).
call_probcli_option(_:Option) :- just_assert_option(Option), !,
   (option(Option) -> true ; assert_option(Option)).
call_probcli_option(_:statistics) :- !, % avoid calling SICS version
   cli_print_statistics.
call_probcli_option(Option) :-
   on_exception(error(existence_error(A,B),E),
       call(Option),treat_existence_error(A,B,E,Option)).

% commands that require no execution; just asserting option(.)
just_assert_option(depth_first).
just_assert_option(breadth_first).
just_assert_option(strict_raise_error).
just_assert_option(no_deadlocks).
just_assert_option(no_invariant_violations).
just_assert_option(no_goal).
just_assert_option(no_ltl).
just_assert_option(no_assertion_violations).
just_assert_option(no_state_errors).
just_assert_option(no_counter_examples).

:- use_module(tools_printing,[format_with_colour/4]).
treat_existence_error(source_sink,File,E,Option) :- !,
     format_with_colour(user_error,[red],
            '* Could not find file ~w~n* for probcli command ~w~n* Detailed error: ~w~n',[File,Option,E]).
treat_existence_error(_,_,E,Option) :-
     format_with_colour(user_error,[red],
            '* probcli command not yet supported in REPL: ~w~n* Error: ~w~n',[Option,E]).

reload_mainfile :- file_loaded(_,MainFile),
      reset_errors,
      print_repl_prompt_s('Reloading and initialising file: '), print_repl_prompt_s(MainFile),nl,
      clear_loaded_files,
      load_main_file(MainFile,0,_),
      get_errors,
      cli_start_animation(0),
      cli_start_initialisation(0).

% REPL EVAL LOOP:
eval_probcli_repl_line(end_of_file) :- !, eval_line(end_of_file).
eval_probcli_repl_line(Line) :- strip_ws(Line,SLine),
   on_exception(E,eval_line(SLine),
                (E=halt(_) -> throw(E) % e.g., coming from :quit; will be caught above
                 ; add_error(repl,'Uncaught Exception in REPL: ',E)
                 )).

% strip whitespace at end and beginning
strip_ws([H|T],Res) :- is_ws(H),!, strip_ws(T,Res).
strip_ws(C,Res) :- reverse(C,CR), strip_ws2(CR,SCR), reverse(SCR,Res).
strip_ws2([H|T],Res) :- is_ws(H),!, strip_ws2(T,Res).
strip_ws2(R,R).

is_ws(32).

:- use_module(performance_messages,[toggle_perfmessages/0]).
eval_line([]) :- !, print_repl_prompt('Type :q or :quit to quit.'),nl,eval0.
eval_line(end_of_file) :- !, halt_exception(0).
% Haskell GHCI like syntax
eval_line(":r") :- !, eval_line("--reload").
eval_line(":reload") :- !, eval_line("--reload").
eval_line("--reload") :- !,
      (reload_mainfile -> true ; get_errors,print_repl_prompt('### Error(s) occured during reload (use :e to jump to first error)'),nl),
      eval0.
% TO DO: other Haskell commands :info E   :l FILE ,  let pattern = expression
eval_line(":prefs") :- !,print_eclipse_prefs, eval0.
eval_line([45|Command]) :- % -command
   generate_atom_list([45|Command],[],ArgV),
   %print(argv(ArgV)),nl,
   % try and parse like commands passed to probcli
   get_options(ArgV,recognised_cli_option,Options,[],fail),
   print_repl_prompt('Executing probcli command: '),print_repl_prompt(Options),nl,!,
   (maplist(prob_cli:call_probcli_option,Options) -> true
     ; print_red('Failed to execute probcli arguments'),nl),
   eval0.
eval_line("+") :- !, add_last_expression_to_unit_tests, eval0.
eval_line("$+") :- !, preferences:temporary_set_preference(expand_avl_upto,-1,CHNG),
    print_last_value,preferences:reset_temporary_preference(expand_avl_upto,CHNG),
    eval0.
%eval_line("$$") :- !, print_last_expression, eval0.   % now in eval_strings
eval_line("$$$") :- !, % $$0 - $$9 commands to print last expression with indentation
   indent_print_last_expression, eval0.
%eval_line("$") :- !, print_last_info, eval0.  % now in eval_strings
eval_line("!trace") :- !, eval_line("^").
eval_line("^") :- !,
    (retract(trace_eval) -> print_repl_prompt('TRACING OFF'),nl
      ; assert(trace_eval), print_repl_prompt('TRACING ON'),nl), eval0.
eval_line("!observe") :- !, toggle_observe_evaluation.
eval_line("!v") :- !, tcltk_turn_debugging_off.
eval_line("!p") :- !, toggle_perfmessages.
eval_line("!perf") :- !, toggle_perfmessages.
eval_line("!profile") :- !, eval_line("%").
eval_line("!print_profile") :- !, eval_line("%%").
eval_line("%") :- !, print_repl_prompt('PROFILING : '), %spy([avl:avl_size/2]),
   (current_prolog_flag(profiling,on)
    -> set_prolog_flag(profiling,off), print('OFF') ;
       set_prolog_flag(profiling,on), print('ON')),
    nl,print_repl_prompt('USE %% to print profile info'),nl,eval0.
eval_line("%%") :- !, nl,print_repl_prompt('PROLOG PROFILE INFORMATION:'), nl,
   on_exception(error(existence_error(_,_),_),print_profile, print_red('CAN ONLY BE USED WHEN RUNNING PROB FROM SOURCE')),
   nl, debug:timer_statistics,
   eval0.
eval_line("!print_coverage") :- !, nl,print_repl_prompt('PROLOG COVERAGE INFORMATION:'), nl,
   (prolog_flag(source_info,on) -> true ; print_red('Only useful when prolog_flag(source_info,on)!'),nl),
   on_exception(error(existence_error(_,_),_),print_coverage, print_red('CAN ONLY BE USED WHEN RUNNING PROB FROM SOURCE')),nl,
   eval0.
eval_line("!profile_reset") :- !, nl,print_repl_prompt('RESETTING PROLOG PROFILE INFORMATION'), nl,
   on_exception(error(existence_error(_,_),_),profile_reset, print_red('CAN ONLY BE USED WHEN RUNNING PROB FROM SOURCE')),
   eval0.
eval_line("%%%") :- !, nl,print('PROFILE INFORMATION (Starting TK Viewer):'), nl,
   on_exception(error(existence_error(_,_),_),
        (use_module(library(gauge)),gauge:view), print_red('CAN ONLY BE USED WHEN RUNNING PROB FROM SOURCE')),
   nl,eval0.
eval_line("!debug") :- !,
   print_repl_prompt('ENTERING PROLOG DEBUG MODE:'),
   on_exception(error(existence_error(_,_),_),debug, print_red('CAN ONLY BE USED WHEN RUNNING PROB FROM SOURCE')),
   nl,
   eval0.
eval_line("@") :- !, get_preference(find_abort_values,OldVal),
  print_repl_prompt('Try more aggressively to detect ill-defined expressions: '),
  (OldVal=true -> Val=false ; Val=true), print(Val),nl,
  temporary_set_preference(find_abort_values,Val) , eval0.
eval_line("!") :- !, toggle_eval_det,eval0.
eval_line("!norm") :- !, toggle_normalising,eval0.
eval_line(Codes) :- parse_eval_command(Codes,CommandName,Argument),!,
       debug_println(9,executing_eval_command(CommandName,Argument)),
       (exec_eval_command(CommandName,Argument) -> eval0
         ; format_with_colour(user_error,[red,bold],'Command ~w failed~n',[CommandName]),
           eval0).
eval_line(ExpressionOrPredicate) :- (trace_eval -> trace ; true),
      (eval_codes(ExpressionOrPredicate,exists,_,_,_,_)
        -> eval0
         ; print_red('Evaluation failed'),nl,eval0).

parse_eval_command([C|Rest],CommandName,Argument) :- [C]=":",
   eval_command(Cmd,CommandName),
   append(Cmd,RestArg,Rest),
   (RestArg = [Letter1|_] -> is_ws(Letter1) /* otherwise command name continues */ ; true),
   strip_ws(RestArg,Argument),
   (eval_command_help(CommandName,[],_), Argument = [_|_]
     -> format_with_colour(user_error,[red],'WARNING: Command ~w does not take arguments!~n',[CommandName])
    ; eval_command_help(CommandName,[_|_],_), Argument = []
     -> format_with_colour(user_error,[red],'WARNING: Command ~w requires arguments!~n',[CommandName])
     ; true).

% TO DO: some of these commands should also be made available in the Tcl/Tk Console
eval_command("q",quit).
eval_command("quit",quit).
eval_command("halt",quit).
eval_command("x",exit).
eval_command("exit",exit).
eval_command("f",find).
eval_command("find",find).
eval_command("*",apropos).
eval_command("apropos",apropos).
eval_command("help",help).
eval_command("h",help).
eval_command("?",help).
eval_command("ctl",ctl).
eval_command("ltl",ltl).
eval_command("reset",reset_animator(hard)). % :reset
eval_command("reset-history",reset_animator(history_only)). % :reset
eval_command("statistics",statistics).
eval_command("stats",statistics). % :stats
eval_command("states",state_space_stats). % :states
eval_command("state",show_state_info). % :state
eval_command("statespace",state_space_display). % :statespace
eval_command("u",unsat_core).
%eval_command("core",unsat_core).
eval_command("show",show_last_as_table). % :show
eval_command("dot",show_last_as_dot(no_dot_viewing)). % :dot
eval_command("dotty",show_last_as_dot(dotty)).
eval_command("dotpdf",show_last_as_dot(dot)).
eval_command("sfdp",show_last_as_dot(sfdp)).
eval_command("browse",browse). % :browse
eval_command("abstract_constants",check_abstract_constants). % :abstract_constants
eval_command("det_check_constants",det_check_constants). % :det_check_constants
eval_command("b",browse).
eval_command("hbrowse",hbrowse). % :hbrowse
eval_command("comp",show_components). % :comp
eval_command("replay",replay_repl_file). % :replay
eval_command("trim",trimcore). % :trim
eval_command("src",show_source). %:src
eval_command("source",show_source). %:source
eval_command("origin",show_origin). %:origin
eval_command("edit",edit_main_file).
eval_command("e",edit_main_file). % :e
eval_command("comment",comment).
eval_command("machine",show_machine_info(statistics)). %:machine
eval_command("machine-stats",show_machine_info(statistics)). %:machine
eval_command("files",show_machine_info(files)). %:files

available_commands(SLC) :-
  findall(Cmd,(eval_command(Cs,_),atom_codes(Cmd,[58|Cs])), LC),
  sort(LC,SLC).

eval_command_help(exit,[],'Exit ProB').
eval_command_help(find,['P'],'Find state in state-space which makes LTL atomic proposition P true; LTL Propositions: {B-Pred}, e(Op), [Op], true, false, sink').
eval_command_help(ltl,['F'],'Check LTL formula F; LTL Operators: G,F,X,U,W,R,not,&,or,=>; LTL Propositions: {B-Pred}, e(Op), [Op], true, false, sink; Past-LTL Operators: Y,H,O,S,T (dual to X,G,F,U,R)').
eval_command_help(ctl,['F'],'Check CTL formula F; CTL Syntax: ExUy,EXx,AXx,EFx,AGx,EX[Op]x,e(Op),{B-Pred}').
eval_command_help(browse,opt('PAT'),'Browse available constants, variables, sets and lets introduced in REPL').
eval_command_help(apropos,['PAT'],'Find constant or variable whose names contains PAT').
eval_command_help(hbrowse,['PAT'],'Browse machine hierarchy for all identifiers whose names contains PAT').
eval_command_help(show_components,[],'Show components of PROPERTIES').
eval_command_help(abstract_constants,[],'Show ABSTRACT_CONSTANTS and check if can be fully evaluated').
eval_command_help(det_check_constants,[],'Check if values of CONSTANTS are forced and explain if they are').
eval_command_help(show_last_as_table,[],'Show last evaluated expression in tabular form').
eval_command_help(show_last_as_dot(_),['F'],'Show expression or predicate F as dot graph').
eval_command_help(unsat_core,[],'Compute Unsatisfiable Core of last evaluated predicate').
eval_command_help(help,opt('CMD'),'Provide help about REPL command CMD').
eval_command_help(replay_repl_file,['FILE'],'Replay FILE of REPL commands').
eval_command_help(reset_animator(_),[],'Reset history and statespace of animator').
eval_command_help(show_source,['ID'],'Show origin and source code definition of identifier ID').
eval_command_help(show_origin,['ID'],'Show origin of identifier ID and try opening in EDITOR').
eval_command_help(show_machine_info(_),[],'Show statistics about loaded machine and files').
eval_command_help(state_space_stats,[],'Show statistics about state space').
eval_command_help(state_space_display,[],'Show complete state space transitions (could be very big !)').
eval_command_help(show_state_info,[],'Show current state').
eval_command_help(statistics,[],'Show statistics about last evaluation').
% -machine_stats : cli_print_machine_info(statistics) -machine_files : cli_print_machine_info(files)
eval_command_help(trim,[],'Trim memory usage of probcli (try and give memory back to the OS)').
% implemented in eval_strings:
eval_command_help(type,['E'],'Show type of expression E').
eval_command_help(cvc4,['P'],'Solve predicate P using CVC4 solver').
eval_command_help(kodkod,['P'],'Solve predicate P using SAT solver via Kodkod').
eval_command_help(z3,['P'],'Solve predicate P using Z3 solver').
eval_command_help(edit_main_file,opt('ID'),'Edit main file (or origin of identifier ID) using EDITOR (path_to_text_editor preference)').
eval_command_help(comment,['STRING'],'provide STRING as a comment (mainly useful for :replay files)').

print_eval_command_help(Codes) :-
   eval_command(Codes,Cmd),
   eval_command_help(Cmd,Args,Descr),
   (Args = []
     -> format('Command ~w~n Syntax :~s~n        ~w~n',[Cmd,Codes,Descr])
     ; Args=[Arg] -> format('Command ~w~n Syntax :~s ~w~n       ~w~n',[Cmd,Codes,Arg,Descr])
     ; Args=opt(Arg) -> format('Command ~w~n Syntax :~s [~w]~n       ~w~n',[Cmd,Codes,Arg,Descr])
      ; format('Command ~w~n Syntax :~s ~w~n        ~w~n',[Cmd,Codes,Args,Descr])).

:- use_module(tools_commands,[show_dot_file/1, show_ps_file/1, gen_dot_output/4]).
:- use_module(state_space,[transition/4]).
exec_eval_command(quit,_) :- !, halt_exception(0).
exec_eval_command(exit,_) :- !,halt.
exec_eval_command(browse,CodesToMatch) :- !,
     (CodesToMatch=[] -> browse  % maybe merge with apropos functionality
      ; exec_eval_command(apropos,CodesToMatch)).
exec_eval_command(find,FORMULA) :-
    atom_codes(APF,FORMULA),cli_find_ltl_ap(APF).
exec_eval_command(apropos,CodesToMatch) :-  /* :* Pattern (apropos command) */
       browse_machine(CodesToMatch).
exec_eval_command(hbrowse,CodesToMatch) :-  /* :* Pattern (apropos command) */
       browse_all_machines(CodesToMatch).
exec_eval_command(show_components,_) :-
       print_property_partitions.
exec_eval_command(check_abstract_constants,_) :-
       check_abstract_constants.
exec_eval_command(det_check_constants,_) :-
       det_check_constants.
exec_eval_command(help,Arg) :-
  (Arg=[] -> eval_help
   ;  print_eval_command_help(Arg) -> true
   ;  (Arg=[58|RA],print_eval_command_help(RA)) -> true  % remove : at front
   ;  format('Cannot provide help about ~s~n',[Arg]),
      available_commands(LC), format('Available commands: ~w~n',[LC])
  ).
exec_eval_command(ctl,FORMULA) :-
    atom_codes(F,FORMULA),
    (cli_ctl_model_check(F,_,Status)
     -> (Status=false -> write_history_to_user_output([show_init,show_states]) ; true)
     ; print('CTL Syntax: ExUy,EXx,AXx,EFx,AGx,EX[Op]x,e(Op),{B-Pred}'),nl).
exec_eval_command(ltl,FORMULA) :-
    atom_codes(F,FORMULA),
    (cli_ltl_model_check(F,_,Status)
     -> (Status=no -> write_history_to_user_output([show_init,show_states]) ; true)
     ; print('LTL Operators: G,F,X,U,W,R,not,&,or,=>'),nl,
       print('LTL Propositions: {B-Pred}, e(Op), [Op], true, false, sink'),nl,
       print('Past-LTL Operators: Y,H,O,S,T (dual to X,G,F,U,R)'),nl
     ).
exec_eval_command(reset_animator(Hard),_) :- !,
    state_space:get_state_space_stats(TotalNodeSum,TotalTransSum,_),
    (Hard=hard ->
        format('Resetting statespace (~w states, ~w transitions)~n',[TotalNodeSum,TotalTransSum]),
        reset_animator
      ; format('Resetting animation history (keeping statespace: ~w states, ~w transitions)~n',[TotalNodeSum,TotalTransSum]),
        tcltk_reset % only resets animation history,...
    ).
exec_eval_command(statistics,_) :- !, print_last_info.
exec_eval_command(state_space_stats,_) :- !, % :states
   state_space:get_state_space_stats(TotalNodeSum,TotalTransSum,ProcessedTotal),
   format('Statespace: ~w states (~w processed) and ~w transitions.~n',[TotalNodeSum,ProcessedTotal,TotalTransSum]).
exec_eval_command(state_space_display,_) :- !, % :statespace
   ( visited_expression(ID,State),
     functor(State,F,N),
     format('State ID ~w (~w/~w)~n',[ID,F,N]),
     transition(ID,OperationTerm,_OpID,ToID),
     functor(OperationTerm,F2,N2),
     format('  -> ~w  (~w/~w)~n',[ToID,F2,N2]),
     fail
    ;
     current_state_id(ID),
     format('Current State ID ~w~n',[ID])
  ).
exec_eval_command(show_state_info,_) :- !, % :state
    current_expression(ID,CurState),
    expand_const_and_vars_to_full_store(CurState,EState),
    format('Current state id ~w : ~n',[ID]),translate:print_state(EState),nl.
exec_eval_command(unsat_core,_) :- !,   % :core :u
    unsat_core_last_expression.
exec_eval_command(trimcore,_) :- !,   % :trim
    prob_trimcore_verbose.
exec_eval_command(show_last_as_table,_) :- !, % :show
    show_last_expression_as_table.
exec_eval_command(show_last_as_dot(Show),Arg) :- !,
    (Arg = [] -> print('*** :dot requires an expression or predicate as argument.'),nl
     ; safe_absolute_file_name('~/probcli_repl.dot',AFile),
       set_eval_dot_file(AFile),
       format('Displaying evaluation result in: ~w~n',[AFile]),
       (eval_codes(Arg,exists,_,_,_,_) -> true ; true), unset_eval_dot_file,
       (  Show=no_dot_viewing -> true
        ; Show=dotty -> show_dot_file(AFile)
        ; safe_absolute_file_name('~/probcli_repl.pdf',PDFFile),
          gen_dot_output(AFile,Show,pdf,PDFFile),
          show_ps_file(PDFFile) % we assume psviewer can also open pdf
            )).
exec_eval_command(replay_repl_file,FILEC) :- !, % :replay
    atom_codes(File,FILEC),
    set_repl_input_file(not_verbose,File).
exec_eval_command(show_source,IDC) :- !, % :src
    atom_codes(ID,IDC),
    show_source(ID).
exec_eval_command(show_origin,IDC) :- !, % :origin
    atom_codes(ID,IDC),
    show_origin(ID).
exec_eval_command(show_machine_info(X),_) :- !, % :machine
    cli_print_machine_info(X).
exec_eval_command(edit_main_file,Arg) :- !, % :e
    (Arg=[] -> edit_main_file ; exec_eval_command(show_origin,Arg)).
exec_eval_command(comment,_Arg) :- !. % do nothing; argument was a comment; mainly useful for :replay files

:- use_module(tools_commands,[edit_file/2]).
edit_main_file :- last_repl_error(File,Line),!,
     format('Showing first error from last command~n',[]),
     edit_file(File,Line).
edit_main_file :- file_loaded(true,MainFile),!,edit_file(MainFile,unknown).
% Note: for the bbedit command we can also specify line numbers bbedit +LINE FILE
edit_main_file :- format_with_colour(user_error,[red],'No file loaded, cannot open EDITOR!~n',[]).



:- use_module(probsrc(error_manager),[extract_file_line_col/6]).
open_file_at_position(OriginTerm) :-
    extract_file_line_col(OriginTerm,FILE,LINE,_COL,_Erow,_Ecol),
    edit_file(FILE,LINE).


:- use_module(probsrc(bmachine),[source_code_for_identifier/6]).
show_source(ID) :- source_code_for_identifier(ID,Kind,_Type,OriginStr,OriginTerm,Source),!,
    translate:translate_subst_or_bexpr(Source,PPS),
    %format('~w: ~w (Type: ~w)~norigin: ~w~nsource: ~w~n',[Kind,ID,_Type,Origin,PPS]).
    format('~w: ~w~norigin: ~w~nsource: ~w~n',[Kind,ID,OriginStr,PPS]),
    (OriginTerm=b(_,_,_),bsyntaxtree:get_texpr_description(OriginTerm,Description)
     -> format('description: ~w~n',[Description]) ; true).
show_source(ID) :- format('### Could not find source for ~w~n',[ID]).

show_origin('') :- last_repl_error(_,_),!, % error occured: show error in editor like :e would
   edit_main_file.
show_origin('') :- !,format('### You need to provided an identifier~n',[]).
show_origin(ID) :- source_code_for_identifier(ID,Kind,_Type,OriginStr,OriginTerm,_Source),!,
    format('~w: ~w~norigin: ~w~n',[Kind,ID,OriginStr]),
    open_file_at_position(OriginTerm).
show_origin(ID) :- format('### Could not find origin for ~w~n',[ID]).

profiling_on :- set_prolog_flag(profiling,on), print('% PROFILING ON'),nl.

% find a state satisfying LTL atomic property
cli_find_ltl_ap(APF) :-
    if(ltl:find_atomic_property_formula(APF,ID),
         (format('Found state (id = ~w) satisfying LTL atomic property.~n',[ID]),
          tcltk_goto_state('LTL FIND',ID)),
         format('No explored state satsifies LTL atomic property.~n',[])).

eval_help :-
       print('ProB Interactive Expression and Predicate Evaluator '), nl,
       print('Type a valid B expressions or predicates, followed by RETURN or ENTER.'),nl,
       print('You can spread input over multiple lines by ending lines with "\\".'),nl,
       browse_machine([]),
       print('You can also type one of the following commands: '),nl,
       (option_verbose ->
          print('    + to save last expression to ProB unit tests.'),nl,
          print('    ! to go to deterministic propagation only mode.'),nl,
          print('    $ to print evaluation time for last expression.'),nl,
          print('    $$ to pretty-print last expression and its type.'),nl,
          print('    $$$ to pretty-print last expression in nested fashion.'),nl,
          print('    !p to toggle performance messages.'),nl,
          print('    !norm to toggle normalisation of results.'),nl
         ; true),
       print('    let x = E to define a new local variable x'),nl,
       print('    :b or :b Prefix to browse the available identifiers'),nl,
       print('    :t E to get the type of an expression and :r to reload the machine'),nl,
       print('    :show to display the last result as a table (if possible)'),nl,
       print('    :list CAT to display information with CAT : {files,variables,help,...}'),nl,
       print('    :* P to display constants/variables containing pattern P'),nl,
       print('    :core Pred to compute the unsat core for Pred'),nl,
       print('    :u to compute the unsat core for last evaulated result'),nl,
       print('    :stats to print the type and evaluation time for last query'),nl,
       print('    -PROBCLIARGS to pass command-line probcli arguments to the REPL'),nl,
       print('    :ctl F or :ltl F to check a CTL or LTL formula.'),nl,
       print('    :f F to find a state satisfying LTL atomic property.'),nl,
       print('    :exec S to execute an operation or substitution S.'),nl,
       print('    :replay FILE to replay a file of commands.'),nl,
       print('    :z3 P, :cvc4 P, :kodkod P to solve predicate P using alternate solver'),nl,
       print('    :forall P to prove predicate P as universally quantified with default solver'),nl,
       (option_verbose ->
       print('    :krt P, :pp P, :ml P to prove predicate P using Atelier-B provers if installed'),nl
         ; true),
       print('    :print P to pretty print predicate in a nested fashion'),nl,
       print('    :min P, :max P to find a minimal/maximal model for predicate P or %x.(P|E)'),nl,
       print('    :prefs to print current value of preferences'),nl,
       print('    :reset to reset the state space of the animator.'),nl, % :reset-history only resets history
       print('    :help CMD to obtain more help about a command.'),nl,
       print('    :state, :statespace, :states,'),nl,
       print('            :machine, :files, :source, :orgin,'),nl,
       print('            :dot, :dotty, :sfdp, :trim, :comp - use :help CMD for more info'),nl,
       print('    :q to exit.'),nl.


browse :- browse_machine([]), browse_repl_lets.

:- use_module(prob2_interface,[get_machine_identifiers/2]).
% the CodesToMatch parameters mimics the apropos command of the Clojure-REPL
browse_machine(CodesToMatch) :-
       get_machine_identifiers(machines,MN), display_match('MACHINES',CodesToMatch,MN),
       (CodesToMatch =[] -> print_sets
         ; get_machine_identifiers(sets,SN), display_match('SETS',CodesToMatch,SN),
           get_machine_identifiers(set_constants,SCN), display_match('SETS-ELEMENTS',CodesToMatch,SCN)
        ),
       get_machine_identifiers(definition_files,DFN),
       (DFN=[] -> true ; display_match('DEFINITIONS FILES',CodesToMatch,DFN)),
       get_machine_identifiers(definitions,DN),
       (DN=[] -> true ; display_match('DEFINITIONS',CodesToMatch,DN)),
       get_machine_identifiers(constants,CN),
       display_match('CONSTANTS',CodesToMatch,CN),
       get_machine_identifiers(variables,VN),
       display_match('VARIABLES',CodesToMatch,VN),
       get_machine_identifiers(operations,Ops),
       display_match('OPERATIONS',CodesToMatch,Ops).

display_match(KIND,CodesToMatch,Ids) :- display_match(KIND,CodesToMatch,Ids,show_empty).
display_match(KIND,CodesToMatch,Ids,ShowEmpty) :-
       include(prob_cli:atom_contains_codes(CodesToMatch),Ids,MatchingIds),
       length(MatchingIds,LenMIds),
       (LenMIds=0, ShowEmpty=show_only_if_match -> true
       ; sort(MatchingIds,SMatchingIds),
         (CodesToMatch=[]
          -> format(' ~w: ~w ~w~n',[KIND,LenMIds,SMatchingIds])
           ; length(Ids,LenIds),
             format('Matching ~w: ~w/~w ~w~n',[KIND,LenMIds,LenIds,SMatchingIds]))
        ).

% check if an atom contains a list of codes in its name
atom_contains_codes([],_) :- !.
atom_contains_codes(Codes,Name) :- atom_codes(Name,NC), append(Codes,_,C),
   append(_,C,NC).

:- use_module(b_global_sets,[b_global_set/1]).
print_sets :- print('Available SETS: '), b_global_set(GS), print_set(GS),fail.
print_sets :- nl.

:- use_module(probsrc(b_global_sets),[is_b_global_constant/3]).
print_set(GS) :- print(GS), \+ is_b_global_constant(GS,_,_),!, print(' ').
print_set(GS) :- print(' = {'), is_b_global_constant(GS,_,Cst), print(Cst), print(' '),fail.
print_set(_) :- print(' } ').

:- use_module(b_machine_hierarchy,[get_machine_identifier_names/7]).
% browse all machines, shows identifiers maybe not visible at top-level
browse_all_machines(CodesToMatch) :-
   format('Searching machine hierarchy for identifiers matching ~s~n',[CodesToMatch]),
   get_machine_identifier_names(Name,Params,Sets,AVars,CVars,AConsts,CConsts),
   format('~nMACHINE ~w~n',[Name]),
   display_match('PARAMS',CodesToMatch,Params,show_only_if_match),
   display_match('SETS',CodesToMatch,Sets,show_only_if_match),
   display_match('ABSTRACT_VARIABLES',CodesToMatch,AVars,show_only_if_match),
   display_match('CONCRETE_VARIABLES',CodesToMatch,CVars,show_only_if_match),
   display_match('ABSTRACT_CONSTANTS',CodesToMatch,AConsts,show_only_if_match),
   display_match('CONCRETE_CONSTANTS',CodesToMatch,CConsts,show_only_if_match),
   fail.
browse_all_machines(_).


:- use_module(bmachine,[b_get_properties_from_machine/1]).
:- use_module(bsyntaxtree,[predicate_components/2]).
print_property_partitions :- print('PARTITIONS OF PROPERTIES'),nl,
   b_get_properties_from_machine(Properties),
   predicate_components(Properties,Comp),
   length(Comp,Len), print(Len), print(' components found in PROPERTIES'),nl,
   nth1(Nr,Comp,component(P,Vars)),
   format('~n& // Component ~w/~w over identifiers ~w~n',[Nr,Len,Vars]),
   translate:print_bexpr(P),nl,fail.
print_property_partitions :- nl, print(' ============== '),nl.

:- use_module(store,[lookup_value_for_existing_id/3]).
:- use_module(b_machine_hierarchy,[abstract_constant/2]).
check_abstract_constants :-
   format('Checking whether abstract constants can be expanded:~n',[]),
   current_expression(_ID,CurState),
   expand_const_and_vars_to_full_store(CurState,EState),
   abstract_constant(AID,_),
   lookup_value_for_existing_id(AID,EState,Val),
   functor(Val,VF,_),
   format(user_output,'~n*** Evaluating ABSTRACT_CONSTANT: (~w)~n',[VF]),
   format_with_colour(user_output,[blue],' ~w~n',[AID]),
   (debug_mode(off) -> true
    ; translate:translate_bvalue(Val,VS), format_with_colour(user_output,[blue],' Stored value = ~w~n',[VS])),
   atom_codes(AID,C),
   % TO DO: provide info if value symbolic and can be expanded fully + add timing
   eval_codes(C,exists,_,_,_,_),nl,fail.
check_abstract_constants.

:- use_module(b_state_model_check,[cbc_constants_det_check/1]).
det_check_constants :- \+ current_state_corresponds_to_setup_constants_b_machine, !,
   format_with_colour(user_error,[red],'This command requires to setup the constants first!~n',[]).
det_check_constants :-
   current_state_id(ID),
   %format('Checking whether constants are forced in state ~w:~n',[ID]),
   cbc_constants_det_check(ID).

% showing relations as tables:

:- use_module(table_tools).
show_last_expression_as_table :- \+ last_expression(_,_Expr,_Value),!,
   print_red('Please evaluate an expression or predicate first.'),nl.
show_last_expression_as_table :-
   get_last_result_value(Expr,_,Value),
   print_value_as_table(Expr,Value).


% a few definitions so that probcli commands work in REPL:
:- use_module(translate,[set_unicode_mode/0, unset_unicode_mode/0]).
:- public pretty_print_internal_rep/3, pretty_print_internal_rep_to_B/1.
pretty_print_internal_rep(PPFILE,TYPES,unicode) :- !,
    set_unicode_mode, b_write_machine_representation_to_file(TYPES,PPFILE),unset_unicode_mode.
pretty_print_internal_rep(PPFILE,TYPES,_) :- b_write_machine_representation_to_file(TYPES,PPFILE).

pretty_print_internal_rep_to_B(PPFILE) :- b_write_eventb_machine_to_classicalb_to_file(PPFILE).


% Simple Animator

interactive_animate_machine :-
                   nl,print('IMPORTANT: Do not use this mode for automatic tools.'),nl,
                   print('The output format can change arbitrarily in future versions.'),nl,
                   print('Please terminate your input with a dot (.) and then type return.'),nl,nl,
                   animate_machine2.
animate_machine2 :-
  print_current_state,
  cli_computeOperations(Ops),
  length(Ops,Max),
  print('Enabled Operations: '),nl,
  print_options(Ops,1),
  print(' ==> '),!,
  read(Nr),
  (number(Nr),Nr>0,Nr=<Max
    -> cli_animateOperationNr(Nr,Ops,0)
    ;  fail
  ),!,
  animate_machine2.
animate_machine2.

print_current_state :- getCurrentStateID(CurID), print('ID ==> '), print(CurID),nl,
  getStateValues(CurID,State),
  print_bindings(State),
  (specfile:b_or_z_mode,\+is_initialised_state(CurID)
  -> print_red(' Not yet initialised.'),print_mode_info, debug_println(10,state(State)) ; nl).

print_mode_info :- animation_mode(M), (animation_minor_mode(MM) -> true ; MM=''),
  format('Animation Mode = ~w [~w]~n',[M,MM]).

cli_computeOperations(Ops) :- option(animate_stats),!, % provide statistics about the animation
   nl,
   start_probcli_timer(Timer),
   computeOperations(Ops),
   stop_probcli_timer(Timer,'Time to compute operations: ').
cli_computeOperations(Ops) :- computeOperations(Ops).

cli_animateOperationNr(Nr,Options,StepNr) :-
   (option(animate_stats)
     -> nth1(Nr,Options,Action),
        truncate_animate_action(Action,TA),
        (StepNr>1 -> format('performing step ~w : ~w~n',[StepNr,TA])
          ; format('performing ~w~n',[TA]))
     ; true),
   executeOperationNr_on_computed_node(Nr).

:- use_module(tools_strings,[truncate_atom/3]).
% optionally truncate animation action atom for printing:
truncate_animate_action(Action,TA) :-
   (option_verbose -> TA = Action
	  ; \+ atom(Action) -> TA = Action
	  ; truncate_atom(Action,100,TA)).

:- use_module(library(random),[random/3]).
perform_random_step(StepNr) :- perform_random_step(_Ops,_Len,_RanChoice,StepNr).
perform_random_step(Ops,Len,RanChoice,StepNr) :-
  cli_computeOperations(Ops),
  getCurrentStateID(CurID), check_for_errors(CurID,StepNr),
  length(Ops,Len), Len>0,
  debug_println(20,perform_random_step(Len,StepNr)),
  L1 is Len+1,
  (do_det_checking, Len>1
    ->  print_error('### Non-deterministic step in animate or init'),
        print_error('### State:'),
        print_current_state,  print_error('### Enabled Operations: '), print_options(Ops,1),
        error_occurred(det_check)
    ; true),
  random(1,L1,RanChoice),
  debug_println(20,random(L1,RanChoice)),
  cli_animateOperationNr(RanChoice,Ops,StepNr).

:- use_module(state_space,[visited_expression/2]).
check_for_errors(CurID,StepNr) :- invariant_violated(CurID),
   \+ option(no_invariant_violations),
   get_preference(do_invariant_checking,true),
   ajoin(['INVARIANT VIOLATED after ',StepNr,' steps (state id ',CurID,').'],ErrMsg),
   format('~w~n',[ErrMsg]),
   visited_expression(CurID,CurState), print_state_silent(CurState),
   error_occurred_with_msg(invariant_violation,ErrMsg),
   fail.
check_for_errors(CurID,_) :- get_state_errors(CurID).
% TO DO: also check for assertion errors, goal,  state_errors with abort

:- use_module(bmachine,[b_machine_has_constants_or_properties/0]).
do_det_checking :- option(det_check),!.
do_det_checking :- option(det_constants_check),getCurrentStateID(root),
   b_or_z_mode, b_machine_has_constants_or_properties.

perform_random_steps(Nr,_) :- \+ number(Nr),!,
   print_error('### Argument to animate not a number'), print_error(Nr),error_occurred(animate).
perform_random_steps(Nr,_) :- Nr<0, !,
   print_error('### Argument to animate is a negative number'), print_error(Nr),error_occurred(animate).
perform_random_steps(0,_) :- !.
perform_random_steps(Nr,ErrorOnDeadlock) :-
   perform_random_initialisation_if_necessary(Steps), % if Nr=1 we currently will also execute the INITIALISATION ! TO DO: fix
   perform_random_steps_aux(Steps,Nr,ErrorOnDeadlock).

perform_random_steps_aux(Nr,Max,_) :- Nr >= Max,!, debug_println(9,performed_random_steps(Nr)).
perform_random_steps_aux(Nr,Max,ErrorOnDeadlock) :-
   N1 is Nr+1,
   (perform_random_step(N1)
     -> perform_random_steps_aux(N1,Max,ErrorOnDeadlock)
     ;  /* deadlock */
      write_xml_element_to_log(deadlock_found,[step/Nr]),
      (ErrorOnDeadlock=true, \+ option(no_deadlocks)) ->
       print_error('### Deadlock occurred during -animate, at step number:'), print_error(Nr),
       error_occurred(animate)
     ; print('% Deadlock occurred during -animate, at step number:'), print(Nr),nl
    ).

perform_random_initialisation_if_necessary(Steps) :-
   b_or_z_mode, getCurrentStateID(State), State=root,!, perform_random_initialisation(Steps).
perform_random_initialisation_if_necessary(0).

perform_random_initialisation :- perform_random_initialisation(_).
perform_random_initialisation(Steps) :- getCurrentStateID(State), State \= root, !,
   print_error('### init can only be used in initial state'), print_error(State),error_occurred(initialisation),
   Steps=0.
perform_random_initialisation(Steps) :- b_mode, b_machine_has_constants_or_properties,!,
   (perform_random_step(Ops,_Len,RanChoice,1)
    -> nth1(RanChoice,Ops,Choice), %print(Choice),nl,
       (Choice = 'PARTIAL_SETUP_CONSTANTS'
         -> error_occurred(setup_constants_inconsistent)
         ; true)
    ; error_occurred(setup_constants_fails),fail), % $setup_constants  TODO: properties unknown or unsat
   perform_random_init_after_setup_constants, Steps=2. % $initialise_machine
perform_random_initialisation(Steps) :- (perform_random_step(1) -> Steps=1 ; error_occurred(initialisation_fails),fail).


perform_random_init_after_setup_constants :- \+ option(initialise), we_need_only_static_assertions(_),!,
         printsilent('% NOT INITIALISING MACHINE (not required)'),nls.
 %  debug_println(20,'% NOT INITIALISING MACHINE (not required)').
perform_random_init_after_setup_constants :-
   (perform_random_step(2)  % 2 is the step nr not the number of steps
    -> true ; error_occurred(initialisation_fails),fail).

:- use_module(enabling_analysis,[tcltk_cbc_enabling_analysis/1, print_enable_table/1, is_timeout_enabling_result/1]).
do_enabling_analysis_csv(EnablingCsvFile,NOW) :-
    start_probcli_timer(Timer1),
    start_xml_feature(enabling_analysis,file,EnablingCsvFile,FINFO),
    tcltk_cbc_enabling_analysis(list(R)),
    stop_probcli_timer(Timer1,'% Finished CBC Enabling Analysis',_TotWallTime),
    print_cbc_stats(R,NOW),
    debug_println(9,writing_to_file(EnablingCsvFile)),
    my_tell(EnablingCsvFile),
    print_enable_table(R),
    told,!,
    stop_xml_feature(enabling_analysis,FINFO).
do_enabling_analysis_csv(EnablingCsvFile,_) :-
    add_error(enabling_analysis,'Enabling analysis failed',EnablingCsvFile),
    stop_xml_group_in_log(enabling_analysis).

print_cbc_stats(Res,_NOW) :- length(Res,Len), Ops is Len-2, % Header + Init
      CBC_Calls is Ops*(Ops+1), % +1 for INITIALISATION
      findall(TO,(member(list([_|T]),Res), member(TO,T),is_timeout_enabling_result(TO)),TOS),
      length(TOS,NrTOS),
      format('% CBC Enabling Stats:~n% Nr of events: ~w~n% Nr of cbc calls: ~w, Timeout results: ~w~n',[Ops,CBC_Calls,NrTOS]),
      write_xml_element_to_log(cbc_enabling_stats,[nr_events/Ops,cbc_calls/CBC_Calls,nr_timeouts/NrTOS]).


:- use_module(enabling_analysis,[feasible_operation_with_timeout/3]).
do_feasibility_analysis(ATimeOut,EnablingCsvFile) :-
    arg_is_number(ATimeOut,TimeOut),
    start_xml_feature(feasibility_analysis,file,EnablingCsvFile,FINFO),
    findall(list([Op,Res]),feasible_operation_with_timeout(Op,TimeOut,Res),R),
    debug_println(9,writing_to_file(EnablingCsvFile)),
    my_tell(EnablingCsvFile),
    print_enable_table([list(['Event','Feasibility'])|R]),
    told,!,
    stop_xml_feature(feasibility_analysis,FINFO).
do_feasibility_analysis(_,EnablingCsvFile) :-
    add_error(feasibility_analysis,'Feasibility analysis failed',EnablingCsvFile),
    stop_xml_group_in_log(feasibility_analysis).

:- use_module(b_read_write_info,[tcltk_read_write_matrix/1]).
generate_read_write_matrix(CsvFile) :-
    tcltk_read_write_matrix(list(Matrix)),
    my_tell(CsvFile),
    print_enable_table(Matrix),
    told,!.
generate_read_write_matrix(CsvFile) :-
    add_error(read_write_matrix,'Generating Read-Write-Matrix failed',CsvFile).


my_tell(File) :- on_exception(error(_E,_), % existence_error(_,_)
                   tell(File), add_error_fail(tell,'File cannot be written to: ',File)).

print_options([],_).
print_options([H|T],N) :-
  print(' '), print(N), print(':'), print(H),nl,
  N1 is N+1,
  print_options(T,N1).

print_nr_list(List) :- print_nr_list(List,0,1,no_repeats).

print_nr_list([],NM1,_,Repeats) :- !, print_repeats(NM1,Repeats).
print_nr_list([H|T],_,N,repeated(H,SinceN)) :- !, N1 is N+1,
   print_nr_list(T,N,N1,repeated(H,SinceN)).
print_nr_list([H|T],NM1,N,Repeats) :- !,
  print_repeats(NM1,Repeats),
  N1 is N+1,
  print_nr_list(T,N,N1,repeated(H,N)).
print_nr_list(X,_,_,_) :- print('### not a list: '), print(X),nl.

print_repeats(N,repeated(H,N)) :- !,
  format(' ~w: ~w~n',[N,H]).
print_repeats(N,repeated(H,Since)) :- !, Repeats is 1+N-Since,
  format(' ~w - ~w: ~w   (~w repetitions)~n',[Since,N,H,Repeats]).
print_repeats(_,_).

print_bindings([]) :- !.
print_bindings([binding(Var,_,PPV)|T]) :- !, print(Var),print('='),print(PPV),
  (T=[] -> true ; print(', '), print_bindings(T)).
print_bindings([binding(Var,_,PPV,_Tag)|T]) :- !, print(Var),print('='),print(PPV),
  (T=[] -> true ; print(', '), print_bindings(T)).
print_bindings(X) :- print('### Internal Error: illegal binding list: '), print(X),nl.

:- dynamic expected_error_occurred/1.
:- dynamic error_did_not_occur/1.
reset_expected_error_occurred :- retractall(expected_error_occurred(_)).
check_all_expected_errors_occurred(NOW) :-
  %error_manager:display_error_statistics,
  get_errors, get_state_space_errors,
  retractall(error_did_not_occur(_)),
  expected_error(Type),
  \+ expected_error_occurred(Type),
  print('*** Expected Error of following type to occur: '), print(Type),nl,
  writeln_log_time(expected_error_did_not_occur(NOW,Type)),
  assert(error_did_not_occur(Type)),
  (option(strict_raise_error) -> definite_error_occurred ; fail).
check_all_expected_errors_occurred(_NOW) :-
   (expected_error(_)
    -> (error_did_not_occur(_) -> print('*** Some expected errors did NOT occur !')
         ; print('All expected errors occurred.')),nl
    ;  true).

expected_error(Type) :- option(expect_error(Type)).
expected_error(Type) :- option(expect_error_pos(Type,_Line,_Col)).

error_occurred(warning(Type)) :- !, error_occurred(Type,warning).
error_occurred(Type) :- error_occurred(Type,error).

get_error_category_and_type(warning(Cat),Category,Type) :- !, Category=Cat,Type=warning.
get_error_category_and_type(C,C,error).

error_occurred_with_msg(Type,Msg) :- error_occurred_with_msg(Type,Msg,not_yet_extracted).
error_occurred_with_msg(warning(Type),Msg,Span) :- !, error_occurred(Type,warning,Span,Msg).
error_occurred_with_msg(Type,Msg,Span) :- error_occurred(Type,error,Span,Msg).

error_occurred(Type,ErrOrWarn) :- error_occurred(Type,ErrOrWarn,not_yet_extracted,'').

error_occurred(Type,ErrOrWarning,ExtractedSpan,Msg) :-
    option(expect_error_pos(Type,Line,Col)),!,
    write_xml_element_to_log(expected_error_occurred,[category/Type, type/ErrOrWarning, message/Msg]),
    assert(expected_error_occurred(Type)),
    (get_error_or_warning_span(ExtractedSpan,Type,EL,EC)
      -> (option(expect_error_pos(Type,EL,EC))
          -> debug_println(9,expect_error_pos_ok(Type,EL,EC))
           ; format('*** Unexpected line ~w and column ~w for error ~w!~n*** Expected line ~w and column ~w.~n',[EL,EC,Type,Line,Col]),
             definite_error_occurred
         )
      ;  format('*** Could not obtain position information for error ~w! Expected line ~w and column ~w.~n',[Type,Line,Col]),
         %display_error_statistics,
         definite_error_occurred).
error_occurred(Type,ErrOrWarning,ExtractedSpan,Msg) :-
    option(expect_error(Type)),!,
    inc_counter(cli_expected_errors),
    get_xml_span(ExtractedSpan,XML),
    write_xml_element_to_log(expected_error_occurred,[category/Type, type/ErrOrWarning, message/Msg|XML]),
    assert(expected_error_occurred(Type)).
error_occurred(Type,ErrOrWarning,ExtractedSpan,Msg) :-
    (probcli_time_stamp(NOW) -> true ; NOW=unknown),
    writeln_log(error_occurred(NOW,Type)),
    get_xml_span(ExtractedSpan,XML),
    (option(optional_error(Type)) ->
        write_xml_element_to_log(optional_error_occurred,[category/Type, type/ErrOrWarning, message/Msg|XML]),
        formatsilent('% Optional error occured: ~w~n',[Type])
     ;
       write_xml_element_to_log(error_occurred,[category/Type, type/ErrOrWarning, message/Msg|XML]),
       (ErrOrWarning = warning -> safe_inc_counter(cli_warnings) ; safe_inc_counter(cli_errors)),
       (option(strict_raise_error) ->
			      print_error('*** Unexpected error occurred ***'),
			      print_error(Type),
            findall(Err,option(expect_error(Err)),Ls), (Ls=[] -> true ; print_error(expected(Ls))),
			      definite_error_occurred
		    ; ErrOrWarning=error,serious_error(Type)
		     -> print_error('*** Serious error occurred ***'),
			      print_error(Type),
			      definite_error_occurred
		     ;  print_probcli_error_non_strict(Type,ErrOrWarning)
       )
    ).

safe_inc_counter(Counter) :-
   on_exception(E,inc_counter(Counter),
     format(user_error,'~n*** Exception in counter library, probably not yet initialized: ~w.~n~n',[E])).


get_xml_span(Span,XML) :- extract_file_line_col(Span,FullFilename,Line,Col,EndLine,EndCol),!,
    XML = [file/FullFilename,start_line/Line,end_line/EndLine,start_col/Col,end_col/EndCol|XT],
    get_xml_add_description(Span,XT).
get_xml_span(Span,XML) :- get_xml_add_description(Span,XML).

get_xml_add_description(Span,XML) :-
    extract_additional_description(Span,Msg),!,
    XML = [additional_description/Msg].
get_xml_add_description(_,[]).

get_error_or_warning_span(not_yet_extracted,Type,EL,EC) :- check_error_span_file_linecol(Type,_File,EL,EC,_,_).
get_error_or_warning_span(not_yet_extracted,Type,EL,EC) :- check_error_span_file_linecol(warning(Type),_File,EL,EC,_,_).
get_error_or_warning_span(Span,_,EL,EC) :- Span \= not_yet_extracted, extract_line_col(Span,EL,EC,_,_).


% a list of serious errors: if these occur; then return code different from 0 even in non-strict mode
serious_error(get_java_command_path).
serious_error(internal_error(_)).

print_probcli_error_non_strict(parse_machine_predicate_error,_) :-
    !. % have already been reported
print_probcli_error_non_strict(Type,ErrOrWarning) :-
  (ErrOrWarning=warning -> print_error('*** warning occurred ***')
     ; print_error('*** error occurred ***')),
  print_error(Type).

definite_error_occurred :- print_error('*** Abnormal termination of probcli !'),
  (file_loaded(_,File) -> print_error('*** for_file'(File)) ; true),
  (current_probcli_command(Cmd) -> print_error('*** for_command'(Cmd)) ; true),
  (probcli_time_stamp(NOW) -> halt_prob(NOW,1)
   ; writeln_log(halt(1)),
     halt_exception(1)
  ).

:- dynamic current_probcli_command/1.
set_current_probcli_command(X) :- retractall(current_probcli_command(_)),
  assert(current_probcli_command(X)).
unset_current_probcli_command :- retractall(current_probcli_command(_)).


halt_prob(NOW,ExitCode) :-
  write_xml_element_to_log(probcli_halted_prematurely,[now/NOW]),
  close_all_xml_groups_in_log_until('probcli-run'),
  stop_xml_probcli_run(NOW),
  halt_exception(ExitCode).


:- dynamic accumulated_infos/3, individual_file_infos/3.
accumulate_infos(Context,Infos) :- option(benchmark_info_csv_output(_)),
   file_loaded(_,File),
   get_additional_infos(Infos,Infos2),
   sort(Infos2,SInfos), % infos is a list of the form Info-Value
   assert(individual_file_infos(File,Context,SInfos)), % store for later csv summary printing
   fail.
accumulate_infos(Context,Infos) :- accumulate_infos_2(Context,Infos).

% also store additional infos if -machine_stats provided; useful for benchmarking/articles
get_additional_infos(I,Res) :- option(cli_print_machine_info(statistics)),!,
    findall(Key-Nr,b_machine_statistics(Key,Nr),Res,I).
get_additional_infos(I,I).

accumulate_infos_2(_,[]).
accumulate_infos_2(Context,[Info|T]) :- get_info(Info,FF,Nr),Nr>0,!,
  (retract(accumulated_infos(Context,FF,OldNr))->true ; OldNr=0),
  N1 is OldNr+Nr,
  assert(accumulated_infos(Context,FF,N1)),
  accumulate_infos_2(Context,T).
accumulate_infos_2(Context,[_|T]) :- accumulate_infos_2(Context,T).
get_info(FF-Nr,FF,Nr).
get_info(FF/Nr,FF,Nr).

:- use_module(tools_io,[safe_intelligent_open_file/3]).
print_accumulated_infos(NrFilesProcessed) :-
    (option(benchmark_info_csv_output(File)) -> safe_intelligent_open_file(File,write,Stream) ; Stream=user_output),
    call_cleanup(pr_acc_infos_aux(Stream,NrFilesProcessed),
                 close(Stream)).
:- use_module(library(system),[ datime/1]).
pr_acc_infos_aux(Stream,NrFilesProcessed) :- NrFilesProcessed>1, accumulated_infos(_,_,_),!,
    print_individual_file_infos_csv(Stream),
    format(Stream,'Analysis summary (~w files processed): ',[NrFilesProcessed]),
    start_xml_group_in_log(summary,files_processed,NrFilesProcessed),
    findall(Context-F-Nr,accumulated_infos(Context,F,Nr),L), sort(L,SL),
    maplist(prob_cli:pracc(Stream),SL),nl(Stream),
    % TO DO: write infos to XML log
    (option(print_version) ->
      datime(datime(Year,Month,Day,Hour,Min,_Sec)),
      format(Stream,'CSV file generated at ~w:~w on the date ~w/~w/~w using probcli:~n',[Hour,Min,Year,Month,Day]),
      print_probcli_version(Stream)
     ; true),
    stop_xml_group_in_log_no_statistics(summary).
pr_acc_infos_aux(_,_NrFilesProcessed).
pracc(Stream,Context-F-Nr) :- format(Stream,'~w:~w:~w ',[Context,F,Nr]).
% print CSV summary of run
print_individual_file_infos_csv(Stream) :-
    individual_file_infos(_,Context,HInfos),!,
    format(Stream,'~nFILE,ANALYSIS,',[]),
    print_titles(HInfos,Stream),nl(Stream),
    % TO DO: ensure Infos and SHInfos identical, else add 0 for missing categories
    ( individual_file_infos(File,Context,Infos),
      format(Stream,'~w,~w,',[File,Context]),
      print_vals(Infos,HInfos,Stream),nl(Stream),
      fail
      ; true
    ).
print_individual_file_infos_csv(_).
print_vals(_,[],_) :- !.
print_vals([H|T],[Header|HT],Stream) :- get_info(Header,Title,_),
   get_info(H,Title,Nr), !,
   write(Stream,Nr),
   (T=[] -> true ; write(Stream,','), print_vals(T,HT,Stream)).
print_vals(Vals,[_|HT],Stream) :- % a value is missing for this file
   write(Stream,'-'),
   (HT=[] -> true ; write(Stream,','), print_vals(Vals,HT,Stream)).
print_titles([],_).
print_titles([H|T],Stream) :- get_info(H,FF,_), write(Stream,FF),
   (T=[] -> true ; write(Stream,','), print_titles(T,Stream)).

write_important_xml_element_to_log(Category,Infos) :-
   include(prob_cli:important_info,Infos,II),
   write_xml_element_to_log(Category,II).
important_info(FF/Nr) :-
   \+ irrelevant_xml_info(FF),
   (Nr=0 -> \+ irrelevant_xml_if_zero(FF) ; true).
irrelevant_xml_info(true_after_expansion).
irrelevant_xml_info(false_after_expansion).
irrelevant_xml_info(unknown_after_expansion).
irrelevant_xml_info(total_after_expansion).
irrelevant_xml_if_zero(timeout).
irrelevant_xml_if_zero(enum_warning).

check_required_infos([],_,_).
check_required_infos([H|T],Infos,ErrType) :-
  (check_single_info(H,Infos)
    -> check_required_infos(T,Infos,ErrType)
     ; translate_err_type(ErrType,ES),
       format_with_colour(user_error,[red],
           '*** Unexpected result while checking: ~w~n***    expected : ~w~n***          in : ~w~n',
            [ES,H,Infos]),
       error_occurred(ErrType)).
translate_err_type(check_assertions,'ASSERTIONS') :- !.
translate_err_type(cli_check_assertions,'ASSERTIONS') :- !.
translate_err_type(check_goal,'GOAL') :- !.
translate_err_type(load_po_file,'PROOF OBLIGATIONS') :- !.
translate_err_type(cli_wd_check,'WD PROOF OBLIGATIONS') :- !.
translate_err_type(X,X).

check_single_info(Label-Nr,Infos) :- !, member(Label-ActualNr,Infos),
   match_info(Nr,ActualNr).
check_single_info(H,List) :- member(H,List).
match_info(X,X).
match_info(comparison_operator(Comp,Nr),ActualNr) :-
   number(Nr), number(ActualNr),call(Comp,ActualNr,Nr).

:- use_module(tools,[max_tagged_integer/1]).
:- public mc_ok_arg/2.
mc_ok_arg(Arg,X) :- Arg==all,!,max_tagged_integer(X).
mc_ok_arg(Arg,N) :- tools:arg_is_number(Arg,N).


:- dynamic option/1.
assert_all_options([]).
assert_all_options([Opt|T]) :- assert_option(Opt),
    assert_all_options(T).

assert_option(Opt) :- assert(option(Opt)),
    (Opt=silent -> set_silent_mode(on),set_error_manager_silent_mode(on) ; true).

reset_options :- retractall(option(_)), set_silent_mode(off), set_error_manager_silent_mode(off).

:- use_module(library(lists)).
% replace a leading double-dash -- by a single dash and replace inner dashes by underscores
normalise_option_atom(X,RX) :- atom(X),!,
    atom_codes(X,CodesX),
    % remove leading dash
    (CodesX=[45,45,H|T], H\=45
     -> maplist(prob_cli:convert_dash_to_underscore,[H|T],HT2),
        RXCodes=[45|HT2]
     ; CodesX = [45|T]
     -> maplist(prob_cli:convert_dash_to_underscore,T,T2),
        RXCodes=[45|T2]
     ; maplist(prob_cli:convert_dash_to_underscore,CodesX,RXCodes)
    ),
    atom_codes(RX,RXCodes).
normalise_option_atom(T,T).

:- public normalise_pref_name/2. % called via recognised_option
% replace dashes by underscores
normalise_pref_name(X,RX) :- atom(X),!,
    atom_codes(X,CodesX),
    maplist(prob_cli:convert_dash_to_underscore,CodesX,C2),
    atom_codes(RX,C2).
normalise_pref_name(T,T).

convert_dash_to_underscore(45,R) :- !, R=95.
convert_dash_to_underscore(X,X).

recognised_cli_option(X,Opt,Args,Condition) :- normalise_option_atom(X,RX),
   recognised_option(RX,Opt,Args,Condition).

recognised_option(X,Opt,[],true) :- recognised_option(X,Opt). % options without arguments
recognised_option(X,Opt,Args,true) :- recognised_option(X,Opt,Args). % options with arguments but no code needed to check arguments

recognised_option('-mc',cli_mc(N),[Arg],prob_cli:mc_ok_arg(Arg,N)).
recognised_option('-model_check',cli_mc(LimitNr),[],true) :- max_tagged_integer(LimitNr).
recognised_option('-timeout',timeout(N),[Arg],tools:arg_is_number(Arg,N)). % for model checking, refinement checking and for disprover per PO
recognised_option('-time_out',timeout(N),[Arg],tools:arg_is_number(Arg,N)).
recognised_option('-s',socket(S,true),[Arg],tools:arg_is_number(Arg,S)).
recognised_option('-cc',coverage(N,N2,just_check_stats),[Arg,Arg2],
                   (arg_is_number_or_wildcard(Arg,N),arg_is_number_or_wildcard(Arg2,N2))).
recognised_option('-hash',check_statespace_hash(H,_),[Arg],tools:arg_is_number(Arg,H)).
recognised_option('-hash64',check_statespace_hash(H,'64bit'),[Arg],tools:arg_is_number(Arg,H)).
recognised_option('-hash32',check_statespace_hash(H,'32bit'),[Arg],tools:arg_is_number(Arg,H)).
recognised_option('-ltllimit',ltl_limit(Nr),[Arg], tools:arg_is_number(Arg,Nr)).
recognised_option('-check_disprover_result',cli_check_disprover_result([true-TNr,false-FNr,unknown-UNr,failure-0]),[T,F,U],
   (arg_is_number_or_wildcard(T,TNr),arg_is_number_or_wildcard(F,FNr),arg_is_number_or_wildcard(U,UNr))).
recognised_option('-aa',cli_check_assertions(all,[true/TNr,false/FNr,unknown/UNr]),[T,F,U],
   (arg_is_number_or_wildcard(T,TNr),arg_is_number_or_wildcard(F,FNr),arg_is_number_or_wildcard(U,UNr))).
recognised_option('-ma',cli_check_assertions(main,[true/TNr,false/FNr,unknown/UNr]),[T,F,U],
   (arg_is_number_or_wildcard(T,TNr),arg_is_number_or_wildcard(F,FNr),arg_is_number_or_wildcard(U,UNr))).
recognised_option('-wd',cli_wd_check(DNr,TNr),[D,T],
   (arg_is_number_or_wildcard(T,TNr),arg_is_number_or_wildcard(D,DNr))).
recognised_option('-kodkod_comparision',kodkod_comparision(Nr),[Arg],tools:arg_is_number(Arg,Nr)).
recognised_option('-kodkod_performance',kodkod_performance(File,Nr),[File,Arg],tools:arg_is_number(Arg,Nr)).
recognised_option('-animate',cli_random_animate(N,true),[Steps],tools:arg_is_number(Steps,N)).
recognised_option('-execute',execute(N,true,current_state),[Steps],tools:arg_is_number(Steps,N)).
recognised_option('-logxml_write_vars',logxml_write_ids(variables,Prefix),[Prefix],true).
recognised_option('-logxml_write_ids',logxml_write_ids(all,Prefix),[Prefix],true).
recognised_option('-zmq_master',zmq_master(Identifier),[Identifier], true).
recognised_option('-cbc_tests', cbc_tests(Depth,EndPred,Output),[Depth,EndPred,Output],
                  tools:check_filename_arg(Output,'cbc_tests')).
recognised_option('-mcm_tests', mcm_tests(Depth,MaxStates,EndPred,Output),[Depth,MaxStates,EndPred,Output],
                  tools:check_filename_arg(Output,'mcm_tests')).
recognised_option('-test_description', test_description(File), [File],
                  tools:check_filename_arg(File,'test_description')).
recognised_option('-all_paths', all_deadlocking_paths(File), [File],
                  tools:check_filename_arg(File,'all_paths')).
recognised_option('-dot',dot_command(Category,File),[Category,File],
                  tools:check_filename_arg(File,'dot')).
recognised_option('-spdot',dot_command(state_space,File),[File], tools:check_filename_arg(File,'spdot')). % we keep this : it is shown in Wiki
% recognised_option('-spmdot',dot_command(signature_merge,File),[File], tools:check_filename_arg(File,'spmdot')).
% recognised_option('-spddot',dot_command(dfa_merge,File),[File], tools:check_filename_arg(File,'spddot')).
% recognised_option('-sgdot',dot_command(state_as_graph,File),[File],  tools:check_filename_arg(File,'sgdot')).
recognised_option('-dotexpr',dot_command_for_expr(Category,Expr,File,[]),[Category,Expr,File],
                  tools:check_filename_arg(File,'dot')).
%recognised_option('-sgedot',dot_command_for_expr(expr_as_graph,Expr,File,[]),[Expr,File], tools:check_filename_arg(File,'sgedot')).
% recognised_option('-sptdot',dot_command_for_expr(transition_diagram,Expr,File,[]),[Expr,File],tools:check_filename_arg(File,'sptdot')).
%recognised_option('-invdot',dot_command(invariant,File),[File],   tools:check_filename_arg(File,'invdot')).
%recognised_option('-propdot',dot_command(properties,File),[File],    tools:check_filename_arg(File,'propdot')).
%recognised_option('-assdot',dot_command(assertions,File),[File],  tools:check_filename_arg(File,'assdot')).
%recognised_option('-deaddot',dot_command(deadlock,File)(File),[File],   tools:check_filename_arg(File,'deaddot')).
recognised_option('-csvhist',evaluate_expression_over_history_to_csv_file(Expr,File),[Expr,File],
                  tools:check_filename_arg(File,'csvhist')).
recognised_option('-read_write_matrix_csv',generate_read_write_matrix_csv(CsvFile),
                  [CsvFile],
                  tools:check_filename_arg(CsvFile,'read_write_matrix_csv')).
recognised_option('-feasibility_analysis_csv',feasibility_analysis_csv(TimeOut,EnablingCsvFile),
                  [TimeOut,EnablingCsvFile],
                  tools:check_filename_arg(EnablingCsvFile,'feasibility_analysis_csv')).
recognised_option('-l',log(Log,prolog),[Log],
                  tools:check_filename_arg(Log,'l')).
recognised_option('-log',log(Log,prolog),[Log],
                  tools:check_filename_arg(Log,'log')).
recognised_option('-logxml',log(Log,xml),[Log],
                  tools:check_filename_arg(Log,'logxml')).
recognised_option('-pp',pretty_print_internal_rep(File,needed,ascii),[File],
                  tools:check_filename_arg(File,'pp')).
recognised_option('-ppunicode',pretty_print_internal_rep(File,needed,unicode),[File],
                  tools:check_filename_arg(File,'pp')).
recognised_option('-ppf',pretty_print_internal_rep(File,all,ascii),[File],
                  tools:check_filename_arg(File,'ppf')).
recognised_option('-ppB',pretty_print_internal_rep_to_B(File),[File],
                  tools:check_filename_arg(File,'ppB')).
recognised_option('-save_state',save_state(Filename),[Filename],
                  tools:check_filename_arg(Filename,'save_state')).
recognised_option('-save',save_state_for_refinement(Filename),[Filename],
                  tools:check_filename_arg(Filename,'save')).
recognised_option('-sptxt',print_values(Filename),[Filename],
                  tools:check_filename_arg(Filename,'sptxt')).
recognised_option('-sstxt',print_all_values(Dirname),[Dirname],
                  tools:check_filename_arg(Dirname,'sstxt')).
recognised_option('-latex',process_latex_file(In,Out),[In,Out],
                  (tools:check_filename_arg(In,'latex'),tools:check_filename_arg(Out,'latex'))).

recognised_option('-zmq_assertions',zmq_assertion(Identifier),[Identifier],true).
recognised_option('-zmq_worker',zmq_worker(Identifier),[Identifier], true).
%recognised_option('-zmq_worker2',zmq_worker2(MasterIP, Port, ProxyID, Logfile),[MasterIP, SPort, SProxyID, Logfile],
%                  tools:(arg_is_number(SPort,Port), arg_is_number(SProxyID, ProxyID))).
recognised_option('-p',set_pref(NPREF,PREFVAL),[PREF,PREFVAL],prob_cli:normalise_pref_name(PREF,NPREF)).
recognised_option('-pref',set_pref(NPREF,PREFVAL),[PREF,PREFVAL],prob_cli:normalise_pref_name(PREF,NPREF)).
recognised_option('-cbc_redundant_invariants',cbc_redundant_invariants(Nr),[X],tools:arg_is_number(X,Nr)).
recognised_option('-expcterrpos',expect_error_pos(Type,LNr,CNr),[Type,Line,Col],
                  (tools:arg_is_number(Line,LNr),tools:arg_is_number(Col,CNr))).

recognised_option('-pref_group',set_preference_group(Group,Val),[Group,Val]).
recognised_option('-prefs',set_prefs_from_file(PREFFILE),[PREFFILE]).
%recognised_option('-plugin',plugin(Plugin), [Plugin]).
recognised_option('-card',set_card(SET,SCOPE),[SET,SCOPE]).
recognised_option('-argv',set_argv(ARGV),[ARGV]).
recognised_option('-goal',set_goal(GOAL),[GOAL]).
recognised_option('-property',add_additional_property(PRED),[PRED]).
recognised_option('-scope',set_searchscope(GOAL),[GOAL]).
recognised_option('-searchscope',set_searchscope(GOAL),[GOAL]).
recognised_option('-eval',eval_string_or_file(string,E,exists,_,norecheck),[E]).
recognised_option('-evalt',eval_string_or_file(string,E,exists,'TRUE',norecheck),[E]).
recognised_option('-evalf',eval_string_or_file(string,E,exists,'FALSE',norecheck),[E]).
recognised_option('-evalt_rc',eval_string_or_file(string,E,exists,'TRUE',recheck(ascii)),[E]).
recognised_option('-evalf_rc',eval_string_or_file(string,E,exists,'FALSE',recheck(ascii)),[E]).
recognised_option('-evalu',eval_string_or_file(string,E,exists,'UNKNOWN',norecheck),[E]).
recognised_option('-evalnwd',eval_string_or_file(string,E,exists,'NOT-WELL-DEFINED',norecheck),[E]).
recognised_option('-eval_file',eval_string_or_file(file,F,exists,_ANY,norecheck),[F]).
recognised_option('-evalt_file',eval_string_or_file(file,F,exists,'TRUE',norecheck),[F]).
recognised_option('-eval_rule_file',eval_string_or_file(file,F,forall,_ANY,norecheck),[F]).
recognised_option('-parsercp',(L),[L]).
recognised_option('-expcterr',expect_error(Type),[Type]).
recognised_option('-expecterr',expect_error(Type),[Type]).
recognised_option('-expect',expect_error(Type),[Type]).
recognised_option('-opterr',optional_error(Type),[Type]).
recognised_option('-his',history(File),[File]).
recognised_option('-his_option',history_option(Option),[Option]).
recognised_option('-cache',cache_storage(D),[D]).
recognised_option('-show_cache',show_cache,[]).

recognised_option('-MAIN',csp_main(ProcessName),[ProcessName]).

recognised_option('-ltlfile',ltl_file(Filename),[Filename]).
recognised_option('-ltlformula',ltl_formula_model_check(Formula,_),[Formula]).
recognised_option('-ltlformulat',ltl_formula_model_check(Formula,true),[Formula]).
recognised_option('-ltlformulaf',ltl_formula_model_check(Formula,false),[Formula]).
recognised_option('-ctlformula',ctl_formula_model_check(Formula,_),[Formula]).
recognised_option('-ctlformulat',ctl_formula_model_check(Formula,true),[Formula]).
recognised_option('-ctlformulaf',ctl_formula_model_check(Formula,false),[Formula]).
recognised_option('-load_state',load_state(Filename),[Filename]).
recognised_option('-refchk',refinement_check(Filename,0,100000),[Filename]).

%recognised_option('-cspref',csp_in_situ_refinement_check(assertRef('False',val_of(AbsP1,no_loc_info_available),Type,val_of(ImplP2,no_loc_info_available),no_loc_info_available),'False'),[AbsP1,Type,ImplP2]).
recognised_option('-cspref',csp_in_situ_refinement_check(AbsP1,Type,ImplP2),[AbsP1,Type,ImplP2]).
% -cspref R [F= Q
recognised_option('-cspdeadlock',csp_checkAssertion(Proc,Model,'deadlock free'),[Proc,Model]).
% -cspdeadlock R F
recognised_option('-cspdeterministic',csp_checkAssertion(Proc,Model,'deterministic'),[Proc,Model]).
% -cspdeterministic R F
recognised_option('-csplivelock',csp_checkAssertion(Proc,'FD','livelock free'),[Proc]).
% -csplivelock R
% -csp_assertion "P [F= Q"
recognised_option('-csp_assertion',check_csp_assertion(Assertion),[Assertion]).
recognised_option('-csp_eval', eval_csp_expression(Expr),[Expr]).
recognised_option('-get_csp_assertions_as_string',csp_get_assertions,[]).
recognised_option('-prologOut',csp_translate_to_file(PlFile),[PlFile]).
recognised_option('-csp_guide',add_csp_guide(Filename),[Filename]).
recognised_option('-get_min_max_coverage',get_min_max_coverage(FileName),[FileName]).
recognised_option('-get_coverage_information',get_coverage_information(FileName),[FileName]).
recognised_option('-vacuity_check',vacuity_check,[]).
recognised_option('-wd_check',cli_wd_check(_,_),[]).
recognised_option('-wd_check_all',cli_wd_check(X,X),[]).
recognised_option('-well_definedness_check',cli_wd_check(_,_),[]).
recognised_option('-lint',cli_lint,[]). % extended static check (ESC)
recognised_option('-cbc',constraint_based_check(OpName),[OpName]).
recognised_option('-cbc_deadlock',cbc_deadlock_check(true),[]).
recognised_option('-cbc_assertions',cbc_assertions(true,[]),[]).
recognised_option('-cbc_main_assertions',cbc_assertions(true,[main_assertions]),[]).
recognised_option('-cbc_assertions_proof',cbc_assertions(false,[]),[]). % do not allow enumeration warnings
recognised_option('-cbc_assertions_tautology_proof',cbc_assertions(false,[tautology_check]),[]). % do not allow enumeration warnings + disregard PROPERTIES, used for Atelier-B proof/disproof; TO DO: also call WD prover
recognised_option('-cbc_assertions_tautology_proof_check',cbc_assertions(false,[tautology_check,contradiction_check]),[]).
recognised_option('-cbc_option',cbc_option(OPT),[OPT]). % should be tautology_check,contradiction_check, unsat_core
recognised_option('-cbc_result_file',cbc_result_file(FILE),[FILE]). % write result to FILE
recognised_option('-cbc_refinement',cbc_refinement,[]).
recognised_option('-cbc_deadlock_pred',cbc_deadlock_check(GoalPred),[GoalPred]).
recognised_option('-cbc_sequence',cbc_sequence(OpSequence,'',single_solution),[OpSequence]).
recognised_option('-cbc_sequence_all',cbc_sequence(OpSequence,'',findall),[OpSequence]).
recognised_option('-cbc_sequence_with_target',cbc_sequence(OpSequence,TargetPredString,single_solution),[OpSequence,TargetPredString]).
recognised_option('-cbc_sequence_with_target_all',cbc_sequence(OpSequence,TargetPredString,findall),[OpSequence,TargetPredString]).
recognised_option('-comment',comment(UserComment),[UserComment]). % not processed by tool, but will be stored in log-file and used by log_analyser
recognised_option('-junit',junit(Dir),[Dir]).
recognised_option('-mcm_cover', mcm_cover(Event),[Event]).
recognised_option('-cbc_cover', cbc_cover(Event),[Event]).
recognised_option('-cbc_cover_match', cbc_cover(match_event(Event)),[Event]). % find events which have Event String occuring somewhere in name
recognised_option('-cbc_cover_all', cbc_cover_all,[]).  % is now default if no cbc_cover provided
recognised_option('-cbc_cover_final', cbc_cover_final,[]).
recognised_option('-bmc', cbc_tests(Depth,'#not_invariant',''),[Depth]).
recognised_option('-bdc', cbc_tests(Depth,'#deadlock',''),[Depth]).
recognised_option('-dot_output',dot_analyse_output_prefix(Path),[Path]).
recognised_option('-evaldot',evaldot(File),[File]).
recognised_option('-enabling_analysis',enabling_analysis_csv(user_output),[]).
recognised_option('-enabling_analysis_csv',enabling_analysis_csv(EnablingCsvFile),[EnablingCsvFile]).
recognised_option('-feasibility_analysis',feasibility_analysis_csv(1000,user_output),[]).
recognised_option('-read_write_matrix',generate_read_write_matrix_csv(user_output),[]).
recognised_option('-state_trace',state_trace(File),[File]).
recognised_option('-typecheckertest',typechecker_test(File),[File]).
recognised_option('-scc_trace',check_scc_for_ltl_formula(LtlFormula,SCC),[LtlFormula,SCC]).
recognised_option('-selfcheck_module',selfcheck(M),[M]).
recognised_option('-mc_mode',depth_breadth_first_mode(M),[M]). % can be mixed, hash, heuristic
recognised_option('-assertion',cli_check_assertions(specific(X),[false/0,unknown/0]),[X]).
recognised_option('-cbc_assertion',cbc_assertions(true,[specific(X)]),[X]). % check only a specific assertion
recognised_option('-symbolic_model_check', cli_symbolic_model_check(Algorithm), [Algorithm]).
recognised_option('-replay',eval_repl([File]),[File]). % used to be -eval
recognised_option('-ltsmin2',ltsmin2(EndpointPath), [EndpointPath]).
recognised_option('-ltsmin_ltl_output',ltsmin_ltl_output(Path), [Path]).
recognised_option('-ltsmin_option', ltsmin_option(X),[X]).
recognised_option('-machine_hash_check',cli_print_machine_info(hash(X)),[X]).
recognised_option('-bench_csv',benchmark_info_csv_output(File),[File]).

recognised_option('-dot_all',dot_generate_for_all_formulas). % generate dot also for true formulas
recognised_option('-animate_all',cli_random_animate(2147483647,false)).
recognised_option('-execute_all',execute(2147483647,false,current_state)).
recognised_option('-execute_all_inits',execute(2147483647,false,from_all_initial_states)).
recognised_option('-animate_stats',animate_stats).
recognised_option('-execute_monitor',execute_monitoring).
recognised_option('-check_goal',check_goal).
recognised_option('-ltlassertions',ltl_assertions).
recognised_option('-assertions',cli_check_assertions(all,[false/0,unknown/0])).
recognised_option('-main_assertions',cli_check_assertions(main,[false/0,unknown/0])).
recognised_option('-properties',cli_check_properties).
recognised_option('-selfcheck',selfcheck(_)).
recognised_option('-pacheck',pa_check). % predicate analysis for Kodkod
recognised_option('-det_check',det_check). % check if animation is deterministic
recognised_option('-det_constants',det_constants_check). % check if animation for setup_constants is deterministic
recognised_option('-bf',breadth_first).
recognised_option('-breadth',breadth_first).
recognised_option('-df',depth_first).
recognised_option('-depth',depth_first).
recognised_option('-strict',strict_raise_error).
recognised_option('-silent',silent).
recognised_option('-statistics',statistics).
recognised_option('-profile_statistics',profile_statistics).
recognised_option('-reset_profile_statistics',reset_profiler). % mainly for use in REPL
recognised_option('-nodead',no_deadlocks).
recognised_option('-no_deadlocks',no_deadlocks).
recognised_option('-noinv',no_invariant_violations).
recognised_option('-no_invariant_violations',no_invariant_violations).
recognised_option('-nogoal',no_goal).
recognised_option('-no_goal',no_goal).
recognised_option('-noltl',no_ltl). % just used for TLC at the moment
recognised_option('-noass',no_assertion_violations).
recognised_option('-no_assertion_violations',no_assertion_violations).
recognised_option('-no_state_errors',no_state_errors). % disable checking for general_errors and transition related state_errors
recognised_option('-nocounter',no_counter_examples).
recognised_option('-no_counter_examples',no_counter_examples).
recognised_option('-disable_time_out',set_preference_group(time_out,disable_time_out)).
recognised_option('-disable_timeout',set_preference_group(time_out,disable_time_out)).
%recognised_option('-POR',with_reduction).
recognised_option('-i',animate).
recognised_option('-repl',eval_repl([])). % used to be -eval
recognised_option('-c',coverage(false)).
recognised_option('-cs',coverage(just_summary)).
recognised_option('-coverage',coverage(false)).
recognised_option('-coverage_summary',coverage(just_summary)).
recognised_option('-machine_stats',cli_print_machine_info(statistics)).
recognised_option('-machine_statistics',cli_print_machine_info(statistics)).
recognised_option('-machine_files',cli_print_machine_info(files)).
recognised_option('-machine_hash',cli_print_machine_info(hash(_))).
recognised_option('-vc',variable_coverage).
recognised_option('-variable_coverage',variable_coverage).
recognised_option('-cv',coverage(true)).
recognised_option('-v',verbose).
recognised_option('-vv',very_verbose).
recognised_option('-verbose',verbose).
recognised_option('-very_verbose',very_verbose).
recognised_option('-profiling_on',profiling_on). % Prolog profiling
recognised_option('-profile',prob_profile). % ProB Operation profiling
recognised_option('-prob_profile',prob_profile). % ProB Operation profiling
recognised_option('-version',print_version).
recognised_option('-svers',print_short_version).
recognised_option('-short_version',print_short_version).
recognised_option('-check_java_version',check_java_version).
recognised_option('-java_version',check_java_version).
recognised_option('-release_java_parser',release_java_parser).
recognised_option('-file_info',file_info).
recognised_option('-t',trace_check).
recognised_option('-trace_check',trace_check).
recognised_option('-init',initialise).
recognised_option('-initialise',initialise).
recognised_option('-ll',log('/tmp/prob_cli_debug.log',prolog)).
recognised_option('-ss',socket(9000,true)). % standard socket 9000
recognised_option('-sf',socket(_,true)). % free socket
recognised_option('-local_socketserver',socket(_,true)). % do not allow remote socket connections
recognised_option('-remote_socketserver',socket(_,false)). % allow remote socket connections
recognised_option('-help',help).
recognised_option('-h',help).
recognised_option('-rc',runtimechecking).
recognised_option('-test_mode',test_mode).
recognised_option('-check_complete',check_complete).
recognised_option('-check_complete_operation_coverage', check_complete_operation_coverage).
recognised_option('-mc_with_tlc', cli_start_mc_with_tlc).
recognised_option('-mc_with_lts_sym', cli_start_sym_mc_with_lts(symbolic)).
recognised_option('-mc_with_lts_seq', cli_start_sym_mc_with_lts(sequential)).
%recognised_option('-csp',csp_pl_mode).
%recognised_option('-cspm',csp_mode).
recognised_option('-core',disprover_options([disprover_option(unsat_core),unsat_core_algorithm/linear])).

recognised_option('-ltsmin',ltsmin).

% some utilities to be able to call the above options directly from repl:
:- public variable_coverage/0, silent/0, coverage/1, help/0.
% predicate to set_verbose_mode
verbose :- tcltk_turn_debugging_on(19).
very_verbose :- tcltk_turn_debugging_on(5).
file_info :- file_loaded(true,MainFile), print_file_info(MainFile).
coverage(ShowEnabledInfo) :- probcli_time_stamp(NOW), cli_show_coverage(ShowEnabledInfo,NOW).
variable_coverage :- probcli_time_stamp(NOW), cli_show_variable_coverage(NOW).
silent :- (option(silent) -> true ; assert_option(silent)).
help :- eval_help.
dot_command(DCommand,DotFile) :- call_dot_command(DCommand,DotFile,[]).
dot_command_for_expr(DECommand,Expr,DotFile,Opts) :- call_dot_command_for_expr(DECommand,Expr,DotFile,Opts).
save_state(StateFile) :- debug_println(20,'% Saving state space to file'),
                   state_space:tcltk_save_state(StateFile).
:- public load_state/1. % for REPL
load_state(StateFile) :- debug_println(20,'% Loading state space from file'),
                   state_space:tcltk_load_state(StateFile).
:- public execute/3. % for REPL
execute(ESteps,ErrOnDeadlock,From) :- cli_execute(ESteps,ErrOnDeadlock,From).

option_verbose :- (option(verbose) -> true ; option(very_verbose)).

% new profiler
%:- use_module(('../extensions/profiler/profiler.pl').
%prob_profiler :- pen.

% old profiler
:- use_module(runtime_profiler,[print_runtime_profile/0]).
:- use_module(source_profiler,[print_source_profile/0]).
:- use_module(memoization,[print_memo_profile/0]).
prob_profile :-
   statistics(walltime,[WT,_]),
   statistics(runtime,[RT,_]),
   format('--------------------------~nProB profile info after ~w ms walltime (~w ms runtime)~n',[WT,RT]),
   print_source_profile,
   print_runtime_profile,
   print_memo_profile.


set_random_seed_to_deterministic_start_seed :-
     % in test_mode we do not change the random number generator's initial seed
     true. %getrand(CurrState),setrand(CurrState). % this seems to be a no-op

:- use_module(library(random),[setrand/1]).
set_new_random_seed :-
   now(TimeStamp), % getting the unix time
   setrand(TimeStamp). % setting new random seed by every execution of probcli

halt_exception :- halt_exception(0).
halt_exception(Code) :- throw(halt(Code)).

% -----------------

start_xml_feature(FeatureName,[CErrs1,CWarns1,CEErrs1]) :-
    debug_format(20,'% Starting ~w~n',[FeatureName]),
    get_counter(cli_errors,CErrs1), get_counter(cli_warnings,CWarns1), get_counter(cli_expected_errors,CEErrs1),
    start_xml_group_in_log(FeatureName).

start_xml_feature(FeatureName,Attr,Value,[CErrs1,CWarns1,CEErrs1]) :-
    debug_format(20,'% Starting ~w (~w=~w)~n',[FeatureName,Attr,Value]),
    get_counter(cli_errors,CErrs1), get_counter(cli_warnings,CWarns1), get_counter(cli_expected_errors,CEErrs1),
    start_xml_group_in_log(FeatureName,Attr,Value).

stop_xml_feature(FeatureName,[CErrs1,CWarns1,CEErrs1]) :-
	get_counter(cli_errors,CErrs2), get_counter(cli_warnings,CWarns2), get_counter(cli_expected_errors,CEErrs2),
	CErrs is CErrs2-CErrs1, CWarns is CWarns2-CWarns1, CEErrs is CEErrs2-CEErrs1,
	(CEErrs>0
	  ->  write_xml_element_to_log('probcli-errors',[errors/CErrs,warnings/CWarns,expected_errors/CEErrs])
	   ;  write_xml_element_to_log('probcli-errors',[errors/CErrs,warnings/CWarns])
	),
  debug_format(20,'% Finished ~w (errors=~w, warnings=~w, expected_errors=~w)~n',[FeatureName,CErrs,CWarns,CEErrs]),
	stop_xml_group_in_log(FeatureName).
	%(CErrs>0 -> (file_loaded(_,MainFile) -> true ; MainFile=unknown), Time=unknown, % TO DO: determine time
    %   create_and_print_junit_result(FeatureName,'Feature', MainFile, Time, error) ; true).
% Note: call stop_xml_group_in_log if the feature stops unexpectedly and you do not have the Info list available

% -----------------

:- public user:runtime_entry/1.
user:runtime_entry(start) :- go_cli.

%save :- save_program('probcli.sav').

:- use_module(eventhandling,[announce_event/1]).
:- announce_event(compile_prob).
