% (c) 2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html

:- module(fuzzer,[fuzz/3, gen_data/2, test/2, test_clpfd/2, test_clpfd1/2]).



:- use_module(library(lists)).



:- use_module(library(timeout)).
run(Call,Result) :- time_out( run_aux(Call,InnerResult), 1000, TORes),
  (TORes= time_out -> Result = time_out ; Result = InnerResult).

run_aux(Call,Result) :- on_exception(Exc, if(Call,Result=ok,Result=fail), (Result=exception(Exc))).

gen_data(Desc,Data) :- gen_data(Desc,Data,[],_).

:- use_module(library(random)).
gen_data(tagged_int,Data) -->
  {prolog_flag(max_tagged_integer,Max),prolog_flag(min_tagged_integer,Min),Max1 is Max+1,
   random(Min,Max1,Data)}.
gen_data(integer(L,U),Data) --> {random(L,U,Data)}.
gen_data(int8,Data) --> {random(-128,128,Data)}.
gen_data(random(L),Data) --> {random_member(Elem,L)}, gen_data(Elem,Data).
gen_data(Module:Call,Module:Data) --> gen_data(Call,Data).
gen_data([H|T],Data) --> {random_member(Elem,[H|T])}, gen_data(Elem,Data).
gen_data(term(F,FArgs),Data) --> l_gen_data(FArgs,DArgs), {Data =.. [F|DArgs]}.
gen_data(var,Data,In,Out) :- random_member(Data,[V|In]), % existing or new variable
   (exact_member(Data,In) -> Out=In ; Out=[V|In]).
gen_data(evar,Data,In,Out) :- random_member(Data,In), Out=In. % existing variable
gen_data(type(L),Data) --> {type_def(L,Desc)}, gen_data(Desc,Data).
gen_data(L,Data) --> {type_def(L,Desc)}, gen_data(Desc,Data).

l_gen_data([],[]) --> [].
l_gen_data([Desc|T],[Data|DT]) --> gen_data(Desc,Data), l_gen_data(T,DT).

exact_member(X,[Y|T]) :-
   ((X==Y) -> true ; exact_member(X,T)).

:- dynamic type_def/2.
% list of data type definitions; extensible at runtime
type_def(list(L),R) :- R=random([ term([],[]), term('.',[L,type(list(L))])]).
type_def(vlist(L),R) :- R=random([ var, term([],[]), term('.',[L,type(list(L))])]).
type_def(opt(Desc),R) :- R=random( [Desc, var]).
type_def(arith(BaseType),R) :- !, R= arith(BaseType,2).
type_def(arith(BaseType,0),R) :- !, R= BaseType.
type_def(arith(BaseType,Lvl),R) :- L1 is Lvl-1,
  R = random([  BaseType,
                term('+',[arith(BaseType,L1),arith(BaseType,L1)]),
                term('-',[arith(BaseType,L1),arith(BaseType,L1)]),
                term('*',[arith(BaseType,L1),arith(BaseType,L1)]) ]).
type_def(goal(G),goal(G,4)).
type_def(goal(G,0),R) :- !, R=G.
type_def(goal(G,Lvl),R) :- L1 is Lvl-1, R=random( [G, term(',',[goal(G,L1),goal(G,L1)])]).
type_def(clpfd_range,R) :- R = term('..',[int8,int8]).
type_def(clpfd(BT),R) :-
     R = random([ term('#=',[arith(BT),arith(BT)]),
                  term('#<',[arith(BT),arith(BT)]),
                  term('#>',[arith(BT),arith(BT)]),
                  term('#>=',[arith(BT),arith(BT)]),
                  term('minimum',[BT,list(BT)]),
                  term('maximum',[BT,list(BT)]),
                  term('in',[var, clpfd_range])
                ]).

fuzz(Description,ErrCall,Error) :-
    fuzz(Description,ErrCall,Error,[ok,fail]).
fuzz(Description,ErrCall,Error,AcceptableResults) :-
   gen_data(Description,Call,[],_Vars),
   print(calling(Call)),
   run(Call,Result), print(result(Result)),nl,
   (nonmember(Result,AcceptableResults)
     -> Error=Result, ErrCall=Call
      ; fuzz(Description,ErrCall,Error,AcceptableResults)
    ).


test(C,R) :- fuzz(term(append,[vlist(opt(tagged_int)),vlist(opt(tagged_int)),vlist(opt(tagged_int))]),C,R).

test_clpfd1(C,R) :-
   fuzz(clpfd:goal(clpfd(random([var,int8]))),C,R).
test_clpfd(C,R) :-
   fuzz(clpfd:goal(clpfd(random([var,int8]))),C,R,
   [ok,fail,time_out,
    exception(error(representation_error(max_clpfd_integer),_)),
    exception(error(representation_error(min_clpfd_integer),_))
    ]).


/*
calling(clpfd:(19*(_2789283*_2789283)+(110-89)*_2790053#= -122,30- -16#=_2791613))result(time_out)
C = clpfd:(19*(_A*_A)+(110-89)*_B#= -122,30- -16#=_C),
R = time_out ?

C = clpfd:(-48+(_A+100)-7#>=_A*(-65- -45+ -96* -120),(minimum(30,[_B,_A]),(_A* -125+_A)*(_C*_A)#>=_D),-11* -60*(_E+94)+(_F* -14+_E* -50)#<(_A* -55-(-51+_B))*(_F+_D-_C*_E),minimum(_A,[])),
R = exception(error(representation_error(max_clpfd_integer),representation_error('x*y=z'(_G,11500,_H),0,max_clpfd_integer))) ?

calling(clpfd:(_60851*_60955+(_60955+_60851)+66*(_62059+118)#>(-35-55+(96+ -62))*_63433))result(ok)
R = exception(error(representation_error(max_clpfd_integer),representation_error('x*x=y'(_B,_C),0,max_clpfd_integer))) ?

C = clpfd:((88*51#= -12*_A+_B*_B,_B+_A+(54-_C)#>=_C*_C-(_C-_A)),_C#=_C+_A-_B),
R = exception(error(representation_error(max_clpfd_integer),
representation_error('x*x=y'(_D,_E),0,max_clpfd_integer))) ?

C = clpfd:(_A*_A-(_A-_A)#=(_A+ -48)*_A,((107#>_A- -100- -33*111,_B*_B+(66-33)#<_A+_C*41),(_D-_C)*(88* -68)#>_D+(36-_E),_C in-44..10),98+(_E-61)#=79*67+_E*86),
R = time_out ?
*/
